diff -pruN 2.3.1-1/.github/workflows/artifacts.yml 2.4-1/.github/workflows/artifacts.yml
--- 2.3.1-1/.github/workflows/artifacts.yml	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/.github/workflows/artifacts.yml	2024-07-10 09:56:14.000000000 +0000
@@ -9,13 +9,13 @@ jobs:
     steps:
     - name: Checkout code
       id: checkout-code
-      uses: actions/checkout@v3
+      uses: actions/checkout@v4
       with:
         submodules: recursive
         fetch-depth: 0
 
     - name: Setup problem matcher
-      uses: Joshua-Ashton/gcc-problem-matcher@v2
+      uses: Joshua-Ashton/gcc-problem-matcher@v3
 
     - name: Build release
       id: build-release
@@ -28,9 +28,9 @@ jobs:
 
     - name: Upload artifacts
       id: upload-artifacts
-      uses: actions/upload-artifact@v3
+      uses: actions/upload-artifact@v4
       with:
-        name: dxvk-${{ env.VERSION_NAME }}
+        name: dxvk-win-${{ env.VERSION_NAME }}
         path: build/dxvk-${{ env.VERSION_NAME }}
         if-no-files-found: error
 
@@ -41,26 +41,43 @@ jobs:
     steps:
     - name: Checkout code
       id: checkout-code
-      uses: actions/checkout@v3
+      uses: actions/checkout@v4
       with:
         submodules: recursive
         fetch-depth: 0
 
     - name: Setup problem matcher
-      uses: Joshua-Ashton/gcc-problem-matcher@v2
+      uses: Joshua-Ashton/gcc-problem-matcher@v3
 
     - name: Build release
       id: build-release
       shell: bash
       run: |
         export VERSION_NAME="${GITHUB_REF##*/}-${GITHUB_SHA##*/}"
-        ./package-native.sh ${VERSION_NAME} build --no-package
+        ./package-native.sh ${VERSION_NAME} build
         echo "VERSION_NAME=${VERSION_NAME}" >> $GITHUB_ENV
 
     - name: Upload artifacts
       id: upload-artifacts
-      uses: actions/upload-artifact@v3
+      uses: actions/upload-artifact@v4
       with:
-        name: dxvk-${{ env.VERSION_NAME }}
-        path: build/dxvk-native-${{ env.VERSION_NAME }}
+        name: dxvk-native-${{ env.VERSION_NAME }}
+        path: build/dxvk-native-${{ env.VERSION_NAME }}.tar.gz
         if-no-files-found: error
+
+  merge-artifacts:
+    runs-on: ubuntu-20.04
+    needs: [artifacts-mingw-w64, artifacts-steamrt-sniper]
+    steps:
+      - name: Get version
+        id: get-version
+        shell: bash
+        run: |
+          echo "VERSION_NAME=${GITHUB_REF##*/}-${GITHUB_SHA##*/}" >> $GITHUB_ENV
+
+      - name: Merge Artifacts
+        uses: actions/upload-artifact/merge@v4
+        with:
+          name: dxvk-${{ env.VERSION_NAME }}
+          pattern: dxvk*
+          delete-merged: true
diff -pruN 2.3.1-1/.github/workflows/test-build-windows.yml 2.4-1/.github/workflows/test-build-windows.yml
--- 2.3.1-1/.github/workflows/test-build-windows.yml	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/.github/workflows/test-build-windows.yml	2024-07-10 09:56:14.000000000 +0000
@@ -9,7 +9,7 @@ jobs:
     steps:
     - name: Checkout code
       id: checkout-code
-      uses: actions/checkout@v3
+      uses: actions/checkout@v4
       with:
         submodules: recursive
 
@@ -32,6 +32,13 @@ jobs:
         Write-Output "VSDEVCMD=${installationPath}\Common7\Tools\VsDevCmd.bat" `
           | Out-File -FilePath "${Env:GITHUB_ENV}" -Append
 
+    - name: Download D3D8 SDK Headers
+      shell: pwsh
+      run: |
+        Invoke-WebRequest -URI https://raw.githubusercontent.com/NovaRain/DXSDK_Collection/master/DXSDK_Aug2007/Include/d3d8.h -OutFile include/d3d8.h
+        Invoke-WebRequest -URI https://raw.githubusercontent.com/NovaRain/DXSDK_Collection/master/DXSDK_Aug2007/Include/d3d8types.h -OutFile include/d3d8types.h
+        Invoke-WebRequest -URI https://raw.githubusercontent.com/NovaRain/DXSDK_Collection/master/DXSDK_Aug2007/Include/d3d8caps.h -OutFile include/d3d8caps.h
+
     - name: Build MSVC x86
       shell: pwsh
       run: |
diff -pruN 2.3.1-1/LICENSE 2.4-1/LICENSE
--- 2.3.1-1/LICENSE	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/LICENSE	2024-07-10 09:56:14.000000000 +0000
@@ -1,5 +1,7 @@
                   Copyright (c) 2017 Philip Rebohle
                   Copyright (c) 2019 Joshua Ashton
+                  Copyright (c) 2019 Robin Kertels
+                  Copyright (c) 2023 Jeffrey Ellison
 
                           zlib/libpng license
 
diff -pruN 2.3.1-1/README.md 2.4-1/README.md
--- 2.3.1-1/README.md	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/README.md	2024-07-10 09:56:14.000000000 +0000
@@ -1,6 +1,6 @@
 # DXVK
 
-A Vulkan-based translation layer for Direct3D 9/10/11 which allows running 3D applications on Linux using Wine.
+A Vulkan-based translation layer for Direct3D 8/9/10/11 which allows running 3D applications on Linux using Wine.
 
 For the current status of the project, please refer to the [project wiki](https://github.com/doitsujin/dxvk/wiki).
 
@@ -9,7 +9,7 @@ The most recent development builds can b
 Release builds can be found [here](https://github.com/doitsujin/dxvk/releases).
 
 ## How to use
-In order to install a DXVK package obtained from the [release](https://github.com/doitsujin/dxvk/releases) page into a given wine prefix, copy or symlink the DLLs into the following directories as follows, then open `winecfg` and manually add DLL overrides for `d3d11`, `d3d10core`, `dxgi`, and `d3d9`.
+In order to install a DXVK package obtained from the [release](https://github.com/doitsujin/dxvk/releases) page into a given wine prefix, copy or symlink the DLLs into the following directories as follows, then open `winecfg` and manually add DLL overrides for `d3d11`, `d3d10core`, `dxgi`, `d3d9` and `d3d8`.
 
 In a default Wine prefix that would be as follows:
 ```
@@ -163,3 +163,22 @@ For non debian based distros, make sure
 does have `--enable-threads=posix` enabled during configure. If your distro does
 ship its mingw-w64-gcc binary with `--enable-threads=win32` you might have to
 recompile locally or open a bug at your distro's bugtracker to ask for it. 
+
+# DXVK Native
+
+DXVK Native is a version of DXVK which allows it to be used natively without Wine.
+
+This is primarily useful for game and application ports to either avoid having to write another rendering backend, or to help with port bringup during development.
+
+[Release builds](https://github.com/doitsujin/dxvk/releases) are built using the Steam Runtime.
+
+### How does it work?
+
+DXVK Native replaces certain Windows-isms with a platform and framework-agnostic replacement, for example, `HWND`s can become `SDL_Window*`s, etc.
+All it takes to do that is to add another WSI backend.
+
+**Note:** DXVK Native requires a backend to be explicitly set via the `DXVK_WSI_DRIVER` environment variable. The current built-in options are `SDL2` and `GLFW`.
+
+DXVK Native comes with a slim set of Windows header definitions required for D3D9/11 and the MinGW headers for D3D9/11.
+In most cases, it will end up being plug and play with your renderer, but there may be certain teething issues such as:
+- `__uuidof(type)` is supported, but `__uuidof(variable)` is not supported. Use `__uuidof_var(variable)` instead.
diff -pruN 2.3.1-1/RELEASE 2.4-1/RELEASE
--- 2.3.1-1/RELEASE	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/RELEASE	2024-07-10 09:56:14.000000000 +0000
@@ -1 +1 @@
-2.3.1
+2.4
diff -pruN 2.3.1-1/debian/changelog 2.4-1/debian/changelog
--- 2.3.1-1/debian/changelog	2024-05-08 09:37:24.000000000 +0000
+++ 2.4-1/debian/changelog	2024-07-11 17:48:33.000000000 +0000
@@ -1,3 +1,12 @@
+dxvk (2.4-1) unstable; urgency=medium
+
+  * NUR:
+    - refresh patches
+  * Update supported D3D features in description.
+  * Bump Standards-Version to 4.7.0 (no changes needed).
+
+ -- Marc Dequènes (Duck) <Duck@DuckCorp.org>  Fri, 12 Jul 2024 02:48:33 +0900
+
 dxvk (2.3.1-1) unstable; urgency=medium
 
   * NUR:
diff -pruN 2.3.1-1/debian/control 2.4-1/debian/control
--- 2.3.1-1/debian/control	2024-05-08 09:37:24.000000000 +0000
+++ 2.4-1/debian/control	2024-07-11 17:48:33.000000000 +0000
@@ -16,7 +16,7 @@ Build-Depends: debhelper-compat (= 13),
                g++-mingw-w64-i686-posix [i386 armel armhf],
                rename,
                llvm,
-Standards-Version: 4.6.2
+Standards-Version: 4.7.0
 Homepage: https://github.com/doitsujin/dxvk
 Vcs-Browser: https://salsa.debian.org/aviau/dxvk
 Vcs-Git: https://salsa.debian.org/aviau/dxvk.git
@@ -31,11 +31,11 @@ Depends: ${shlibs:Depends},
          dxvk-wine64-development (>= ${source:Version}) | dxvk-wine32-development (>= ${source:Version}),
 Recommends: dxvk-wine32-development,
             dxvk-wine64-development,
-Description: Vulkan-based translation layer for Direct3D 9/10/11
- DXVK is a Vulkan-based translation layer for Direct3D 9/10/11 which
+Description: Vulkan-based translation layer for Direct3D 8/9/10/11
+ DXVK is a Vulkan-based translation layer for Direct3D 8/9/10/11 which
  allows running 3D applications on Linux using Wine.
  .
- DXVK aims to provide support for D3D11, feature level 11_1, and
+ DXVK aims to provide support for D3D11, feature level 12_1, and
  D3D10, feature level 10_1.
  .
  This package contains the dxvk-setup script that configures
@@ -47,11 +47,11 @@ Multi-Arch: foreign
 Recommends: dxvk
 Depends: ${shlibs:Depends},
          ${misc:Depends},
-Description: Vulkan-based translation layer for Direct3D 9/10/11 - wine32 build
- DXVK is a Vulkan-based translation layer for Direct3D 9/10/11 which
+Description: Vulkan-based translation layer for Direct3D 8/9/10/11 - wine32 build
+ DXVK is a Vulkan-based translation layer for Direct3D 8/9/10/11 which
  allows running 3D applications on Linux using Wine.
  .
- DXVK aims to provide support for D3D11, feature level 11_1, and
+ DXVK aims to provide support for D3D11, feature level 12_1, and
  D3D10, feature level 10_1.
  .
  This package contains the wine32-development build.
@@ -62,11 +62,11 @@ Multi-Arch: foreign
 Recommends: dxvk
 Depends: ${shlibs:Depends},
          ${misc:Depends},
-Description: Vulkan-based translation layer for Direct3D 9/10/11 - wine64 build
- DXVK is a Vulkan-based translation layer for Direct3D 9/10/11 which
+Description: Vulkan-based translation layer for Direct3D 8/9/10/11 - wine64 build
+ DXVK is a Vulkan-based translation layer for Direct3D 8/9/10/11 which
  allows running 3D applications on Linux using Wine.
  .
- DXVK aims to provide support for D3D11, feature level 11_1, and
+ DXVK aims to provide support for D3D11, feature level 12_1, and
  D3D10, feature level 10_1.
  .
  This package contains the wine64-development build.
diff -pruN 2.3.1-1/debian/patches/dxvk-version.patch 2.4-1/debian/patches/dxvk-version.patch
--- 2.3.1-1/debian/patches/dxvk-version.patch	2024-05-08 09:37:24.000000000 +0000
+++ 2.4-1/debian/patches/dxvk-version.patch	2024-07-11 17:48:33.000000000 +0000
@@ -5,7 +5,7 @@ Forwarded: not-needed
 
 --- a/meson.build
 +++ b/meson.build
-@@ -174,7 +174,7 @@
+@@ -190,7 +190,7 @@
  )
  
  dxvk_version = vcs_tag(
diff -pruN 2.3.1-1/dxvk.conf 2.4-1/dxvk.conf
--- 2.3.1-1/dxvk.conf	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/dxvk.conf	2024-07-10 09:56:14.000000000 +0000
@@ -1,3 +1,10 @@
+# Device filter. Only exposes devices whose Vulkan device name contains
+# the given string. May be useful to force an application to run on a
+# specific GPU, but not applications launched by that application.
+
+# dxvk.deviceFilter = ""
+
+
 # Expose the HDR10 ColorSpace (DXGI_COLOR_SPACE_RGB_FULL_G2084_NONE_P2020)
 # to the application by default.
 # This shows to the game that the global Windows 'HDR Mode' is enabled.
@@ -10,6 +17,7 @@
 
 # dxgi.enableHDR = True
 
+
 # Create the VkSurface on the first call to IDXGISwapChain::Present,
 # rather than when creating the swap chain. Some games that start
 # rendering with a different graphics API may require this option,
@@ -35,7 +43,12 @@
 # bugs in games that have physics or other simulation tied to their frame
 # rate, but do not provide their own limiter.
 #
-# Supported values : Any non-negative integer
+# Supported values
+# -1: Always disables the limiter
+#  0: Default behaviour. Limits the frame rate to the selected display
+#     refresh rate when vertical synchronization is enabled if the
+#     actual display mode does not match the game's one.
+#  n: Limit to n frames per second.
 
 # dxgi.maxFrameRate = 0
 # d3d9.maxFrameRate = 0
@@ -312,6 +325,20 @@
 # d3d11.exposeDriverCommandLists = True
 
 
+# Reproducible Command Stream
+#
+# Ensure that for the same D3D commands the output VK commands
+# don't change between runs. Useful for comparative benchmarking,
+# can negatively affect performance and can break some games
+# that don't use queries correctly.
+#
+# Supported values:
+# - True/False
+
+# d3d11.reproducibleCommandStream = False
+# d3d9.reproducibleCommandStream = False
+
+
 # Sets number of pipeline compiler threads.
 # 
 # If the graphics pipeline library feature is enabled, the given
@@ -499,6 +526,7 @@
 # Supported values:
 # - True/False
 
+# d3d11.longMad = False
 # d3d9.longMad = False
 
 # Device Local Constant Buffers
@@ -514,7 +542,9 @@
 
 # Support DF formats
 #
-# Support the vendor extension DF floating point depth formats
+# Support the vendor extension DF floating point depth formats on AMD and Intel.
+# Note that this config is ignored and disabled by default on Nvidia, or when
+# spoofing a Nvidia GPU, as it does not support these formats natively.
 #
 # Supported values:
 # - True/False
@@ -540,14 +570,16 @@
 
 # d3d9.supportX4R4G4B4 = True
 
-# Support D32
+# Support D16_LOCKABLE
 #
-# Support the D32 format.
+# Support the D16_LOCKABLE format.
+# Always enabled on AMD, or when spoofing an AMD GPU
+# via customVendorId, disabled by default on Nvidia and Intel.
 #
 # Supported values:
 # - True/False
 
-# d3d9.supportD32 = True
+# d3d9.supportD16Lockable = False
 
 # Disable A8 as a Render Target
 #
@@ -644,18 +676,6 @@
 
 # d3d9.textureMemory = 100
 
-# Always enumerate all monitors on each dxgi output
-#
-# Used to avoid performance degradation in some games
-# (will be deprecated once QueryDisplayConfig optimization
-# is in Proton Wine).
-#
-# Supported values:
-# - True/False
-
-# dxgi.useMonitorFallback = False
-
-
 # Hide integrated graphics from applications
 #
 # Only has an effect when dedicated GPUs are present on the system. It is
@@ -689,3 +709,16 @@
 # - True/False
 
 # d3d9.countLosableResources = True
+
+# Use NVIDIA Shadow Buffers
+# 
+# Vendor behavior for GeForce3 and GeForce4 cards that allows
+# sampling depth textures with non-normalized Z coordinates
+# and applies hardware shadow filtering.
+#
+# Supported values:
+# - True/False
+
+# d3d8.useShadowBuffers = False
+
+
diff -pruN 2.3.1-1/include/native/meson.build 2.4-1/include/native/meson.build
--- 2.3.1-1/include/native/meson.build	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/include/native/meson.build	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,19 @@
+install_subdir(
+  'directx',
+  install_dir: get_option('includedir') / 'dxvk',
+  strip_directory: true,
+  exclude_files: '.git'
+)
+
+install_subdir(
+  'windows',
+  install_dir: get_option('includedir') / 'dxvk',
+  strip_directory: true,
+)
+
+install_headers(
+  'wsi/native_wsi.h',
+  'wsi/native_sdl2.h',
+  'wsi/native_glfw.h',
+  subdir: 'dxvk/wsi',
+)
diff -pruN 2.3.1-1/include/native/wsi/native_sdl2.h 2.4-1/include/native/wsi/native_sdl2.h
--- 2.3.1-1/include/native/wsi/native_sdl2.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/include/native/wsi/native_sdl2.h	2024-07-10 09:56:14.000000000 +0000
@@ -1,6 +1,6 @@
 #include <windows.h>
 
-#include <SDL2/SDL.h>
+#include <SDL.h>
 
 namespace dxvk::wsi {
 
@@ -22,4 +22,4 @@ namespace dxvk::wsi {
     return reinterpret_cast<HMONITOR>(static_cast<intptr_t>(displayId + 1));
   }
 
-}
\ No newline at end of file
+}
diff -pruN 2.3.1-1/meson.build 2.4-1/meson.build
--- 2.3.1-1/meson.build	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/meson.build	2024-07-10 09:56:14.000000000 +0000
@@ -1,5 +1,6 @@
-project('dxvk', ['c', 'cpp'], version : 'v2.3.1', meson_version : '>= 0.58', default_options : [ 'cpp_std=c++17', 'warning_level=2' ])
+project('dxvk', ['c', 'cpp'], version : 'v2.4.0', meson_version : '>= 0.58', default_options : [ 'cpp_std=c++17', 'warning_level=2' ])
 
+pkg = import('pkgconfig')
 cpu_family = target_machine.cpu_family()
 platform   = target_machine.system()
 fs = import('fs')
@@ -54,6 +55,8 @@ dep_displayinfo = dependency(
 )
 
 if platform == 'windows'
+  dxvk_so_version = {'name_prefix': ''}
+
   compiler_args += [
     '-DNOMINMAX',
     '-D_WIN32_WINNT=0xa00',
@@ -115,10 +118,21 @@ if platform == 'windows'
     )
   endif
 
-  dxvk_wsi = 'win32'
   dxvk_name_prefix = ''
   compiler_args += ['-DDXVK_WSI_WIN32']
 else
+  dxvk_abi_version = '0'
+  dxvk_version_raw = meson.project_version().strip('v').split('.')
+  dxvk_version = [ dxvk_version_raw[0] ]
+  foreach i : [ 1, 2 ]
+    padded = dxvk_version_raw[i]
+    if padded.to_int() < 10
+      padded = '0' + padded
+    endif
+    dxvk_version += [ padded ]
+  endforeach
+  dxvk_so_version = {'version': dxvk_abi_version + '.' + dxvk_version[0] + dxvk_version[1] + dxvk_version[2]}
+
   wrc           = find_program('touch')
   wrc_generator = generator(wrc, output : [ '@BASENAME@_ignored.h' ], arguments : [ '@OUTPUT@' ] )
 
@@ -128,17 +142,20 @@ else
     './include/native/directx'
   ]
 
-  dxvk_wsi = get_option('dxvk_native_wsi')
-  
-  if dxvk_wsi == 'sdl2'
-    lib_sdl2 = cpp.find_library('SDL2')
+  lib_sdl2 = dependency('SDL2', required: false)
+  lib_glfw = dependency('glfw', required: false)
+  if lib_sdl2.found()
     compiler_args += ['-DDXVK_WSI_SDL2']
-  elif dxvk_wsi == 'glfw'
-    lib_glfw = cpp.find_library('glfw')
+  endif
+  if lib_glfw.found()
     compiler_args += ['-DDXVK_WSI_GLFW']
   endif
+  if (not lib_sdl2.found() and not lib_glfw.found())
+    error('SDL2 or GLFW are required to build dxvk-native')
+  endif
   
-  dxvk_name_prefix = 'libdxvk_'
+  dxvk_name_prefix = 'dxvk_'
+  dxvk_pkg_prefix = 'dxvk-'
 
   link_args += [
     '-static-libgcc',
@@ -154,13 +171,12 @@ add_project_link_arguments(cpp.get_suppo
 add_project_link_arguments(cc.get_supported_link_arguments(link_args), language: 'c')
 
 exe_ext = ''
-dll_ext = ''
 def_spec_ext = '.def'
 
 glsl_compiler = find_program('glslang', 'glslangValidator')
 glsl_args = [
   '--quiet',
-  '--target-env', 'vulkan1.2',
+  '--target-env', 'vulkan1.3',
   '--vn', '@BASENAME@',
   '--depfile', '@DEPFILE@',
   '@INPUT@',
@@ -179,4 +195,8 @@ dxvk_version = vcs_tag(
   output: 'version.h',
 )
 
+if platform != 'windows'
+  subdir('include/native')
+endif
+
 subdir('src')
diff -pruN 2.3.1-1/meson_options.txt 2.4-1/meson_options.txt
--- 2.3.1-1/meson_options.txt	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/meson_options.txt	2024-07-10 09:56:14.000000000 +0000
@@ -1,4 +1,5 @@
 option('enable_dxgi',  type : 'boolean', value : true, description: 'Build DXGI')
+option('enable_d3d8',  type : 'boolean', value : true, description: 'Build D3D8')
 option('enable_d3d9',  type : 'boolean', value : true, description: 'Build D3D9')
 option('enable_d3d10', type : 'boolean', value : true, description: 'Build D3D10')
 option('enable_d3d11', type : 'boolean', value : true, description: 'Build D3D11')
diff -pruN 2.3.1-1/package-native.sh 2.4-1/package-native.sh
--- 2.3.1-1/package-native.sh	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/package-native.sh	2024-07-10 09:56:14.000000000 +0000
@@ -25,6 +25,8 @@ shift 2
 opt_nopackage=0
 opt_devbuild=0
 opt_buildid=false
+opt_64_only=0
+opt_32_only=0
 
 CC=${CC:="gcc"}
 CXX=${CXX:="g++"}
@@ -41,6 +43,12 @@ while [ $# -gt 0 ]; do
   "--build-id")
     opt_buildid=true
     ;;
+  "--64-only")
+    opt_64_only=1
+    ;;
+  "--32-only")
+    opt_32_only=1
+    ;;
   *)
     echo "Unrecognized option: $1" >&2
     exit 1
@@ -81,8 +89,12 @@ function package {
   rm -R "dxvk-native-$DXVK_VERSION"
 }
 
-build_arch 64 lib
-build_arch 32 lib32
+if [ $opt_32_only -eq 0 ]; then
+  build_arch 64 lib
+fi
+if [ $opt_64_only -eq 0 ]; then
+  build_arch 32 lib32
+fi
 
 if [ $opt_nopackage -eq 0 ]; then
   package
diff -pruN 2.3.1-1/src/d3d10/meson.build 2.4-1/src/d3d10/meson.build
--- 2.3.1-1/src/d3d10/meson.build	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d10/meson.build	2024-07-10 09:56:14.000000000 +0000
@@ -15,16 +15,23 @@ else
   d3d10_d3d11_dep = d3d11_dep
 endif
 
-d3d10_core_dll = shared_library('d3d10core'+dll_ext, d3d10_core_src, d3d10_core_res,
-  name_prefix         : dxvk_name_prefix,
+d3d10_core_dll = shared_library(dxvk_name_prefix+'d3d10core', d3d10_core_src, d3d10_core_res,
   dependencies        : [ d3d10_d3d11_dep ],
   include_directories : dxvk_include_path,
   install             : true,
   vs_module_defs      : 'd3d10core'+def_spec_ext,
   link_args           : d3d10_core_ld_args,
   link_depends        : [ d3d10_core_link_depends ],
+  kwargs              : dxvk_so_version,
 )
 
 d3d10_core_dep = declare_dependency(
   link_with           : [ d3d10_core_dll ],
 )
+
+if platform != 'windows'
+  pkg.generate(d3d10_core_dll,
+    filebase: dxvk_pkg_prefix + 'd3d10core',
+    subdirs:  'dxvk',
+  )
+endif
diff -pruN 2.3.1-1/src/d3d11/d3d11_context_imm.cpp 2.4-1/src/d3d11/d3d11_context_imm.cpp
--- 2.3.1-1/src/d3d11/d3d11_context_imm.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d11/d3d11_context_imm.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -19,6 +19,7 @@ namespace dxvk {
     m_csThread(Device, Device->createContext(DxvkContextType::Primary)),
     m_maxImplicitDiscardSize(pParent->GetOptions()->maxImplicitDiscardSize),
     m_submissionFence(new sync::CallbackFence()),
+    m_flushTracker(pParent->GetOptions()->reproducibleCommandStream),
     m_multithread(this, false, pParent->GetOptions()->enableContextLock),
     m_videoContext(this, Device) {
     EmitCs([
diff -pruN 2.3.1-1/src/d3d11/d3d11_options.cpp 2.4-1/src/d3d11/d3d11_options.cpp
--- 2.3.1-1/src/d3d11/d3d11_options.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d11/d3d11_options.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -30,8 +30,9 @@ namespace dxvk {
     this->deferSurfaceCreation  = config.getOption<bool>("dxgi.deferSurfaceCreation", false);
     this->numBackBuffers        = config.getOption<int32_t>("dxgi.numBackBuffers", 0);
     this->maxFrameLatency       = config.getOption<int32_t>("dxgi.maxFrameLatency", 0);
-    this->maxFrameRate          = config.getOption<int32_t>("dxgi.maxFrameRate", 0);
     this->exposeDriverCommandLists = config.getOption<bool>("d3d11.exposeDriverCommandLists", true);
+    this->longMad               = config.getOption<bool>("d3d11.longMad", false);
+    this->reproducibleCommandStream = config.getOption<bool>("d3d11.reproducibleCommandStream", false);
 
     // Clamp LOD bias so that people don't abuse this in unintended ways
     this->samplerLodBias = dxvk::fclamp(this->samplerLodBias, -2.0f, 1.0f);
diff -pruN 2.3.1-1/src/d3d11/d3d11_options.h 2.4-1/src/d3d11/d3d11_options.h
--- 2.3.1-1/src/d3d11/d3d11_options.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d11/d3d11_options.h	2024-07-10 09:56:14.000000000 +0000
@@ -80,9 +80,6 @@ namespace dxvk {
     /// a higher value. May help with frame timing issues.
     int32_t maxFrameLatency;
 
-    /// Limit frame rate
-    int32_t maxFrameRate;
-
     /// Limit discardable resource size
     VkDeviceSize maxImplicitDiscardSize;
 
@@ -120,6 +117,14 @@ namespace dxvk {
 
     /// Shader dump path
     std::string shaderDumpPath;
+
+    /// Should we make our Mads a FFma or do it the long way with an FMul and an FAdd?
+    bool longMad;
+
+    /// Ensure that for the same D3D commands the output VK commands
+    /// don't change between runs. Useful for comparative benchmarking,
+    /// can negatively affect performance.
+    bool reproducibleCommandStream;
   };
   
 }
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d11/d3d11_swapchain.cpp 2.4-1/src/d3d11/d3d11_swapchain.cpp
--- 2.3.1-1/src/d3d11/d3d11_swapchain.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d11/d3d11_swapchain.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -99,7 +99,8 @@ namespace dxvk {
 
     if (riid == __uuidof(IUnknown)
      || riid == __uuidof(IDXGIVkSwapChain)
-     || riid == __uuidof(IDXGIVkSwapChain1)) {
+     || riid == __uuidof(IDXGIVkSwapChain1)
+     || riid == __uuidof(IDXGIVkSwapChain2)) {
       *ppvObject = ref(this);
       return S_OK;
     }
@@ -347,6 +348,15 @@ namespace dxvk {
   }
 
 
+  void STDMETHODCALLTYPE D3D11SwapChain::SetTargetFrameRate(
+          double                    FrameRate) {
+    m_targetFrameRate = FrameRate;
+
+    if (m_presenter != nullptr)
+      m_presenter->setFrameRateLimit(m_targetFrameRate);
+  }
+
+
   HRESULT D3D11SwapChain::PresentImage(UINT SyncInterval) {
     // Flush pending rendering commands before
     auto immediateContext = m_parent->GetContext();
@@ -496,7 +506,7 @@ namespace dxvk {
     presenterDesc.fullScreenExclusive = PickFullscreenMode();
 
     m_presenter = new Presenter(m_device, m_frameLatencySignal, presenterDesc);
-    m_presenter->setFrameRateLimit(m_parent->GetOptions()->maxFrameRate);
+    m_presenter->setFrameRateLimit(m_targetFrameRate);
   }
 
 
diff -pruN 2.3.1-1/src/d3d11/d3d11_swapchain.h 2.4-1/src/d3d11/d3d11_swapchain.h
--- 2.3.1-1/src/d3d11/d3d11_swapchain.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d11/d3d11_swapchain.h	2024-07-10 09:56:14.000000000 +0000
@@ -13,7 +13,7 @@ namespace dxvk {
   class D3D11Device;
   class D3D11DXGIDevice;
 
-  class D3D11SwapChain : public ComObject<IDXGIVkSwapChain1> {
+  class D3D11SwapChain : public ComObject<IDXGIVkSwapChain2> {
     constexpr static uint32_t DefaultFrameLatency = 1;
   public:
 
@@ -86,6 +86,9 @@ namespace dxvk {
     void STDMETHODCALLTYPE GetFrameStatistics(
             DXGI_VK_FRAME_STATISTICS* pFrameStatistics);
 
+    void STDMETHODCALLTYPE SetTargetFrameRate(
+            double                    FrameRate);
+
   private:
 
     enum BindingIds : uint32_t {
@@ -116,18 +119,20 @@ namespace dxvk {
 
     std::vector<Rc<DxvkImageView>> m_imageViews;
 
-    uint64_t                m_frameId      = DXGI_MAX_SWAP_CHAIN_BUFFERS;
-    uint32_t                m_frameLatency = DefaultFrameLatency;
-    uint32_t                m_frameLatencyCap = 0;
-    HANDLE                  m_frameLatencyEvent = nullptr;
-    Rc<sync::CallbackFence> m_frameLatencySignal;
+    uint64_t                  m_frameId      = DXGI_MAX_SWAP_CHAIN_BUFFERS;
+    uint32_t                  m_frameLatency = DefaultFrameLatency;
+    uint32_t                  m_frameLatencyCap = 0;
+    HANDLE                    m_frameLatencyEvent = nullptr;
+    Rc<sync::CallbackFence>   m_frameLatencySignal;
 
-    bool                    m_dirty = true;
+    bool                      m_dirty = true;
 
-    VkColorSpaceKHR         m_colorspace = VK_COLOR_SPACE_SRGB_NONLINEAR_KHR;
+    VkColorSpaceKHR           m_colorspace = VK_COLOR_SPACE_SRGB_NONLINEAR_KHR;
 
     std::optional<VkHdrMetadataEXT> m_hdrMetadata;
-    bool m_dirtyHdrMetadata = true;
+    bool                      m_dirtyHdrMetadata = true;
+
+    double                    m_targetFrameRate = 0.0;
 
     dxvk::mutex               m_frameStatisticsLock;
     DXGI_VK_FRAME_STATISTICS  m_frameStatistics = { };
diff -pruN 2.3.1-1/src/d3d11/d3d11_texture.cpp 2.4-1/src/d3d11/d3d11_texture.cpp
--- 2.3.1-1/src/d3d11/d3d11_texture.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d11/d3d11_texture.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -58,9 +58,6 @@ namespace dxvk {
           "\n  MiscFlags:  ", m_desc.MiscFlags,
           "\n  FeatureLevel:  ", pDevice->GetFeatureLevel()));
 
-      if (m_desc.MiscFlags & D3D11_RESOURCE_MISC_SHARED_KEYEDMUTEX)
-        Logger::warn("D3D11_RESOURCE_MISC_SHARED_KEYEDMUTEX: not supported.");
-
       imageInfo.shared = true;
       imageInfo.sharing.mode = hSharedHandle == INVALID_HANDLE_VALUE ? DxvkSharedHandleMode::Export : DxvkSharedHandleMode::Import;
       imageInfo.sharing.type = (m_desc.MiscFlags & D3D11_RESOURCE_MISC_SHARED_NTHANDLE)
diff -pruN 2.3.1-1/src/d3d11/d3d11_video.cpp 2.4-1/src/d3d11/d3d11_video.cpp
--- 2.3.1-1/src/d3d11/d3d11_video.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d11/d3d11_video.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -1262,12 +1262,28 @@ namespace dxvk {
         viewport.height = float(cStreamState.dstRect.bottom) - viewport.y;
       }
 
+      VkExtent3D viewExtent = cViews[0]->mipLevelExtent(0);
+
+      VkRect2D srcRect;
+      srcRect.offset = { 0, 0 };
+      srcRect.extent = { viewExtent.width, viewExtent.height };
+
+      if (cStreamState.srcRectEnabled) {
+        srcRect.offset.x      = cStreamState.srcRect.left;
+        srcRect.offset.y      = cStreamState.srcRect.top;
+        srcRect.extent.width  = cStreamState.srcRect.right - srcRect.offset.x;
+        srcRect.extent.height = cStreamState.srcRect.bottom - srcRect.offset.y;
+      }
+
       UboData uboData = { };
       uboData.colorMatrix[0][0] = 1.0f;
       uboData.colorMatrix[1][1] = 1.0f;
       uboData.colorMatrix[2][2] = 1.0f;
-      uboData.coordMatrix[0][0] = 1.0f;
-      uboData.coordMatrix[1][1] = 1.0f;
+      uboData.coordMatrix[0][0] = float(srcRect.extent.width) / float(viewExtent.width);
+      uboData.coordMatrix[1][1] = float(srcRect.extent.height) / float(viewExtent.height);
+      uboData.coordMatrix[2][0] = float(srcRect.offset.x) / float(viewExtent.width);
+      uboData.coordMatrix[2][1] = float(srcRect.offset.y) / float(viewExtent.height);
+      uboData.srcRect = srcRect;
       uboData.yMin = 0.0f;
       uboData.yMax = 1.0f;
       uboData.isPlanar = cViews[1] != nullptr;
@@ -1290,17 +1306,14 @@ namespace dxvk {
       ctx->bindShader<VK_SHADER_STAGE_FRAGMENT_BIT>(Rc<DxvkShader>(m_fs));
 
       ctx->bindUniformBuffer(VK_SHADER_STAGE_FRAGMENT_BIT, 0, DxvkBufferSlice(m_ubo));
-      ctx->bindResourceSampler(VK_SHADER_STAGE_FRAGMENT_BIT, 1, Rc<DxvkSampler>(m_sampler));
 
       for (uint32_t i = 0; i < cViews.size(); i++)
-        ctx->bindResourceImageView(VK_SHADER_STAGE_FRAGMENT_BIT, 2 + i, Rc<DxvkImageView>(cViews[i]));
+        ctx->bindResourceImageView(VK_SHADER_STAGE_FRAGMENT_BIT, 1 + i, Rc<DxvkImageView>(cViews[i]));
 
       ctx->draw(3, 1, 0, 0);
 
-      ctx->bindResourceSampler(VK_SHADER_STAGE_FRAGMENT_BIT, 1, nullptr);
-
       for (uint32_t i = 0; i < cViews.size(); i++)
-        ctx->bindResourceImageView(VK_SHADER_STAGE_FRAGMENT_BIT, 2 + i, nullptr);
+        ctx->bindResourceImageView(VK_SHADER_STAGE_FRAGMENT_BIT, 1 + i, nullptr);
     });
   }
 
@@ -1315,38 +1328,14 @@ namespace dxvk {
   }
 
 
-  void D3D11VideoContext::CreateSampler() {
-    DxvkSamplerCreateInfo samplerInfo;
-    samplerInfo.magFilter       = VK_FILTER_LINEAR;
-    samplerInfo.minFilter       = VK_FILTER_LINEAR;
-    samplerInfo.mipmapMode      = VK_SAMPLER_MIPMAP_MODE_NEAREST;
-    samplerInfo.mipmapLodBias   = 0.0f;
-    samplerInfo.mipmapLodMin    = 0.0f;
-    samplerInfo.mipmapLodMax    = 0.0f;
-    samplerInfo.useAnisotropy   = VK_FALSE;
-    samplerInfo.maxAnisotropy   = 1.0f;
-    samplerInfo.addressModeU    = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
-    samplerInfo.addressModeV    = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
-    samplerInfo.addressModeW    = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
-    samplerInfo.compareToDepth  = VK_FALSE;
-    samplerInfo.compareOp       = VK_COMPARE_OP_ALWAYS;
-    samplerInfo.reductionMode   = VK_SAMPLER_REDUCTION_MODE_WEIGHTED_AVERAGE;
-    samplerInfo.borderColor     = VkClearColorValue();
-    samplerInfo.usePixelCoord   = VK_FALSE;
-    samplerInfo.nonSeamless     = VK_FALSE;
-    m_sampler = m_device->createSampler(samplerInfo);
-  }
-
-
   void D3D11VideoContext::CreateShaders() {
     SpirvCodeBuffer vsCode(d3d11_video_blit_vert);
     SpirvCodeBuffer fsCode(d3d11_video_blit_frag);
 
-    const std::array<DxvkBindingInfo, 4> fsBindings = {{
+    const std::array<DxvkBindingInfo, 3> fsBindings = {{
       { VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 0, VK_IMAGE_VIEW_TYPE_MAX_ENUM, VK_SHADER_STAGE_FRAGMENT_BIT, VK_ACCESS_UNIFORM_READ_BIT, VK_TRUE },
-      { VK_DESCRIPTOR_TYPE_SAMPLER,        1, VK_IMAGE_VIEW_TYPE_MAX_ENUM, VK_SHADER_STAGE_FRAGMENT_BIT, 0 },
+      { VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE,  1, VK_IMAGE_VIEW_TYPE_2D,       VK_SHADER_STAGE_FRAGMENT_BIT, VK_ACCESS_SHADER_READ_BIT },
       { VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE,  2, VK_IMAGE_VIEW_TYPE_2D,       VK_SHADER_STAGE_FRAGMENT_BIT, VK_ACCESS_SHADER_READ_BIT },
-      { VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE,  3, VK_IMAGE_VIEW_TYPE_2D,       VK_SHADER_STAGE_FRAGMENT_BIT, VK_ACCESS_SHADER_READ_BIT },
     }};
 
     DxvkShaderCreateInfo vsInfo;
@@ -1368,7 +1357,6 @@ namespace dxvk {
     if (std::exchange(m_resourcesCreated, true))
       return;
 
-    CreateSampler();
     CreateUniformBuffer();
     CreateShaders();
   }
diff -pruN 2.3.1-1/src/d3d11/d3d11_video.h 2.4-1/src/d3d11/d3d11_video.h
--- 2.3.1-1/src/d3d11/d3d11_video.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d11/d3d11_video.h	2024-07-10 09:56:14.000000000 +0000
@@ -584,6 +584,7 @@ namespace dxvk {
     struct alignas(16) UboData {
       float colorMatrix[3][4];
       float coordMatrix[3][2];
+      VkRect2D srcRect;
       float yMin, yMax;
       VkBool32 isPlanar;
     };
@@ -593,7 +594,6 @@ namespace dxvk {
     Rc<DxvkDevice>          m_device;
     Rc<DxvkShader>          m_vs;
     Rc<DxvkShader>          m_fs;
-    Rc<DxvkSampler>         m_sampler;
     Rc<DxvkBuffer>          m_ubo;
 
     VkExtent2D m_dstExtent = { 0u, 0u };
@@ -613,8 +613,6 @@ namespace dxvk {
 
     void CreateUniformBuffer();
 
-    void CreateSampler();
-
     void CreateShaders();
 
     void CreateResources();
diff -pruN 2.3.1-1/src/d3d11/meson.build 2.4-1/src/d3d11/meson.build
--- 2.3.1-1/src/d3d11/meson.build	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d11/meson.build	2024-07-10 09:56:14.000000000 +0000
@@ -77,18 +77,25 @@ else
   d3d11_dxgi_dep = dxgi_dep
 endif
 
-d3d11_dll = shared_library('d3d11'+dll_ext, dxgi_common_src + d3d11_src + d3d10_src,
+d3d11_dll = shared_library(dxvk_name_prefix+'d3d11', dxgi_common_src + d3d11_src + d3d10_src,
     glsl_generator.process(d3d11_shaders), d3d11_res,
-  name_prefix         : dxvk_name_prefix,
   dependencies        : [ d3d11_dxgi_dep, dxbc_dep, dxvk_dep ],
   include_directories : dxvk_include_path,
   install             : true,
   vs_module_defs      : 'd3d11'+def_spec_ext,
   link_args           : d3d11_ld_args,
   link_depends        : [ d3d11_link_depends ],
+  kwargs              : dxvk_so_version,
 )
 
 d3d11_dep = declare_dependency(
   link_with           : [ d3d11_dll ],
   include_directories : [ dxvk_include_path ],
 )
+
+if platform != 'windows'
+  pkg.generate(d3d11_dll,
+    filebase: dxvk_pkg_prefix + 'd3d11',
+    subdirs:  'dxvk',
+  )
+endif
diff -pruN 2.3.1-1/src/d3d11/shaders/d3d11_video_blit_frag.frag 2.4-1/src/d3d11/shaders/d3d11_video_blit_frag.frag
--- 2.3.1-1/src/d3d11/shaders/d3d11_video_blit_frag.frag	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d11/shaders/d3d11_video_blit_frag.frag	2024-07-10 09:56:14.000000000 +0000
@@ -1,5 +1,7 @@
 #version 450
 
+#extension GL_EXT_samplerless_texture_functions : require
+
 // Can't use matrix types here since even a two-row
 // matrix will be padded to 16 bytes per column for
 // absolutely no reason
@@ -11,6 +13,8 @@ uniform ubo_t {
   vec2 coord_matrix_c1;
   vec2 coord_matrix_c2;
   vec2 coord_matrix_c3;
+  uvec2 src_offset;
+  uvec2 src_extent;
   float y_min;
   float y_max;
   bool is_planar;
@@ -19,9 +23,8 @@ uniform ubo_t {
 layout(location = 0) in vec2 i_texcoord;
 layout(location = 0) out vec4 o_color;
 
-layout(set = 0, binding = 1) uniform sampler s_sampler;
-layout(set = 0, binding = 2) uniform texture2D s_inputY;
-layout(set = 0, binding = 3) uniform texture2D s_inputCbCr;
+layout(set = 0, binding = 1) uniform texture2D s_inputY;
+layout(set = 0, binding = 2) uniform texture2D s_inputCbCr;
 
 void main() {
   // Transform input texture coordinates to
@@ -31,25 +34,61 @@ void main() {
     coord_matrix_c2,
     coord_matrix_c3);
 
-  vec2 coord = coord_matrix * vec3(i_texcoord, 1.0f);
-
-  // Fetch source image color
-  vec4 color = vec4(0.0f, 0.0f, 0.0f, 1.0f);
-
-  if (is_planar) {
-    color.g  = texture(sampler2D(s_inputY,    s_sampler), coord).r;
-    color.rb = texture(sampler2D(s_inputCbCr, s_sampler), coord).gr;
-    color.g  = clamp((color.g - y_min) / (y_max - y_min), 0.0f, 1.0f);
-  } else {
-    color = texture(sampler2D(s_inputY, s_sampler), coord);
-  }
-
-  // Color space transformation
+  // Load color space transform
   mat3x4 color_matrix = mat3x4(
     color_matrix_r1,
     color_matrix_r2,
     color_matrix_r3);
 
-  o_color.rgb = vec4(color.rgb, 1.0f) * color_matrix;
-  o_color.a = color.a;
+  // Compute actual pixel coordinates to sample. We filter
+  // manually in order to avoid bleeding from pixels outside
+  // the source rectangle.
+  vec2 abs_size_y = vec2(textureSize(s_inputY, 0));
+  vec2 abs_size_c = vec2(textureSize(s_inputCbCr, 0));
+
+  vec2 coord = coord_matrix * vec3(i_texcoord, 1.0f);
+  coord -= 0.5f / abs_size_y;
+
+  vec2 size_factor = abs_size_c / abs_size_y;
+
+  vec2 src_lo = vec2(src_offset);
+  vec2 src_hi = vec2(src_offset + src_extent - 1u);
+
+  vec2 abs_coord = coord * abs_size_y;
+  vec2 fract_coord = fract(clamp(abs_coord, src_lo, src_hi));
+
+  vec4 accum = vec4(0.0f, 0.0f, 0.0f, 0.0f);
+
+  for (int i = 0; i < 4; i++) {
+    ivec2 offset = ivec2(i & 1, i >> 1);
+
+    // Compute exact pixel coordinates for the current
+    // iteration and clamp it to the source rectangle.
+    vec2 fetch_coord = clamp(abs_coord + vec2(offset), src_lo, src_hi);
+
+    // Fetch actual pixel color in source color space
+    vec4 color;
+
+    if (is_planar) {
+      color.g  = texelFetch(s_inputY, ivec2(fetch_coord), 0).r;
+      color.rb = texelFetch(s_inputCbCr, ivec2(fetch_coord * size_factor), 0).gr;
+      color.g  = clamp((color.g - y_min) / (y_max - y_min), 0.0f, 1.0f);
+      color.a = 1.0f;
+    } else {
+      color = texelFetch(s_inputY, ivec2(fetch_coord), 0);
+    }
+
+    // Transform color space before accumulation
+    color.rgb = vec4(color.rgb, 1.0f) * color_matrix;
+
+    // Filter and accumulate final pixel color
+    vec2 factor = fract_coord;
+
+    if (offset.x == 0) factor.x = 1.0f - factor.x;
+    if (offset.y == 0) factor.y = 1.0f - factor.y;
+
+    accum += factor.x * factor.y * color;
+  }
+
+  o_color = accum;
 }
diff -pruN 2.3.1-1/src/d3d8/d3d8.def 2.4-1/src/d3d8/d3d8.def
--- 2.3.1-1/src/d3d8/d3d8.def	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8.def	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,3 @@
+LIBRARY D3D8.DLL
+EXPORTS
+  Direct3DCreate8 @ 5
diff -pruN 2.3.1-1/src/d3d8/d3d8.sym 2.4-1/src/d3d8/d3d8.sym
--- 2.3.1-1/src/d3d8/d3d8.sym	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8.sym	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,7 @@
+{
+  global:
+    Direct3DCreate8;
+
+  local:
+    *;
+};
diff -pruN 2.3.1-1/src/d3d8/d3d8_batch.h 2.4-1/src/d3d8/d3d8_batch.h
--- 2.3.1-1/src/d3d8/d3d8_batch.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_batch.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,239 @@
+#pragma once
+
+#include "d3d8_include.h"
+#include "d3d8_buffer.h"
+#include "d3d8_format.h"
+
+#include <vector>
+#include <cstdint>
+#include <climits>
+
+namespace dxvk {
+
+  inline constexpr size_t            D3DPT_COUNT   = size_t(D3DPT_TRIANGLEFAN) + 1;
+  inline constexpr D3DPRIMITIVETYPE  D3DPT_INVALID = D3DPRIMITIVETYPE(0);
+
+  // Vertex buffer that can handle many tiny locks while
+  // still maintaing the lock ordering of direct-mapped buffers.
+  class D3D8BatchBuffer final : public D3D8VertexBuffer {
+  public:
+    D3D8BatchBuffer(
+        D3D8Device*                         pDevice,
+        D3DPOOL                             Pool,
+        DWORD                               Usage,
+        UINT                                Length,
+        DWORD                               FVF)
+      : D3D8VertexBuffer(pDevice, nullptr, Pool, Usage)
+      , m_data(Length)
+      , m_fvf(FVF) {
+    }
+
+    HRESULT STDMETHODCALLTYPE Lock(
+            UINT   OffsetToLock,
+            UINT   SizeToLock,
+            BYTE** ppbData,
+            DWORD  Flags) {
+      *ppbData = m_data.data() + OffsetToLock;
+      return D3D_OK;
+    }
+
+    HRESULT STDMETHODCALLTYPE Unlock() {
+      return D3D_OK;
+    }
+
+    HRESULT STDMETHODCALLTYPE GetDesc(D3DVERTEXBUFFER_DESC* pDesc) {
+      pDesc->Format = D3DFMT_VERTEXDATA;
+      pDesc->Type   = D3DRTYPE_VERTEXBUFFER;
+      pDesc->Usage  = m_usage;
+      pDesc->Pool   = m_pool;
+      pDesc->Size   = m_data.size();
+      pDesc->FVF    = m_fvf;
+      return D3D_OK;
+    }
+
+    void STDMETHODCALLTYPE PreLoad() {
+    }
+
+    const void* GetPtr(UINT byteOffset = 0) const {
+      return m_data.data() + byteOffset;
+    }
+
+    UINT Size() const {
+      return m_data.size();
+    }
+
+  private:
+    std::vector<BYTE> m_data;
+    DWORD             m_fvf;
+  };
+
+  
+  // Main handler for batching D3D8 draw calls.
+  class D3D8Batcher {
+
+    struct Batch {
+      D3DPRIMITIVETYPE PrimitiveType = D3DPT_INVALID;
+      std::vector<uint16_t> Indices;
+      UINT Offset = 0;
+      UINT MinVertex = UINT_MAX;
+      UINT MaxVertex = 0;
+      UINT PrimitiveCount = 0;
+      UINT DrawCallCount = 0;
+    };
+
+  public:
+    D3D8Batcher(D3D8Device* pDevice8, Com<d3d9::IDirect3DDevice9>&& pDevice9)
+      : m_device8(pDevice8)
+      , m_device(std::move(pDevice9)) {
+    }
+
+    inline D3D8BatchBuffer* CreateVertexBuffer(UINT Length, DWORD Usage, DWORD FVF, D3DPOOL Pool) {
+      return ref(new D3D8BatchBuffer(m_device8, Pool, Usage, Length, FVF));
+    }
+
+    inline void StateChange() {
+      if (likely(m_batches.empty()))
+        return;
+      for (auto& draw : m_batches) {
+
+        if (draw.PrimitiveType == D3DPT_INVALID)
+          continue;
+
+        for (auto& index : draw.Indices)
+          index -= draw.MinVertex;
+
+        m_device->DrawIndexedPrimitiveUP(
+          d3d9::D3DPRIMITIVETYPE(draw.PrimitiveType),
+          0,
+          draw.MaxVertex - draw.MinVertex,
+          draw.PrimitiveCount,
+          draw.Indices.data(),
+          d3d9::D3DFMT_INDEX16,
+          m_stream->GetPtr(draw.MinVertex * m_stride),
+          m_stride);
+        
+        m_device->SetStreamSource(0, D3D8VertexBuffer::GetD3D9Nullable(m_stream), 0, m_stride);
+        m_device->SetIndices(D3D8IndexBuffer::GetD3D9Nullable(m_indices));
+        
+        draw.PrimitiveType = D3DPRIMITIVETYPE(0);
+        draw.Offset = 0;
+        draw.MinVertex = UINT_MAX;
+        draw.MaxVertex = 0;
+        draw.PrimitiveCount = 0;
+        draw.DrawCallCount = 0;
+      }
+    }
+
+    inline void EndFrame() {
+      // Nothing to be done.
+    }
+
+    inline HRESULT DrawPrimitive(
+            D3DPRIMITIVETYPE PrimitiveType,
+            UINT             StartVertex,
+            UINT             PrimitiveCount) {
+
+      // None of this linestrip or fan malarkey
+      D3DPRIMITIVETYPE batchedPrimType = PrimitiveType;
+      switch (PrimitiveType) {
+        case D3DPT_LINESTRIP:     batchedPrimType = D3DPT_LINELIST; break;
+        case D3DPT_TRIANGLEFAN:   batchedPrimType = D3DPT_TRIANGLELIST; break;
+        default: break;
+      }
+
+      Batch* batch = &m_batches[size_t(batchedPrimType)];
+      batch->PrimitiveType = batchedPrimType;
+
+      //UINT vertices = GetVertexCount8(PrimitiveType, PrimitiveCount);
+
+      switch (PrimitiveType) {
+        case D3DPT_POINTLIST:
+          batch->Indices.resize(batch->Offset + PrimitiveCount);
+          for (UINT i = 0; i < PrimitiveCount; i++)
+            batch->Indices[batch->Offset++] = (StartVertex + i);
+          break;
+        case D3DPT_LINELIST:
+          batch->Indices.resize(batch->Offset + PrimitiveCount * 2);
+          for (UINT i = 0; i < PrimitiveCount; i++) {
+            batch->Indices[batch->Offset++] = (StartVertex + i * 2 + 0);
+            batch->Indices[batch->Offset++] = (StartVertex + i * 2 + 1);
+          }
+          break;
+        case D3DPT_LINESTRIP:
+          batch->Indices.resize(batch->Offset + PrimitiveCount * 2);
+          for (UINT i = 0; i < PrimitiveCount; i++) {
+            batch->Indices[batch->Offset++] = (StartVertex + i + 0);
+            batch->Indices[batch->Offset++] = (StartVertex + i + 1);
+          }
+          break;
+        case D3DPT_TRIANGLELIST:
+          batch->Indices.resize(batch->Offset + PrimitiveCount * 3);
+          for (UINT i = 0; i < PrimitiveCount; i++) {
+            batch->Indices[batch->Offset++] = (StartVertex + i * 3 + 0);
+            batch->Indices[batch->Offset++] = (StartVertex + i * 3 + 1);
+            batch->Indices[batch->Offset++] = (StartVertex + i * 3 + 2);
+          }
+          break;
+        case D3DPT_TRIANGLESTRIP:
+          // Join with degenerate triangle
+          // 1 2 3, 3 4, 4 5 6
+          batch->Indices.resize(batch->Offset + PrimitiveCount + 2);
+          if (batch->Offset > 0) {
+            batch->Indices[batch->Offset + 1] = batch->Indices[batch->Offset-2];
+            batch->Indices[batch->Offset += 2] = StartVertex;
+          }
+          for (UINT i = 0; i < PrimitiveCount; i++) {
+            batch->Indices[batch->Offset++] = (StartVertex + i + 0);
+          }
+          break;
+        // 1 2 3 4 5 6 7 -> 1 2 3, 1 3 4, 1 4 5, 1 5 6, 1 6 7
+        case D3DPT_TRIANGLEFAN:
+          batch->Indices.resize(batch->Offset + PrimitiveCount * 3);
+          for (UINT i = 0; i < PrimitiveCount; i++) {
+            batch->Indices[batch->Offset++] = (StartVertex + 0);
+            batch->Indices[batch->Offset++] = (StartVertex + i + 1);
+            batch->Indices[batch->Offset++] = (StartVertex + i + 2);
+          }
+          break;
+        default:
+          return D3DERR_INVALIDCALL;
+      }
+      batch->MinVertex = std::min(batch->MinVertex, StartVertex);
+      if (!batch->Indices.empty())
+        batch->MaxVertex = std::max(batch->MaxVertex, UINT(batch->Indices.back() + 1));
+      batch->PrimitiveCount += PrimitiveCount;
+      batch->DrawCallCount++;
+      return D3D_OK;
+    }
+
+    inline void SetStream(UINT num, D3D8VertexBuffer* stream, UINT stride) {
+      if (unlikely(num != 0)) {
+        StateChange();
+        return;
+      }
+      if (unlikely(m_stream != stream || m_stride != stride)) {
+        StateChange();
+        m_stream = static_cast<D3D8BatchBuffer*>(stream);
+        m_stride = stride;
+      }
+    }
+
+    inline void SetIndices(D3D8IndexBuffer* indices, INT baseVertexIndex) {
+      if (m_indices != indices || m_baseVertexIndex != baseVertexIndex) {
+        StateChange();
+        m_indices = indices;
+        m_baseVertexIndex = baseVertexIndex;
+      }
+    }
+
+  private:
+    D3D8Device*                     m_device8;
+    Com<d3d9::IDirect3DDevice9>     m_device;
+
+    D3D8BatchBuffer*                m_stream = nullptr;
+    UINT                            m_stride = 0;
+    D3D8IndexBuffer*                m_indices = nullptr;
+    INT                             m_baseVertexIndex = 0;
+    std::array<Batch, D3DPT_COUNT>  m_batches;
+  };
+}
diff -pruN 2.3.1-1/src/d3d8/d3d8_buffer.h 2.4-1/src/d3d8/d3d8_buffer.h
--- 2.3.1-1/src/d3d8/d3d8_buffer.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_buffer.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,102 @@
+#pragma once
+
+#include "d3d8_include.h"
+#include "d3d8_resource.h"
+
+namespace dxvk {
+
+  template <typename D3D9, typename D3D8>
+  class D3D8Buffer : public D3D8Resource<D3D9, D3D8> {
+
+  public:
+
+    D3D8Buffer(
+            D3D8Device*     pDevice,
+            Com<D3D9>&&     pBuffer,
+            D3DPOOL         Pool,
+            DWORD           Usage)
+      : D3D8Resource<D3D9, D3D8> (pDevice, std::move(pBuffer))
+      , m_pool                   (Pool)
+      , m_usage                  (Usage) {
+    }
+
+    HRESULT STDMETHODCALLTYPE Lock(
+            UINT   OffsetToLock,
+            UINT   SizeToLock,
+            BYTE** ppbData,
+            DWORD  Flags) {
+      return this->GetD3D9()->Lock(
+        OffsetToLock,
+        SizeToLock,
+        reinterpret_cast<void**>(ppbData),
+        Flags);
+    }
+
+    HRESULT STDMETHODCALLTYPE Unlock() {
+      return this->GetD3D9()->Unlock();
+    }
+
+    void STDMETHODCALLTYPE PreLoad() {
+      this->GetD3D9()->PreLoad();
+    }
+
+  protected:
+    // This is the D3D8 pool, not necessarily what's given to D3D9.
+    const D3DPOOL m_pool;
+    // This is the D3D8 usage, not necessarily what's given to D3D9.
+    const DWORD   m_usage;
+  };
+
+
+  using D3D8VertexBufferBase = D3D8Buffer<d3d9::IDirect3DVertexBuffer9, IDirect3DVertexBuffer8>;
+  class D3D8VertexBuffer : public D3D8VertexBufferBase {
+
+  public:
+
+    D3D8VertexBuffer(
+        D3D8Device*                         pDevice,
+        Com<d3d9::IDirect3DVertexBuffer9>&& pBuffer,
+        D3DPOOL                             Pool,
+        DWORD                               Usage)
+      : D3D8VertexBufferBase(pDevice, std::move(pBuffer), Pool, Usage) {
+    }
+
+    D3DRESOURCETYPE STDMETHODCALLTYPE GetType() final { return D3DRTYPE_VERTEXBUFFER; }
+
+    HRESULT STDMETHODCALLTYPE GetDesc(D3DVERTEXBUFFER_DESC* pDesc) {
+      HRESULT hr = GetD3D9()->GetDesc(reinterpret_cast<d3d9::D3DVERTEXBUFFER_DESC*>(pDesc));
+      if (!FAILED(hr)) {
+        pDesc->Pool = m_pool;
+        pDesc->Usage = m_usage;
+      }
+      return hr;
+    }
+
+  };
+
+  using D3D8IndexBufferBase = D3D8Buffer<d3d9::IDirect3DIndexBuffer9, IDirect3DIndexBuffer8>;
+  class D3D8IndexBuffer final : public D3D8IndexBufferBase {
+
+  public:
+
+    D3D8IndexBuffer(
+        D3D8Device*                        pDevice,
+        Com<d3d9::IDirect3DIndexBuffer9>&& pBuffer,
+        D3DPOOL                            Pool,
+        DWORD                              Usage)
+      : D3D8IndexBufferBase(pDevice, std::move(pBuffer), Pool, Usage) {
+    }
+
+    D3DRESOURCETYPE STDMETHODCALLTYPE GetType() final { return D3DRTYPE_INDEXBUFFER; }
+
+    HRESULT STDMETHODCALLTYPE GetDesc(D3DINDEXBUFFER_DESC* pDesc) final {
+      HRESULT hr = GetD3D9()->GetDesc(reinterpret_cast<d3d9::D3DINDEXBUFFER_DESC*>(pDesc));
+      if (!FAILED(hr)) {
+        pDesc->Pool = m_pool;
+        pDesc->Usage = m_usage;
+      }
+      return hr;
+    }
+
+  };
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/d3d8_caps.h 2.4-1/src/d3d8/d3d8_caps.h
--- 2.3.1-1/src/d3d8/d3d8_caps.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_caps.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,8 @@
+#pragma once
+
+namespace dxvk::d8caps {
+
+  inline constexpr uint32_t MAX_TEXTURE_STAGES = 8;
+  inline constexpr uint32_t MAX_STREAMS = 16;
+
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/d3d8_d3d9_util.h 2.4-1/src/d3d8/d3d8_d3d9_util.h
--- 2.3.1-1/src/d3d8/d3d8_d3d9_util.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_d3d9_util.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,164 @@
+#pragma once
+
+// Utility functions for converting
+// between DirectX8 and DirectX9 types.
+
+#include "d3d8_include.h"
+#include "d3d8_format.h"
+#include "d3d8_options.h"
+
+#include <utility>
+
+namespace dxvk {
+
+  // (8<-9) D3DCAPSX: Writes to D3DCAPS8 from D3DCAPS9
+  inline void ConvertCaps8(const d3d9::D3DCAPS9& caps9, D3DCAPS8* pCaps8) {
+
+    // should be aligned
+    std::memcpy(pCaps8, &caps9, sizeof(D3DCAPS8));
+
+    // Max supported shader model is PS 1.4 and VS 1.1
+    pCaps8->VertexShaderVersion = D3DVS_VERSION(1, 1);
+    pCaps8->PixelShaderVersion  = D3DPS_VERSION(1, 4);
+
+    // This was removed by D3D9. We can probably render windowed.
+    pCaps8->Caps2       |= D3DCAPS2_CANRENDERWINDOWED;
+
+    // Replaced by D3DPRASTERCAPS_DEPTHBIAS in D3D9
+    pCaps8->RasterCaps  |= D3DPRASTERCAPS_ZBIAS;
+
+
+    // Remove D3D9-specific caps:
+    pCaps8->Caps2             &= ~D3DCAPS2_CANAUTOGENMIPMAP;
+
+    pCaps8->Caps3             &= ~D3DCAPS3_LINEAR_TO_SRGB_PRESENTATION
+                               & ~D3DCAPS3_COPY_TO_VIDMEM
+                               & ~D3DCAPS3_COPY_TO_SYSTEMMEM;
+
+    pCaps8->PrimitiveMiscCaps &= ~D3DPMISCCAPS_INDEPENDENTWRITEMASKS
+                               & ~D3DPMISCCAPS_PERSTAGECONSTANT
+                               & ~D3DPMISCCAPS_FOGANDSPECULARALPHA
+                               & ~D3DPMISCCAPS_SEPARATEALPHABLEND
+                               & ~D3DPMISCCAPS_MRTINDEPENDENTBITDEPTHS
+                               & ~D3DPMISCCAPS_MRTPOSTPIXELSHADERBLENDING
+                               & ~D3DPMISCCAPS_FOGVERTEXCLAMPED
+                               & ~D3DPMISCCAPS_POSTBLENDSRGBCONVERT;
+
+    pCaps8->RasterCaps        &= ~D3DPRASTERCAPS_SCISSORTEST
+                               & ~D3DPRASTERCAPS_SLOPESCALEDEPTHBIAS
+                               & ~D3DPRASTERCAPS_DEPTHBIAS
+                               & ~D3DPRASTERCAPS_MULTISAMPLE_TOGGLE;
+
+    pCaps8->SrcBlendCaps      &= ~D3DPBLENDCAPS_INVSRCCOLOR2
+                               & ~D3DPBLENDCAPS_SRCCOLOR2;
+
+    pCaps8->LineCaps          &= ~D3DLINECAPS_ANTIALIAS;
+
+    pCaps8->StencilCaps       &= ~D3DSTENCILCAPS_TWOSIDED;
+  }
+
+  // (9<-8) D3DD3DPRESENT_PARAMETERS: Returns D3D9's params given an input for D3D8
+  inline d3d9::D3DPRESENT_PARAMETERS ConvertPresentParameters9(const D3DPRESENT_PARAMETERS* pParams) {
+
+    d3d9::D3DPRESENT_PARAMETERS params;
+    params.BackBufferWidth = pParams->BackBufferWidth;
+    params.BackBufferHeight = pParams->BackBufferHeight;
+    params.BackBufferFormat = d3d9::D3DFORMAT(pParams->BackBufferFormat);
+    params.BackBufferCount = pParams->BackBufferCount;
+
+    params.MultiSampleType = d3d9::D3DMULTISAMPLE_TYPE(pParams->MultiSampleType);
+    params.MultiSampleQuality = 0; // (D3D8: no MultiSampleQuality), TODO: get a value for this
+
+
+    UINT PresentationInterval = pParams->FullScreen_PresentationInterval;
+
+    if (pParams->Windowed) {
+
+      if (unlikely(PresentationInterval != D3DPRESENT_INTERVAL_DEFAULT)) {
+        // TODO: what does dx8 do if windowed app sets FullScreen_PresentationInterval?
+        Logger::warn(str::format(
+          "D3D8 Application is windowed yet requested FullScreen_PresentationInterval ", PresentationInterval,
+          " (should be D3DPRESENT_INTERVAL_DEFAULT). This will be ignored."));
+      }
+
+      // D3D8: For windowed swap chain, the back buffer is copied to the window immediately.
+      PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
+    }
+
+    D3DSWAPEFFECT SwapEffect = pParams->SwapEffect;
+
+    // D3DSWAPEFFECT_COPY_VSYNC has been removed
+    if (SwapEffect == D3DSWAPEFFECT_COPY_VSYNC) {
+
+      SwapEffect = D3DSWAPEFFECT_COPY;
+
+      // D3D8: In windowed mode, D3DSWAPEFFECT_COPY_VSYNC enables VSYNC.
+      // In fullscreen, D3DPRESENT_INTERVAL_IMMEDIATE is meaningless.
+      if (pParams->Windowed || (PresentationInterval & D3DPRESENT_INTERVAL_IMMEDIATE) != 0) {
+        PresentationInterval = D3DPRESENT_INTERVAL_ONE;
+        // TODO: what does dx8 do if multiple D3DPRESENT_INTERVAL flags are set? 
+      }
+    }
+
+    params.SwapEffect = d3d9::D3DSWAPEFFECT(SwapEffect);
+    params.hDeviceWindow = pParams->hDeviceWindow;
+    params.Windowed = pParams->Windowed;
+    params.EnableAutoDepthStencil = pParams->EnableAutoDepthStencil;
+    params.AutoDepthStencilFormat = d3d9::D3DFORMAT(pParams->AutoDepthStencilFormat);
+    params.Flags = pParams->Flags;
+
+    params.FullScreen_RefreshRateInHz = pParams->FullScreen_RefreshRateInHz;
+
+    // FullScreen_PresentationInterval -> PresentationInterval
+    params.PresentationInterval = PresentationInterval;
+
+    return params;
+  }
+
+  // (8<-9) Convert D3DSURFACE_DESC
+  inline void ConvertSurfaceDesc8(const d3d9::D3DSURFACE_DESC* pSurf9, D3DSURFACE_DESC* pSurf8) {
+    pSurf8->Format  = D3DFORMAT(pSurf9->Format);
+    pSurf8->Type    = D3DRESOURCETYPE(pSurf9->Type);
+    pSurf8->Usage   = pSurf9->Usage;
+    pSurf8->Pool    = D3DPOOL(pSurf9->Pool);
+    pSurf8->Size    = getSurfaceSize(pSurf8->Format, pSurf9->Width, pSurf9->Height);
+
+    pSurf8->MultiSampleType = D3DMULTISAMPLE_TYPE(pSurf9->MultiSampleType);
+    // DX8: No multisample quality
+    pSurf8->Width   = pSurf9->Width;
+    pSurf8->Height  = pSurf9->Height;
+  }
+
+  // (8<-9) Convert D3DVOLUME_DESC
+  inline void ConvertVolumeDesc8(const d3d9::D3DVOLUME_DESC* pVol9, D3DVOLUME_DESC* pVol8) {
+    pVol8->Format = D3DFORMAT(pVol9->Format);
+    pVol8->Type   = D3DRESOURCETYPE(pVol9->Type);
+    pVol8->Usage  = pVol9->Usage;
+    pVol8->Pool   = D3DPOOL(pVol9->Pool);
+    pVol8->Size   = getSurfaceSize(pVol8->Format, pVol9->Width, pVol9->Height) * pVol9->Depth;
+    pVol8->Width  = pVol9->Width;
+    pVol8->Height = pVol9->Height;
+    pVol8->Depth  = pVol9->Depth;
+  }
+
+  // If this D3DTEXTURESTAGESTATETYPE has been remapped to a d3d9::D3DSAMPLERSTATETYPE
+  // it will be returned, otherwise returns -1
+  inline d3d9::D3DSAMPLERSTATETYPE GetSamplerStateType9(const D3DTEXTURESTAGESTATETYPE StageType) {
+    switch (StageType) {
+      // 13-21:
+      case D3DTSS_ADDRESSU:       return d3d9::D3DSAMP_ADDRESSU;
+      case D3DTSS_ADDRESSV:       return d3d9::D3DSAMP_ADDRESSV;
+      case D3DTSS_BORDERCOLOR:    return d3d9::D3DSAMP_BORDERCOLOR;
+      case D3DTSS_MAGFILTER:      return d3d9::D3DSAMP_MAGFILTER;
+      case D3DTSS_MINFILTER:      return d3d9::D3DSAMP_MINFILTER;
+      case D3DTSS_MIPFILTER:      return d3d9::D3DSAMP_MIPFILTER;
+      case D3DTSS_MIPMAPLODBIAS:  return d3d9::D3DSAMP_MIPMAPLODBIAS;
+      case D3DTSS_MAXMIPLEVEL:    return d3d9::D3DSAMP_MIPFILTER;
+      case D3DTSS_MAXANISOTROPY:  return d3d9::D3DSAMP_MAXANISOTROPY;
+      // 25:
+      case D3DTSS_ADDRESSW:       return d3d9::D3DSAMP_ADDRESSW;
+      default:                    return d3d9::D3DSAMPLERSTATETYPE(-1);
+    }
+  }
+}
+
diff -pruN 2.3.1-1/src/d3d8/d3d8_device.cpp 2.4-1/src/d3d8/d3d8_device.cpp
--- 2.3.1-1/src/d3d8/d3d8_device.cpp	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_device.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,1753 @@
+#include "d3d8_device.h"
+#include "d3d8_interface.h"
+#include "d3d8_shader.h"
+
+#ifdef MSC_VER
+#pragma fenv_access (on)
+#endif
+
+namespace dxvk {
+
+  static constexpr DWORD isFVF(DWORD Handle) {
+    return (Handle & D3DFVF_RESERVED0) == 0;
+  }
+
+  static constexpr DWORD getShaderHandle(DWORD Index) {
+    return (Index << 1) | D3DFVF_RESERVED0;
+  }
+
+  static constexpr DWORD getShaderIndex(DWORD Handle) {
+    if ((Handle & D3DFVF_RESERVED0) != 0) {
+      return (Handle & ~(D3DFVF_RESERVED0)) >> 1;
+    } else {
+      return Handle;
+    }
+  }
+
+  struct D3D8VertexShaderInfo {
+    d3d9::IDirect3DVertexDeclaration9*  pVertexDecl   = nullptr;
+    d3d9::IDirect3DVertexShader9*       pVertexShader = nullptr;
+    std::vector<DWORD>                  declaration;
+    std::vector<DWORD>                  function;
+  };
+
+  D3D8Device::D3D8Device(
+    D3D8Interface*                pParent,
+    Com<d3d9::IDirect3DDevice9>&& pDevice,
+    D3DDEVTYPE                    DeviceType,
+    HWND                          hFocusWindow,
+    DWORD                         BehaviorFlags,
+    D3DPRESENT_PARAMETERS*        pParams)
+    : D3D8DeviceBase(std::move(pDevice))
+    , m_d3d8Options(pParent->GetOptions())
+    , m_parent(pParent)
+    , m_presentParams(*pParams)
+    , m_deviceType(DeviceType)
+    , m_window(hFocusWindow)
+    , m_behaviorFlags(BehaviorFlags) {
+    // Get the bridge interface to D3D9.
+    if (FAILED(GetD3D9()->QueryInterface(__uuidof(IDxvkD3D8Bridge), (void**)&m_bridge))) {
+      throw DxvkError("D3D8Device: ERROR! Failed to get D3D9 Bridge. d3d9.dll might not be DXVK!");
+    }
+
+    m_bridge->SetAPIName("D3D8");
+    m_bridge->SetD3D8CompatibilityMode(true);
+
+    ResetState();
+    RecreateBackBuffersAndAutoDepthStencil();
+    
+    if (m_d3d8Options.batching)
+      m_batcher = new D3D8Batcher(this, GetD3D9());
+  }
+
+  D3D8Device::~D3D8Device() {
+    if (m_batcher)
+      delete m_batcher;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetInfo(DWORD DevInfoID, void* pDevInfoStruct, DWORD DevInfoStructSize) {
+    Logger::debug(str::format("D3D8Device::GetInfo: ", DevInfoID));
+
+    if (unlikely(pDevInfoStruct == nullptr || DevInfoStructSize == 0))
+      return D3DERR_INVALIDCALL;
+
+    HRESULT res;
+    d3d9::IDirect3DQuery9* pQuery = nullptr;
+    
+    switch (DevInfoID) {
+      // pre-D3D8 queries
+      case 0:
+      case D3DDEVINFOID_TEXTUREMANAGER:
+      case D3DDEVINFOID_D3DTEXTUREMANAGER:
+      case D3DDEVINFOID_TEXTURING:
+        return E_FAIL;
+      
+      case D3DDEVINFOID_VCACHE:
+        // The query will return D3D_OK on Nvidia and D3DERR_NOTAVAILABLE on AMD/Intel
+        // in D3D9, however in the case of the latter we'll need to return a
+        // zeroed out query result and S_FALSE. This behavior has been observed both
+        // on modern native AMD drivers and D3D8-era native ATI drivers.
+        res = GetD3D9()->CreateQuery(d3d9::D3DQUERYTYPE_VCACHE, &pQuery);
+
+        struct D3DDEVINFO_VCACHE {
+          DWORD         Pattern;
+          DWORD         OptMethod;
+          DWORD         CacheSize;
+          DWORD         MagicNumber;
+        };
+
+        if(FAILED(res)) {
+          if (DevInfoStructSize != sizeof(D3DDEVINFO_VCACHE))
+            return D3DERR_INVALIDCALL;
+
+          memset(pDevInfoStruct, 0, sizeof(D3DDEVINFO_VCACHE));
+          return S_FALSE;
+        }
+
+        break;
+      case D3DDEVINFOID_RESOURCEMANAGER:
+        // May not be implemented by D9VK.
+        res = GetD3D9()->CreateQuery(d3d9::D3DQUERYTYPE_RESOURCEMANAGER, &pQuery);
+        break;
+      case D3DDEVINFOID_VERTEXSTATS:
+        res = GetD3D9()->CreateQuery(d3d9::D3DQUERYTYPE_VERTEXSTATS, &pQuery);
+        break;
+
+      default:
+        Logger::warn(str::format("D3D8Device::GetInfo: Unsupported device info ID: ", DevInfoID));
+        return E_FAIL;
+    }
+
+    if (unlikely(FAILED(res)))
+      goto done;
+    
+    // Immediately issue the query.
+    // D3D9 will begin it automatically before ending.
+    res = pQuery->Issue(D3DISSUE_END);
+    if (unlikely(FAILED(res))) {
+      goto done;
+    }
+
+    // TODO: Will immediately issuing the query without doing any API calls
+    // actually yield meaingful results? And should we flush or let it mellow?
+    res = pQuery->GetData(pDevInfoStruct, DevInfoStructSize, D3DGETDATA_FLUSH);
+
+  done:
+    if (pQuery != nullptr)
+      pQuery->Release();
+    
+    if (unlikely(FAILED(res))) {
+      if (res == D3DERR_NOTAVAILABLE) // unsupported
+        return E_FAIL;
+      else // any unknown error
+        return S_FALSE;
+    }
+    return res;
+  }
+
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::TestCooperativeLevel() {
+    // Equivalent of D3D11/DXGI present tests.
+    return GetD3D9()->TestCooperativeLevel();
+  }
+
+  UINT STDMETHODCALLTYPE D3D8Device::GetAvailableTextureMem() {
+    return GetD3D9()->GetAvailableTextureMem();
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::ResourceManagerDiscardBytes(DWORD bytes) { 
+    return GetD3D9()->EvictManagedResources();
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetDirect3D(IDirect3D8** ppD3D8) {
+    if (ppD3D8 == nullptr)
+      return D3DERR_INVALIDCALL;
+
+    *ppD3D8 = m_parent.ref();
+    return D3D_OK;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetDeviceCaps(D3DCAPS8* pCaps) {
+    d3d9::D3DCAPS9 caps9;
+    HRESULT res = GetD3D9()->GetDeviceCaps(&caps9);
+    dxvk::ConvertCaps8(caps9, pCaps);
+    return res;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetDisplayMode(D3DDISPLAYMODE* pMode) {
+    // swap chain 0
+    return GetD3D9()->GetDisplayMode(0, (d3d9::D3DDISPLAYMODE*)pMode);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetCreationParameters(D3DDEVICE_CREATION_PARAMETERS* pParameters) {
+    return GetD3D9()->GetCreationParameters((d3d9::D3DDEVICE_CREATION_PARAMETERS*)pParameters);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetCursorProperties(
+          UINT               XHotSpot,
+          UINT               YHotSpot,
+          IDirect3DSurface8* pCursorBitmap) {
+    D3D8Surface* surf = static_cast<D3D8Surface*>(pCursorBitmap);
+    return GetD3D9()->SetCursorProperties(XHotSpot, YHotSpot, D3D8Surface::GetD3D9Nullable(surf));
+  }
+
+  void    STDMETHODCALLTYPE D3D8Device::SetCursorPosition(UINT XScreenSpace, UINT YScreenSpace, DWORD Flags) {
+    GetD3D9()->SetCursorPosition(XScreenSpace, YScreenSpace, Flags);
+  }
+
+  // Microsoft d3d8.h in the DirectX 9 SDK uses a different function signature...
+  void    STDMETHODCALLTYPE D3D8Device::SetCursorPosition(int X, int Y, DWORD Flags) {
+    GetD3D9()->SetCursorPosition(X, Y, Flags);
+  }
+
+  BOOL    STDMETHODCALLTYPE D3D8Device::ShowCursor(BOOL bShow) {
+    return GetD3D9()->ShowCursor(bShow);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::CreateAdditionalSwapChain(
+      D3DPRESENT_PARAMETERS* pPresentationParameters,
+      IDirect3DSwapChain8** ppSwapChain) {
+    
+    Com<d3d9::IDirect3DSwapChain9> pSwapChain9;
+    d3d9::D3DPRESENT_PARAMETERS params = ConvertPresentParameters9(pPresentationParameters);
+    HRESULT res = GetD3D9()->CreateAdditionalSwapChain(
+      &params,
+      &pSwapChain9
+    );
+    
+    *ppSwapChain = ref(new D3D8SwapChain(this, std::move(pSwapChain9)));
+
+    return res;
+  }
+
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::Reset(D3DPRESENT_PARAMETERS* pPresentationParameters) {
+    StateChange();
+
+    m_presentParams = *pPresentationParameters;
+    ResetState();
+
+    d3d9::D3DPRESENT_PARAMETERS params = ConvertPresentParameters9(pPresentationParameters);
+    HRESULT res = GetD3D9()->Reset(&params);
+
+    if (FAILED(res))
+      return res;
+
+    RecreateBackBuffersAndAutoDepthStencil();
+
+    return res;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::Present(
+    const RECT* pSourceRect,
+    const RECT* pDestRect,
+          HWND hDestWindowOverride,
+    const RGNDATA* pDirtyRegion) {
+    m_batcher->EndFrame();
+    StateChange();
+    return GetD3D9()->Present(pSourceRect, pDestRect, hDestWindowOverride, pDirtyRegion);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetBackBuffer(
+          UINT iBackBuffer,
+          D3DBACKBUFFER_TYPE Type,
+          IDirect3DSurface8** ppBackBuffer) {
+    InitReturnPtr(ppBackBuffer);
+    
+    if (iBackBuffer >= m_backBuffers.size() || m_backBuffers[iBackBuffer] == nullptr) {
+      Com<d3d9::IDirect3DSurface9> pSurface9;
+      HRESULT res = GetD3D9()->GetBackBuffer(0, iBackBuffer, (d3d9::D3DBACKBUFFER_TYPE)Type, &pSurface9);
+
+      if (FAILED(res)) return res;
+      
+      m_backBuffers[iBackBuffer] = new D3D8Surface(this, std::move(pSurface9));
+      *ppBackBuffer = m_backBuffers[iBackBuffer].ref();
+
+      return res;
+    }
+
+    *ppBackBuffer = m_backBuffers[iBackBuffer].ref();
+    return D3D_OK;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetRasterStatus(D3DRASTER_STATUS* pRasterStatus) {
+    return GetD3D9()->GetRasterStatus(0, (d3d9::D3DRASTER_STATUS*)pRasterStatus);
+  }
+
+  void STDMETHODCALLTYPE D3D8Device::SetGammaRamp(DWORD Flags, const D3DGAMMARAMP* pRamp) {
+    StateChange();
+    // For swap chain 0
+    GetD3D9()->SetGammaRamp(0, Flags, reinterpret_cast<const d3d9::D3DGAMMARAMP*>(pRamp));
+  }
+
+  void STDMETHODCALLTYPE D3D8Device::GetGammaRamp(D3DGAMMARAMP* pRamp) {
+    // For swap chain 0
+    GetD3D9()->GetGammaRamp(0, reinterpret_cast<d3d9::D3DGAMMARAMP*>(pRamp));
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::CreateTexture(
+          UINT                Width,
+          UINT                Height,
+          UINT                Levels,
+          DWORD               Usage,
+          D3DFORMAT           Format,
+          D3DPOOL             Pool,
+          IDirect3DTexture8** ppTexture) {
+    InitReturnPtr(ppTexture);
+
+    // Nvidia & Intel workaround for The Lord of the Rings: The Fellowship of the Ring
+    if (m_d3d8Options.placeP8InScratch && Format == D3DFMT_P8)
+      Pool = D3DPOOL_SCRATCH;
+
+    Com<d3d9::IDirect3DTexture9> pTex9 = nullptr;
+    HRESULT res = GetD3D9()->CreateTexture(
+      Width,
+      Height,
+      Levels,
+      Usage,
+      d3d9::D3DFORMAT(Format),
+      d3d9::D3DPOOL(Pool),
+      &pTex9,
+      NULL);
+
+    if (FAILED(res))
+      return res;
+
+    *ppTexture = ref(new D3D8Texture2D(this, std::move(pTex9)));
+
+    return res;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::CreateVolumeTexture(
+          UINT                      Width,
+          UINT                      Height,
+          UINT                      Depth,
+          UINT                      Levels,
+          DWORD                     Usage,
+          D3DFORMAT                 Format,
+          D3DPOOL                   Pool,
+          IDirect3DVolumeTexture8** ppVolumeTexture) {
+    Com<d3d9::IDirect3DVolumeTexture9> pVolume9 = nullptr;
+    HRESULT res = GetD3D9()->CreateVolumeTexture(
+      Width, Height, Depth, Levels,
+      Usage,
+      d3d9::D3DFORMAT(Format),
+      d3d9::D3DPOOL(Pool),
+      &pVolume9,
+      NULL);
+
+    *ppVolumeTexture = ref(new D3D8Texture3D(this, std::move(pVolume9)));
+
+    return res;
+
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::CreateCubeTexture(
+        UINT                      EdgeLength,
+          UINT                    Levels,
+          DWORD                   Usage,
+          D3DFORMAT               Format,
+          D3DPOOL                 Pool,
+          IDirect3DCubeTexture8** ppCubeTexture) {
+    Com<d3d9::IDirect3DCubeTexture9> pCube9 = nullptr;
+    HRESULT res = GetD3D9()->CreateCubeTexture(
+      EdgeLength,
+      Levels,
+      Usage,
+      d3d9::D3DFORMAT(Format),
+      d3d9::D3DPOOL(Pool),
+      &pCube9,
+      NULL);
+
+    *ppCubeTexture = ref(new D3D8TextureCube(this, std::move(pCube9)));
+
+    return res;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::CreateVertexBuffer(
+          UINT                     Length,
+          DWORD                    Usage,
+          DWORD                    FVF,
+          D3DPOOL                  Pool,
+          IDirect3DVertexBuffer8** ppVertexBuffer) {
+    InitReturnPtr(ppVertexBuffer);
+
+    if (ShouldBatch()) {
+      *ppVertexBuffer = m_batcher->CreateVertexBuffer(Length, Usage, FVF, Pool);
+      return D3D_OK;
+    }
+
+    Com<d3d9::IDirect3DVertexBuffer9> pVertexBuffer9 = nullptr;
+
+    HRESULT res = GetD3D9()->CreateVertexBuffer(Length, Usage, FVF, d3d9::D3DPOOL(Pool), &pVertexBuffer9, NULL);
+
+    if (!FAILED(res))
+      *ppVertexBuffer = ref(new D3D8VertexBuffer(this, std::move(pVertexBuffer9), Pool, Usage));
+  
+    return res;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::CreateIndexBuffer(
+          UINT                    Length,
+          DWORD                   Usage,
+          D3DFORMAT               Format,
+          D3DPOOL                 Pool,
+          IDirect3DIndexBuffer8** ppIndexBuffer) {
+    InitReturnPtr(ppIndexBuffer);
+    Com<d3d9::IDirect3DIndexBuffer9> pIndexBuffer9 = nullptr;
+    
+    HRESULT res = GetD3D9()->CreateIndexBuffer(Length, Usage, d3d9::D3DFORMAT(Format), d3d9::D3DPOOL(Pool), &pIndexBuffer9, NULL);
+    
+    if (!FAILED(res))
+      *ppIndexBuffer = ref(new D3D8IndexBuffer(this, std::move(pIndexBuffer9), Pool, Usage));
+    
+    return res;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::CreateRenderTarget(
+          UINT                Width,
+          UINT                Height,
+          D3DFORMAT           Format,
+          D3DMULTISAMPLE_TYPE MultiSample,
+          BOOL                Lockable,
+          IDirect3DSurface8** ppSurface) {
+    Com<d3d9::IDirect3DSurface9> pSurf9 = nullptr;
+    HRESULT res = GetD3D9()->CreateRenderTarget(
+      Width,
+      Height,
+      d3d9::D3DFORMAT(Format),
+      d3d9::D3DMULTISAMPLE_TYPE(MultiSample),
+      0,    // TODO: CreateRenderTarget MultisampleQuality
+      Lockable,
+      &pSurf9,
+      NULL);
+
+    *ppSurface = ref(new D3D8Surface(this, std::move(pSurf9)));
+
+    return res;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::CreateDepthStencilSurface(
+          UINT                Width,
+          UINT                Height,
+          D3DFORMAT           Format,
+          D3DMULTISAMPLE_TYPE MultiSample,
+          IDirect3DSurface8** ppSurface) {
+    Com<d3d9::IDirect3DSurface9> pSurf9 = nullptr;
+    HRESULT res = GetD3D9()->CreateDepthStencilSurface(
+      Width,
+      Height,
+      d3d9::D3DFORMAT(Format),
+      d3d9::D3DMULTISAMPLE_TYPE(MultiSample),
+      0,    // TODO: CreateDepthStencilSurface MultisampleQuality
+      true, // TODO: CreateDepthStencilSurface Discard
+      &pSurf9,
+      NULL);
+
+    *ppSurface = ref(new D3D8Surface(this, std::move(pSurf9)));
+
+    return res;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::CreateImageSurface(
+          UINT                Width,
+          UINT                Height,
+          D3DFORMAT           Format,
+          IDirect3DSurface8** ppSurface) {
+    // FIXME: Handle D3DPOOL_SCRATCH in CopyRects
+    D3DPOOL pool = isUnsupportedSurfaceFormat(Format) ? D3DPOOL_SCRATCH : D3DPOOL_SYSTEMMEM;
+
+    Com<d3d9::IDirect3DSurface9> pSurf = nullptr;
+    HRESULT res = GetD3D9()->CreateOffscreenPlainSurface(
+      Width,
+      Height,
+      d3d9::D3DFORMAT(Format),
+      d3d9::D3DPOOL(pool),
+      &pSurf,
+      NULL);
+
+    *ppSurface = ref(new D3D8Surface(this, std::move(pSurf)));
+
+    return res;
+  }
+
+  // Copies texture rect in system mem using memcpy.
+  // Rects must be congruent, but need not be aligned.
+  HRESULT copyTextureBuffers(
+      D3D8Surface*                  src,
+      D3D8Surface*                  dst,
+      const d3d9::D3DSURFACE_DESC&  srcDesc,
+      const d3d9::D3DSURFACE_DESC&  dstDesc,
+      const RECT&                   srcRect,
+      const RECT&                   dstRect) {
+    HRESULT res = D3D_OK;
+    D3DLOCKED_RECT srcLocked, dstLocked;
+
+    // CopyRects cannot perform format conversions.
+    if (srcDesc.Format != dstDesc.Format)
+      return D3DERR_INVALIDCALL;
+
+    bool compressed = isDXT(srcDesc.Format);
+
+    res = src->LockRect(&srcLocked, &srcRect, D3DLOCK_READONLY);
+    if (FAILED(res))
+      return res;
+    
+    res = dst->LockRect(&dstLocked, &dstRect, 0);
+    if (FAILED(res)) {
+      src->UnlockRect();
+      return res;
+    }
+
+    auto rows = srcRect.bottom  - srcRect.top;
+    auto cols = srcRect.right   - srcRect.left;
+    auto bpp  = srcLocked.Pitch / srcDesc.Width;
+
+    if (!compressed
+     && srcRect.left    == 0
+     && srcRect.right   == LONG(srcDesc.Width)
+     && srcDesc.Width   == dstDesc.Width
+     && srcLocked.Pitch == dstLocked.Pitch) {
+
+      // If copying the entire texture into a congruent destination,
+      // we can do this in one continuous copy.
+      std::memcpy(dstLocked.pBits, srcLocked.pBits, srcLocked.Pitch * rows);
+
+    } else {
+      // Bytes per row of the rect
+      auto amplitude = cols * bpp;
+
+      // Handle DXT compressed textures.
+      // TODO: Are rects always 4x4 aligned?
+      if (compressed) {
+        // Assume that DXT blocks are 4x4 pixels.
+        constexpr UINT blockWidth  = 4;
+        constexpr UINT blockHeight = 4;
+
+        // Compute rect dimensions in 4x4 blocks
+        UINT rectWidthBlocks  = cols / blockWidth;
+        UINT rectHeightBlocks = rows / blockHeight;
+
+        // Compute total texture width in blocks
+        // to derive block size in bytes using the pitch.
+        UINT texWidthBlocks = std::max(srcDesc.Width / blockWidth, 1u);
+        UINT bytesPerBlock  = srcLocked.Pitch / texWidthBlocks;
+
+        // Copy H/4 rows of W/4 blocks
+        amplitude = rectWidthBlocks * bytesPerBlock;
+        rows      = rectHeightBlocks;
+      }
+
+      // Copy one row at a time
+      size_t srcOffset = 0, dstOffset = 0;
+      for (auto i = 0; i < rows; i++) {
+        std::memcpy(
+          (uint8_t*)dstLocked.pBits + dstOffset,
+          (uint8_t*)srcLocked.pBits + srcOffset,
+          amplitude);
+        srcOffset += srcLocked.Pitch;
+        dstOffset += dstLocked.Pitch;
+      }
+    }
+
+    res = dst->UnlockRect();
+    res = src->UnlockRect();
+    return res;
+  }
+
+  /**
+   * \brief D3D8 CopyRects implementation
+   *
+   * \details
+   * The following table shows the possible combinations of source
+   * and destination surface pools, and how we handle each of them.
+   *
+   *     ┌────────────┬───────────────────────────┬───────────────────────┬───────────────────────┬──────────┐
+   *     │ Src/Dst    │ DEFAULT                   │ MANAGED               │ SYSTEMMEM             │ SCRATCH  │
+   *     ├────────────┼───────────────────────────┼───────────────────────┼───────────────────────┼──────────┤
+   *     │ DEFAULT    │  StretchRect              │  GetRenderTargetData  │  GetRenderTargetData  │ -        │
+   *     │ MANAGED    │  UpdateTextureFromBuffer  │  memcpy               │  memcpy               │ -        │
+   *     │ SYSTEMMEM  │  UpdateSurface            │  memcpy               │  memcpy               │ -        │
+   *     │ SCRATCH    │  -                        │  -                    │  -                    │ -        │
+   *     └────────────┴───────────────────────────┴───────────────────────┴───────────────────────┴──────────┘
+   */
+  HRESULT STDMETHODCALLTYPE D3D8Device::CopyRects(
+          IDirect3DSurface8*  pSourceSurface,
+    const RECT*               pSourceRectsArray,
+          UINT                cRects,
+          IDirect3DSurface8*  pDestinationSurface,
+    const POINT*              pDestPointsArray) {
+    if (pSourceSurface == NULL || pDestinationSurface == NULL) {
+      return D3DERR_INVALIDCALL;
+    }
+
+    // TODO: No format conversion, no stretching, no clipping.
+    // All src/dest rectangles must fit within the dest surface.
+
+    Com<D3D8Surface> src = static_cast<D3D8Surface*>(pSourceSurface);
+    Com<D3D8Surface> dst = static_cast<D3D8Surface*>(pDestinationSurface);
+
+    d3d9::D3DSURFACE_DESC srcDesc, dstDesc;
+    src->GetD3D9()->GetDesc(&srcDesc);
+    dst->GetD3D9()->GetDesc(&dstDesc);
+
+    // This method cannot be applied to surfaces whose formats
+    // are classified as depth stencil formats.
+    if (unlikely(isDepthStencilFormat(D3DFORMAT(srcDesc.Format)) ||
+                 isDepthStencilFormat(D3DFORMAT(dstDesc.Format)))) {
+      return D3DERR_INVALIDCALL;
+    }
+
+    StateChange();
+
+    // If pSourceRectsArray is NULL, then the entire surface is copied
+    RECT rect;
+    POINT point = { 0, 0 };
+    if (pSourceRectsArray == NULL) {
+      cRects = 1;
+      rect.top    = rect.left = 0;
+      rect.right  = srcDesc.Width;
+      rect.bottom = srcDesc.Height;
+      pSourceRectsArray = &rect;
+
+      pDestPointsArray = &point;
+    }
+
+    for (UINT i = 0; i < cRects; i++) {
+
+      RECT srcRect, dstRect;
+      srcRect = pSourceRectsArray[i];
+
+      // True if the copy is asymmetric
+      bool asymmetric = true;
+      // True if the copy requires stretching (not technically supported)
+      bool stretch = true;
+      // True if the copy is not perfectly aligned (supported)
+      bool offset = true;
+
+      if (pDestPointsArray != NULL) {
+        dstRect.left    = pDestPointsArray[i].x;
+        dstRect.right   = dstRect.left + (srcRect.right - srcRect.left);
+        dstRect.top     = pDestPointsArray[i].y;
+        dstRect.bottom  = dstRect.top + (srcRect.bottom - srcRect.top);
+        asymmetric  = dstRect.left  != srcRect.left  || dstRect.top    != srcRect.top
+                   || dstRect.right != srcRect.right || dstRect.bottom != srcRect.bottom;
+
+        stretch     = (dstRect.right-dstRect.left) != (srcRect.right-srcRect.left)
+                   || (dstRect.bottom-dstRect.top) != (srcRect.bottom-srcRect.top);
+
+        offset      = !stretch && asymmetric;
+      } else {
+        dstRect     = srcRect;
+        asymmetric  = stretch = offset = false;
+      }
+
+      POINT dstPt = { dstRect.left, dstRect.top };
+
+      auto unhandled = [&] {
+        Logger::warn(str::format("CopyRects: Hit unhandled case from src pool ", srcDesc.Pool, " to dst pool ", dstDesc.Pool));
+        return D3DERR_INVALIDCALL;
+      };
+
+      auto logError = [&] (HRESULT res) {
+        if (FAILED(res)) {
+          // Only a debug message because some games mess up CopyRects every frame in a way
+          // that fails on native too but are perfectly fine with it.
+          Logger::debug(str::format("CopyRects: FAILED to copy from src pool ", srcDesc.Pool, " to dst pool ", dstDesc.Pool));
+        }
+        return res;
+      };
+
+      switch (dstDesc.Pool) {
+
+        // Dest: DEFAULT
+        case d3d9::D3DPOOL_DEFAULT:
+          switch (srcDesc.Pool) {
+            case d3d9::D3DPOOL_DEFAULT: {
+              // DEFAULT -> DEFAULT: use StretchRect
+              return logError(GetD3D9()->StretchRect(
+                src->GetD3D9(),
+                &srcRect,
+                dst->GetD3D9(),
+                &dstRect,
+                d3d9::D3DTEXF_NONE
+              ));
+            }
+            case d3d9::D3DPOOL_MANAGED: {
+              // MANAGED -> DEFAULT: UpdateTextureFromBuffer
+              return logError(m_bridge->UpdateTextureFromBuffer(
+                src->GetD3D9(),
+                dst->GetD3D9(),
+                &srcRect,
+                &dstPt
+              ));
+            }
+            case d3d9::D3DPOOL_SYSTEMMEM: {
+              // SYSTEMMEM -> DEFAULT: use UpdateSurface
+              return logError(GetD3D9()->UpdateSurface(
+                src->GetD3D9(),
+                &srcRect,
+                dst->GetD3D9(),
+                &dstPt
+              ));
+            }
+            case d3d9::D3DPOOL_SCRATCH:
+            default: {
+              // TODO: Unhandled case.
+              return unhandled();
+            }
+          } break;
+
+        // Dest: MANAGED
+        case d3d9::D3DPOOL_MANAGED:
+          switch (srcDesc.Pool) {
+            case d3d9::D3DPOOL_DEFAULT: {
+              // TODO: Copy on GPU (handle MANAGED similarly to SYSTEMMEM for now)
+
+              // Get temporary off-screen surface for stretching.
+              Com<d3d9::IDirect3DSurface9> pBlitImage = dst->GetBlitImage();
+
+              // Stretch the source RT to the temporary surface.
+              HRESULT res = GetD3D9()->StretchRect(
+                src->GetD3D9(),
+                &srcRect,
+                pBlitImage.ptr(),
+                &dstRect,
+                d3d9::D3DTEXF_NONE);
+
+              if (FAILED(res)) {
+                return logError(res);
+              }
+
+              // Now sync the rendertarget data into main memory.
+              return logError(GetD3D9()->GetRenderTargetData(pBlitImage.ptr(), dst->GetD3D9()));
+            }
+            case d3d9::D3DPOOL_MANAGED:
+            case d3d9::D3DPOOL_SYSTEMMEM: {
+              // SYSTEMMEM -> MANAGED: LockRect / memcpy
+
+              if (stretch) {
+                return logError(D3DERR_INVALIDCALL);
+              }
+
+              return logError(copyTextureBuffers(src.ptr(), dst.ptr(), srcDesc, dstDesc, srcRect, dstRect));
+            }
+            case d3d9::D3DPOOL_SCRATCH:
+            default: {
+              // TODO: Unhandled case.
+              return unhandled();
+            }
+          } break;
+
+        // DEST: SYSTEMMEM
+        case d3d9::D3DPOOL_SYSTEMMEM: {
+
+          // RT (DEFAULT) -> SYSTEMMEM: Use GetRenderTargetData as fast path if possible
+          if ((srcDesc.Usage & D3DUSAGE_RENDERTARGET || m_renderTarget.ptr() == src.ptr())) {
+
+            // GetRenderTargetData works if the formats and sizes match
+            if (srcDesc.MultiSampleType == d3d9::D3DMULTISAMPLE_NONE
+                && srcDesc.Width  == dstDesc.Width
+                && srcDesc.Height == dstDesc.Height
+                && srcDesc.Format == dstDesc.Format
+                && !asymmetric) {
+              return logError(GetD3D9()->GetRenderTargetData(src->GetD3D9(), dst->GetD3D9()));
+            }
+          }
+
+          switch (srcDesc.Pool) {
+            case d3d9::D3DPOOL_DEFAULT: {
+              // Get temporary off-screen surface for stretching.
+              Com<d3d9::IDirect3DSurface9> pBlitImage = dst->GetBlitImage();
+
+              // Stretch the source RT to the temporary surface.
+              HRESULT res = GetD3D9()->StretchRect(
+                src->GetD3D9(),
+                &srcRect,
+                pBlitImage.ptr(),
+                &dstRect,
+                d3d9::D3DTEXF_NONE);
+              if (FAILED(res)) {
+                return logError(res);
+              }
+
+              // Now sync the rendertarget data into main memory.
+              return logError(GetD3D9()->GetRenderTargetData(pBlitImage.ptr(), dst->GetD3D9()));
+            }
+
+            // SYSMEM/MANAGED -> SYSMEM: LockRect / memcpy
+            case d3d9::D3DPOOL_MANAGED:
+            case d3d9::D3DPOOL_SYSTEMMEM: {
+              if (stretch) {
+                return logError(D3DERR_INVALIDCALL);
+              }
+
+              return logError(copyTextureBuffers(src.ptr(), dst.ptr(), srcDesc, dstDesc, srcRect, dstRect));
+            }
+            case d3d9::D3DPOOL_SCRATCH:
+            default: {
+              // TODO: Unhandled case.
+              return unhandled();
+            }
+          } break;
+        }
+
+        // DEST: SCRATCH
+        case d3d9::D3DPOOL_SCRATCH:
+        default: {
+          // TODO: Unhandled case.
+          return unhandled();
+        }
+      }
+    }
+
+    return D3DERR_INVALIDCALL;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::UpdateTexture(
+          IDirect3DBaseTexture8* pSourceTexture,
+          IDirect3DBaseTexture8* pDestinationTexture) {
+    D3D8Texture2D* src = static_cast<D3D8Texture2D*>(pSourceTexture);
+    D3D8Texture2D* dst = static_cast<D3D8Texture2D*>(pDestinationTexture);
+
+    StateChange();
+    return GetD3D9()->UpdateTexture(D3D8Texture2D::GetD3D9Nullable(src), D3D8Texture2D::GetD3D9Nullable(dst));
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetFrontBuffer(IDirect3DSurface8* pDestSurface) {
+    if (unlikely(pDestSurface == nullptr))
+      return D3DERR_INVALIDCALL;
+
+    Com<D3D8Surface> surf = static_cast<D3D8Surface*>(pDestSurface);
+
+    StateChange();
+    // This actually gets a copy of the front buffer and writes it to pDestSurface
+    return GetD3D9()->GetFrontBufferData(0, D3D8Surface::GetD3D9Nullable(surf));
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetRenderTarget(IDirect3DSurface8* pRenderTarget, IDirect3DSurface8* pNewZStencil) {
+    HRESULT res;
+
+    if (pRenderTarget != NULL) {
+      D3D8Surface* surf = static_cast<D3D8Surface*>(pRenderTarget);
+
+      if(likely(m_renderTarget.ptr() != surf)) {
+        StateChange();
+        res = GetD3D9()->SetRenderTarget(0, surf->GetD3D9());
+
+        if (FAILED(res)) return res;
+
+        if (likely(m_renderTarget != surf))
+          m_renderTarget = surf;
+      }
+    }
+
+    // SetDepthStencilSurface is a separate call
+    D3D8Surface* zStencil = static_cast<D3D8Surface*>(pNewZStencil);
+
+    if(likely(m_depthStencil.ptr() != zStencil)) {
+      StateChange();
+      res = GetD3D9()->SetDepthStencilSurface(D3D8Surface::GetD3D9Nullable(zStencil));
+
+      if (FAILED(res)) return res;
+
+      if (likely(m_depthStencil != zStencil))
+          m_depthStencil = zStencil;
+    }
+
+    return D3D_OK;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetRenderTarget(IDirect3DSurface8** ppRenderTarget) {
+    InitReturnPtr(ppRenderTarget);
+
+    if (unlikely(m_renderTarget == nullptr)) {
+      Com<d3d9::IDirect3DSurface9> pRT9 = nullptr;
+      HRESULT res = GetD3D9()->GetRenderTarget(0, &pRT9); // use RT index 0
+
+      if(FAILED(res)) return res;
+
+      m_renderTarget = new D3D8Surface(this, std::move(pRT9));
+
+      *ppRenderTarget = m_renderTarget.ref();
+      return res;
+    }
+
+    *ppRenderTarget = m_renderTarget.ref();
+    return D3D_OK;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetDepthStencilSurface(IDirect3DSurface8** ppZStencilSurface) {
+    InitReturnPtr(ppZStencilSurface);
+    
+    if (unlikely(m_depthStencil == nullptr)) {
+      Com<d3d9::IDirect3DSurface9> pStencil9 = nullptr;
+      HRESULT res = GetD3D9()->GetDepthStencilSurface(&pStencil9);
+
+      if(FAILED(res)) return res;
+
+      m_depthStencil = new D3D8Surface(this, std::move(pStencil9));
+
+      *ppZStencilSurface = m_depthStencil.ref();
+      return res;
+    }
+
+    *ppZStencilSurface = m_depthStencil.ref();
+    return D3D_OK;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::BeginScene() { return GetD3D9()->BeginScene(); }
+  
+  HRESULT STDMETHODCALLTYPE D3D8Device::EndScene() { StateChange(); return GetD3D9()->EndScene(); }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::Clear(
+          DWORD    Count,
+    const D3DRECT* pRects,
+          DWORD    Flags,
+          D3DCOLOR Color,
+          float    Z,
+          DWORD    Stencil) {
+    StateChange();
+    return GetD3D9()->Clear(Count, pRects, Flags, Color, Z, Stencil);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetTransform(D3DTRANSFORMSTATETYPE State, const D3DMATRIX* pMatrix) {
+    StateChange();
+    return GetD3D9()->SetTransform(d3d9::D3DTRANSFORMSTATETYPE(State), pMatrix);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetTransform(D3DTRANSFORMSTATETYPE State, D3DMATRIX* pMatrix) {
+    return GetD3D9()->GetTransform(d3d9::D3DTRANSFORMSTATETYPE(State), pMatrix);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::MultiplyTransform(D3DTRANSFORMSTATETYPE TransformState, const D3DMATRIX* pMatrix) {
+    StateChange();
+    return GetD3D9()->MultiplyTransform(d3d9::D3DTRANSFORMSTATETYPE(TransformState), pMatrix);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetViewport(const D3DVIEWPORT8* pViewport) {
+    StateChange();
+    return GetD3D9()->SetViewport(reinterpret_cast<const d3d9::D3DVIEWPORT9*>(pViewport));
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetViewport(D3DVIEWPORT8* pViewport) {
+    return GetD3D9()->GetViewport(reinterpret_cast<d3d9::D3DVIEWPORT9*>(pViewport));
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetMaterial(const D3DMATERIAL8* pMaterial) {
+    StateChange();
+    return GetD3D9()->SetMaterial((const d3d9::D3DMATERIAL9*)pMaterial);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetMaterial(D3DMATERIAL8* pMaterial) {
+    return GetD3D9()->GetMaterial((d3d9::D3DMATERIAL9*)pMaterial);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetLight(DWORD Index, const D3DLIGHT8* pLight) {
+    StateChange();
+    return GetD3D9()->SetLight(Index, (const d3d9::D3DLIGHT9*)pLight);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetLight(DWORD Index, D3DLIGHT8* pLight) {
+    return GetD3D9()->GetLight(Index, (d3d9::D3DLIGHT9*)pLight);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::LightEnable(DWORD Index, BOOL Enable) {
+    StateChange();
+    return GetD3D9()->LightEnable(Index, Enable);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetLightEnable(DWORD Index, BOOL* pEnable) {
+    return GetD3D9()->GetLightEnable(Index, pEnable);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetClipPlane(DWORD Index, const float* pPlane) {
+    StateChange();
+    return GetD3D9()->SetClipPlane(Index, pPlane);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetClipPlane(DWORD Index, float* pPlane) {
+    return GetD3D9()->GetClipPlane(Index, pPlane);
+  }
+  
+  HRESULT STDMETHODCALLTYPE D3D8Device::CreateStateBlock(
+          D3DSTATEBLOCKTYPE     Type,
+          DWORD*                pToken) {
+    Com<d3d9::IDirect3DStateBlock9> pStateBlock9;
+    HRESULT res = GetD3D9()->CreateStateBlock(d3d9::D3DSTATEBLOCKTYPE(Type), &pStateBlock9);
+
+    m_token++;
+    m_stateBlocks.emplace(std::piecewise_construct,
+                          std::forward_as_tuple(m_token),
+                          std::forward_as_tuple(this, Type, pStateBlock9.ref()));
+    *pToken = m_token;
+
+    return res;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::CaptureStateBlock(DWORD Token) {
+    auto stateBlockIter = m_stateBlocks.find(Token);
+
+    if (unlikely(stateBlockIter == m_stateBlocks.end())) {
+      Logger::err("Invalid token passed to CaptureStateBlock");
+      return D3DERR_INVALIDCALL;
+    }
+
+    return stateBlockIter->second.Capture();
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::ApplyStateBlock(DWORD Token) {
+    StateChange();
+
+    auto stateBlockIter = m_stateBlocks.find(Token);
+
+    if (unlikely(stateBlockIter == m_stateBlocks.end())) {
+      Logger::err("Invalid token passed to ApplyStateBlock");
+      return D3DERR_INVALIDCALL;
+    }
+
+    return stateBlockIter->second.Apply();
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::DeleteStateBlock(DWORD Token) {
+    // "Applications cannot delete a device-state block while another is being recorded"
+    if (unlikely(ShouldRecord()))
+      return D3DERR_INVALIDCALL;
+
+    auto stateBlockIter = m_stateBlocks.find(Token);
+
+    if (unlikely(stateBlockIter == m_stateBlocks.end())) {
+      Logger::err("Invalid token passed to DeleteStateBlock");
+      return D3DERR_INVALIDCALL;
+    }
+
+    m_stateBlocks.erase(stateBlockIter);
+
+    // native apparently does drop the token counter in
+    // situations where the token being removed is the
+    // last allocated token, which allows some reuse
+    if (m_token == Token)
+      m_token--;
+
+    return D3D_OK;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::BeginStateBlock() {
+    if (unlikely(m_recorder != nullptr))
+      return D3DERR_INVALIDCALL;
+
+    m_token++;
+    auto stateBlockIterPair = m_stateBlocks.emplace(std::piecewise_construct,
+                                                    std::forward_as_tuple(m_token),
+                                                    std::forward_as_tuple(this));
+    m_recorder = &stateBlockIterPair.first->second;
+    m_recorderToken = m_token;
+
+    return GetD3D9()->BeginStateBlock();
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::EndStateBlock(DWORD* pToken) {
+    if (unlikely(pToken == nullptr || m_recorder == nullptr))
+      return D3DERR_INVALIDCALL;
+
+    Com<d3d9::IDirect3DStateBlock9> pStateBlock;
+    HRESULT res = GetD3D9()->EndStateBlock(&pStateBlock);
+
+    m_recorder->SetD3D9(std::move(pStateBlock));
+
+    *pToken = m_recorderToken;
+
+    m_recorder = nullptr;
+    m_recorderToken = 0;
+
+    return res;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetClipStatus(const D3DCLIPSTATUS8* pClipStatus) {
+    StateChange();
+    return GetD3D9()->SetClipStatus(reinterpret_cast<const d3d9::D3DCLIPSTATUS9*>(pClipStatus));
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetClipStatus(D3DCLIPSTATUS8* pClipStatus) {
+    return GetD3D9()->GetClipStatus(reinterpret_cast<d3d9::D3DCLIPSTATUS9*>(pClipStatus));
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetTexture(DWORD Stage, IDirect3DBaseTexture8** ppTexture) {
+    InitReturnPtr(ppTexture);
+
+    *ppTexture = m_textures[Stage].ref();
+
+    return D3D_OK;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetTexture(DWORD Stage, IDirect3DBaseTexture8* pTexture) {
+    if (unlikely(Stage >= d8caps::MAX_TEXTURE_STAGES))
+      return D3DERR_INVALIDCALL;
+
+    if (unlikely(ShouldRecord()))
+      return m_recorder->SetTexture(Stage, pTexture);
+
+    D3D8Texture2D* tex = static_cast<D3D8Texture2D*>(pTexture);
+
+    if(unlikely(m_textures[Stage].ptr() == tex))
+      return D3D_OK;
+
+    StateChange();
+
+    m_textures[Stage] = tex;
+
+    return GetD3D9()->SetTexture(Stage, D3D8Texture2D::GetD3D9Nullable(tex));
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetTextureStageState(
+          DWORD                    Stage,
+          D3DTEXTURESTAGESTATETYPE Type,
+          DWORD*                   pValue) {
+    d3d9::D3DSAMPLERSTATETYPE stateType = GetSamplerStateType9(Type);
+
+    if (stateType != -1) {
+      // if the type has been remapped to a sampler state type:
+      return GetD3D9()->GetSamplerState(Stage, stateType, pValue);
+    }
+    else {
+      return GetD3D9()->GetTextureStageState(Stage, d3d9::D3DTEXTURESTAGESTATETYPE(Type), pValue);
+    }
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetTextureStageState(
+          DWORD                    Stage,
+          D3DTEXTURESTAGESTATETYPE Type,
+          DWORD                    Value) {
+    d3d9::D3DSAMPLERSTATETYPE stateType = GetSamplerStateType9(Type);
+
+    StateChange();
+    if (stateType != -1) {
+      // if the type has been remapped to a sampler state type:
+      return GetD3D9()->SetSamplerState(Stage, stateType, Value);
+    } else {
+      return GetD3D9()->SetTextureStageState(Stage, d3d9::D3DTEXTURESTAGESTATETYPE(Type), Value);
+    }
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::ValidateDevice(DWORD* pNumPasses) {
+    return GetD3D9()->ValidateDevice(pNumPasses);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetPaletteEntries(UINT PaletteNumber, const PALETTEENTRY* pEntries) {
+    StateChange();
+    return GetD3D9()->SetPaletteEntries(PaletteNumber, pEntries);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetPaletteEntries(UINT PaletteNumber, PALETTEENTRY* pEntries) {
+    return GetD3D9()->GetPaletteEntries(PaletteNumber, pEntries);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetCurrentTexturePalette(UINT PaletteNumber) {
+    StateChange();
+    return GetD3D9()->SetCurrentTexturePalette(PaletteNumber);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetCurrentTexturePalette(UINT* PaletteNumber) {
+    return GetD3D9()->GetCurrentTexturePalette(PaletteNumber);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::DrawPrimitive(
+          D3DPRIMITIVETYPE PrimitiveType,
+          UINT             StartVertex,
+          UINT             PrimitiveCount) {
+    if (ShouldBatch())
+      return m_batcher->DrawPrimitive(PrimitiveType, StartVertex, PrimitiveCount);
+    return GetD3D9()->DrawPrimitive(d3d9::D3DPRIMITIVETYPE(PrimitiveType), StartVertex, PrimitiveCount);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::DrawIndexedPrimitive(
+          D3DPRIMITIVETYPE PrimitiveType,
+          UINT             MinVertexIndex,
+          UINT             NumVertices,
+          UINT             StartIndex,
+          UINT             PrimitiveCount) {
+    return GetD3D9()->DrawIndexedPrimitive(
+      d3d9::D3DPRIMITIVETYPE(PrimitiveType),
+      m_baseVertexIndex, // set by SetIndices
+      MinVertexIndex,
+      NumVertices,
+      StartIndex,
+      PrimitiveCount);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::DrawPrimitiveUP(
+          D3DPRIMITIVETYPE PrimitiveType,
+          UINT             PrimitiveCount,
+    const void*            pVertexStreamZeroData,
+          UINT             VertexStreamZeroStride) {
+    StateChange();
+
+    // Stream 0 is set to null by this call
+    m_streams[0] = D3D8VBO {nullptr, 0};
+
+    return GetD3D9()->DrawPrimitiveUP(d3d9::D3DPRIMITIVETYPE(PrimitiveType), PrimitiveCount, pVertexStreamZeroData, VertexStreamZeroStride);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::DrawIndexedPrimitiveUP(
+          D3DPRIMITIVETYPE PrimitiveType,
+          UINT             MinVertexIndex,
+          UINT             NumVertices,
+          UINT             PrimitiveCount,
+    const void*            pIndexData,
+          D3DFORMAT        IndexDataFormat,
+    const void*            pVertexStreamZeroData,
+          UINT             VertexStreamZeroStride) {
+    StateChange();
+
+    // Stream 0 and the index buffer are set to null by this call
+    m_streams[0] = D3D8VBO {nullptr, 0};
+    m_indices = nullptr;
+    m_baseVertexIndex = 0;
+    
+    return GetD3D9()->DrawIndexedPrimitiveUP(
+      d3d9::D3DPRIMITIVETYPE(PrimitiveType),
+      MinVertexIndex,
+      NumVertices,
+      PrimitiveCount,
+      pIndexData,
+      d3d9::D3DFORMAT(IndexDataFormat),
+      pVertexStreamZeroData,
+      VertexStreamZeroStride);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::ProcessVertices(
+      UINT                         SrcStartIndex,
+      UINT                         DestIndex,
+      UINT                         VertexCount,
+      IDirect3DVertexBuffer8*      pDestBuffer,
+      DWORD                        Flags) {
+    if (unlikely(!pDestBuffer))
+      return D3DERR_INVALIDCALL;
+    D3D8VertexBuffer* buffer = static_cast<D3D8VertexBuffer*>(pDestBuffer);
+    return GetD3D9()->ProcessVertices(
+      SrcStartIndex,
+      DestIndex,
+      VertexCount,
+      buffer->GetD3D9(),
+      nullptr,
+      Flags
+    );
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetVertexShaderConstant(
+          DWORD StartRegister,
+    const void* pConstantData,
+          DWORD ConstantCount) {
+    StateChange();
+    // ConstantCount is actually the same as Vector4fCount
+    return GetD3D9()->SetVertexShaderConstantF(StartRegister, reinterpret_cast<const float*>(pConstantData), ConstantCount);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetVertexShaderConstant(DWORD Register, void* pConstantData, DWORD ConstantCount) {
+    return GetD3D9()->GetVertexShaderConstantF(Register, (float*)pConstantData, ConstantCount);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetStreamSource(
+          UINT                    StreamNumber,
+          IDirect3DVertexBuffer8* pStreamData,
+          UINT                    Stride) {
+    if (unlikely(StreamNumber >= d8caps::MAX_STREAMS))
+      return D3DERR_INVALIDCALL;
+
+    D3D8VertexBuffer* buffer = static_cast<D3D8VertexBuffer*>(pStreamData);
+
+    if (ShouldBatch())
+      m_batcher->SetStream(StreamNumber, buffer, Stride);
+
+    m_streams[StreamNumber] = D3D8VBO {buffer, Stride};
+
+    // DXVK: Never fails
+    return GetD3D9()->SetStreamSource(StreamNumber, D3D8VertexBuffer::GetD3D9Nullable(buffer), 0, Stride);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetStreamSource(
+          UINT                     StreamNumber,
+          IDirect3DVertexBuffer8** ppStreamData,
+          UINT*                    pStride) {
+    InitReturnPtr(ppStreamData);
+
+    if (likely(pStride != nullptr))
+      *pStride = 0;
+
+    if (unlikely(ppStreamData == nullptr || pStride == nullptr))
+      return D3DERR_INVALIDCALL;
+
+    if (unlikely(StreamNumber >= d8caps::MAX_STREAMS))
+      return D3DERR_INVALIDCALL;
+    
+    const D3D8VBO& vbo = m_streams[StreamNumber];
+
+    *ppStreamData = vbo.buffer.ref();
+    *pStride      = vbo.stride;
+    
+    return D3D_OK;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetIndices(IDirect3DIndexBuffer8* pIndexData, UINT BaseVertexIndex) {
+    if (unlikely(ShouldRecord()))
+      return m_recorder->SetIndices(pIndexData, BaseVertexIndex);
+
+    // used by DrawIndexedPrimitive
+    m_baseVertexIndex = static_cast<INT>(BaseVertexIndex);
+
+    D3D8IndexBuffer* buffer = static_cast<D3D8IndexBuffer*>(pIndexData);
+
+    if (ShouldBatch())
+      m_batcher->SetIndices(buffer, m_baseVertexIndex);
+
+    m_indices = buffer;
+
+    // DXVK: Never fails
+    return GetD3D9()->SetIndices(D3D8IndexBuffer::GetD3D9Nullable(buffer));
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetIndices(
+          IDirect3DIndexBuffer8** ppIndexData,
+          UINT* pBaseVertexIndex) {
+    InitReturnPtr(ppIndexData);
+
+    *ppIndexData      = m_indices.ref();
+    *pBaseVertexIndex = m_baseVertexIndex;
+
+    return D3D_OK;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetPixelShaderConstant(DWORD Register, void* pConstantData, DWORD ConstantCount) {
+    return GetD3D9()->GetPixelShaderConstantF(Register, (float*)pConstantData, ConstantCount);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetPixelShaderConstant(
+          DWORD StartRegister,
+    const void* pConstantData,
+          DWORD ConstantCount) {
+    StateChange();
+    // ConstantCount is actually the same as Vector4fCount
+    return GetD3D9()->SetPixelShaderConstantF(StartRegister, reinterpret_cast<const float*>(pConstantData), ConstantCount);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::DrawRectPatch(
+          UINT               Handle,
+    const float*             pNumSegs,
+    const D3DRECTPATCH_INFO* pRectPatchInfo) {
+    return GetD3D9()->DrawRectPatch(Handle, pNumSegs, reinterpret_cast<const d3d9::D3DRECTPATCH_INFO*>(pRectPatchInfo));
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::DrawTriPatch(
+          UINT              Handle,
+    const float*            pNumSegs,
+    const D3DTRIPATCH_INFO* pTriPatchInfo) {
+    return GetD3D9()->DrawTriPatch(Handle, pNumSegs, reinterpret_cast<const d3d9::D3DTRIPATCH_INFO*>(pTriPatchInfo));
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::DeletePatch(UINT Handle) {
+    return GetD3D9()->DeletePatch(Handle);
+  }
+
+
+  // Render States //
+
+  // ZBIAS can be an integer from 0 to 1 and needs to be remapped to float
+  static constexpr float ZBIAS_SCALE     = -0.000005f;
+  static constexpr float ZBIAS_SCALE_INV = 1 / ZBIAS_SCALE;
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetRenderState(D3DRENDERSTATETYPE State, DWORD Value) {
+    d3d9::D3DRENDERSTATETYPE State9 = (d3d9::D3DRENDERSTATETYPE)State;
+    bool stateChange = true;
+
+    switch (State) {
+      // Most render states translate 1:1 to D3D9
+      default:
+        break;
+
+      // TODO: D3DRS_LINEPATTERN - vkCmdSetLineRasterizationModeEXT
+      case D3DRS_LINEPATTERN: {
+        [[maybe_unused]]
+        D3DLINEPATTERN pattern = bit::cast<D3DLINEPATTERN>(Value);
+        stateChange = false;
+      } break;
+
+      // Not supported by D3D8.
+      case D3DRS_ZVISIBLE:
+        stateChange = false;
+        break;
+
+      // TODO: Not implemented by D9VK. Try anyway.
+      case D3DRS_EDGEANTIALIAS:
+        State9 = d3d9::D3DRS_ANTIALIASEDLINEENABLE;
+        break;
+
+      case D3DRS_ZBIAS:
+        State9 = d3d9::D3DRS_DEPTHBIAS;
+        Value  = bit::cast<DWORD>(float(Value) * ZBIAS_SCALE);
+        break;
+
+      case D3DRS_SOFTWAREVERTEXPROCESSING:
+        // D3D9 can return D3DERR_INVALIDCALL, but we don't care.
+        if (!(m_behaviorFlags & D3DCREATE_MIXED_VERTEXPROCESSING))
+          return D3D_OK;
+
+        // This was a very easy footgun for D3D8 applications.
+        if (unlikely(ShouldRecord()))
+          return m_recorder->SetSoftwareVertexProcessing(Value);
+
+        return GetD3D9()->SetSoftwareVertexProcessing(Value);
+
+      // TODO: D3DRS_PATCHSEGMENTS
+      case D3DRS_PATCHSEGMENTS:
+        stateChange = false;
+        break;
+    }
+
+    if (stateChange) {
+      DWORD value;
+      GetRenderState(State, &value);
+      if (value != Value)
+        StateChange();
+    }
+
+    return GetD3D9()->SetRenderState(State9, Value);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetRenderState(D3DRENDERSTATETYPE State, DWORD* pValue) {
+    d3d9::D3DRENDERSTATETYPE State9 = (d3d9::D3DRENDERSTATETYPE)State;
+
+    switch (State) {
+      // Most render states translate 1:1 to D3D9
+      default:
+        break;
+
+      // TODO: D3DRS_LINEPATTERN
+      case D3DRS_LINEPATTERN:
+        break;
+
+      // Not supported by D3D8.
+      case D3DRS_ZVISIBLE:
+        break;
+
+      case D3DRS_EDGEANTIALIAS:
+        State9 = d3d9::D3DRS_ANTIALIASEDLINEENABLE;
+        break;
+
+      case D3DRS_ZBIAS: {
+        float bias  = 0;
+        HRESULT res = GetD3D9()->GetRenderState(d3d9::D3DRS_DEPTHBIAS, (DWORD*)&bias);
+        *pValue     = bit::cast<DWORD>(bias * ZBIAS_SCALE_INV);
+        return res;
+      } break;
+
+      case D3DRS_SOFTWAREVERTEXPROCESSING:
+        return GetD3D9()->GetSoftwareVertexProcessing();
+
+      // TODO: D3DRS_PATCHSEGMENTS
+      case D3DRS_PATCHSEGMENTS:
+        break;
+    }
+
+    return GetD3D9()->GetRenderState(State9, pValue);
+  }
+
+  // Vertex Shaders //
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::CreateVertexShader(
+        const DWORD* pDeclaration,
+        const DWORD* pFunction,
+              DWORD* pHandle,
+              DWORD Usage ) {
+    D3D8VertexShaderInfo& info = m_vertexShaders.emplace_back();
+
+    // Store D3D8 bytecodes in the shader info
+    if (pDeclaration != nullptr)
+      for (UINT i = 0; pDeclaration[i+1] != D3DVSD_END(); i++)
+        info.declaration.push_back(pDeclaration[i]);
+
+    if (pFunction != nullptr)
+      for (UINT i = 0; pFunction[i+1] != D3DVS_END(); i++)
+        info.function.push_back(pFunction[i]);
+    
+    D3D9VertexShaderCode result = TranslateVertexShader8(pDeclaration, pFunction, m_d3d8Options);
+
+    // Create vertex declaration
+    HRESULT res = GetD3D9()->CreateVertexDeclaration(result.declaration, &(info.pVertexDecl));
+    if (FAILED(res))
+      return res;
+
+    if (pFunction != nullptr) {
+      res = GetD3D9()->CreateVertexShader(result.function.data(), &(info.pVertexShader));
+    } else {
+      // pFunction is NULL: fixed function pipeline
+      info.pVertexShader = nullptr;
+    }
+
+    // Set bit to indicate this is not an FVF
+    *pHandle = getShaderHandle(m_vertexShaders.size() - 1);
+
+    return res;
+  }
+
+  inline D3D8VertexShaderInfo* getVertexShaderInfo(D3D8Device* device, DWORD Handle) {
+    
+    Handle = getShaderIndex(Handle);
+
+    if (unlikely(Handle >= device->m_vertexShaders.size())) {
+      Logger::debug(str::format("getVertexShaderInfo: Invalid vertex shader index ", std::hex, Handle));
+      return nullptr;
+    }
+
+    D3D8VertexShaderInfo& info = device->m_vertexShaders[Handle];
+
+    if (unlikely(!info.pVertexDecl && !info.pVertexShader)) {
+      Logger::debug(str::format("getVertexShaderInfo: Application provided deleted vertex shader ", std::hex, Handle));
+      return nullptr;
+    }
+
+    return &info;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetVertexShader( DWORD Handle ) {
+    if (unlikely(ShouldRecord())) {
+      return m_recorder->SetVertexShader(Handle);
+    }
+
+    // Check for extra bit that indicates this is not an FVF
+    if (!isFVF(Handle)) {
+
+      D3D8VertexShaderInfo* info = getVertexShaderInfo(this, Handle);
+
+      if (!info)
+        return D3DERR_INVALIDCALL;
+
+      StateChange();
+
+      // Cache current shader
+      m_currentVertexShader = Handle;
+      
+      GetD3D9()->SetVertexDeclaration(info->pVertexDecl);
+      return GetD3D9()->SetVertexShader(info->pVertexShader);
+
+    } else if (m_currentVertexShader != Handle) {
+      StateChange();
+
+      // Cache current FVF
+      m_currentVertexShader = Handle;
+
+      //GetD3D9()->SetVertexDeclaration(nullptr);
+      GetD3D9()->SetVertexShader(nullptr);
+      return GetD3D9()->SetFVF( Handle );
+    }
+    return D3D_OK;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetVertexShader(DWORD* pHandle) {
+    // Return cached shader
+    *pHandle = m_currentVertexShader;
+
+    return D3D_OK;
+
+    /*
+    // Slow path. Use to debug cached shader validation. //
+    
+    d3d9::IDirect3DVertexShader9* pVertexShader;
+    HRESULT res = GetD3D9()->GetVertexShader(&pVertexShader);
+
+    if (FAILED(res) || pVertexShader == nullptr) {
+      return GetD3D9()->GetFVF(pHandle);
+    }
+
+    for (unsigned int i = 0; i < m_vertexShaders.size(); i++) {
+      D3D8VertexShaderInfo& info = m_vertexShaders[i];
+
+      if (info.pVertexShader == pVertexShader) {
+        *pHandle = getShaderHandle(DWORD(i));
+        return res;
+      }
+    }
+
+    return res;
+    */
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::DeleteVertexShader(DWORD Handle) {
+    if (!isFVF(Handle)) {
+      D3D8VertexShaderInfo* info = getVertexShaderInfo(this, Handle);
+
+      if (!info)
+        return D3DERR_INVALIDCALL;
+
+      if (info->pVertexDecl)
+        info->pVertexDecl->Release();
+      if (info->pVertexShader)
+        info->pVertexShader->Release();
+
+      info->declaration.clear();
+      info->function.clear();
+    }
+
+    return D3D_OK;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetVertexShaderDeclaration(DWORD Handle, void* pData, DWORD* pSizeOfData) {
+    D3D8VertexShaderInfo* pInfo = getVertexShaderInfo(this, Handle);
+
+    if (unlikely(!pInfo))
+      return D3DERR_INVALIDCALL;
+    
+    UINT SizeOfData = *pSizeOfData;
+    
+    // Get actual size
+    UINT ActualSize = pInfo->declaration.size() * sizeof(DWORD);
+    
+    if (pData == nullptr) {
+      *pSizeOfData = ActualSize;
+      return D3D_OK;
+    }
+
+    // D3D8-specific behavior
+    if (SizeOfData < ActualSize) {
+      *pSizeOfData = ActualSize;
+      return D3DERR_MOREDATA;
+    }
+
+    memcpy(pData, pInfo->declaration.data(), ActualSize);
+    return D3D_OK;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetVertexShaderFunction(DWORD Handle, void* pData, DWORD* pSizeOfData) {
+    D3D8VertexShaderInfo* pInfo = getVertexShaderInfo(this, Handle);
+
+    if (unlikely(!pInfo))
+      return D3DERR_INVALIDCALL;
+    
+    UINT SizeOfData = *pSizeOfData;
+    
+    // Get actual size
+    UINT ActualSize = pInfo->function.size() * sizeof(DWORD);
+    
+    if (pData == nullptr) {
+      *pSizeOfData = ActualSize;
+      return D3D_OK;
+    }
+
+    // D3D8-specific behavior
+    if (SizeOfData < ActualSize) {
+      *pSizeOfData = ActualSize;
+      return D3DERR_MOREDATA;
+    }
+
+    memcpy(pData, pInfo->function.data(), ActualSize);
+    return D3D_OK;
+
+  }
+
+  // Pixel Shaders //
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::CreatePixelShader(
+        const DWORD* pFunction,
+              DWORD* pHandle) {
+    d3d9::IDirect3DPixelShader9* pPixelShader;
+    
+    HRESULT res = GetD3D9()->CreatePixelShader(pFunction, &pPixelShader);
+
+    m_pixelShaders.push_back(pPixelShader);
+
+    // Still set the shader bit, to prevent conflicts with NULL.
+    *pHandle = getShaderHandle(m_pixelShaders.size() - 1);
+
+    return res;
+  }
+
+  inline d3d9::IDirect3DPixelShader9* getPixelShaderPtr(D3D8Device* device, DWORD Handle) {
+    Handle = getShaderIndex(Handle);
+
+    if (unlikely(Handle >= device->m_pixelShaders.size())) {
+      Logger::debug(str::format("getPixelShaderPtr: Invalid pixel shader index ", std::hex, Handle));
+      return nullptr;
+    }
+
+    d3d9::IDirect3DPixelShader9* pPixelShader = device->m_pixelShaders[Handle];
+
+    if (unlikely(pPixelShader == nullptr)) {
+      Logger::debug(str::format("getPixelShaderPtr: Application provided deleted pixel shader ", std::hex, Handle));
+      return nullptr;
+    }
+
+    return pPixelShader;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::SetPixelShader(DWORD Handle) {
+    if (unlikely(ShouldRecord())) {
+      return m_recorder->SetPixelShader(Handle);
+    }
+
+    if (Handle == DWORD(NULL)) {
+      StateChange();
+      m_currentPixelShader = DWORD(NULL);
+      return GetD3D9()->SetPixelShader(nullptr);
+    }
+
+    d3d9::IDirect3DPixelShader9* pPixelShader = getPixelShaderPtr(this, Handle);
+
+    if (unlikely(!pPixelShader)) {
+      return D3DERR_INVALIDCALL;
+    }
+
+    StateChange();
+
+    // Cache current pixel shader
+    m_currentPixelShader = Handle;
+
+    return GetD3D9()->SetPixelShader(pPixelShader);
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetPixelShader(DWORD* pHandle) {
+    // Return cached shader
+    *pHandle = m_currentPixelShader;
+
+    return D3D_OK;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::DeletePixelShader(DWORD Handle) {
+    d3d9::IDirect3DPixelShader9* pPixelShader = getPixelShaderPtr(this, Handle);
+
+    if (unlikely(!pPixelShader)) {
+      return D3DERR_INVALIDCALL;
+    }
+
+    pPixelShader->Release();
+
+    m_pixelShaders[getShaderIndex(Handle)] = nullptr;
+
+    return D3D_OK;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Device::GetPixelShaderFunction(DWORD Handle, void* pData, DWORD* pSizeOfData) {
+    d3d9::IDirect3DPixelShader9* pPixelShader = getPixelShaderPtr(this, Handle);
+
+    if (unlikely(!pPixelShader))
+      return D3DERR_INVALIDCALL;
+
+    UINT SizeOfData = *pSizeOfData;
+    
+    // Get actual size
+    UINT ActualSize = 0;
+    pPixelShader->GetFunction(nullptr, &ActualSize);
+    
+    if (pData == nullptr) {
+      *pSizeOfData = ActualSize;
+      return D3D_OK;
+    }
+
+    // D3D8-specific behavior
+    if (SizeOfData < ActualSize) {
+      *pSizeOfData = ActualSize;
+      return D3DERR_MOREDATA;
+    }
+
+    return pPixelShader->GetFunction(pData, &SizeOfData);
+  }
+
+}
diff -pruN 2.3.1-1/src/d3d8/d3d8_device.h 2.4-1/src/d3d8/d3d8_device.h
--- 2.3.1-1/src/d3d8/d3d8_device.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_device.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,456 @@
+#pragma once
+
+#include "d3d8_include.h"
+#include "d3d8_texture.h"
+#include "d3d8_buffer.h"
+#include "d3d8_swapchain.h"
+#include "d3d8_state_block.h"
+#include "d3d8_d3d9_util.h"
+#include "d3d8_caps.h"
+#include "d3d8_batch.h"
+
+#include "../d3d9/d3d9_bridge.h"
+
+#include <array>
+#include <vector>
+#include <type_traits>
+#include <unordered_map>
+
+namespace dxvk {
+
+  class D3D8Interface;
+
+  struct D3D8VertexShaderInfo;
+
+  using D3D8DeviceBase = D3D8WrappedObject<d3d9::IDirect3DDevice9, IDirect3DDevice8>;
+  class D3D8Device final : public D3D8DeviceBase {
+
+    friend class D3D8StateBlock;
+  public:
+
+    D3D8Device(
+      D3D8Interface*                pParent,
+      Com<d3d9::IDirect3DDevice9>&& pDevice,
+      D3DDEVTYPE                    DeviceType,
+      HWND                          hFocusWindow,
+      DWORD                         BehaviorFlags,
+      D3DPRESENT_PARAMETERS*        pParams);
+
+    ~D3D8Device();
+      
+    HRESULT STDMETHODCALLTYPE TestCooperativeLevel();
+
+    UINT    STDMETHODCALLTYPE GetAvailableTextureMem();
+
+    HRESULT STDMETHODCALLTYPE ResourceManagerDiscardBytes(DWORD bytes);
+
+    HRESULT STDMETHODCALLTYPE GetDirect3D(IDirect3D8** ppD3D8);
+
+    HRESULT STDMETHODCALLTYPE GetDeviceCaps(D3DCAPS8* pCaps);
+
+    HRESULT STDMETHODCALLTYPE GetDisplayMode(D3DDISPLAYMODE* pMode);
+
+    HRESULT STDMETHODCALLTYPE GetCreationParameters(D3DDEVICE_CREATION_PARAMETERS* pParameters);
+
+    HRESULT STDMETHODCALLTYPE SetCursorProperties(
+      UINT               XHotSpot,
+      UINT               YHotSpot,
+      IDirect3DSurface8* pCursorBitmap);
+
+    void    STDMETHODCALLTYPE SetCursorPosition(UINT XScreenSpace, UINT YScreenSpace, DWORD Flags);
+
+    // Microsoft d3d8.h in the DirectX 9 SDK uses a different function signature...
+    void    STDMETHODCALLTYPE SetCursorPosition(int X, int Y, DWORD Flags);
+
+    BOOL    STDMETHODCALLTYPE ShowCursor(BOOL bShow);
+
+    HRESULT STDMETHODCALLTYPE CreateAdditionalSwapChain(
+        D3DPRESENT_PARAMETERS* pPresentationParameters,
+        IDirect3DSwapChain8** ppSwapChain);
+
+    HRESULT STDMETHODCALLTYPE Reset(D3DPRESENT_PARAMETERS* pPresentationParameters);
+
+    HRESULT STDMETHODCALLTYPE Present(
+      const RECT* pSourceRect,
+      const RECT* pDestRect,
+            HWND hDestWindowOverride,
+      const RGNDATA* pDirtyRegion);
+
+    HRESULT STDMETHODCALLTYPE GetBackBuffer(
+            UINT iBackBuffer,
+            D3DBACKBUFFER_TYPE Type,
+            IDirect3DSurface8** ppBackBuffer);
+
+    HRESULT STDMETHODCALLTYPE GetRasterStatus(D3DRASTER_STATUS* pRasterStatus);
+
+    void STDMETHODCALLTYPE SetGammaRamp(DWORD Flags, const D3DGAMMARAMP* pRamp);
+
+    void STDMETHODCALLTYPE GetGammaRamp(D3DGAMMARAMP* pRamp);
+
+    HRESULT STDMETHODCALLTYPE CreateTexture(
+            UINT                Width,
+            UINT                Height,
+            UINT                Levels,
+            DWORD               Usage,
+            D3DFORMAT           Format,
+            D3DPOOL             Pool,
+            IDirect3DTexture8** ppTexture);
+
+    HRESULT STDMETHODCALLTYPE CreateVolumeTexture(
+            UINT                      Width,
+            UINT                      Height,
+            UINT                      Depth,
+            UINT                      Levels,
+            DWORD                     Usage,
+            D3DFORMAT                 Format,
+            D3DPOOL                   Pool,
+            IDirect3DVolumeTexture8** ppVolumeTexture);
+
+    HRESULT STDMETHODCALLTYPE CreateCubeTexture(
+          UINT                      EdgeLength,
+            UINT                    Levels,
+            DWORD                   Usage,
+            D3DFORMAT               Format,
+            D3DPOOL                 Pool,
+            IDirect3DCubeTexture8** ppCubeTexture);
+
+    HRESULT STDMETHODCALLTYPE CreateVertexBuffer(
+            UINT                     Length,
+            DWORD                    Usage,
+            DWORD                    FVF,
+            D3DPOOL                  Pool,
+            IDirect3DVertexBuffer8** ppVertexBuffer);
+
+    HRESULT STDMETHODCALLTYPE CreateIndexBuffer(
+            UINT                    Length,
+            DWORD                   Usage,
+            D3DFORMAT               Format,
+            D3DPOOL                 Pool,
+            IDirect3DIndexBuffer8** ppIndexBuffer);
+
+    HRESULT STDMETHODCALLTYPE CreateRenderTarget(
+            UINT                Width,
+            UINT                Height,
+            D3DFORMAT           Format,
+            D3DMULTISAMPLE_TYPE MultiSample,
+            BOOL                Lockable,
+            IDirect3DSurface8** ppSurface);
+
+    HRESULT STDMETHODCALLTYPE CreateDepthStencilSurface(
+            UINT                Width,
+            UINT                Height,
+            D3DFORMAT           Format,
+            D3DMULTISAMPLE_TYPE MultiSample,
+            IDirect3DSurface8** ppSurface);
+
+    HRESULT STDMETHODCALLTYPE CreateImageSurface(UINT Width, UINT Height, D3DFORMAT Format, IDirect3DSurface8** ppSurface);
+
+    HRESULT STDMETHODCALLTYPE CopyRects(
+            IDirect3DSurface8*  pSourceSurface,
+      const RECT*               pSourceRectsArray,
+            UINT                cRects,
+            IDirect3DSurface8*  pDestinationSurface,
+      const POINT*              pDestPointsArray);
+
+    HRESULT STDMETHODCALLTYPE UpdateTexture(
+            IDirect3DBaseTexture8* pSourceTexture,
+            IDirect3DBaseTexture8* pDestinationTexture);
+
+    HRESULT STDMETHODCALLTYPE GetFrontBuffer(IDirect3DSurface8* pDestSurface);
+
+    HRESULT STDMETHODCALLTYPE SetRenderTarget(IDirect3DSurface8* pRenderTarget, IDirect3DSurface8* pNewZStencil);
+
+    HRESULT STDMETHODCALLTYPE GetRenderTarget(IDirect3DSurface8** ppRenderTarget);
+
+    HRESULT STDMETHODCALLTYPE GetDepthStencilSurface(IDirect3DSurface8** ppZStencilSurface);
+
+    HRESULT STDMETHODCALLTYPE BeginScene();
+
+    HRESULT STDMETHODCALLTYPE EndScene();
+
+    HRESULT STDMETHODCALLTYPE Clear(
+            DWORD    Count,
+      const D3DRECT* pRects,
+            DWORD    Flags,
+            D3DCOLOR Color,
+            float    Z,
+            DWORD    Stencil);
+
+    HRESULT STDMETHODCALLTYPE SetTransform(D3DTRANSFORMSTATETYPE State, const D3DMATRIX* pMatrix);
+
+    HRESULT STDMETHODCALLTYPE GetTransform(D3DTRANSFORMSTATETYPE State, D3DMATRIX* pMatrix);
+
+    HRESULT STDMETHODCALLTYPE MultiplyTransform(D3DTRANSFORMSTATETYPE TransformState, const D3DMATRIX* pMatrix);
+
+    HRESULT STDMETHODCALLTYPE SetViewport(const D3DVIEWPORT8* pViewport);
+
+    HRESULT STDMETHODCALLTYPE GetViewport(D3DVIEWPORT8* pViewport);
+
+    HRESULT STDMETHODCALLTYPE SetMaterial(const D3DMATERIAL8* pMaterial);
+
+    HRESULT STDMETHODCALLTYPE GetMaterial(D3DMATERIAL8* pMaterial);
+
+    HRESULT STDMETHODCALLTYPE SetLight(DWORD Index, const D3DLIGHT8* pLight);
+
+    HRESULT STDMETHODCALLTYPE GetLight(DWORD Index, D3DLIGHT8* pLight);
+
+    HRESULT STDMETHODCALLTYPE LightEnable(DWORD Index, BOOL Enable);
+
+    HRESULT STDMETHODCALLTYPE GetLightEnable(DWORD Index, BOOL* pEnable);
+
+    HRESULT STDMETHODCALLTYPE SetClipPlane(DWORD Index, const float* pPlane);
+
+    HRESULT STDMETHODCALLTYPE GetClipPlane(DWORD Index, float* pPlane);
+
+    HRESULT STDMETHODCALLTYPE SetRenderState(D3DRENDERSTATETYPE State, DWORD Value);
+
+    HRESULT STDMETHODCALLTYPE GetRenderState(D3DRENDERSTATETYPE State, DWORD* pValue);
+    
+    HRESULT STDMETHODCALLTYPE CreateStateBlock(
+            D3DSTATEBLOCKTYPE     Type,
+            DWORD*                pToken);
+
+    HRESULT STDMETHODCALLTYPE CaptureStateBlock(DWORD Token);
+
+    HRESULT STDMETHODCALLTYPE ApplyStateBlock(DWORD Token);
+
+    HRESULT STDMETHODCALLTYPE DeleteStateBlock(DWORD Token);
+
+    HRESULT STDMETHODCALLTYPE BeginStateBlock();
+
+    HRESULT STDMETHODCALLTYPE EndStateBlock(DWORD* pToken);
+
+    HRESULT STDMETHODCALLTYPE SetClipStatus(const D3DCLIPSTATUS8* pClipStatus);
+
+    HRESULT STDMETHODCALLTYPE GetClipStatus(D3DCLIPSTATUS8* pClipStatus);
+
+    HRESULT STDMETHODCALLTYPE GetTexture(DWORD Stage, IDirect3DBaseTexture8** ppTexture);
+
+    HRESULT STDMETHODCALLTYPE SetTexture(DWORD Stage, IDirect3DBaseTexture8* pTexture);
+
+    HRESULT STDMETHODCALLTYPE GetTextureStageState(
+            DWORD                    Stage,
+            D3DTEXTURESTAGESTATETYPE Type,
+            DWORD*                   pValue);
+
+    HRESULT STDMETHODCALLTYPE SetTextureStageState(
+            DWORD                    Stage,
+            D3DTEXTURESTAGESTATETYPE Type,
+            DWORD                    Value);
+
+    HRESULT STDMETHODCALLTYPE ValidateDevice(DWORD* pNumPasses);
+
+    HRESULT STDMETHODCALLTYPE GetInfo(DWORD DevInfoID, void* pDevInfoStruct, DWORD DevInfoStructSize);
+
+    HRESULT STDMETHODCALLTYPE SetPaletteEntries(UINT PaletteNumber, const PALETTEENTRY* pEntries);
+
+    HRESULT STDMETHODCALLTYPE GetPaletteEntries(UINT PaletteNumber, PALETTEENTRY* pEntries);
+
+    HRESULT STDMETHODCALLTYPE SetCurrentTexturePalette(UINT PaletteNumber);
+
+    HRESULT STDMETHODCALLTYPE GetCurrentTexturePalette(UINT* PaletteNumber);
+
+    HRESULT STDMETHODCALLTYPE DrawPrimitive(
+            D3DPRIMITIVETYPE PrimitiveType,
+            UINT             StartVertex,
+            UINT             PrimitiveCount);
+
+    HRESULT STDMETHODCALLTYPE DrawIndexedPrimitive(
+            D3DPRIMITIVETYPE PrimitiveType,
+            UINT             MinVertexIndex,
+            UINT             NumVertices,
+            UINT             StartIndex,
+            UINT             PrimitiveCount);
+
+    HRESULT STDMETHODCALLTYPE DrawPrimitiveUP(
+            D3DPRIMITIVETYPE PrimitiveType,
+            UINT             PrimitiveCount,
+      const void*            pVertexStreamZeroData,
+            UINT             VertexStreamZeroStride);
+
+    HRESULT STDMETHODCALLTYPE DrawIndexedPrimitiveUP(
+            D3DPRIMITIVETYPE PrimitiveType,
+            UINT             MinVertexIndex,
+            UINT             NumVertices,
+            UINT             PrimitiveCount,
+      const void*            pIndexData,
+            D3DFORMAT        IndexDataFormat,
+      const void*            pVertexStreamZeroData,
+            UINT             VertexStreamZeroStride);
+
+    HRESULT STDMETHODCALLTYPE ProcessVertices(
+        UINT                         SrcStartIndex,
+        UINT                         DestIndex,
+        UINT                         VertexCount,
+        IDirect3DVertexBuffer8*      pDestBuffer,
+        DWORD                        Flags);
+
+    HRESULT STDMETHODCALLTYPE CreateVertexShader(
+      const DWORD*  pDeclaration,
+      const DWORD*  pFunction,
+            DWORD*  pHandle,
+            DWORD   Usage);
+
+    HRESULT STDMETHODCALLTYPE SetVertexShader(DWORD Handle);
+
+    HRESULT STDMETHODCALLTYPE GetVertexShader(DWORD* pHandle);
+
+    HRESULT STDMETHODCALLTYPE DeleteVertexShader(DWORD Handle);
+
+    HRESULT STDMETHODCALLTYPE SetVertexShaderConstant(
+            DWORD StartRegister,
+      const void* pConstantData,
+            DWORD ConstantCount);
+
+    HRESULT STDMETHODCALLTYPE GetVertexShaderConstant(DWORD Register, void* pConstantData, DWORD ConstantCount);
+
+    HRESULT STDMETHODCALLTYPE GetVertexShaderDeclaration(DWORD Handle, void* pData, DWORD* pSizeOfData);
+    
+    HRESULT STDMETHODCALLTYPE GetVertexShaderFunction(DWORD Handle, void* pData, DWORD* pSizeOfData);
+
+    HRESULT STDMETHODCALLTYPE SetStreamSource(
+            UINT                    StreamNumber,
+            IDirect3DVertexBuffer8* pStreamData,
+            UINT                    Stride);
+
+    HRESULT STDMETHODCALLTYPE GetStreamSource(
+            UINT                     StreamNumber,
+            IDirect3DVertexBuffer8** ppStreamData,
+            UINT*                    pStride);
+
+    HRESULT STDMETHODCALLTYPE SetIndices(IDirect3DIndexBuffer8* pIndexData, UINT BaseVertexIndex);
+
+    HRESULT STDMETHODCALLTYPE GetIndices(
+            IDirect3DIndexBuffer8** ppIndexData,
+            UINT* pBaseVertexIndex);
+
+    HRESULT STDMETHODCALLTYPE CreatePixelShader(
+      const DWORD* pFunction, 
+            DWORD* pHandle);
+
+    HRESULT STDMETHODCALLTYPE SetPixelShader(DWORD Handle);
+
+    HRESULT STDMETHODCALLTYPE GetPixelShader(DWORD* pHandle);
+
+    HRESULT STDMETHODCALLTYPE DeletePixelShader(THIS_ DWORD Handle);
+
+    HRESULT STDMETHODCALLTYPE GetPixelShaderConstant(DWORD Register, void* pConstantData, DWORD ConstantCount);
+
+    HRESULT STDMETHODCALLTYPE SetPixelShaderConstant(
+            DWORD StartRegister,
+      const void* pConstantData,
+            DWORD ConstantCount);
+
+    HRESULT STDMETHODCALLTYPE GetPixelShaderFunction(DWORD Handle, void* pData, DWORD* pSizeOfData);
+
+    HRESULT STDMETHODCALLTYPE DrawRectPatch(
+            UINT               Handle,
+      const float*             pNumSegs,
+      const D3DRECTPATCH_INFO* pRectPatchInfo);
+
+    HRESULT STDMETHODCALLTYPE DrawTriPatch(
+            UINT              Handle,
+      const float*            pNumSegs,
+      const D3DTRIPATCH_INFO* pTriPatchInfo);
+
+    HRESULT STDMETHODCALLTYPE DeletePatch(UINT Handle);
+
+  public: // Internal Methods //
+
+    inline bool ShouldRecord() { return m_recorder != nullptr; }
+    inline bool ShouldBatch()  { return m_batcher  != nullptr; }
+
+    /**
+     * Marks any state change in the device, so we can signal
+     * the batcher to emit draw calls. StateChange should be
+     * called immediately before changing any D3D9 state.
+     */
+    inline void StateChange() {
+      if (ShouldBatch())
+        m_batcher->StateChange();
+    }
+
+    inline void ResetState() {
+      // Mirrors how D3D9 handles the BackBufferCount
+      m_presentParams.BackBufferCount = std::max(m_presentParams.BackBufferCount, 1u);
+
+      // Purge cached objects
+      // TODO: Some functions may need to be called here (e.g. SetTexture, etc.)
+      // in case Reset can be recorded by state blocks and other things.
+      m_textures.fill(nullptr);
+      m_streams.fill(D3D8VBO());
+      m_indices = nullptr;
+      m_renderTarget = nullptr;
+      m_depthStencil = nullptr;
+
+      m_backBuffers.clear();
+      m_backBuffers.resize(m_presentParams.BackBufferCount);
+      
+      m_autoDepthStencil = nullptr;
+    }
+
+    inline void RecreateBackBuffersAndAutoDepthStencil() {
+      for (UINT i = 0; i < m_presentParams.BackBufferCount; i++) {
+        Com<d3d9::IDirect3DSurface9> pSurface9;
+        GetD3D9()->GetBackBuffer(0, i, d3d9::D3DBACKBUFFER_TYPE_MONO, &pSurface9);
+        m_backBuffers[i] = new D3D8Surface(this, std::move(pSurface9));
+      }
+
+      Com<d3d9::IDirect3DSurface9> pStencil9 = nullptr;
+      GetD3D9()->GetDepthStencilSurface(&pStencil9);
+      m_autoDepthStencil = new D3D8Surface(this, std::move(pStencil9));
+
+      m_renderTarget = m_backBuffers[0];
+      m_depthStencil = m_autoDepthStencil;
+    }
+
+    friend d3d9::IDirect3DPixelShader9* getPixelShaderPtr(D3D8Device* device, DWORD Handle);
+    friend D3D8VertexShaderInfo*        getVertexShaderInfo(D3D8Device* device, DWORD Handle);
+
+  private:
+
+    Com<IDxvkD3D8Bridge>  m_bridge;
+    const D3D8Options&    m_d3d8Options;
+
+    Com<D3D8Interface>    m_parent;
+
+    D3DPRESENT_PARAMETERS m_presentParams;
+
+    D3D8StateBlock*                            m_recorder = nullptr;
+    DWORD                                      m_recorderToken = 0;
+    DWORD                                      m_token    = 0;
+    std::unordered_map<DWORD, D3D8StateBlock>  m_stateBlocks;
+    D3D8Batcher*                               m_batcher  = nullptr;
+
+    struct D3D8VBO {
+      Com<D3D8VertexBuffer, false>   buffer = nullptr;
+      UINT                           stride = 0;
+    };
+    
+    // Remember to fill() these in the constructor!
+    std::array<Com<D3D8Texture2D, false>, d8caps::MAX_TEXTURE_STAGES>  m_textures;
+    std::array<D3D8VBO, d8caps::MAX_STREAMS>                           m_streams;
+
+    Com<D3D8IndexBuffer, false>        m_indices;
+    INT                                m_baseVertexIndex = 0;
+
+    // TODO: Which of these should be a private ref
+    std::vector<Com<D3D8Surface, false>> m_backBuffers;
+    Com<D3D8Surface, false>              m_autoDepthStencil;
+
+    Com<D3D8Surface, false>     m_renderTarget;
+    Com<D3D8Surface, false>     m_depthStencil;
+
+    std::vector<D3D8VertexShaderInfo>           m_vertexShaders;
+    std::vector<d3d9::IDirect3DPixelShader9*>   m_pixelShaders;
+    DWORD                                       m_currentVertexShader  = 0; // can be FVF or vs index (marked by D3DFVF_RESERVED0)
+    DWORD                                       m_currentPixelShader   = 0;
+
+    D3DDEVTYPE            m_deviceType;
+    HWND                  m_window;
+
+    DWORD                 m_behaviorFlags;
+
+  };
+
+}
diff -pruN 2.3.1-1/src/d3d8/d3d8_device_child.h 2.4-1/src/d3d8/d3d8_device_child.h
--- 2.3.1-1/src/d3d8/d3d8_device_child.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_device_child.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,71 @@
+#pragma once
+
+// Common methods for device-tied objects.
+// - AddRef, Release from IUnknown
+// - GetDevice from various classes including IDirect3DResource8
+
+#include "d3d8_include.h"
+#include "d3d8_wrapped_object.h"
+
+namespace dxvk {
+
+  class D3D8Device;
+
+  template <typename D3D9, typename D3D8>
+  class D3D8DeviceChild : public D3D8WrappedObject<D3D9, D3D8> {
+
+  public:
+
+    D3D8DeviceChild(D3D8Device* pDevice, Com<D3D9>&& Object)
+      : D3D8WrappedObject<D3D9, D3D8>(std::move(Object))
+      , m_parent( pDevice ) { }
+
+    ULONG STDMETHODCALLTYPE AddRef() {
+      uint32_t refCount = this->m_refCount++;
+      if (unlikely(!refCount)) {
+        this->AddRefPrivate();
+        GetDevice()->AddRef();
+      }
+
+      return refCount + 1;
+    }
+    
+    ULONG STDMETHODCALLTYPE Release() {
+      // ignore Release calls on objects with 0 refCount
+      if(unlikely(!this->m_refCount))
+        return this->m_refCount;
+
+      uint32_t refCount = --this->m_refCount;
+      if (unlikely(!refCount)) {
+        auto* pDevice = GetDevice();
+        this->ReleasePrivate();
+        pDevice->Release();
+      }
+      return refCount;
+    }
+
+    HRESULT STDMETHODCALLTYPE GetDevice(IDirect3DDevice8** ppDevice) {
+      InitReturnPtr(ppDevice);
+
+      if (ppDevice == nullptr)
+        return D3DERR_INVALIDCALL;
+
+      *ppDevice = ref(GetDevice());
+      return D3D_OK;
+    }
+
+    IDirect3DDevice8* GetDevice() {
+      return reinterpret_cast<IDirect3DDevice8*>(m_parent);
+    }
+
+    D3D8Device* GetParent() {
+      return m_parent;
+    }
+
+  protected:
+
+    D3D8Device* m_parent;
+
+  };
+
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/d3d8_format.h 2.4-1/src/d3d8/d3d8_format.h
--- 2.3.1-1/src/d3d8/d3d8_format.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_format.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,220 @@
+#pragma once
+
+#include "d3d8_include.h"
+
+namespace dxvk {
+  constexpr bool isDXT(D3DFORMAT fmt) {
+    return fmt == D3DFMT_DXT1
+        || fmt == D3DFMT_DXT2
+        || fmt == D3DFMT_DXT3
+        || fmt == D3DFMT_DXT4
+        || fmt == D3DFMT_DXT5;
+  }
+
+  constexpr bool isDXT(d3d9::D3DFORMAT fmt) {
+    return isDXT(D3DFORMAT(fmt));
+  }
+
+  constexpr bool isUnsupportedSurfaceFormat(D3DFORMAT fmt) {
+    // mirror what dxvk doesn't support in terms of d3d9 surface formats
+    return fmt == D3DFMT_R8G8B8
+        || fmt == D3DFMT_R3G3B2
+        || fmt == D3DFMT_A8R3G3B2
+        || fmt == D3DFMT_A8P8
+        || fmt == D3DFMT_P8;
+        // not included in the d3d8 spec
+        //|| fmt == D3DFMT_CXV8U8;
+  }
+
+  constexpr bool isSupportedDepthStencilFormat(D3DFORMAT fmt) {
+    // native d3d8 doesn't support D3DFMT_D32, D3DFMT_D15S1 or D3DFMT_D24X4S4
+    return fmt == D3DFMT_D16_LOCKABLE
+        || fmt == D3DFMT_D16
+        //|| fmt == D3DFMT_D32
+        //|| fmt == D3DFMT_D15S1
+        //|| fmt == D3DFMT_D24X4S4
+        || fmt == D3DFMT_D24S8
+        || fmt == D3DFMT_D24X8;
+  }
+
+  constexpr bool isDepthStencilFormat(D3DFORMAT fmt) {
+    return fmt == D3DFMT_D16_LOCKABLE
+        || fmt == D3DFMT_D16
+        || fmt == D3DFMT_D32
+        || fmt == D3DFMT_D15S1
+        || fmt == D3DFMT_D24X4S4
+        || fmt == D3DFMT_D24S8
+        || fmt == D3DFMT_D24X8;
+  }
+
+  // Get bytes per pixel (or 4x4 block for DXT)
+  constexpr UINT getFormatStride(D3DFORMAT fmt) {
+    switch (fmt) {
+      default:
+      case D3DFMT_UNKNOWN:
+        return 0;
+      case D3DFMT_R3G3B2:
+      case D3DFMT_A8:
+      case D3DFMT_P8:
+      case D3DFMT_L8:
+      case D3DFMT_A4L4:
+        return 1;
+      case D3DFMT_R5G6B5:
+      case D3DFMT_X1R5G5B5:
+      case D3DFMT_A1R5G5B5:
+      case D3DFMT_A4R4G4B4:
+      case D3DFMT_A8R3G3B2:
+      case D3DFMT_X4R4G4B4:
+      case D3DFMT_A8P8:
+      case D3DFMT_A8L8:
+      case D3DFMT_V8U8:
+      case D3DFMT_L6V5U5:
+      case D3DFMT_D16_LOCKABLE:
+      case D3DFMT_D15S1:
+      case D3DFMT_D16:
+      case D3DFMT_UYVY:
+      case D3DFMT_YUY2:
+        return 2;
+      case D3DFMT_R8G8B8:
+        return 3;
+      case D3DFMT_A8R8G8B8:
+      case D3DFMT_X8R8G8B8:
+      case D3DFMT_A2B10G10R10:
+      //case D3DFMT_A8B8G8R8:
+      //case D3DFMT_X8B8G8R8:
+      case D3DFMT_G16R16:
+      case D3DFMT_X8L8V8U8:
+      case D3DFMT_Q8W8V8U8:
+      case D3DFMT_V16U16:
+      case D3DFMT_W11V11U10:
+      case D3DFMT_A2W10V10U10:
+      case D3DFMT_D32:
+      case D3DFMT_D24S8:
+      case D3DFMT_D24X8:
+      case D3DFMT_D24X4S4:
+        return 4;
+      case D3DFMT_DXT1:
+        return 8;
+      case D3DFMT_DXT2:
+      case D3DFMT_DXT3:
+      case D3DFMT_DXT4:
+      case D3DFMT_DXT5:
+        return 16;
+    }
+  }
+
+  constexpr uint32_t GetVertexCount8(D3DPRIMITIVETYPE type, UINT count) {
+    switch (type) {
+      default:
+      case D3DPT_TRIANGLELIST:  return count * 3;
+      case D3DPT_POINTLIST:     return count;
+      case D3DPT_LINELIST:      return count * 2;
+      case D3DPT_LINESTRIP:     return count + 1;
+      case D3DPT_TRIANGLESTRIP: return count + 2;
+      case D3DPT_TRIANGLEFAN:   return count + 2;
+    }
+  }
+
+  // Essentially the same logic as D3D9VertexDecl::SetFVF
+  constexpr UINT GetFVFStride(DWORD FVF) {
+    uint32_t texCount = 0;
+
+    uint32_t betas = 0;
+    uint8_t betaIdx = 0xFF;
+
+    UINT size = 0;
+
+    switch (FVF & D3DFVF_POSITION_MASK) {
+      case D3DFVF_XYZ:
+      case D3DFVF_XYZB1:
+      case D3DFVF_XYZB2:
+      case D3DFVF_XYZB3:
+      case D3DFVF_XYZB4:
+      case D3DFVF_XYZB5:
+        size += sizeof(float) * 3;
+
+        if ((FVF & D3DFVF_POSITION_MASK) == D3DFVF_XYZ)
+          break;
+
+        betas = (((FVF & D3DFVF_XYZB5) - D3DFVF_XYZB1) >> 1) + 1;
+        if (FVF & D3DFVF_LASTBETA_D3DCOLOR)
+          betaIdx = sizeof(D3DCOLOR);
+        else if (FVF & D3DFVF_LASTBETA_UBYTE4)
+          betaIdx = sizeof(BYTE) * 4;
+        else if ((FVF & D3DFVF_XYZB5) == D3DFVF_XYZB5)
+          betaIdx = sizeof(float);
+
+        if (betaIdx != 0xFF)
+          betas--;
+
+        if (betas > 0) {
+          if (betas <= 4)
+            size += sizeof(float) * betas;
+        }
+
+        if (betaIdx != 0xFF) {
+          size += betaIdx;
+        }
+        break;
+
+      case D3DFVF_XYZW:
+      case D3DFVF_XYZRHW:
+        size += sizeof(float) * 4;
+        break;
+
+      default:
+        break;
+    }
+
+    if (FVF & D3DFVF_NORMAL) {
+      size += sizeof(float) * 3;
+    }
+    if (FVF & D3DFVF_PSIZE) {
+      size += sizeof(float);
+    }
+    if (FVF & D3DFVF_DIFFUSE) {
+      size += sizeof(D3DCOLOR);
+    }
+    if (FVF & D3DFVF_SPECULAR) {
+      size += sizeof(D3DCOLOR);
+    }
+
+    texCount = (FVF & D3DFVF_TEXCOUNT_MASK) >> D3DFVF_TEXCOUNT_SHIFT;
+    texCount = std::min(texCount, 8u);
+
+    for (uint32_t i = 0; i < texCount; i++) {
+      switch ((FVF >> (16 + i * 2)) & 0x3) {
+        case D3DFVF_TEXTUREFORMAT1:
+          size += sizeof(float);
+          break;
+
+        case D3DFVF_TEXTUREFORMAT2:
+          size += sizeof(float) * 2;
+          break;
+
+        case D3DFVF_TEXTUREFORMAT3:
+          size += sizeof(float) * 3;
+          break;
+
+        case D3DFVF_TEXTUREFORMAT4:
+          size += sizeof(float) * 4;
+          break;
+
+        default:
+          break;
+      }
+    }
+
+    return size;
+  }
+
+
+  constexpr UINT getSurfaceSize(D3DFORMAT Format, UINT Width, UINT Height) {
+    if (isDXT(Format)) {
+      Width = ((Width + 3) >> 2);
+      Height = ((Height + 3) >> 2);
+    }
+    return Width * Height * getFormatStride(Format);
+  }
+
+}
diff -pruN 2.3.1-1/src/d3d8/d3d8_include.h 2.4-1/src/d3d8/d3d8_include.h
--- 2.3.1-1/src/d3d8/d3d8_include.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_include.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,200 @@
+#pragma once
+
+#ifndef _MSC_VER
+#ifdef _WIN32_WINNT
+#undef _WIN32_WINNT
+#endif
+#define _WIN32_WINNT 0x0A00
+#endif
+
+#include <stdint.h>
+#include <d3d8.h>
+
+// Declare __uuidof for D3D8 interfaces
+#ifdef __CRT_UUID_DECL
+__CRT_UUID_DECL(IDirect3D8,              0x1DD9E8DA,0x1C77,0x4D40,0xB0,0xCF,0x98,0xFE,0xFD,0xFF,0x95,0x12);
+__CRT_UUID_DECL(IDirect3DDevice8,        0x7385E5DF,0x8FE8,0x41D5,0x86,0xB6,0xD7,0xB4,0x85,0x47,0xB6,0xCF);
+__CRT_UUID_DECL(IDirect3DResource8,      0x1B36BB7B,0x09B7,0x410A,0xB4,0x45,0x7D,0x14,0x30,0xD7,0xB3,0x3F);
+__CRT_UUID_DECL(IDirect3DVertexBuffer8,  0x8AEEEAC7,0x05F9,0x44D4,0xB5,0x91,0x00,0x0B,0x0D,0xF1,0xCB,0x95);
+__CRT_UUID_DECL(IDirect3DVolume8,        0xBD7349F5,0x14F1,0x42E4,0x9C,0x79,0x97,0x23,0x80,0xDB,0x40,0xC0);
+__CRT_UUID_DECL(IDirect3DSwapChain8,     0x928C088B,0x76B9,0x4C6B,0xA5,0x36,0xA5,0x90,0x85,0x38,0x76,0xCD);
+__CRT_UUID_DECL(IDirect3DSurface8,       0xB96EEBCA,0xB326,0x4EA5,0x88,0x2F,0x2F,0xF5,0xBA,0xE0,0x21,0xDD);
+__CRT_UUID_DECL(IDirect3DIndexBuffer8,   0x0E689C9A,0x053D,0x44A0,0x9D,0x92,0xDB,0x0E,0x3D,0x75,0x0F,0x86);
+__CRT_UUID_DECL(IDirect3DBaseTexture8,   0xB4211CFA,0x51B9,0x4A9F,0xAB,0x78,0xDB,0x99,0xB2,0xBB,0x67,0x8E);
+__CRT_UUID_DECL(IDirect3DTexture8,       0xE4CDD575,0x2866,0x4F01,0xB1,0x2E,0x7E,0xEC,0xE1,0xEC,0x93,0x58);
+__CRT_UUID_DECL(IDirect3DCubeTexture8,   0x3EE5B968,0x2ACA,0x4C34,0x8B,0xB5,0x7E,0x0C,0x3D,0x19,0xB7,0x50);
+__CRT_UUID_DECL(IDirect3DVolumeTexture8, 0x4B8AAAFA,0x140F,0x42BA,0x91,0x31,0x59,0x7E,0xAF,0xAA,0x2E,0xAD);
+#elif defined(_MSC_VER)
+interface DECLSPEC_UUID("1DD9E8DA-1C77-4D40-B0CF-98FEFDFF9512") IDirect3D8;
+interface DECLSPEC_UUID("7385E5DF-8FE8-41D5-86B6-D7B48547B6CF") IDirect3DDevice8;
+interface DECLSPEC_UUID("1B36BB7B-09B7-410A-B445-7D1430D7B33F") IDirect3DResource8;
+interface DECLSPEC_UUID("8AEEEAC7-05F9-44D4-B591-000B0DF1CB95") IDirect3DVertexBuffer8;
+interface DECLSPEC_UUID("BD7349F5-14F1-42E4-9C79-972380DB40C0") IDirect3DVolume8;
+interface DECLSPEC_UUID("928C088B-76B9-4C6B-A536-A590853876CD") IDirect3DSwapChain8;
+interface DECLSPEC_UUID("B96EEBCA-B326-4EA5-882F-2FF5BAE021DD") IDirect3DSurface8;
+interface DECLSPEC_UUID("0E689C9A-053D-44A0-9D92-DB0E3D750F86") IDirect3DIndexBuffer8;
+interface DECLSPEC_UUID("B4211CFA-51B9-4A9F-AB78-DB99B2BB678E") IDirect3DBaseTexture8;
+interface DECLSPEC_UUID("E4CDD575-2866-4F01-B12E-7EECE1EC9358") IDirect3DTexture8;
+interface DECLSPEC_UUID("3EE5B968-2ACA-4C34-8BB5-7E0C3D19B750") IDirect3DCubeTexture8;
+interface DECLSPEC_UUID("4B8AAAFA-140F-42BA-9131-597EAFAA2EAD") IDirect3DVolumeTexture8;
+#endif
+
+// Undefine D3D8 macros
+#undef DIRECT3D_VERSION
+#undef D3D_SDK_VERSION
+
+#undef D3DCS_ALL            // parentheses added in D3D9
+#undef D3DFVF_POSITION_MASK // changed from 0x00E to 0x400E in D3D9
+#undef D3DFVF_RESERVED2     // reduced from 4 to 2 in DX9
+
+#undef D3DSP_REGNUM_MASK    // changed from 0x00000FFF to 0x000007FF in D3D9
+
+
+#if defined(__MINGW32__) || defined(__GNUC__)
+
+// Avoid redundant definitions (add D3D*_DEFINED macros here)
+#define D3DRECT_DEFINED
+#define D3DMATRIX_DEFINED
+
+// Temporarily override __CRT_UUID_DECL to allow usage in d3d9 namespace
+#pragma push_macro("__CRT_UUID_DECL")
+#ifdef __CRT_UUID_DECL
+#undef __CRT_UUID_DECL
+#endif
+
+#ifdef __MINGW32__
+#define __CRT_UUID_DECL(type,l,w1,w2,b1,b2,b3,b4,b5,b6,b7,b8)         \
+}                                                                     \
+  extern "C++" template<> struct __mingw_uuidof_s<d3d9::type> { static constexpr IID __uuid_inst = {l,w1,w2, {b1,b2,b3,b4,b5,b6,b7,b8}}; }; \
+  extern "C++" template<> constexpr const GUID &__mingw_uuidof<d3d9::type>() { return __mingw_uuidof_s<d3d9::type>::__uuid_inst; }          \
+  extern "C++" template<> constexpr const GUID &__mingw_uuidof<d3d9::type*>() { return __mingw_uuidof_s<d3d9::type>::__uuid_inst; }         \
+namespace d3d9 {
+
+#elif defined(__GNUC__)
+#define __CRT_UUID_DECL(type, a, b, c, d, e, f, g, h, i, j, k) \
+}                                                                                                                               \
+  extern "C++" { template <> constexpr GUID __uuidof_helper<d3d9::type>() { return GUID{a,b,c,{d,e,f,g,h,i,j,k}}; } }           \
+  extern "C++" { template <> constexpr GUID __uuidof_helper<d3d9::type*>() { return __uuidof_helper<d3d9::type>(); } }          \
+  extern "C++" { template <> constexpr GUID __uuidof_helper<const d3d9::type*>() { return __uuidof_helper<d3d9::type>(); } }    \
+  extern "C++" { template <> constexpr GUID __uuidof_helper<d3d9::type&>() { return __uuidof_helper<d3d9::type>(); } }          \
+  extern "C++" { template <> constexpr GUID __uuidof_helper<const d3d9::type&>() { return __uuidof_helper<d3d9::type>(); } }    \
+namespace d3d9 {
+#endif
+
+#endif // defined(__MINGW32__) || defined(__GNUC__)
+
+
+/**
+* \brief Direct3D 9
+* 
+* All D3D9 interfaces are included within
+* a namespace, so as not to collide with
+* D3D8 interfaces.
+*/
+namespace d3d9 {
+#include <d3d9.h>
+}
+
+// Indicates d3d9:: namespace is in-use.
+#define DXVK_D3D9_NAMESPACE
+
+#if defined(__MINGW32__) || defined(__GNUC__)
+#pragma pop_macro("__CRT_UUID_DECL")
+#endif
+
+//for some reason we need to specify __declspec(dllexport) for MinGW
+#if defined(__WINE__) || !defined(_WIN32)
+  #define DLLEXPORT __attribute__((visibility("default")))
+#else
+  #define DLLEXPORT
+#endif
+
+
+#include "../util/com/com_guid.h"
+#include "../util/com/com_object.h"
+#include "../util/com/com_pointer.h"
+
+#include "../util/log/log.h"
+#include "../util/log/log_debug.h"
+
+#include "../util/util_error.h"
+#include "../util/util_likely.h"
+#include "../util/util_string.h"
+
+// Missed definitions in Wine/MinGW.
+
+#ifndef D3DPRESENT_BACK_BUFFERS_MAX_EX
+#define D3DPRESENT_BACK_BUFFERS_MAX_EX 30
+#endif
+
+#ifndef D3DSI_OPCODE_MASK
+#define D3DSI_OPCODE_MASK 0x0000FFFF
+#endif
+
+#ifndef D3DSP_TEXTURETYPE_MASK
+#define D3DSP_TEXTURETYPE_MASK 0x78000000
+#endif
+
+#ifndef D3DUSAGE_AUTOGENMIPMAP
+#define D3DUSAGE_AUTOGENMIPMAP 0x00000400L
+#endif
+
+#ifndef D3DSP_DCL_USAGE_MASK
+#define D3DSP_DCL_USAGE_MASK 0x0000000f
+#endif
+
+#ifndef D3DSP_OPCODESPECIFICCONTROL_MASK
+#define D3DSP_OPCODESPECIFICCONTROL_MASK 0x00ff0000
+#endif
+
+#ifndef D3DSP_OPCODESPECIFICCONTROL_SHIFT
+#define D3DSP_OPCODESPECIFICCONTROL_SHIFT 16
+#endif
+
+#ifndef D3DCURSOR_IMMEDIATE_UPDATE
+#define D3DCURSOR_IMMEDIATE_UPDATE             0x00000001L
+#endif
+
+#ifndef D3DPRESENT_FORCEIMMEDIATE
+#define D3DPRESENT_FORCEIMMEDIATE              0x00000100L
+#endif
+
+// From d3dtypes.h
+
+#ifndef D3DDEVINFOID_TEXTUREMANAGER
+#define D3DDEVINFOID_TEXTUREMANAGER    1
+#endif
+
+#ifndef D3DDEVINFOID_D3DTEXTUREMANAGER
+#define D3DDEVINFOID_D3DTEXTUREMANAGER 2
+#endif
+
+#ifndef D3DDEVINFOID_TEXTURING
+#define D3DDEVINFOID_TEXTURING         3
+#endif
+
+// From d3dhal.h
+
+#ifndef D3DDEVINFOID_VCACHE
+#define D3DDEVINFOID_VCACHE            4
+#endif
+
+// MinGW headers are broken. Who'dve guessed?
+#ifndef _MSC_VER
+
+// Missing from d3d8types.h
+#ifndef D3DDEVINFOID_RESOURCEMANAGER
+#define D3DDEVINFOID_RESOURCEMANAGER    5
+#endif
+
+#ifndef D3DDEVINFOID_VERTEXSTATS
+#define D3DDEVINFOID_VERTEXSTATS        6   // Aka D3DDEVINFOID_D3DVERTEXSTATS
+#endif
+
+#else // _MSC_VER
+
+// These are enum typedefs in the MinGW headers, but not defined by Microsoft
+#define D3DVSDT_TYPE     DWORD
+#define D3DVSDE_REGISTER DWORD
+
+#endif
diff -pruN 2.3.1-1/src/d3d8/d3d8_interface.cpp 2.4-1/src/d3d8/d3d8_interface.cpp
--- 2.3.1-1/src/d3d8/d3d8_interface.cpp	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_interface.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,136 @@
+#include "d3d8_interface.h"
+
+#include "d3d8_device.h"
+#include "d3d8_texture.h"
+
+#include <cstring>
+
+namespace dxvk
+{
+  D3D8Interface::D3D8Interface() {
+    m_d3d9 = d3d9::Direct3DCreate9(D3D_SDK_VERSION);
+
+    // Get the bridge interface to D3D9.
+    if (FAILED(m_d3d9->QueryInterface(__uuidof(IDxvkD3D8InterfaceBridge), (void**)&m_bridge))) {
+      throw DxvkError("D3D8Device: ERROR! Failed to get D3D9 Bridge. d3d9.dll might not be DXVK!");
+    }
+
+    m_d3d8Options = D3D8Options(*m_bridge->GetConfig());
+
+    m_adapterCount = m_d3d9->GetAdapterCount();
+    m_adapterModeCounts.resize(m_adapterCount);
+    m_adapterModes.reserve(m_adapterCount);
+
+    for (UINT adapter = 0; adapter < m_adapterCount; adapter++) {
+      m_adapterModes.emplace_back();
+
+      // cache adapter modes and mode counts for each d3d9 format
+      for (d3d9::D3DFORMAT fmt : ADAPTER_FORMATS) {
+
+        const UINT modeCount = m_d3d9->GetAdapterModeCount(adapter, fmt);
+        for (UINT mode = 0; mode < modeCount; mode++) {
+
+          m_adapterModes[adapter].emplace_back();
+          m_d3d9->EnumAdapterModes(adapter, fmt, mode, &(m_adapterModes[adapter].back()));
+
+          // can't use modeCount as it's only for one fmt
+          m_adapterModeCounts[adapter]++;
+        }
+      }
+    }
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Interface::QueryInterface(REFIID riid, void** ppvObject) {
+    if (ppvObject == nullptr)
+      return E_POINTER;
+
+    *ppvObject = nullptr;
+
+    if (riid == __uuidof(IUnknown)
+     || riid == __uuidof(IDirect3D8)) {
+      *ppvObject = ref(this);
+      return S_OK;
+    }
+
+    Logger::warn("D3D8Interface::QueryInterface: Unknown interface query");
+    Logger::warn(str::format(riid));
+    return E_NOINTERFACE;
+  }
+
+  HRESULT STDMETHODCALLTYPE D3D8Interface::GetAdapterIdentifier(
+          UINT Adapter,
+          DWORD Flags,
+          D3DADAPTER_IDENTIFIER8* pIdentifier) {
+    // This flag now has the opposite effect.
+    // Either way, WHQLevel will be 1 with Direct3D9Ex
+    if (Flags & D3DENUM_NO_WHQL_LEVEL)
+      Flags &= ~D3DENUM_WHQL_LEVEL;
+    else
+      Flags |= D3DENUM_WHQL_LEVEL;
+
+    d3d9::D3DADAPTER_IDENTIFIER9 identifier9;
+    HRESULT res = m_d3d9->GetAdapterIdentifier(Adapter, Flags, &identifier9);
+
+    strncpy(pIdentifier->Driver, identifier9.Driver, MAX_DEVICE_IDENTIFIER_STRING);
+    strncpy(pIdentifier->Description, identifier9.Description, MAX_DEVICE_IDENTIFIER_STRING);
+
+    pIdentifier->DriverVersion    = identifier9.DriverVersion;
+    pIdentifier->VendorId         = identifier9.VendorId;
+    pIdentifier->DeviceId         = identifier9.DeviceId;
+    pIdentifier->SubSysId         = identifier9.SubSysId;
+    pIdentifier->Revision         = identifier9.Revision;
+    pIdentifier->DeviceIdentifier = identifier9.DeviceIdentifier;
+
+    pIdentifier->WHQLLevel = identifier9.WHQLLevel;
+    
+    return res;
+  }
+
+  HRESULT __stdcall D3D8Interface::EnumAdapterModes(
+          UINT Adapter,
+          UINT Mode,
+          D3DDISPLAYMODE* pMode) {
+    if (Adapter >= m_adapterCount || Mode >= m_adapterModeCounts[Adapter] || pMode == nullptr) {
+      return D3DERR_INVALIDCALL;
+    }
+
+    pMode->Width        = m_adapterModes[Adapter][Mode].Width;
+    pMode->Height       = m_adapterModes[Adapter][Mode].Height;
+    pMode->RefreshRate  = m_adapterModes[Adapter][Mode].RefreshRate;
+    pMode->Format       = D3DFORMAT(m_adapterModes[Adapter][Mode].Format);
+    return D3D_OK;
+  }
+
+  HRESULT __stdcall D3D8Interface::CreateDevice(
+        UINT Adapter,
+        D3DDEVTYPE DeviceType,
+        HWND hFocusWindow,
+        DWORD BehaviorFlags,
+        D3DPRESENT_PARAMETERS* pPresentationParameters,
+        IDirect3DDevice8** ppReturnedDeviceInterface) {
+    Com<d3d9::IDirect3DDevice9> pDevice9 = nullptr;
+    d3d9::D3DPRESENT_PARAMETERS params = ConvertPresentParameters9(pPresentationParameters);
+    HRESULT res = m_d3d9->CreateDevice(
+      Adapter,
+      (d3d9::D3DDEVTYPE)DeviceType,
+      hFocusWindow,
+      BehaviorFlags,
+      &params,
+      &pDevice9
+    );
+
+    if (FAILED(res)) {
+      return res;
+    }
+
+    *ppReturnedDeviceInterface = ref(new D3D8Device(
+      this, std::move(pDevice9),
+      DeviceType, hFocusWindow, BehaviorFlags,
+      pPresentationParameters
+    ));
+
+    return res;
+  }
+
+
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/d3d8_interface.h 2.4-1/src/d3d8/d3d8_interface.h
--- 2.3.1-1/src/d3d8/d3d8_interface.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_interface.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,163 @@
+#pragma once
+
+#include "d3d8_include.h"
+#include "d3d8_d3d9_util.h"
+#include "d3d8_options.h"
+#include "d3d8_format.h"
+#include "../d3d9/d3d9_bridge.h"
+
+namespace dxvk {
+
+  /**
+  * \brief D3D8 interface implementation
+  *
+  * Implements the IDirect3DDevice8 interfaces
+  * which provides the way to get adapters and create other objects such as \ref IDirect3DDevice8.
+  * similar to \ref DxgiFactory but for D3D8.
+  */
+  class D3D8Interface final : public ComObjectClamp<IDirect3D8> {
+
+    static constexpr d3d9::D3DFORMAT ADAPTER_FORMATS[] = {
+      d3d9::D3DFMT_A1R5G5B5,
+      //d3d9::D3DFMT_A2R10G10B10, (not in D3D8)
+      d3d9::D3DFMT_A8R8G8B8,
+      d3d9::D3DFMT_R5G6B5,
+      d3d9::D3DFMT_X1R5G5B5,
+      d3d9::D3DFMT_X8R8G8B8
+    };
+
+  public:
+    D3D8Interface();
+
+    HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void** ppvObject);
+
+    HRESULT STDMETHODCALLTYPE RegisterSoftwareDevice(void* pInitializeFunction) {
+      return m_d3d9->RegisterSoftwareDevice(pInitializeFunction);
+    }
+
+    UINT STDMETHODCALLTYPE GetAdapterCount() {
+      return m_d3d9->GetAdapterCount();
+    }
+
+    HRESULT STDMETHODCALLTYPE GetAdapterIdentifier(
+      UINT                    Adapter,
+      DWORD                   Flags,
+      D3DADAPTER_IDENTIFIER8* pIdentifier);
+
+    UINT STDMETHODCALLTYPE GetAdapterModeCount(UINT Adapter) {
+      return m_adapterModeCounts[Adapter];
+    }
+
+    HRESULT STDMETHODCALLTYPE EnumAdapterModes(
+      UINT            Adapter,
+      UINT            Mode,
+      D3DDISPLAYMODE* pMode);
+
+    HRESULT STDMETHODCALLTYPE GetAdapterDisplayMode(UINT Adapter, D3DDISPLAYMODE* pMode) {
+      return m_d3d9->GetAdapterDisplayMode(Adapter, (d3d9::D3DDISPLAYMODE*)pMode);
+    }
+
+    HRESULT STDMETHODCALLTYPE CheckDeviceType(
+        UINT        Adapter,
+        D3DDEVTYPE  DevType,
+        D3DFORMAT   AdapterFormat,
+        D3DFORMAT   BackBufferFormat,
+        BOOL        bWindowed) {
+      return m_d3d9->CheckDeviceType(
+          Adapter,
+          (d3d9::D3DDEVTYPE)DevType,
+          (d3d9::D3DFORMAT)AdapterFormat,
+          (d3d9::D3DFORMAT)BackBufferFormat,
+          bWindowed
+      );
+    }
+
+    HRESULT STDMETHODCALLTYPE CheckDeviceFormat(
+        UINT            Adapter,
+        D3DDEVTYPE      DeviceType,
+        D3DFORMAT       AdapterFormat,
+        DWORD           Usage,
+        D3DRESOURCETYPE RType,
+        D3DFORMAT       CheckFormat) {
+      return m_d3d9->CheckDeviceFormat(
+        Adapter,
+        (d3d9::D3DDEVTYPE)DeviceType,
+        (d3d9::D3DFORMAT)AdapterFormat,
+        Usage,
+        (d3d9::D3DRESOURCETYPE)RType,
+        (d3d9::D3DFORMAT)CheckFormat
+      );
+    }
+
+    HRESULT STDMETHODCALLTYPE CheckDeviceMultiSampleType(
+            UINT                Adapter,
+            D3DDEVTYPE          DeviceType,
+            D3DFORMAT           SurfaceFormat,
+            BOOL                Windowed,
+            D3DMULTISAMPLE_TYPE MultiSampleType) {
+      DWORD* pQualityLevels = nullptr;
+      return m_d3d9->CheckDeviceMultiSampleType(
+        Adapter,
+        (d3d9::D3DDEVTYPE)DeviceType,
+        (d3d9::D3DFORMAT)SurfaceFormat,
+        Windowed,
+        (d3d9::D3DMULTISAMPLE_TYPE)MultiSampleType,
+        pQualityLevels
+      );
+    }
+
+    HRESULT STDMETHODCALLTYPE CheckDepthStencilMatch(
+        UINT Adapter,
+        D3DDEVTYPE DeviceType,
+        D3DFORMAT AdapterFormat,
+        D3DFORMAT RenderTargetFormat,
+        D3DFORMAT DepthStencilFormat) {
+      if (isSupportedDepthStencilFormat(DepthStencilFormat))
+        return m_d3d9->CheckDepthStencilMatch(
+          Adapter,
+          (d3d9::D3DDEVTYPE)DeviceType,
+          (d3d9::D3DFORMAT)AdapterFormat,
+          (d3d9::D3DFORMAT)RenderTargetFormat,
+          (d3d9::D3DFORMAT)DepthStencilFormat
+        );
+
+      return D3DERR_NOTAVAILABLE;
+    }
+
+    HRESULT STDMETHODCALLTYPE GetDeviceCaps(
+        UINT Adapter,
+        D3DDEVTYPE DeviceType,
+        D3DCAPS8* pCaps) {
+      d3d9::D3DCAPS9 caps9;
+      HRESULT res = m_d3d9->GetDeviceCaps(Adapter, (d3d9::D3DDEVTYPE)DeviceType, &caps9);
+      dxvk::ConvertCaps8(caps9, pCaps);
+      return res;
+    }
+
+    HMONITOR STDMETHODCALLTYPE GetAdapterMonitor(UINT Adapter) {
+      return m_d3d9->GetAdapterMonitor(Adapter);
+    }
+
+    HRESULT STDMETHODCALLTYPE CreateDevice(
+        UINT Adapter,
+        D3DDEVTYPE DeviceType,
+        HWND hFocusWindow,
+        DWORD BehaviorFlags,
+        D3DPRESENT_PARAMETERS* pPresentationParameters,
+        IDirect3DDevice8** ppReturnedDeviceInterface);
+
+
+    const D3D8Options& GetOptions() { return m_d3d8Options; }
+
+  private:
+
+    UINT                                            m_adapterCount;
+    std::vector<UINT>                               m_adapterModeCounts;
+    std::vector<std::vector<d3d9::D3DDISPLAYMODE>>  m_adapterModes;
+
+    d3d9::IDirect3D9*                               m_d3d9;
+    Com<IDxvkD3D8InterfaceBridge>                   m_bridge;
+    D3D8Options                                     m_d3d8Options;
+  };
+
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/d3d8_main.cpp 2.4-1/src/d3d8/d3d8_main.cpp
--- 2.3.1-1/src/d3d8/d3d8_main.cpp	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_main.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,24 @@
+#include "d3d8_interface.h"
+
+namespace dxvk {
+  Logger Logger::s_instance("d3d8.log");
+
+  HRESULT CreateD3D8(IDirect3D8** ppDirect3D8) {
+    if (!ppDirect3D8)
+      return D3DERR_INVALIDCALL;
+
+    *ppDirect3D8 = ref(new D3D8Interface());
+    return D3D_OK;
+  }
+}
+
+extern "C" {  
+  DLLEXPORT IDirect3D8* __stdcall Direct3DCreate8(UINT nSDKVersion) {
+    dxvk::Logger::trace("Direct3DCreate8 called");
+
+    IDirect3D8* pDirect3D = nullptr;
+    dxvk::CreateD3D8(&pDirect3D);
+
+    return pDirect3D;
+  }
+}
diff -pruN 2.3.1-1/src/d3d8/d3d8_options.cpp 2.4-1/src/d3d8/d3d8_options.cpp
--- 2.3.1-1/src/d3d8/d3d8_options.cpp	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_options.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,55 @@
+#include "d3d8_options.h"
+#include "../d3d9/d3d9_bridge.h"
+#include "../util/config/config.h"
+#include "../util/util_string.h"
+
+#include <charconv>
+
+namespace dxvk {
+  static inline uint32_t parseDword(std::string_view str) {
+    uint32_t value = UINT32_MAX;
+    std::from_chars(str.data(), str.data() + str.size(), value);
+    return value;
+  }
+
+  void D3D8Options::parseVsDecl(const std::string& decl) {
+    if (decl.empty())
+      return;
+
+    if (decl.find_first_of("0123456789") == std::string::npos) {
+      Logger::warn(str::format("D3D8: Invalid forceVsDecl value: ", decl));
+      Logger::warn("D3D8: Expected numbers.");
+      return;
+    }
+
+    if (decl.find_first_of(":,;") == std::string::npos) {
+      Logger::warn(str::format("D3D8: Invalid forceVsDecl value: ", decl));
+      Logger::warn("D3D8: Expected a comma-separated list of colon-separated number pairs.");
+      return;
+    }
+
+    std::vector<std::string_view> decls = str::split(decl, ":,;");
+
+    if (decls.size() % 2 != 0) {
+      Logger::warn(str::format("D3D8: Invalid forceVsDecl value: ", decl));
+      Logger::warn("D3D8: Expected an even number of numbers.");
+      return;
+    }
+
+    for (size_t i = 0; i < decls.size(); i += 2) {
+      uint32_t reg = parseDword(decls[i]);
+      uint32_t type = parseDword(decls[i+1]);
+      if (reg > D3DVSDE_NORMAL2) {
+        Logger::warn(str::format("D3D8: Invalid forceVsDecl register number: ", decls[i]));
+        return;
+      }
+      if (type > D3DVSDT_SHORT4) {
+        Logger::warn(str::format("D3D8: Invalid forceVsDecl type: ", decls[i+1]));
+        return;
+      }
+
+      forceVsDecl.emplace_back(D3DVSDE_REGISTER(reg), D3DVSDT_TYPE(type));
+    }
+
+  }
+}
diff -pruN 2.3.1-1/src/d3d8/d3d8_options.h 2.4-1/src/d3d8/d3d8_options.h
--- 2.3.1-1/src/d3d8/d3d8_options.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_options.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,48 @@
+#pragma once
+
+#include "d3d8_include.h"
+#include "../d3d9/d3d9_bridge.h"
+#include "../util/config/config.h"
+
+namespace dxvk {
+  struct D3D8Options {
+
+    /// Some games rely on undefined behavior by using undeclared vertex shader inputs.
+    /// The simplest way to fix them is to simply modify their vertex shader decl.
+    /// 
+    /// This option takes a comma-separated list of colon-separated number pairs, where
+    /// the first number is a D3DVSDE_REGISTER value, the second is a D3DVSDT_TYPE value.
+    ///   e.g. "0:2,3:2,7:1" for float3 position : v0, float3 normal : v3, float2 uv : v7
+    std::vector<std::pair<D3DVSDE_REGISTER, D3DVSDT_TYPE>> forceVsDecl;
+
+    /// Specialized drawcall batcher, typically for games that draw a lot of similar
+    /// geometry in separate drawcalls (sometimes even one triangle at a time).
+    ///
+    /// May hurt performance outside of specifc games that benefit from it.
+    bool batching = false;
+
+    /// The Lord of the Rings: The Fellowship of the Ring tries to create a P8 texture
+    /// in D3DPOOL_MANAGED on Nvidia and Intel, which fails, but has a separate code
+    /// path for ATI/AMD that creates it in D3DPOOL_SCRATCH instead, which works.
+    ///
+    /// The internal logic determining this path doesn't seem to be d3d-related, but
+    /// the game works universally if we mimic its own ATI/AMD workaround during P8
+    /// texture creation.
+    ///
+    /// Early Nvidia GPUs, such as the GeForce 4 generation cards, included and exposed
+    /// P8 texture support. However, it was no longer advertised with cards in the FX series
+    /// and above. Most likely ATI/AMD drivers never supported P8 in the first place.
+    bool placeP8InScratch = false;
+
+    D3D8Options() {}
+    D3D8Options(const Config& config) {
+      auto forceVsDeclStr     = config.getOption<std::string>("d3d8.forceVsDecl",            "");
+      batching                = config.getOption<bool>       ("d3d8.batching",               batching);
+      placeP8InScratch        = config.getOption<bool>       ("d3d8.placeP8InScratch",       placeP8InScratch);
+
+      parseVsDecl(forceVsDeclStr);
+    }
+
+    void parseVsDecl(const std::string& decl);
+  };
+}
diff -pruN 2.3.1-1/src/d3d8/d3d8_resource.h 2.4-1/src/d3d8/d3d8_resource.h
--- 2.3.1-1/src/d3d8/d3d8_resource.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_resource.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,100 @@
+#pragma once 
+
+/** Implements IDirect3DResource8
+* 
+* - SetPrivateData, GetPrivateData, FreePrivateData
+* - SetPriority, GetPriority
+* 
+* - Subclasses provide: PreLoad, GetType
+*/
+
+#include "d3d8_device_child.h"
+#include "../util/com/com_private_data.h"
+
+namespace dxvk {
+  
+  template <typename D3D9, typename D3D8>
+  class D3D8Resource : public D3D8DeviceChild<D3D9, D3D8> {
+
+  public:
+
+    D3D8Resource(D3D8Device* pDevice, Com<D3D9>&& Object)
+      : D3D8DeviceChild<D3D9, D3D8>(pDevice, std::move(Object))
+      , m_priority                 ( 0 ) { }
+
+    HRESULT STDMETHODCALLTYPE SetPrivateData(
+            REFGUID     refguid,
+      const void*       pData,
+            DWORD       SizeOfData,
+            DWORD       Flags) final {
+      HRESULT hr;
+      if (Flags & D3DSPD_IUNKNOWN) {
+        IUnknown* unknown =
+          const_cast<IUnknown*>(
+            reinterpret_cast<const IUnknown*>(pData));
+        hr = m_privateData.setInterface(
+          refguid, unknown);
+      }
+      else
+        hr = m_privateData.setData(
+          refguid, SizeOfData, pData);
+
+      if (FAILED(hr))
+        return D3DERR_INVALIDCALL;
+
+      return D3D_OK;
+    }
+
+    HRESULT STDMETHODCALLTYPE GetPrivateData(
+            REFGUID     refguid,
+            void*       pData,
+            DWORD*      pSizeOfData) final {
+      HRESULT hr = m_privateData.getData(
+        refguid, reinterpret_cast<UINT*>(pSizeOfData), pData);
+
+      if (FAILED(hr))
+        return D3DERR_INVALIDCALL;
+
+      return D3D_OK;
+    }
+
+    HRESULT STDMETHODCALLTYPE FreePrivateData(REFGUID refguid) final {
+      HRESULT hr = m_privateData.setData(refguid, 0, nullptr);
+
+      if (FAILED(hr))
+        return D3DERR_INVALIDCALL;
+
+      return D3D_OK;
+    }
+
+    DWORD STDMETHODCALLTYPE SetPriority(DWORD PriorityNew) {
+      DWORD oldPriority = m_priority;
+      m_priority = PriorityNew;
+      return oldPriority;
+    }
+
+    DWORD STDMETHODCALLTYPE GetPriority() {
+      return m_priority;
+    }
+
+    virtual IUnknown* GetInterface(REFIID riid) override try {
+      return D3D8DeviceChild<D3D9, D3D8>::GetInterface(riid);
+    } catch (HRESULT err) {
+      if (riid == __uuidof(IDirect3DResource8))
+        return this;
+      
+      throw err;
+    }
+
+  protected:
+
+    DWORD m_priority;
+
+  private:
+
+    ComPrivateData m_privateData;
+
+  };
+
+
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/d3d8_shader.cpp 2.4-1/src/d3d8/d3d8_shader.cpp
--- 2.3.1-1/src/d3d8/d3d8_shader.cpp	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_shader.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,336 @@
+
+#include "d3d8_shader.h"
+
+#define VSD_SHIFT_MASK(token, field) ((token & field ## MASK) >> field ## SHIFT)
+#define VSD_ENCODE(token, field) ((token << field ## _SHIFT) & field ## _MASK)
+
+// Magic number from D3DVSD_SKIP(...)
+#define VSD_SKIP_FLAG 0x10000000
+
+// This bit is set on all parameter (non-instruction) tokens.
+#define VS_BIT_PARAM 0x80000000
+
+namespace dxvk {
+
+  static constexpr int D3D8_NUM_VERTEX_INPUT_REGISTERS = 17;
+
+  /**
+   * Standard mapping of vertex input registers v0-v16 to D3D9 usages and usage indices
+   * (See D3DVSDE_REGISTER values in d3d8types.h or DirectX 8 docs for vertex shader input registers vN)
+   *
+   * \cite https://learn.microsoft.com/en-us/windows/win32/direct3d9/mapping-between-a-directx-9-declaration-and-directx-8
+  */
+  static constexpr BYTE D3D8_VERTEX_INPUT_REGISTERS[D3D8_NUM_VERTEX_INPUT_REGISTERS][2] = {
+    {d3d9::D3DDECLUSAGE_POSITION, 0},      // dcl_position     v0
+    {d3d9::D3DDECLUSAGE_BLENDWEIGHT, 0},   // dcl_blendweight  v1
+    {d3d9::D3DDECLUSAGE_BLENDINDICES, 0},  // dcl_blendindices v2
+    {d3d9::D3DDECLUSAGE_NORMAL, 0},        // dcl_normal       v3
+    {d3d9::D3DDECLUSAGE_PSIZE, 0},         // dcl_psize        v4
+    {d3d9::D3DDECLUSAGE_COLOR, 0},         // dcl_color        v5 ; diffuse
+    {d3d9::D3DDECLUSAGE_COLOR, 1},         // dcl_color1       v6 ; specular
+    {d3d9::D3DDECLUSAGE_TEXCOORD, 0},      // dcl_texcoord0    v7
+    {d3d9::D3DDECLUSAGE_TEXCOORD, 1},      // dcl_texcoord1    v8
+    {d3d9::D3DDECLUSAGE_TEXCOORD, 2},      // dcl_texcoord2    v9
+    {d3d9::D3DDECLUSAGE_TEXCOORD, 3},      // dcl_texcoord3    v10
+    {d3d9::D3DDECLUSAGE_TEXCOORD, 4},      // dcl_texcoord4    v11
+    {d3d9::D3DDECLUSAGE_TEXCOORD, 5},      // dcl_texcoord5    v12
+    {d3d9::D3DDECLUSAGE_TEXCOORD, 6},      // dcl_texcoord6    v13
+    {d3d9::D3DDECLUSAGE_TEXCOORD, 7},      // dcl_texcoord7    v14
+    {d3d9::D3DDECLUSAGE_POSITION, 1},      // dcl_position1    v15 ; position 2
+    {d3d9::D3DDECLUSAGE_NORMAL, 1},        // dcl_normal1      v16 ; normal 2
+  };
+
+  /** Width in bytes of each d3d9::D3DDECLTYPE or d3d8 D3DVSDT_TYPE */
+  static constexpr BYTE D3D9_DECL_TYPE_SIZES[d3d9::MAXD3DDECLTYPE + 1] = {
+    4,  // FLOAT1
+    8,  // FLOAT2
+    12, // FLOAT3
+    16, // FLOAT4
+    4,  // D3DCOLOR
+
+    4,  // UBYTE4
+    4,  // SHORT2
+    8,  // SHORT4
+
+    // The following are for vs2.0+ //
+    4,  // UBYTE4N
+    4,  // SHORT2N
+    8,  // SHORT4N
+    4,  // USHORT2N
+    8,  // USHORT4N
+    6,  // UDEC3
+    6,  // DEC3N
+    8,  // FLOAT16_2
+    16, // FLOAT16_4
+
+    0   // UNUSED
+  };
+
+  /**
+   * Encodes a \ref DxsoShaderInstruction
+   *
+   * \param [in]  opcode  DxsoOpcode
+   * \cite https://learn.microsoft.com/en-us/windows-hardware/drivers/display/instruction-token
+   */
+  constexpr DWORD encodeInstruction(d3d9::D3DSHADER_INSTRUCTION_OPCODE_TYPE opcode) {
+    DWORD token   = 0;
+    token        |= opcode & 0xFFFF;  // bits 0:15
+    return token;
+  }
+
+  /**
+   * Encodes a \ref DxsoRegister
+   *
+   * \param [in]  regType  DxsoRegisterType
+   * \cite https://learn.microsoft.com/en-us/windows-hardware/drivers/display/destination-parameter-token
+   */
+  constexpr DWORD encodeDestRegister(d3d9::D3DSHADER_PARAM_REGISTER_TYPE type, UINT reg) {
+    DWORD token = 0;
+    token    |= reg & 0x7FF;                  // bits 0:10   num
+    token    |= ((type & 0x07) << 28);        // bits 28:30  type[0:2]
+    token    |= ((type & 0x18) >>  3) << 11;  // bits 11:12  type[3:4]
+    // UINT addrMode : 1;                     // bit  13     hasRelative
+    token    |= 0b1111 << 16;                 // bits 16:19  DxsoRegMask
+    // UINT resultModifier : 3;               // bits 20:23
+    // UINT resultShift : 3;                  // bits 24:27
+    token    |= 1 << 31;                      // bit  31     always 1
+    return token;
+  }
+
+  /**
+   * Encodes a \ref DxsoDeclaration
+   *
+   * \cite https://learn.microsoft.com/en-us/windows-hardware/drivers/display/dcl-instruction
+   */
+  constexpr DWORD encodeDeclaration(d3d9::D3DDECLUSAGE usage, DWORD index) {
+    DWORD token = 0;
+    token |= VSD_ENCODE(usage, D3DSP_DCL_USAGE);       // bits 0:4   DxsoUsage (TODO: missing MSB)
+    token |= VSD_ENCODE(index, D3DSP_DCL_USAGEINDEX);  // bits 16:19 usageIndex
+    token |= 1 << 31;                                  // bit 31     always 1
+    return token;
+  }
+
+  /**
+   * Converts a D3D8 vertex shader + declaration
+   * to a D3D9 vertex shader + declaration.
+  */
+  D3D9VertexShaderCode TranslateVertexShader8(
+      const DWORD*        pDeclaration,
+      const DWORD*        pFunction,
+      const D3D8Options&  options) {
+    using d3d9::D3DDECLTYPE;
+    using d3d9::D3DDECLTYPE_UNUSED;
+
+    D3D9VertexShaderCode result;
+
+    std::vector<DWORD>& tokens = result.function;
+    std::vector<DWORD> defs; // Constant definitions
+
+    // shaderInputRegisters:
+    // set bit N to enable input register vN
+    DWORD shaderInputRegisters = 0;
+
+    d3d9::D3DVERTEXELEMENT9* vertexElements = result.declaration;
+    unsigned int elementIdx = 0;
+
+    // These are used for pDeclaration and pFunction
+    int i = 0;
+    DWORD token;
+
+    std::stringstream dbg;
+    dbg << "Vertex Declaration Tokens:\n\t";
+
+    WORD currentStream = 0;
+    WORD currentOffset = 0;
+
+    auto addVertexElement = [&] (D3DVSDE_REGISTER reg, D3DVSDT_TYPE type) {
+      vertexElements[elementIdx].Stream     = currentStream;
+      vertexElements[elementIdx].Offset     = currentOffset;
+      vertexElements[elementIdx].Method     = d3d9::D3DDECLMETHOD_DEFAULT;
+      vertexElements[elementIdx].Type       = D3DDECLTYPE(type); // (D3DVSDT_TYPE values map directly to D3DDECLTYPE)
+      vertexElements[elementIdx].Usage      = D3D8_VERTEX_INPUT_REGISTERS[reg][0];
+      vertexElements[elementIdx].UsageIndex = D3D8_VERTEX_INPUT_REGISTERS[reg][1];
+
+      // Increase stream offset
+      currentOffset += D3D9_DECL_TYPE_SIZES[type];
+
+      // Enable register vn
+      shaderInputRegisters |= 1 << reg;
+
+      // Finished with this element
+      elementIdx++;
+    };
+
+    // Remap d3d8 decl tokens to d3d9 vertex elements,
+    // and enable bits on shaderInputRegisters for each.
+    if (options.forceVsDecl.size() == 0) do {
+      token = pDeclaration[i++];
+
+      D3DVSD_TOKENTYPE tokenType = D3DVSD_TOKENTYPE(VSD_SHIFT_MASK(token, D3DVSD_TOKENTYPE));
+
+      switch (tokenType) {
+        case D3DVSD_TOKEN_NOP:
+          dbg << "NOP";
+          break;
+        case D3DVSD_TOKEN_STREAM: {
+          dbg << "STREAM ";
+
+          // TODO: D3DVSD_STREAM_TESS
+          if (token & D3DVSD_STREAMTESSMASK) {
+            dbg << "TESS";
+          }
+
+          DWORD streamNum = VSD_SHIFT_MASK(token, D3DVSD_STREAMNUMBER);
+
+          currentStream = WORD(streamNum);
+          currentOffset = 0; // reset offset
+
+          dbg << ", num=" << streamNum;
+          break;
+        }
+        case D3DVSD_TOKEN_STREAMDATA: {
+
+          dbg << "STREAMDATA ";
+
+          // D3DVSD_SKIP
+          if (token & VSD_SKIP_FLAG) {
+            auto skipCount = VSD_SHIFT_MASK(token, D3DVSD_SKIPCOUNT);
+            dbg << "SKIP " << " count=" << skipCount;
+            currentOffset += WORD(skipCount) * sizeof(DWORD);
+            break;
+          }
+
+          // D3DVSD_REG
+          DWORD dataLoadType = VSD_SHIFT_MASK(token, D3DVSD_DATALOADTYPE);
+
+          if ( dataLoadType == 0 ) { // vertex
+            D3DVSDT_TYPE     type = D3DVSDT_TYPE(VSD_SHIFT_MASK(token, D3DVSD_DATATYPE));
+            D3DVSDE_REGISTER reg  = D3DVSDE_REGISTER(VSD_SHIFT_MASK(token, D3DVSD_VERTEXREG));
+
+            addVertexElement(reg, type);
+
+            dbg << "type=" << type << ", register=" << reg;
+          } else {
+            // TODO: When would this bit be 1?
+            dbg << "D3DVSD_DATALOADTYPE " << dataLoadType;
+          }
+          break;
+        }
+        case D3DVSD_TOKEN_TESSELLATOR:
+          dbg << "TESSELLATOR " << std::hex << token;
+          // TODO: D3DVSD_TOKEN_TESSELLATOR
+          break;
+        case D3DVSD_TOKEN_CONSTMEM: {
+          dbg << "CONSTMEM ";
+          DWORD count     = VSD_SHIFT_MASK(token, D3DVSD_CONSTCOUNT);
+          DWORD regCount  = count * 4;
+          DWORD addr      = VSD_SHIFT_MASK(token, D3DVSD_CONSTADDRESS);
+          DWORD rs        = VSD_SHIFT_MASK(token, D3DVSD_CONSTRS);
+
+          dbg << "count=" << count << ", addr=" << addr << ", rs=" << rs;
+
+          // Add a DEF instruction for each constant
+          for (DWORD j = 0; j < regCount; j += 4) {
+            defs.push_back(encodeInstruction(d3d9::D3DSIO_DEF));
+            defs.push_back(encodeDestRegister(d3d9::D3DSPR_CONST2, addr));
+            defs.push_back(pDeclaration[i+j+0]);
+            defs.push_back(pDeclaration[i+j+1]);
+            defs.push_back(pDeclaration[i+j+2]);
+            defs.push_back(pDeclaration[i+j+3]);
+            addr++;
+          }
+          i += regCount;
+          break;
+        }
+        case D3DVSD_TOKEN_EXT: {
+          dbg << "EXT " << std::hex << token << " ";
+          DWORD extInfo = VSD_SHIFT_MASK(token, D3DVSD_EXTINFO);
+          DWORD extCount = VSD_SHIFT_MASK(token, D3DVSD_EXTCOUNT);
+          dbg << "info=" << extInfo << ", count=" << extCount;
+          break;
+        }
+        case D3DVSD_TOKEN_END: {
+          vertexElements[elementIdx++] = D3DDECL_END();
+          dbg << "END";
+          break;
+        }
+        default:
+          dbg << "UNKNOWN TYPE";
+          break;
+      }
+      dbg << "\n\t";
+      //dbg << std::hex << token << " ";
+    } while (token != D3DVSD_END());
+
+    Logger::debug(dbg.str());
+
+    // If forceVsDecl is set, use that decl instead.
+    if (options.forceVsDecl.size() > 0) {
+      for (auto [reg, type] : options.forceVsDecl) {
+        addVertexElement(reg, type);
+      }
+      vertexElements[elementIdx++] = D3DDECL_END();
+    }
+
+    if (pFunction != nullptr) {
+      // Copy first token (version)
+      tokens.push_back(pFunction[0]);
+
+      DWORD vsMajor = D3DSHADER_VERSION_MAJOR(pFunction[0]);
+      DWORD vsMinor = D3DSHADER_VERSION_MINOR(pFunction[0]);
+      Logger::debug(str::format("VS version: ", vsMajor, ".", vsMinor));
+
+      // Insert dcl instructions
+      for (int vn = 0; vn < D3D8_NUM_VERTEX_INPUT_REGISTERS; vn++) {
+
+        // If bit N is set then we need to dcl register vN
+        if ((shaderInputRegisters & (1 << vn)) != 0) {
+
+          Logger::debug(str::format("\tShader Input Regsiter: v", vn));
+
+          DWORD usage = D3D8_VERTEX_INPUT_REGISTERS[vn][0];
+          DWORD index = D3D8_VERTEX_INPUT_REGISTERS[vn][1];
+
+          tokens.push_back(encodeInstruction(d3d9::D3DSIO_DCL));                  // dcl opcode
+          tokens.push_back(encodeDeclaration(d3d9::D3DDECLUSAGE(usage), index));  // usage token
+          tokens.push_back(encodeDestRegister(d3d9::D3DSPR_INPUT, vn));           // dest register num
+        }
+      }
+
+      // Copy constant defs
+      for (DWORD def : defs) {
+        tokens.push_back(def);
+      }
+
+      // Copy shader tokens from input,
+      // skip first token (we already copied it)
+      i = 1;
+      do {
+        token = pFunction[i++];
+
+        DWORD opcode = token & D3DSI_OPCODE_MASK;
+
+        // Instructions
+        if ((token & VS_BIT_PARAM) == 0) {
+
+          // RSQ swizzle fixup
+          if (opcode == D3DSIO_RSQ) {
+            tokens.push_back(token);                            // instr
+            tokens.push_back(token = pFunction[i++]);           // dest
+            token = pFunction[i++];                             // src0
+
+            // If no swizzling is done, then use the w-component.
+            // See d8vk#43 for more information as this may need to change in some cases.
+            if (((token & D3DVS_NOSWIZZLE) == D3DVS_NOSWIZZLE)) {
+              token &= ~D3DVS_SWIZZLE_MASK;
+              token |= (D3DVS_X_W | D3DVS_Y_W | D3DVS_Z_W | D3DVS_W_W);
+            }
+          }
+        }
+        tokens.push_back(token);
+      } while (token != D3DVS_END());
+    }
+
+    return result;
+  }
+}
diff -pruN 2.3.1-1/src/d3d8/d3d8_shader.h 2.4-1/src/d3d8/d3d8_shader.h
--- 2.3.1-1/src/d3d8/d3d8_shader.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_shader.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,18 @@
+#pragma once
+
+#include "d3d8_include.h"
+#include "d3d8_options.h"
+
+namespace dxvk {
+
+    struct D3D9VertexShaderCode {
+      d3d9::D3DVERTEXELEMENT9 declaration[MAXD3DDECLLENGTH + 1];
+      std::vector<DWORD> function;
+    };
+
+    D3D9VertexShaderCode TranslateVertexShader8(
+      const DWORD*        pDeclaration,
+      const DWORD*        pFunction,
+      const D3D8Options&  overrides);
+
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/d3d8_state_block.cpp 2.4-1/src/d3d8/d3d8_state_block.cpp
--- 2.3.1-1/src/d3d8/d3d8_state_block.cpp	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_state_block.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,49 @@
+#include "d3d8_device.h"
+#include "d3d8_state_block.h"
+
+HRESULT dxvk::D3D8StateBlock::Capture() {
+  if (unlikely(m_stateBlock == nullptr))
+    return D3DERR_INVALIDCALL;
+
+  if (m_capture.vs) m_device->GetVertexShader(&m_vertexShader);
+  if (m_capture.ps) m_device->GetPixelShader(&m_pixelShader);
+
+  for (DWORD stage = 0; stage < m_textures.size(); stage++) {
+    if (m_capture.textures.get(stage))
+      m_textures[stage] = m_device->m_textures[stage].ptr();
+  }
+
+  if (m_capture.indices) {
+    m_baseVertexIndex = m_device->m_baseVertexIndex;
+    m_indices = m_device->m_indices.ptr();
+  }
+
+  if (m_capture.swvp)
+    m_device->GetRenderState(D3DRS_SOFTWAREVERTEXPROCESSING, (DWORD*)&m_isSWVP);
+
+  return m_stateBlock->Capture();
+}
+
+HRESULT dxvk::D3D8StateBlock::Apply() {
+  if (unlikely(m_stateBlock == nullptr))
+    return D3DERR_INVALIDCALL;
+
+  HRESULT res = m_stateBlock->Apply();
+
+  if (m_capture.vs) m_device->SetVertexShader(m_vertexShader);
+  if (m_capture.ps) m_device->SetPixelShader(m_pixelShader);
+
+  for (DWORD stage = 0; stage < m_textures.size(); stage++) {
+    if (m_capture.textures.get(stage))
+      m_device->SetTexture(stage, m_textures[stage]);
+  }
+
+  if (m_capture.indices)
+    m_device->SetIndices(m_indices, m_baseVertexIndex);
+
+  // This was a very easy footgun for D3D8 applications.
+  if (m_capture.swvp)
+    m_device->SetRenderState(D3DRS_SOFTWAREVERTEXPROCESSING, m_isSWVP);
+
+  return res;
+}
diff -pruN 2.3.1-1/src/d3d8/d3d8_state_block.h 2.4-1/src/d3d8/d3d8_state_block.h
--- 2.3.1-1/src/d3d8/d3d8_state_block.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_state_block.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,134 @@
+#pragma once
+
+#include "d3d8_caps.h"
+#include "d3d8_include.h"
+#include "d3d8_device.h"
+#include "d3d8_device_child.h"
+
+#include <array>
+
+namespace dxvk {
+
+  struct D3D8StateCapture {
+    bool vs       : 1;
+    bool ps       : 1;
+    bool indices  : 1;
+    bool swvp     : 1;
+
+    bit::bitset<d8caps::MAX_TEXTURE_STAGES> textures;
+
+    D3D8StateCapture()
+      : vs(false)
+      , ps(false)
+      , indices(false)
+      , swvp(false) {
+      // Ensure all bits are initialized to false
+      textures.clearAll();
+    }
+  };
+
+  // Wrapper class for D3D9 state blocks. Captures D3D8-specific state.
+  class D3D8StateBlock  {
+
+  public:
+
+    D3D8StateBlock(
+            D3D8Device*                       pDevice,
+            D3DSTATEBLOCKTYPE                 Type,
+            Com<d3d9::IDirect3DStateBlock9>&& pStateBlock)
+      : m_device(pDevice)
+      , m_stateBlock(std::move(pStateBlock))
+      , m_type(Type) {
+      if (Type == D3DSBT_VERTEXSTATE || Type == D3DSBT_ALL) {
+        // Lights, D3DTSS_TEXCOORDINDEX and D3DTSS_TEXTURETRANSFORMFLAGS,
+        // vertex shader, VS constants, and various render states.
+        m_capture.vs = true;
+      }
+
+      if (Type == D3DSBT_PIXELSTATE || Type == D3DSBT_ALL) {
+        // Pixel shader, PS constants, and various RS/TSS states.
+        m_capture.ps = true;
+      }
+
+      if (Type == D3DSBT_ALL) {
+        m_capture.indices = true;
+        m_capture.swvp    = true;
+        m_capture.textures.setAll();
+      }
+
+      m_textures.fill(nullptr);
+    }
+
+    ~D3D8StateBlock() {}
+
+    // Construct a state block without a D3D9 object
+    D3D8StateBlock(D3D8Device* pDevice)
+      : D3D8StateBlock(pDevice, D3DSTATEBLOCKTYPE(0), nullptr) {
+    }
+
+    // Attach a D3D9 object to a state block that doesn't have one yet
+    void SetD3D9(Com<d3d9::IDirect3DStateBlock9>&& pStateBlock) {
+      if (likely(m_stateBlock == nullptr)) {
+        m_stateBlock = std::move(pStateBlock);
+      } else {
+        Logger::err("D3D8StateBlock::SetD3D9 called when m_stateBlock has already been initialized");
+      }
+    }
+
+    HRESULT Capture();
+
+    HRESULT Apply();
+
+    inline HRESULT SetVertexShader(DWORD Handle) {
+      m_vertexShader  = Handle;
+      m_capture.vs    = true;
+      return D3D_OK;
+    }
+
+    inline HRESULT SetPixelShader(DWORD Handle) {
+      m_pixelShader = Handle;
+      m_capture.ps  = true;
+      return D3D_OK;
+    }
+
+    inline HRESULT SetTexture(DWORD Stage, IDirect3DBaseTexture8* pTexture) {
+      m_textures[Stage] = pTexture;
+      m_capture.textures.set(Stage, true);
+      return D3D_OK;
+    }
+
+    inline HRESULT SetIndices(IDirect3DIndexBuffer8* pIndexData, UINT BaseVertexIndex) {
+      m_indices         = pIndexData;
+      m_baseVertexIndex = BaseVertexIndex;
+      m_capture.indices = true;
+      return D3D_OK;
+    }
+
+    inline HRESULT SetSoftwareVertexProcessing(bool value) {
+      m_isSWVP       = value;
+      m_capture.swvp = true;
+      return D3D_OK;
+    }
+
+  private:
+    D3D8Device*                     m_device;
+    Com<d3d9::IDirect3DStateBlock9> m_stateBlock;
+    D3DSTATEBLOCKTYPE               m_type;
+
+  private: // State Data //
+
+    D3D8StateCapture m_capture;
+
+    DWORD m_vertexShader; // vs
+    DWORD m_pixelShader;  // ps
+
+    std::array<IDirect3DBaseTexture8*, d8caps::MAX_TEXTURE_STAGES>  m_textures; // textures
+
+    IDirect3DIndexBuffer8*  m_indices = nullptr;  // indices
+    UINT                    m_baseVertexIndex;    // indices
+
+    bool m_isSWVP;  // D3DRS_SOFTWAREVERTEXPROCESSING
+  };
+
+
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/d3d8_subresource.h 2.4-1/src/d3d8/d3d8_subresource.h
--- 2.3.1-1/src/d3d8/d3d8_subresource.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_subresource.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,61 @@
+#pragma once
+
+#include "d3d8_resource.h"
+
+namespace dxvk {
+
+  // Base class for Surfaces and Volumes,
+  // which can be attached to Textures.
+
+  template <typename D3D9, typename D3D8>
+  class D3D8Subresource : public D3D8Resource<D3D9, D3D8> {
+
+    using Resource = D3D8Resource<D3D9, D3D8>;
+
+  public:
+
+    D3D8Subresource(
+            D3D8Device*             pDevice,
+            Com<D3D9>&&             Object,
+            IDirect3DBaseTexture8*  pBaseTexture)
+    : Resource(pDevice, std::move(Object)),
+      m_container(pBaseTexture) {
+    }
+
+    ~D3D8Subresource() {
+    }
+
+    // Refing subresources implicitly refs the container texture,
+    ULONG STDMETHODCALLTYPE AddRef() final {
+      if (m_container != nullptr)
+        return m_container->AddRef();
+
+      return Resource::AddRef();
+    }
+
+    // and releasing them implicitly releases the texture.
+    ULONG STDMETHODCALLTYPE Release() final {
+      if (m_container != nullptr)
+        return m_container->Release();
+
+      return Resource::Release();
+    }
+
+    // Clients can grab the container if they want
+    HRESULT STDMETHODCALLTYPE GetContainer(REFIID riid, void** ppContainer) final {
+      if (m_container != nullptr)
+        return m_container->QueryInterface(riid, ppContainer);
+
+      return this->GetDevice()->QueryInterface(riid, ppContainer);
+    }
+
+    inline IDirect3DBaseTexture8* GetBaseTexture() {
+      return m_container;
+    }
+
+  protected:
+
+    IDirect3DBaseTexture8*  m_container;
+  };
+
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/d3d8_surface.cpp 2.4-1/src/d3d8/d3d8_surface.cpp
--- 2.3.1-1/src/d3d8/d3d8_surface.cpp	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_surface.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,26 @@
+
+#include "d3d8_surface.h"
+#include "d3d8_device.h"
+
+namespace dxvk {
+
+    Com<d3d9::IDirect3DSurface9> D3D8Surface::CreateBlitImage() {
+      d3d9::D3DSURFACE_DESC desc;
+      GetD3D9()->GetDesc(&desc);
+
+      // NOTE: This adds a D3DPOOL_DEFAULT resource to the
+      // device, which counts as losable during device reset
+      Com<d3d9::IDirect3DSurface9> image = nullptr;
+      HRESULT res = GetParent()->GetD3D9()->CreateRenderTarget(
+        desc.Width, desc.Height, desc.Format,
+        d3d9::D3DMULTISAMPLE_NONE, 0,
+        FALSE,
+        &image,
+        NULL);      
+      
+      if (FAILED(res))
+        throw new DxvkError("D3D8: Failed to create blit image");
+      
+      return image;
+    }
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/d3d8_surface.h 2.4-1/src/d3d8/d3d8_surface.h
--- 2.3.1-1/src/d3d8/d3d8_surface.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_surface.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,81 @@
+#pragma once
+
+#include "d3d8_include.h"
+#include "d3d8_subresource.h"
+#include "d3d8_d3d9_util.h"
+
+namespace dxvk {
+
+  // TODO: all inherited methods in D3D8Surface should be final like in d9vk
+
+  using D3D8SurfaceBase = D3D8Subresource<d3d9::IDirect3DSurface9, IDirect3DSurface8>;
+  class D3D8Surface final : public D3D8SurfaceBase {
+
+  public:
+
+    D3D8Surface(
+            D3D8Device*                     pDevice,
+            IDirect3DBaseTexture8*          pTexture,
+            Com<d3d9::IDirect3DSurface9>&&  pSurface)
+      : D3D8SurfaceBase (pDevice, std::move(pSurface), pTexture) {
+    }
+
+    // A surface does not need to be attached to a texture
+    D3D8Surface(
+            D3D8Device*                     pDevice,
+            Com<d3d9::IDirect3DSurface9>&&  pSurface)
+      : D3D8Surface (pDevice, nullptr, std::move(pSurface)) {
+    }
+
+    D3DRESOURCETYPE STDMETHODCALLTYPE GetType() {
+      return D3DRESOURCETYPE(GetD3D9()->GetType());
+    }
+
+    HRESULT STDMETHODCALLTYPE GetDesc(D3DSURFACE_DESC* pDesc) {
+      d3d9::D3DSURFACE_DESC desc;
+      HRESULT res = GetD3D9()->GetDesc(&desc);
+      ConvertSurfaceDesc8(&desc, pDesc);
+      return res;
+    }
+
+    HRESULT STDMETHODCALLTYPE LockRect(D3DLOCKED_RECT* pLockedRect, CONST RECT* pRect, DWORD Flags) {
+      return GetD3D9()->LockRect((d3d9::D3DLOCKED_RECT*)pLockedRect, pRect, Flags);
+    }
+
+    HRESULT STDMETHODCALLTYPE UnlockRect() {
+      return GetD3D9()->UnlockRect();
+    }
+
+    HRESULT STDMETHODCALLTYPE GetDC(HDC* phDC) {
+      return GetD3D9()->GetDC(phDC);
+    }
+
+    HRESULT STDMETHODCALLTYPE ReleaseDC(HDC hDC) {
+      return GetD3D9()->ReleaseDC(hDC);
+    }
+
+  public:
+
+    /**
+     * \brief Allocate or reuse an image of the same size
+     * as this texture for performing blit into system mem.
+     * 
+     * TODO: Consider creating only one texture to
+     * encompass all surface levels of a texture.
+     */
+    Com<d3d9::IDirect3DSurface9> GetBlitImage() {
+      if (unlikely(m_blitImage == nullptr)) {
+        m_blitImage = CreateBlitImage();
+      }
+
+      return m_blitImage;
+    }
+
+
+  private:
+    Com<d3d9::IDirect3DSurface9> CreateBlitImage();
+
+    Com<d3d9::IDirect3DSurface9> m_blitImage = nullptr;
+
+  };
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/d3d8_swapchain.h 2.4-1/src/d3d8/d3d8_swapchain.h
--- 2.3.1-1/src/d3d8/d3d8_swapchain.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_swapchain.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,42 @@
+#pragma once
+
+#include "d3d8_device_child.h"
+#include "d3d8_surface.h"
+#include "d3d8_d3d9_util.h"
+
+namespace dxvk {
+  
+  using D3D8SwapChainBase = D3D8DeviceChild<d3d9::IDirect3DSwapChain9, IDirect3DSwapChain8>;
+  class D3D8SwapChain final : public D3D8SwapChainBase {
+
+  public:
+
+    D3D8SwapChain(
+      D3D8Device*                      pDevice,
+      Com<d3d9::IDirect3DSwapChain9>&& pSwapChain)
+      : D3D8SwapChainBase(pDevice, std::move(pSwapChain)) {}
+
+    HRESULT STDMETHODCALLTYPE Present(const RECT *src, const RECT *dst, HWND hWnd, const RGNDATA *dirtyRegion) final {
+        return GetD3D9()->Present(src, dst, hWnd, dirtyRegion, 0);
+    }
+    
+    HRESULT STDMETHODCALLTYPE GetBackBuffer(UINT BackBuffer, D3DBACKBUFFER_TYPE Type, IDirect3DSurface8** ppBackBuffer) final {
+      // Same logic as in D3D8Device::GetBackBuffer
+      if (unlikely(m_backBuffer == nullptr)) {
+        Com<d3d9::IDirect3DSurface9> pSurface9;
+        HRESULT res = GetD3D9()->GetBackBuffer(BackBuffer, (d3d9::D3DBACKBUFFER_TYPE)Type, &pSurface9);
+
+        m_backBuffer = new D3D8Surface(GetParent(), std::move(pSurface9));
+        *ppBackBuffer = m_backBuffer.ref();
+        return res;
+      }
+
+      *ppBackBuffer = m_backBuffer.ref();
+      return D3D_OK;
+    }
+  private:
+    Com<D3D8Surface> m_backBuffer = nullptr;
+
+  };
+
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/d3d8_texture.h 2.4-1/src/d3d8/d3d8_texture.h
--- 2.3.1-1/src/d3d8/d3d8_texture.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_texture.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,233 @@
+#pragma once
+
+#include "d3d8_resource.h"
+#include "d3d8_surface.h"
+#include "d3d8_volume.h"
+
+#include "d3d8_d3d9_util.h"
+
+#include <vector>
+#include <new>
+
+namespace dxvk {
+
+  template <typename SubresourceType, typename D3D9, typename D3D8>
+  class D3D8BaseTexture : public D3D8Resource<D3D9, D3D8> {
+
+  public:
+
+    constexpr static UINT CUBE_FACES = 6;
+
+    using SubresourceType8 = typename SubresourceType::D3D8;
+    using SubresourceType9 = typename SubresourceType::D3D9;
+
+    D3D8BaseTexture(
+            D3D8Device*                         pDevice,
+            Com<D3D9>&&                         pBaseTexture,
+            UINT                                SubresourceCount)
+        : D3D8Resource<D3D9, D3D8> ( pDevice, std::move(pBaseTexture) ) {
+      m_subresources.resize(SubresourceCount, nullptr);
+    }
+
+    ~D3D8BaseTexture() {
+      for (size_t i = 0; i < m_subresources.size(); i++)
+        if (m_subresources[i] != nullptr)
+          m_subresources[i] = nullptr;
+    }
+
+    virtual IUnknown* GetInterface(REFIID riid) final override try {
+      return D3D8Resource<D3D9, D3D8>::GetInterface(riid);
+    } catch (HRESULT err) {
+      if (riid == __uuidof(IDirect3DBaseTexture8))
+        return this;
+      
+      throw err;
+    }
+
+    void STDMETHODCALLTYPE PreLoad() final {
+      this->GetD3D9()->PreLoad();
+    }
+
+    DWORD STDMETHODCALLTYPE SetLOD(DWORD LODNew) final {
+      return this->GetD3D9()->SetLOD(LODNew);
+    }
+
+    DWORD STDMETHODCALLTYPE GetLOD() final {
+      return this->GetD3D9()->GetLOD();
+    }
+
+    DWORD STDMETHODCALLTYPE GetLevelCount() final {
+      return this->GetD3D9()->GetLevelCount();
+    }
+
+  protected:
+
+    HRESULT STDMETHODCALLTYPE GetSubresource(UINT Index, SubresourceType8** ppSubresource) {
+      InitReturnPtr(ppSubresource);
+
+      if (unlikely(Index >= m_subresources.size()))
+        return D3DERR_INVALIDCALL;
+      
+      if (m_subresources[Index] == nullptr) {
+        try {
+          Com<SubresourceType9> subresource = LookupSubresource(Index);
+
+          // Cache the subresource
+          m_subresources[Index] = new SubresourceType(this->m_parent, this, std::move(subresource));
+        } catch (HRESULT res) {
+          return res;
+        }
+      }
+
+      *ppSubresource = m_subresources[Index].ref();
+      return D3D_OK;
+    }
+
+  private:
+
+    Com<SubresourceType9> LookupSubresource(UINT Index) {
+      Com<SubresourceType9> ptr = nullptr;
+      HRESULT res = D3DERR_INVALIDCALL;
+      if constexpr (std::is_same_v<D3D8, IDirect3DTexture8>) {
+        res = this->GetD3D9()->GetSurfaceLevel(Index, &ptr); 
+      } else if constexpr (std::is_same_v<D3D8, IDirect3DVolume8>) {
+        res = this->GetD3D9()->GetVolumeLevel(Index, &ptr);
+      } else if constexpr (std::is_same_v<D3D8, IDirect3DCubeTexture8>) {
+        res = this->GetD3D9()->GetCubeMapSurface(d3d9::D3DCUBEMAP_FACES(Index % CUBE_FACES), Index / CUBE_FACES, &ptr);
+      }
+      if (FAILED(res))
+        throw res;
+      return ptr;
+    }
+
+    std::vector<Com<SubresourceType, false>> m_subresources;
+
+  };
+
+  using D3D8Texture2DBase = D3D8BaseTexture<D3D8Surface, d3d9::IDirect3DTexture9, IDirect3DTexture8>;
+  class D3D8Texture2D final : public D3D8Texture2DBase {
+
+  public:
+
+    D3D8Texture2D(
+            D3D8Device*                    pDevice,
+            Com<d3d9::IDirect3DTexture9>&& pTexture)
+      : D3D8Texture2DBase(pDevice, std::move(pTexture), pTexture->GetLevelCount()) {
+    }
+
+    D3DRESOURCETYPE STDMETHODCALLTYPE GetType() final { return D3DRTYPE_TEXTURE; }
+
+    HRESULT STDMETHODCALLTYPE GetLevelDesc(UINT Level, D3DSURFACE_DESC* pDesc) {
+      d3d9::D3DSURFACE_DESC surf;
+      HRESULT res = GetD3D9()->GetLevelDesc(Level, &surf);
+      ConvertSurfaceDesc8(&surf, pDesc);
+      return res;
+    }
+
+    HRESULT STDMETHODCALLTYPE GetSurfaceLevel(UINT Level, IDirect3DSurface8** ppSurfaceLevel) {
+      return GetSubresource(Level, ppSurfaceLevel);
+    }
+
+    HRESULT STDMETHODCALLTYPE LockRect(UINT Level, D3DLOCKED_RECT* pLockedRect, CONST RECT* pRect, DWORD Flags) {
+      return GetD3D9()->LockRect(Level, reinterpret_cast<d3d9::D3DLOCKED_RECT*>(pLockedRect), pRect, Flags);
+    }
+
+    HRESULT STDMETHODCALLTYPE UnlockRect(UINT Level) {
+      return GetD3D9()->UnlockRect(Level);
+    }
+
+    HRESULT STDMETHODCALLTYPE AddDirtyRect(CONST RECT* pDirtyRect) {
+      return GetD3D9()->AddDirtyRect(pDirtyRect);
+    }
+
+  };
+
+  using D3D8Texture3DBase = D3D8BaseTexture<D3D8Volume, d3d9::IDirect3DVolumeTexture9, IDirect3DVolumeTexture8>;
+  class D3D8Texture3D final : public D3D8Texture3DBase {
+
+  public:
+
+    D3D8Texture3D(
+          D3D8Device*                           pDevice,
+          Com<d3d9::IDirect3DVolumeTexture9>&&  pVolumeTexture)
+      : D3D8Texture3DBase(pDevice, std::move(pVolumeTexture), pVolumeTexture->GetLevelCount()) {}
+
+    D3DRESOURCETYPE STDMETHODCALLTYPE GetType() final { return D3DRTYPE_VOLUMETEXTURE; }
+
+    HRESULT STDMETHODCALLTYPE GetLevelDesc(UINT Level, D3DVOLUME_DESC *pDesc) {
+      d3d9::D3DVOLUME_DESC vol;
+      HRESULT res = GetD3D9()->GetLevelDesc(Level, &vol);
+      ConvertVolumeDesc8(&vol, pDesc);
+      return res;
+    }
+
+    HRESULT STDMETHODCALLTYPE GetVolumeLevel(UINT Level, IDirect3DVolume8** ppVolumeLevel) {
+      return GetSubresource(Level, ppVolumeLevel);
+    }
+
+    HRESULT STDMETHODCALLTYPE LockBox(UINT Level, D3DLOCKED_BOX* pLockedBox, CONST D3DBOX* pBox, DWORD Flags) {
+      return GetD3D9()->LockBox(
+        Level,
+        reinterpret_cast<d3d9::D3DLOCKED_BOX*>(pLockedBox),
+        reinterpret_cast<const d3d9::D3DBOX*>(pBox),
+        Flags
+      );
+    }
+
+    HRESULT STDMETHODCALLTYPE UnlockBox(UINT Level) {
+      return GetD3D9()->UnlockBox(Level);
+    }
+
+    HRESULT STDMETHODCALLTYPE AddDirtyBox(CONST D3DBOX* pDirtyBox) {
+      return GetD3D9()->AddDirtyBox(reinterpret_cast<const d3d9::D3DBOX*>(pDirtyBox));
+    }
+
+  };
+
+  using D3D8TextureCubeBase = D3D8BaseTexture<D3D8Surface, d3d9::IDirect3DCubeTexture9, IDirect3DCubeTexture8>;
+  class D3D8TextureCube final : public D3D8TextureCubeBase {
+
+  public:
+
+    D3D8TextureCube(
+            D3D8Device*                         pDevice,
+            Com<d3d9::IDirect3DCubeTexture9>&&  pTexture)
+      : D3D8TextureCubeBase(pDevice, std::move(pTexture), pTexture->GetLevelCount() * CUBE_FACES) {
+    }
+
+    D3DRESOURCETYPE STDMETHODCALLTYPE GetType() final { return D3DRTYPE_CUBETEXTURE; }
+
+    HRESULT STDMETHODCALLTYPE GetLevelDesc(UINT Level, D3DSURFACE_DESC* pDesc) {
+      d3d9::D3DSURFACE_DESC surf;
+      HRESULT res = GetD3D9()->GetLevelDesc(Level, &surf);
+      ConvertSurfaceDesc8(&surf, pDesc);
+      return res;
+    }
+
+    HRESULT STDMETHODCALLTYPE GetCubeMapSurface(D3DCUBEMAP_FACES Face, UINT Level, IDirect3DSurface8** ppSurfaceLevel) {
+      return GetSubresource((Level * CUBE_FACES) + Face, ppSurfaceLevel);
+    }
+
+    HRESULT STDMETHODCALLTYPE LockRect(
+            D3DCUBEMAP_FACES Face,
+            UINT Level,
+            D3DLOCKED_RECT* pLockedRect,
+            const RECT* pRect,
+            DWORD Flags) {
+      return GetD3D9()->LockRect(
+        d3d9::D3DCUBEMAP_FACES(Face),
+        Level,
+        reinterpret_cast<d3d9::D3DLOCKED_RECT*>(pLockedRect),
+        pRect,
+        Flags);
+    }
+
+    HRESULT STDMETHODCALLTYPE UnlockRect(D3DCUBEMAP_FACES Face, UINT Level) {
+      return GetD3D9()->UnlockRect(d3d9::D3DCUBEMAP_FACES(Face), Level);
+    }
+
+    HRESULT STDMETHODCALLTYPE AddDirtyRect(D3DCUBEMAP_FACES Face, const RECT* pDirtyRect) {
+      return GetD3D9()->AddDirtyRect(d3d9::D3DCUBEMAP_FACES(Face), pDirtyRect);
+    }
+  };
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/d3d8_volume.h 2.4-1/src/d3d8/d3d8_volume.h
--- 2.3.1-1/src/d3d8/d3d8_volume.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_volume.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,40 @@
+#pragma once
+
+#include "d3d8_subresource.h"
+#include "d3d8_d3d9_util.h"
+
+namespace dxvk {
+  
+  using D3D8VolumeBase = D3D8Subresource<d3d9::IDirect3DVolume9, IDirect3DVolume8>;
+  class D3D8Volume final : public D3D8VolumeBase {
+
+  public:
+
+    D3D8Volume(
+          D3D8Device*                   pDevice,
+          IDirect3DVolumeTexture8*      pTexture,
+          Com<d3d9::IDirect3DVolume9>&& pVolume)
+      : D3D8VolumeBase(pDevice, std::move(pVolume), pTexture) {}
+
+    HRESULT STDMETHODCALLTYPE GetDesc(D3DVOLUME_DESC* pDesc) {
+      d3d9::D3DVOLUME_DESC desc;
+      HRESULT res = GetD3D9()->GetDesc(&desc);
+      ConvertVolumeDesc8(&desc, pDesc);
+      return res;
+    }
+
+    HRESULT STDMETHODCALLTYPE LockBox(D3DLOCKED_BOX* pLockedBox, CONST D3DBOX* pBox, DWORD Flags) final {
+      return GetD3D9()->LockBox(
+        reinterpret_cast<d3d9::D3DLOCKED_BOX*>(pLockedBox),
+        reinterpret_cast<const d3d9::D3DBOX*>(pBox),
+        Flags
+      );
+    }
+
+    HRESULT STDMETHODCALLTYPE UnlockBox() final {
+      return GetD3D9()->UnlockBox();
+    }
+
+  };
+
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/d3d8_wrapped_object.h 2.4-1/src/d3d8/d3d8_wrapped_object.h
--- 2.3.1-1/src/d3d8/d3d8_wrapped_object.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/d3d8_wrapped_object.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,68 @@
+#pragma once
+
+#include "d3d8_include.h"
+
+namespace dxvk {
+
+  template <typename D3D9Type, typename D3D8Type>
+  class D3D8WrappedObject : public ComObjectClamp<D3D8Type> {
+
+  public:
+
+    using D3D9 = D3D9Type;
+    using D3D8 = D3D8Type;
+
+    D3D8WrappedObject(Com<D3D9>&& object)
+      : m_d3d9(std::move(object)) {
+    }
+
+    D3D9* GetD3D9() {
+      return m_d3d9.ptr();
+    }
+
+    // For cases where the object may be null.
+    static D3D9* GetD3D9Nullable(D3D8WrappedObject* self) {
+      if (unlikely(self == NULL)) {
+        return NULL;
+      }
+      return self->m_d3d9.ptr();
+    }
+
+    template <typename T>
+    static D3D9* GetD3D9Nullable(Com<T>& self) {
+      return GetD3D9Nullable(self.ptr());
+    }
+
+    virtual IUnknown* GetInterface(REFIID riid) {
+      if (riid == __uuidof(IUnknown))
+        return this;
+      if (riid == __uuidof(D3D8))
+        return this;
+      
+      throw E_NOINTERFACE;
+    }
+
+    HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void** ppvObject) final {
+      if (ppvObject == nullptr)
+        return E_POINTER;
+
+      *ppvObject = nullptr;
+
+      try {
+        *ppvObject = ref(this->GetInterface(riid));
+        return S_OK;
+      } catch (HRESULT err) {
+        Logger::warn("D3D8WrappedObject::QueryInterface: Unknown interface query");
+        Logger::warn(str::format(riid));
+        return err;
+      }
+    }
+
+
+  private:
+
+    Com<D3D9> m_d3d9;
+
+  };
+
+}
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d8/meson.build 2.4-1/src/d3d8/meson.build
--- 2.3.1-1/src/d3d8/meson.build	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/meson.build	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,42 @@
+d3d8_res = wrc_generator.process('version.rc')
+
+d3d8_src = [
+  'd3d8_main.cpp',
+  'd3d8_interface.cpp',
+  'd3d8_device.cpp',
+  'd3d8_options.cpp',
+  'd3d8_surface.cpp',
+  'd3d8_state_block.cpp',
+  'd3d8_shader.cpp'
+]
+
+d3d8_ld_args      = []
+d3d8_link_depends = []
+
+if platform != 'windows'
+  lib_d3d9 = d3d9_dep
+  d3d8_ld_args      += [ '-Wl,--version-script', join_paths(meson.current_source_dir(), 'd3d8.sym') ]
+  d3d8_link_depends += files('d3d8.sym')
+endif
+
+d3d8_dll = shared_library(dxvk_name_prefix+'d3d8', d3d8_src, d3d8_res,
+  dependencies        : [ lib_d3d9, util_dep, dxso_dep, dxvk_dep ],
+  include_directories : dxvk_include_path,
+  install             : true,
+  vs_module_defs      : 'd3d8'+def_spec_ext,
+  link_args           : d3d8_ld_args,
+  link_depends        : [ d3d8_link_depends ],
+  kwargs              : dxvk_so_version,
+)
+
+d3d8_dep = declare_dependency(
+  link_with           : [ d3d8_dll ],
+  include_directories : [ dxvk_include_path ],
+)
+
+if platform != 'windows'
+  pkg.generate(d3d8_dll,
+    filebase: dxvk_pkg_prefix + 'd3d8',
+    subdirs:  'dxvk',
+  )
+endif
diff -pruN 2.3.1-1/src/d3d8/version.rc 2.4-1/src/d3d8/version.rc
--- 2.3.1-1/src/d3d8/version.rc	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/d3d8/version.rc	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,31 @@
+#include <windows.h>
+
+// DLL version information.
+VS_VERSION_INFO    VERSIONINFO
+FILEVERSION        10,0,17763,1
+PRODUCTVERSION     10,0,17763,1
+FILEFLAGSMASK      VS_FFI_FILEFLAGSMASK
+FILEFLAGS          0
+FILEOS VOS_NT_WINDOWS32
+FILETYPE VFT_DLL
+FILESUBTYPE VFT2_UNKNOWN
+BEGIN
+  BLOCK "StringFileInfo"
+  BEGIN
+    BLOCK "080904b0"
+    BEGIN
+      VALUE "CompanyName",      "DXVK"
+      VALUE "FileDescription",  "Direct3D 8 Runtime"
+      VALUE "FileVersion",      "10.0.17763.1 (WinBuild.160101.0800)"
+      VALUE "InternalName",     "D3D8.dll"
+      VALUE "LegalCopyright",   "zlib/libpng license"
+      VALUE "OriginalFilename", "D3D8.dll"
+      VALUE "ProductName",      "DXVK"
+      VALUE "ProductVersion",   "10.0.17763.1"
+    END
+  END
+  BLOCK "VarFileInfo"
+  BEGIN
+    VALUE "Translation", 0x0809, 1200
+  END
+END
diff -pruN 2.3.1-1/src/d3d9/d3d9_adapter.cpp 2.4-1/src/d3d9/d3d9_adapter.cpp
--- 2.3.1-1/src/d3d9/d3d9_adapter.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_adapter.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -162,6 +162,9 @@ namespace dxvk {
     if (mapping.FormatSrgb  == VK_FORMAT_UNDEFINED && srgb)
       return D3DERR_NOTAVAILABLE;
 
+    if (RType == D3DRTYPE_CUBETEXTURE && mapping.Aspect != VK_IMAGE_ASPECT_COLOR_BIT)
+      return D3DERR_NOTAVAILABLE;
+
     if (RType == D3DRTYPE_VERTEXBUFFER || RType == D3DRTYPE_INDEXBUFFER)
       return D3D_OK;
 
@@ -224,11 +227,15 @@ namespace dxvk {
     if (!IsDepthFormat(DepthStencilFormat))
       return D3DERR_NOTAVAILABLE;
 
+    auto dsfMapping = GetFormatMapping(DepthStencilFormat);
+    if (dsfMapping.FormatColor == VK_FORMAT_UNDEFINED)
+      return D3DERR_NOTAVAILABLE;
+
     if (RenderTargetFormat == dxvk::D3D9Format::NULL_FORMAT)
       return D3D_OK;
 
-    auto mapping = ConvertFormatUnfixed(RenderTargetFormat);
-    if (mapping.FormatColor == VK_FORMAT_UNDEFINED)
+    auto rtfMapping = GetFormatMapping(RenderTargetFormat);
+    if (rtfMapping.FormatColor == VK_FORMAT_UNDEFINED)
       return D3DERR_NOTAVAILABLE;
 
     return D3D_OK;
@@ -271,6 +278,8 @@ namespace dxvk {
 
     auto& options = m_parent->GetOptions();
 
+    const VkPhysicalDeviceLimits& limits = m_adapter->deviceProperties().limits;
+
     // TODO: Actually care about what the adapter supports here.
     // ^ For Intel and older cards most likely here.
 
@@ -531,7 +540,7 @@ namespace dxvk {
     // Max Vertex Blend Matrix Index
     pCaps->MaxVertexBlendMatrixIndex = 0;
     // Max Point Size
-    pCaps->MaxPointSize              = 256.0f;
+    pCaps->MaxPointSize              = limits.pointSizeRange[1];
     // Max Primitive Count
     pCaps->MaxPrimitiveCount         = 0x00555555;
     // Max Vertex Index
diff -pruN 2.3.1-1/src/d3d9/d3d9_bridge.h 2.4-1/src/d3d9/d3d9_bridge.h
--- 2.3.1-1/src/d3d9/d3d9_bridge.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_bridge.h	2024-07-10 09:56:14.000000000 +0000
@@ -3,8 +3,6 @@
 #include <windows.h>
 #include "../util/config/config.h"
 
-#include "../vulkan/vulkan_loader.h"
-
 /**
  * The D3D9 bridge allows D3D8 to access DXVK internals.
  * For Vulkan interop without needing DXVK internals, see d3d9_interop.h.
@@ -111,7 +109,7 @@ namespace dxvk {
             void** ppvObject);
 
     const Config* GetConfig() const;
-    
+
   protected:
     D3D9InterfaceEx* m_interface;
   };
diff -pruN 2.3.1-1/src/d3d9/d3d9_common_texture.cpp 2.4-1/src/d3d9/d3d9_common_texture.cpp
--- 2.3.1-1/src/d3d9/d3d9_common_texture.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_common_texture.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -118,6 +118,7 @@ namespace dxvk {
 
   HRESULT D3D9CommonTexture::NormalizeTextureProperties(
           D3D9DeviceEx*             pDevice,
+          D3DRESOURCETYPE           ResourceType,
           D3D9_COMMON_TEXTURE_DESC* pDesc) {
     auto* options = pDevice->GetOptions();
 
@@ -131,6 +132,11 @@ namespace dxvk {
         options->disableA8RT)
       return D3DERR_INVALIDCALL;
 
+    // Cube textures with depth formats are not supported on any native
+    // driver, and allowing them triggers a broken code path in Gothic 3.
+    if (ResourceType == D3DRTYPE_CUBETEXTURE && mapping.Aspect != VK_IMAGE_ASPECT_COLOR_BIT)
+      return D3DERR_INVALIDCALL;
+
     // If the mapping is invalid then lets return invalid
     // Some edge cases:
     // NULL format does not map to anything, but should succeed
diff -pruN 2.3.1-1/src/d3d9/d3d9_common_texture.h 2.4-1/src/d3d9/d3d9_common_texture.h
--- 2.3.1-1/src/d3d9/d3d9_common_texture.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_common_texture.h	2024-07-10 09:56:14.000000000 +0000
@@ -179,11 +179,14 @@ namespace dxvk {
      * Fills in undefined values and validates the texture
      * parameters. Any error returned by this method should
      * be forwarded to the application.
+     * \param [in] pDevice D3D9 device
+     * \param [in] ResourceType Resource type
      * \param [in,out] pDesc Texture description
      * \returns \c S_OK if the parameters are valid
      */
     static HRESULT NormalizeTextureProperties(
             D3D9DeviceEx*              pDevice,
+            D3DRESOURCETYPE            ResourceType,
             D3D9_COMMON_TEXTURE_DESC*  pDesc);
 
     /**
diff -pruN 2.3.1-1/src/d3d9/d3d9_constant_buffer.cpp 2.4-1/src/d3d9/d3d9_constant_buffer.cpp
--- 2.3.1-1/src/d3d9/d3d9_constant_buffer.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_constant_buffer.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -13,7 +13,7 @@ namespace dxvk {
           DxsoProgramType       ShaderStage,
           DxsoConstantBuffers   BufferType,
           VkDeviceSize          Size)
-  : D3D9ConstantBuffer(pDevice, getBufferUsage(pDevice, ShaderStage, BufferType), GetShaderStage(ShaderStage),
+  : D3D9ConstantBuffer(pDevice, VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT | VK_BUFFER_USAGE_STORAGE_BUFFER_BIT, GetShaderStage(ShaderStage),
       computeResourceSlotId(ShaderStage, DxsoBindingType::ConstantBuffer, BufferType),
       Size) {
 
@@ -135,21 +135,4 @@ namespace dxvk {
       device->properties().extRobustness2.robustUniformBufferAccessSizeAlignment);
   }
 
-
-  VkBufferUsageFlags D3D9ConstantBuffer::getBufferUsage(
-          D3D9DeviceEx*         pDevice,
-          DxsoProgramType       ShaderStage,
-          DxsoConstantBuffers   BufferType) {
-    VkBufferUsageFlags result = VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT;
-
-    // We won't always need this, but the only buffer that
-    // this applies to is so large that it will not matter.
-    if (pDevice->CanSWVP()
-     && ShaderStage == DxsoProgramType::VertexShader
-     && BufferType == DxsoConstantBuffers::VSConstantBuffer)
-      result |= VK_BUFFER_USAGE_STORAGE_BUFFER_BIT;
-
-    return result;
-  }
-
 }
diff -pruN 2.3.1-1/src/d3d9/d3d9_constant_buffer.h 2.4-1/src/d3d9/d3d9_constant_buffer.h
--- 2.3.1-1/src/d3d9/d3d9_constant_buffer.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_constant_buffer.h	2024-07-10 09:56:14.000000000 +0000
@@ -80,11 +80,6 @@ namespace dxvk {
 
     VkDeviceSize getAlignment(const Rc<DxvkDevice>& device) const;
 
-    static VkBufferUsageFlags getBufferUsage(
-            D3D9DeviceEx*         pDevice,
-            DxsoProgramType       ShaderStage,
-            DxsoConstantBuffers   BufferType);
-
   };
 
 }
\ No newline at end of file
diff -pruN 2.3.1-1/src/d3d9/d3d9_cursor.cpp 2.4-1/src/d3d9/d3d9_cursor.cpp
--- 2.3.1-1/src/d3d9/d3d9_cursor.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_cursor.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -26,14 +26,14 @@ namespace dxvk {
 
 
   HRESULT D3D9Cursor::SetHardwareCursor(UINT XHotSpot, UINT YHotSpot, const CursorBitmap& bitmap) {
-    DWORD mask[32];
+    CursorMask mask;
     std::memset(mask, ~0, sizeof(mask));
 
     ICONINFO info;
     info.fIcon    = FALSE;
     info.xHotspot = XHotSpot;
     info.yHotspot = YHotSpot;
-    info.hbmMask  = ::CreateBitmap(HardwareCursorWidth, HardwareCursorHeight, 1, 1,  mask);
+    info.hbmMask  = ::CreateBitmap(HardwareCursorWidth, HardwareCursorHeight, 1, 1,  &mask[0]);
     info.hbmColor = ::CreateBitmap(HardwareCursorWidth, HardwareCursorHeight, 1, 32, &bitmap[0]);
 
     if (m_hCursor != nullptr)
diff -pruN 2.3.1-1/src/d3d9/d3d9_cursor.h 2.4-1/src/d3d9/d3d9_cursor.h
--- 2.3.1-1/src/d3d9/d3d9_cursor.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_cursor.h	2024-07-10 09:56:14.000000000 +0000
@@ -11,11 +11,20 @@ namespace dxvk {
 
   // Format Size of 4 bytes (ARGB)
   using CursorBitmap = uint8_t[HardwareCursorHeight * HardwareCursorPitch];
+  // Monochrome mask (1 bit)
+  using CursorMask = uint8_t[HardwareCursorHeight * HardwareCursorWidth / 8];
 
   class D3D9Cursor {
 
   public:
 
+#ifdef _WIN32
+    ~D3D9Cursor() {
+      if (m_hCursor != nullptr)
+        ::DestroyCursor(m_hCursor);
+    }
+#endif
+
     void UpdateCursor(int X, int Y);
 
     BOOL ShowCursor(BOOL bShow);
diff -pruN 2.3.1-1/src/d3d9/d3d9_device.cpp 2.4-1/src/d3d9/d3d9_device.cpp
--- 2.3.1-1/src/d3d9/d3d9_device.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_device.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -56,6 +56,7 @@ namespace dxvk {
     , m_csThread        ( dxvkDevice, dxvkDevice->createContext(DxvkContextType::Primary) )
     , m_csChunk         ( AllocCsChunk() )
     , m_submissionFence (new sync::Fence())
+    , m_flushTracker    (m_d3d9Options.reproducibleCommandStream)
     , m_d3d9Interop     ( this )
     , m_d3d9On12        ( this )
     , m_d3d8Bridge      ( this ) {
@@ -594,7 +595,7 @@ namespace dxvk {
                             || (Usage & D3DUSAGE_DYNAMIC)
                             || IsVendorFormat(EnumerateFormat(Format));
 
-    if (FAILED(D3D9CommonTexture::NormalizeTextureProperties(this, &desc)))
+    if (FAILED(D3D9CommonTexture::NormalizeTextureProperties(this, D3DRTYPE_TEXTURE, &desc)))
       return D3DERR_INVALIDCALL;
 
     try {
@@ -664,7 +665,7 @@ namespace dxvk {
                             || (Usage & D3DUSAGE_DYNAMIC)
                             || IsVendorFormat(EnumerateFormat(Format));
 
-    if (FAILED(D3D9CommonTexture::NormalizeTextureProperties(this, &desc)))
+    if (FAILED(D3D9CommonTexture::NormalizeTextureProperties(this, D3DRTYPE_VOLUMETEXTURE, &desc)))
       return D3DERR_INVALIDCALL;
 
     try {
@@ -721,7 +722,7 @@ namespace dxvk {
                             || (Usage & D3DUSAGE_DYNAMIC)
                             || IsVendorFormat(EnumerateFormat(Format));
 
-    if (FAILED(D3D9CommonTexture::NormalizeTextureProperties(this, &desc)))
+    if (FAILED(D3D9CommonTexture::NormalizeTextureProperties(this, D3DRTYPE_CUBETEXTURE, &desc)))
       return D3DERR_INVALIDCALL;
 
     try {
@@ -1143,6 +1144,9 @@ namespace dxvk {
     if (unlikely((srcSubresource.aspectMask & (VK_IMAGE_ASPECT_DEPTH_BIT | VK_IMAGE_ASPECT_STENCIL_BIT)) && m_flags.test(D3D9DeviceFlag::InScene)))
       return D3DERR_INVALIDCALL;
 
+    if (unlikely(Filter != D3DTEXF_NONE && Filter != D3DTEXF_LINEAR && Filter != D3DTEXF_POINT))
+      return D3DERR_INVALIDCALL;
+
     VkExtent3D srcExtent = srcImage->mipLevelExtent(srcSubresource.mipLevel);
     VkExtent3D dstExtent = dstImage->mipLevelExtent(dstSubresource.mipLevel);
 
@@ -1221,8 +1225,51 @@ namespace dxvk {
       uint32_t(blitInfo.dstOffsets[1].y - blitInfo.dstOffsets[0].y),
       uint32_t(blitInfo.dstOffsets[1].z - blitInfo.dstOffsets[0].z) };
 
+    bool srcIsDepth = IsDepthFormat(srcFormat);
+    bool dstIsDepth = IsDepthFormat(dstFormat);
+    if (unlikely(srcIsDepth || dstIsDepth)) {
+      if (unlikely(!srcIsDepth || !dstIsDepth))
+        return D3DERR_INVALIDCALL;
+
+      if (unlikely(srcTextureInfo->Desc()->Discard || dstTextureInfo->Desc()->Discard))
+        return D3DERR_INVALIDCALL;
+
+      if (unlikely(srcCopyExtent.width != srcExtent.width || srcCopyExtent.height != srcExtent.height))
+        return D3DERR_INVALIDCALL;
+
+      if (unlikely(m_flags.test(D3D9DeviceFlag::InScene)))
+        return D3DERR_INVALIDCALL;
+    }
+
     // Copies would only work if the extents match. (ie. no stretching)
     bool stretch = srcCopyExtent != dstCopyExtent;
+
+    bool dstHasRTUsage = (dstTextureInfo->Desc()->Usage & (D3DUSAGE_RENDERTARGET | D3DUSAGE_DEPTHSTENCIL)) != 0;
+    bool dstIsSurface = dstTextureInfo->GetType() == D3DRTYPE_SURFACE;
+    if (stretch) {
+      if (unlikely(pSourceSurface == pDestSurface))
+        return D3DERR_INVALIDCALL;
+
+      if (unlikely(dstIsDepth))
+        return D3DERR_INVALIDCALL;
+
+      // The docs say that stretching is only allowed if the destination is either a render target surface or a render target texture.
+      // However in practice, using an offscreen plain surface in D3DPOOL_DEFAULT as the destination works fine.
+      // Using a texture without USAGE_RENDERTARGET as destination however does not.
+      if (unlikely(!dstIsSurface && !dstHasRTUsage))
+        return D3DERR_INVALIDCALL;
+    } else {
+      bool srcIsSurface = srcTextureInfo->GetType() == D3DRTYPE_SURFACE;
+      bool srcHasRTUsage = (srcTextureInfo->Desc()->Usage & (D3DUSAGE_RENDERTARGET | D3DUSAGE_DEPTHSTENCIL)) != 0;
+      // Non-stretching copies are only allowed if:
+      // - the destination is either a render target surface or a render target texture
+      // - both destination and source are depth stencil surfaces
+      // - both destination and source are offscreen plain surfaces.
+      // The only way to get a surface with resource type D3DRTYPE_SURFACE without USAGE_RT or USAGE_DS is CreateOffscreenPlainSurface.
+      if (unlikely((!dstHasRTUsage && (!dstIsSurface || !srcIsSurface || srcHasRTUsage)) && !m_isD3D8Compatible))
+        return D3DERR_INVALIDCALL;
+    }
+
     fastPath &= !stretch;
 
     if (!fastPath || needsResolve) {
@@ -2908,14 +2955,19 @@ namespace dxvk {
     auto slice = dst->GetBufferSlice<D3D9_COMMON_BUFFER_TYPE_REAL>();
          slice = slice.subSlice(offset, slice.length() - offset);
 
+    D3D9CompactVertexElements elements;
+    for (const D3DVERTEXELEMENT9& element : decl->GetElements()) {
+      elements.emplace_back(element);
+    }
+
     EmitCs([this,
-      cDecl          = ref(decl),
-      cVertexCount   = VertexCount,
-      cStartIndex    = SrcStartIndex,
-      cInstanceCount = GetInstanceCount(),
-      cBufferSlice   = slice
+      cVertexElements = std::move(elements),
+      cVertexCount    = VertexCount,
+      cStartIndex     = SrcStartIndex,
+      cInstanceCount  = GetInstanceCount(),
+      cBufferSlice    = slice
     ](DxvkContext* ctx) mutable {
-      Rc<DxvkShader> shader = m_swvpEmulator.GetShaderModule(this, cDecl);
+      Rc<DxvkShader> shader = m_swvpEmulator.GetShaderModule(this, std::move(cVertexElements));
 
       auto drawInfo = GenerateDrawInfo(D3DPT_POINTLIST, cVertexCount, cInstanceCount);
 
@@ -3798,7 +3850,7 @@ namespace dxvk {
     desc.IsAttachmentOnly   = TRUE;
     desc.IsLockable         = Lockable;
 
-    if (FAILED(D3D9CommonTexture::NormalizeTextureProperties(this, &desc)))
+    if (FAILED(D3D9CommonTexture::NormalizeTextureProperties(this, D3DRTYPE_TEXTURE, &desc)))
       return D3DERR_INVALIDCALL;
 
     try {
@@ -3846,7 +3898,7 @@ namespace dxvk {
     // Docs: Off-screen plain surfaces are always lockable, regardless of their pool types.
     desc.IsLockable         = TRUE;
 
-    if (FAILED(D3D9CommonTexture::NormalizeTextureProperties(this, &desc)))
+    if (FAILED(D3D9CommonTexture::NormalizeTextureProperties(this, D3DRTYPE_TEXTURE, &desc)))
       return D3DERR_INVALIDCALL;
 
     if (pSharedHandle != nullptr && Pool != D3DPOOL_DEFAULT)
@@ -3901,7 +3953,7 @@ namespace dxvk {
     // Docs don't say anything, so just assume it's lockable.
     desc.IsLockable         = TRUE;
 
-    if (FAILED(D3D9CommonTexture::NormalizeTextureProperties(this, &desc)))
+    if (FAILED(D3D9CommonTexture::NormalizeTextureProperties(this, D3DRTYPE_TEXTURE, &desc)))
       return D3DERR_INVALIDCALL;
 
     try {
@@ -5158,7 +5210,7 @@ namespace dxvk {
     return D3D_OK;
   }
 
-  
+
 
   void D3D9DeviceEx::UploadDynamicSysmemBuffers(
           UINT&                   FirstVertexIndex,
@@ -5190,25 +5242,58 @@ namespace dxvk {
     // First we calculate the size of that UP buffer slice
     // and store all sizes and offsets into it.
 
-    uint32_t upBufferSize = 0;
-    std::array<uint32_t, caps::MaxStreams> vboUPBufferOffsets = {};
-    std::array<uint32_t, caps::MaxStreams> vboUPBufferSizes = {};
-    for (uint32_t i = 0; i < caps::MaxStreams && dynamicSysmemVBOs; i++) {
-      vboUPBufferOffsets[i] = upBufferSize;
+    struct VBOCopy {
+      uint32_t srcOffset;
+      uint32_t dstOffset;
+      uint32_t copyBufferLength;
+      uint32_t copyElementCount;
+      uint32_t copyElementSize;
+      uint32_t copyElementStride;
+    };
+    uint32_t totalUpBufferSize = 0;
+    std::array<VBOCopy, caps::MaxStreams> vboCopies = {};
 
+    for (uint32_t i = 0; i < caps::MaxStreams && dynamicSysmemVBOs; i++) {
       auto* vbo = GetCommonBuffer(m_state.vertexBuffers[i].vertexBuffer);
       if (likely(vbo == nullptr)) {
-        vboUPBufferSizes[i] = 0;
         continue;
       }
-      const uint32_t vertexStride = m_state.vertexDecl->GetSize(i);
-      uint32_t offset = (FirstVertexIndex + BaseVertexIndex) * vertexStride;
+      const uint32_t vertexSize = m_state.vertexDecl->GetSize(i);
+      const uint32_t vertexStride = m_state.vertexBuffers[i].stride;
+      const uint32_t srcStride = vertexStride;
+      const uint32_t dstStride = std::min(vertexStride, vertexSize);
+
+      uint32_t elementCount = NumVertices;
+      if (m_state.streamFreq[i] & D3DSTREAMSOURCE_INSTANCEDATA) {
+        elementCount = GetInstanceCount();
+      }
+      const uint32_t vboOffset = m_state.vertexBuffers[i].offset;
+      const uint32_t vertexOffset = (FirstVertexIndex + BaseVertexIndex) * srcStride;
       const uint32_t vertexBufferSize = vbo->Desc()->Size;
-      if (offset < vertexBufferSize) {
-        const uint32_t vertexDataSize = std::min(NumVertices * vertexStride, vertexBufferSize - offset);
-        vboUPBufferSizes[i] = vertexDataSize;
-        upBufferSize += vertexDataSize;
+      const uint32_t srcOffset = vboOffset + vertexOffset;
+
+      if (unlikely(srcOffset > vertexBufferSize)) {
+        // All vertices are out of bounds
+        vboCopies[i].copyBufferLength = 0;
+      } else if (unlikely(srcOffset + elementCount * srcStride > vertexBufferSize)) {
+        // Some vertices are (partially) out of bounds
+        uint32_t boundVertexBufferRange = vertexBufferSize - vboOffset;
+        elementCount = boundVertexBufferRange / srcStride;
+        // Copy all complete vertices
+        vboCopies[i].copyBufferLength = elementCount * dstStride;
+        // Copy the remaining partial vertex
+        vboCopies[i].copyBufferLength += std::min(dstStride, boundVertexBufferRange % srcStride);
+      } else {
+        // No vertices are out of bounds
+        vboCopies[i].copyBufferLength = elementCount * dstStride;
       }
+
+      vboCopies[i].copyElementCount = elementCount;
+      vboCopies[i].copyElementStride = srcStride;
+      vboCopies[i].copyElementSize = dstStride;
+      vboCopies[i].srcOffset = srcOffset;
+      vboCopies[i].dstOffset = totalUpBufferSize;
+      totalUpBufferSize += vboCopies[i].copyBufferLength;
     }
 
     uint32_t iboUPBufferSize = 0;
@@ -5221,13 +5306,13 @@ namespace dxvk {
         uint32_t indexBufferSize = ibo->Desc()->Size;
         if (offset < indexBufferSize) {
           iboUPBufferSize = std::min(NumIndices * indexStride, indexBufferSize - offset);
-          iboUPBufferOffset = upBufferSize;
-          upBufferSize += iboUPBufferSize;
+          iboUPBufferOffset = totalUpBufferSize;
+          totalUpBufferSize += iboUPBufferSize;
         }
       }
     }
 
-    if (unlikely(upBufferSize == 0)) {
+    if (unlikely(totalUpBufferSize == 0)) {
       *pDynamicVBOs = false;
       if (pDynamicIBO)
         *pDynamicIBO = false;
@@ -5235,41 +5320,45 @@ namespace dxvk {
       return;
     }
 
-    auto upSlice = AllocUPBuffer(upBufferSize);
+    auto upSlice = AllocUPBuffer(totalUpBufferSize);
 
     // Now copy the actual data and bind it.
-    if (dynamicSysmemVBOs) {
-      for (uint32_t i = 0; i < caps::MaxStreams; i++) {
-        if (unlikely(vboUPBufferSizes[i] == 0)) {
-          EmitCs([
-            cStream = i
-          ](DxvkContext* ctx) {
-            ctx->bindVertexBuffer(cStream, DxvkBufferSlice(), 0);
-          });
-          m_flags.set(D3D9DeviceFlag::DirtyVertexBuffers);
-          continue;
-        }
-
-        auto* vbo = GetCommonBuffer(m_state.vertexBuffers[i].vertexBuffer);
-
-        const uint32_t vertexStride = m_state.vertexDecl->GetSize(i);
-        uint32_t offset = (BaseVertexIndex + FirstVertexIndex) * vertexStride + m_state.vertexBuffers[i].offset;
+    for (uint32_t i = 0; i < caps::MaxStreams && dynamicSysmemVBOs; i++) {
+      const VBOCopy& copy = vboCopies[i];
 
-        uint8_t* data = reinterpret_cast<uint8_t*>(upSlice.mapPtr) + vboUPBufferOffsets[i];
-        uint8_t* src = reinterpret_cast<uint8_t*>(vbo->GetMappedSlice().mapPtr) + offset;
-        std::memcpy(data, src, vboUPBufferSizes[i]);
+      if (likely(copy.copyBufferLength != 0)) {
+        const auto* vbo = GetCommonBuffer(m_state.vertexBuffers[i].vertexBuffer);
+        uint8_t* data = reinterpret_cast<uint8_t*>(upSlice.mapPtr) + copy.dstOffset;
+        const uint8_t* src = reinterpret_cast<uint8_t*>(vbo->GetMappedSlice().mapPtr) + copy.srcOffset;
 
-        auto vboSlice = upSlice.slice.subSlice(vboUPBufferOffsets[i], vboUPBufferSizes[i]);
-        EmitCs([
-          cStream      = i,
-          cBufferSlice = std::move(vboSlice),
-          cStride      = vertexStride
-        ](DxvkContext* ctx) mutable {
-          ctx->bindVertexBuffer(cStream, std::move(cBufferSlice), cStride);
-        });
-        m_flags.set(D3D9DeviceFlag::DirtyVertexBuffers);
+        if (likely(copy.copyElementStride == copy.copyElementSize)) {
+          std::memcpy(data, src, copy.copyBufferLength);
+        } else {
+          for (uint32_t j = 0; j < copy.copyElementCount; j++) {
+            std::memcpy(data + j * copy.copyElementSize, src + j * copy.copyElementStride, copy.copyElementSize);
+          }
+          if (unlikely(copy.copyBufferLength > copy.copyElementCount * copy.copyElementSize)) {
+            // Partial vertex at the end
+            std::memcpy(
+              data + copy.copyElementCount * copy.copyElementSize,
+              src + copy.copyElementCount * copy.copyElementStride,
+              copy.copyBufferLength - copy.copyElementCount * copy.copyElementSize);
+          }
+        }
       }
 
+      auto vboSlice = upSlice.slice.subSlice(copy.dstOffset, copy.copyBufferLength);
+      EmitCs([
+        cStream      = i,
+        cBufferSlice = std::move(vboSlice),
+        cStride      = copy.copyElementSize
+      ](DxvkContext* ctx) mutable {
+        ctx->bindVertexBuffer(cStream, std::move(cBufferSlice), cStride);
+      });
+      m_flags.set(D3D9DeviceFlag::DirtyVertexBuffers);
+    }
+
+    if (dynamicSysmemVBOs) {
       // Change the draw call parameters to reflect the changed vertex buffers
       if (NumIndices != 0) {
         BaseVertexIndex = -FirstVertexIndex;
@@ -5292,7 +5381,7 @@ namespace dxvk {
         uint8_t* data = reinterpret_cast<uint8_t*>(upSlice.mapPtr) + iboUPBufferOffset;
         uint8_t* src = reinterpret_cast<uint8_t*>(ibo->GetMappedSlice().mapPtr) + offset;
         std::memcpy(data, src, iboUPBufferSize);
-        
+
         auto iboSlice = upSlice.slice.subSlice(iboUPBufferOffset, iboUPBufferSize);
         EmitCs([
           cBufferSlice = std::move(iboSlice),
@@ -7235,6 +7324,7 @@ namespace dxvk {
         key.Data.Contents.TransformFlags  |= transformFlags << (i * 3);
         key.Data.Contents.TexcoordFlags   |= indexFlags     << (i * 3);
         key.Data.Contents.TexcoordIndices |= index          << (i * 3);
+        key.Data.Contents.Projected       |= ((m_state.textureStages[i][DXVK_TSS_TEXTURETRANSFORMFLAGS] & D3DTTFF_PROJECTED) == D3DTTFF_PROJECTED) << i;
       }
 
       key.Data.Contents.TexcoordDeclMask = m_state.vertexDecl != nullptr ? m_state.vertexDecl->GetTexcoordMask() : 0;
@@ -7405,6 +7495,8 @@ namespace dxvk {
 
         stage.Projected      = (ttff & D3DTTFF_PROJECTED) ? 1      : 0;
         stage.ProjectedCount = (ttff & D3DTTFF_PROJECTED) ? count  : 0;
+
+        stage.SampleDref = (m_depthTextures & (1 << idx)) != 0;
       }
 
       auto& stage0 = key.Stages[0].Contents;
@@ -7687,6 +7779,8 @@ namespace dxvk {
     rs[D3DRS_CLIPPLANEENABLE] = 0;
     m_flags.set(D3D9DeviceFlag::DirtyClipPlanes);
 
+    const VkPhysicalDeviceLimits& limits = m_dxvkDevice->adapter()->deviceProperties().limits;
+
     rs[D3DRS_POINTSPRITEENABLE]          = FALSE;
     rs[D3DRS_POINTSCALEENABLE]           = FALSE;
     rs[D3DRS_POINTSCALE_A]               = bit::cast<DWORD>(1.0f);
@@ -7694,7 +7788,7 @@ namespace dxvk {
     rs[D3DRS_POINTSCALE_C]               = bit::cast<DWORD>(0.0f);
     rs[D3DRS_POINTSIZE]                  = bit::cast<DWORD>(1.0f);
     rs[D3DRS_POINTSIZE_MIN]              = bit::cast<DWORD>(1.0f);
-    rs[D3DRS_POINTSIZE_MAX]              = bit::cast<DWORD>(64.0f);
+    rs[D3DRS_POINTSIZE_MAX]              = bit::cast<DWORD>(limits.pointSizeRange[1]);
     UpdatePushConstant<D3D9RenderStateItem::PointSize>();
     UpdatePushConstant<D3D9RenderStateItem::PointSizeMin>();
     UpdatePushConstant<D3D9RenderStateItem::PointSizeMax>();
@@ -7891,7 +7985,7 @@ namespace dxvk {
       // Docs: Also note that - unlike textures - swap chain back buffers, render targets [..] can be locked
       desc.IsLockable         = TRUE;
 
-      if (FAILED(D3D9CommonTexture::NormalizeTextureProperties(this, &desc)))
+      if (FAILED(D3D9CommonTexture::NormalizeTextureProperties(this, D3DRTYPE_TEXTURE, &desc)))
         return D3DERR_NOTAVAILABLE;
 
       m_autoDepthStencil = new D3D9Surface(this, &desc, nullptr, nullptr);
diff -pruN 2.3.1-1/src/d3d9/d3d9_fixed_function.cpp 2.4-1/src/d3d9/d3d9_fixed_function.cpp
--- 2.3.1-1/src/d3d9/d3d9_fixed_function.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_fixed_function.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -15,6 +15,7 @@ namespace dxvk {
   D3D9FixedFunctionOptions::D3D9FixedFunctionOptions(const D3D9Options* options) {
     invariantPosition = options->invariantPosition;
     forceSampleRateShading = options->forceSampleRateShading;
+    drefScaling = options->drefScaling;
   }
 
   uint32_t DoFixedFunctionFog(D3D9ShaderSpecConstantManager& spec, SpirvModule& spvModule, const D3D9FogContext& fogCtx) {
@@ -336,7 +337,7 @@ namespace dxvk {
   }
 
 
-  uint32_t SetupRenderStateBlock(SpirvModule& spvModule, uint32_t count) {
+  uint32_t SetupRenderStateBlock(SpirvModule& spvModule) {
     uint32_t floatType = spvModule.defFloatType(32);
     uint32_t uintType  = spvModule.defIntType(32, 0);
     uint32_t vec3Type  = spvModule.defVectorType(floatType, 3);
@@ -357,7 +358,7 @@ namespace dxvk {
       floatType,
     }};
 
-    uint32_t rsStruct = spvModule.defStructTypeUnique(count, rsMembers.data());
+    uint32_t rsStruct = spvModule.defStructTypeUnique(rsMembers.size(), rsMembers.data());
     uint32_t rsBlock = spvModule.newVar(
       spvModule.defPointerType(rsStruct, spv::StorageClassPushConstant),
       spv::StorageClassPushConstant);
@@ -369,9 +370,6 @@ namespace dxvk {
 
     uint32_t memberIdx = 0;
     auto SetMemberName = [&](const char* name, uint32_t offset) {
-      if (memberIdx >= count)
-        return;
-
       spvModule.setDebugMemberName   (rsStruct, memberIdx, name);
       spvModule.memberDecorateOffset (rsStruct, memberIdx, offset);
       memberIdx++;
@@ -781,8 +779,6 @@ namespace dxvk {
     uint32_t              m_inputMask = 0u;
     uint32_t              m_outputMask = 0u;
     uint32_t              m_flatShadingMask = 0u;
-    uint32_t              m_pushConstOffset = 0u;
-    uint32_t              m_pushConstSize = 0u;
 
     DxsoProgramType       m_programType;
     D3D9FFShaderKeyVS     m_vsKey;
@@ -892,8 +888,8 @@ namespace dxvk {
     info.inputMask = m_inputMask;
     info.outputMask = m_outputMask;
     info.flatShadingInputs = m_flatShadingMask;
-    info.pushConstOffset = m_pushConstOffset;
-    info.pushConstSize = m_pushConstSize;
+    info.pushConstStages = VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT;
+    info.pushConstSize = sizeof(D3D9RenderStateInfo);
 
     return new DxvkShader(info, m_module.compile());
   }
@@ -1111,37 +1107,69 @@ namespace dxvk {
       const uint32_t wIndex = 3;
 
       uint32_t flags = (m_vsKey.Data.Contents.TransformFlags >> (i * 3)) & 0b111;
-      uint32_t count;
+
+      // Passing 0xffffffff results in it getting clamped to the dimensions of the texture coords and getting treated as PROJECTED
+      // but D3D9 does not apply the transformation matrix.
+      bool applyTransform = flags > D3DTTFF_COUNT1 && flags <= D3DTTFF_COUNT4;
+
+      uint32_t count = std::min(flags, 4u);
+
+      // A projection component index of 4 means we won't do projection
+      uint32_t projIndex = count != 0 ? count - 1 : 4;
+
       switch (inputFlags) {
         default:
         case (DXVK_TSS_TCI_PASSTHRU >> TCIOffset):
           transformed = m_vs.in.TEXCOORD[inputIndex & 0xFF];
-          // flags is actually the number of elements that get passed
-          // to the rasterizer.
-          count = flags;
-          if (texcoordCount) {
-            // Clamp by the number of elements in the texcoord input.
-            if (!count || count > texcoordCount)
-              count = texcoordCount;
+
+          if (texcoordCount < 4) {
+            // Vulkan sets the w component to 1.0 if that's not provided by the vertex buffer, D3D9 expects 0 here
+            transformed = m_module.opCompositeInsert(m_vec4Type, m_module.constf32(0), transformed, 1, &wIndex);
           }
-          else
-            flags = D3DTTFF_DISABLE;
+
+          if (applyTransform && !m_vsKey.Data.Contents.HasPositionT) {
+            /*This doesn't happen every time and I cannot figure out the difference between when it does and doesn't.
+            Keep it disabled for now, it's more likely that games rely on the zero texcoord than the weird 1 here.
+            if (texcoordCount <= 1) {
+              // y gets padded to 1 for some reason
+              uint32_t idx = 1;
+              transformed = m_module.opCompositeInsert(m_vec4Type, m_module.constf32(1), transformed, 1, &idx);
+            }*/
+
+            if (texcoordCount >= 1 && texcoordCount < 4) {
+              // The first component after the last one thats backed by a vertex buffer gets padded to 1 for some reason.
+              uint32_t idx = texcoordCount;
+              transformed = m_module.opCompositeInsert(m_vec4Type, m_module.constf32(1), transformed, 1, &idx);
+            }
+          } else if (texcoordCount != 0 && !applyTransform) {
+            // COUNT0, COUNT1, COUNT > 4 => take count from vertex decl if that's not zero
+            count = texcoordCount;
+          }
+
+          projIndex = count != 0 ? count - 1 : 4;
           break;
 
         case (DXVK_TSS_TCI_CAMERASPACENORMAL >> TCIOffset):
           transformed = outNrm;
-          count = 4;
+          if (!applyTransform) {
+            count = 3;
+            projIndex = 4;
+          }
           break;
 
         case (DXVK_TSS_TCI_CAMERASPACEPOSITION >> TCIOffset):
-          transformed = m_module.opCompositeInsert(m_vec4Type, m_module.constf32(1.0f), vtx, 1, &wIndex);
-          count = 4;
+          transformed = vtx;
+          if (!applyTransform) {
+            Logger::warn(str::format("!applyTransform flags: ", flags, " projidx: ", projIndex));
+            count = 3;
+            projIndex = 4;
+          }
           break;
 
         case (DXVK_TSS_TCI_CAMERASPACEREFLECTIONVECTOR >> TCIOffset): {
           uint32_t vtx3 = m_module.opVectorShuffle(m_vec3Type, vtx, vtx, 3, indices.data());
           vtx3 = m_module.opNormalize(m_vec3Type, vtx3);
-          
+
           uint32_t reflection = m_module.opReflect(m_vec3Type, vtx3, normal);
 
           std::array<uint32_t, 4> transformIndices;
@@ -1150,7 +1178,10 @@ namespace dxvk {
           transformIndices[3] = m_module.constf32(1.0f);
 
           transformed = m_module.opCompositeConstruct(m_vec4Type, transformIndices.size(), transformIndices.data());
-          count = 4;
+          if (!applyTransform) {
+            count = 3;
+            projIndex = 4;
+          }
           break;
         }
 
@@ -1174,35 +1205,27 @@ namespace dxvk {
           transformIndices[3] = m_module.constf32(1.0f);
 
           transformed = m_module.opCompositeConstruct(m_vec4Type, transformIndices.size(), transformIndices.data());
-          count = 4;
           break;
         }
       }
 
-      uint32_t type = flags;
-      if (type != D3DTTFF_DISABLE) {
-        if (!m_vsKey.Data.Contents.HasPositionT) {
-          for (uint32_t j = count; j < 4; j++) {
-            // If we're outside the component count of the vertex decl for this texcoord then we pad with zeroes.
-            // Otherwise, pad with ones.
-
-            // Very weird quirk in order to get texcoord transforms to work like they do in native.
-            // In future, maybe we could sort this out properly by chopping matrices of different sizes, but thats
-            // a project for another day.
-            uint32_t texcoordCount = (m_vsKey.Data.Contents.TexcoordDeclMask >> (3 * inputIndex)) & 0x7;
-            uint32_t value = j > texcoordCount ? m_module.constf32(0) : m_module.constf32(1);
-            transformed = m_module.opCompositeInsert(m_vec4Type, value, transformed, 1, &j);
-          }
-
-          transformed = m_module.opVectorTimesMatrix(m_vec4Type, transformed, m_vs.constants.texcoord[i]);
-        }
+      if (applyTransform && !m_vsKey.Data.Contents.HasPositionT) {
+        transformed = m_module.opVectorTimesMatrix(m_vec4Type, transformed, m_vs.constants.texcoord[i]);
+      }
 
-        // Pad the unused section of it with the value for projection.
-        uint32_t lastIdx = count - 1;
-        uint32_t projValue = m_module.opCompositeExtract(m_floatType, transformed, 1, &lastIdx);
+      if (m_vsKey.Data.Contents.Projected && projIndex < 4) {
+        // The projection idx is always based on the flags, even when the input mode is not DXVK_TSS_TCI_PASSTHRU.
+        uint32_t projValue = m_module.opCompositeExtract(m_floatType, transformed, 1, &projIndex);
+
+        // The w component is only used for projection or unused, so always insert the component that's supposed to be divided by there.
+        // The fragment shader will then decide whether to project or not.
+        transformed = m_module.opCompositeInsert(m_vec4Type, projValue, transformed, 1, &wIndex);
+      }
 
-        for (uint32_t j = count; j < 4; j++)
-          transformed = m_module.opCompositeInsert(m_vec4Type, projValue, transformed, 1, &j);
+      uint32_t totalComponents = (m_vsKey.Data.Contents.Projected && projIndex < 4) ? 3 : 4;
+      for (uint32_t i = count; i < totalComponents; i++) {
+        // Discard the components that exceed the specified D3DTTFF_COUNT
+        transformed = m_module.opCompositeInsert(m_vec4Type, m_module.constf32(0), transformed, 1, &i);
       }
 
       m_module.opStore(m_vs.out.TEXCOORD[i], transformed);
@@ -1315,6 +1338,8 @@ namespace dxvk {
         uint32_t midDot = m_module.opDot(m_floatType, normal, mid);
                  midDot = m_module.opFClamp(m_floatType, midDot, m_module.constf32(0.0f), m_module.constf32(1.0f));
         uint32_t doSpec = m_module.opFOrdGreaterThan(bool_t, midDot, m_module.constf32(0.0f));
+                 doSpec = m_module.opLogicalAnd(bool_t, doSpec, m_module.opFOrdGreaterThan(m_floatType, hitDot, m_module.constf32(0.0f)));
+
         uint32_t specularness = m_module.opPow(m_floatType, midDot, m_vs.constants.materialPower);
                  specularness = m_module.opFMul(m_floatType, specularness, atten);
                  specularness = m_module.opSelect(m_floatType, doSpec, specularness, m_module.constf32(0.0f));
@@ -1384,20 +1409,7 @@ namespace dxvk {
 
 
   void D3D9FFShaderCompiler::setupRenderStateInfo() {
-    uint32_t count;
-
-    if (m_programType == DxsoProgramType::PixelShader) {
-      m_pushConstOffset = 0;
-      m_pushConstSize   = offsetof(D3D9RenderStateInfo, pointSize);
-      count = 5;
-    }
-    else {
-      m_pushConstOffset = offsetof(D3D9RenderStateInfo, pointSize);
-      m_pushConstSize   = sizeof(float) * 6;
-      count = 11;
-    }
-
-    m_rsBlock = SetupRenderStateBlock(m_module, count);
+    m_rsBlock = SetupRenderStateBlock(m_module);
   }
 
 
@@ -1789,6 +1801,11 @@ namespace dxvk {
         return coords;
       };
 
+      auto ScalarReplicate = [&](uint32_t reg) {
+        std::array<uint32_t, 4> replicant = { reg, reg, reg, reg };
+        return m_module.opCompositeConstruct(m_vec4Type, replicant.size(), replicant.data());
+      };
+
       auto GetTexture = [&]() {
         if (!processedTexture) {
           SpirvImageOperands imageOperands;
@@ -1808,22 +1825,17 @@ namespace dxvk {
           texcoord = m_module.opVectorShuffle(texcoord_t,
             texcoord, texcoord, texcoordCnt, indices.data());
 
-          uint32_t projIdx = m_fsKey.Stages[i].Contents.ProjectedCount;
-          if (projIdx == 0 || projIdx > texcoordCnt) {
-            projIdx = 4; // Always use w if ProjectedCount is 0.
-          }
-          --projIdx;
-
+          bool shouldProject = m_fsKey.Stages[i].Contents.Projected;
           uint32_t projValue = 0;
 
-          if (m_fsKey.Stages[i].Contents.Projected) {
+          if (shouldProject) {
+            // Always use w, the vertex shader puts the correct value there.
+            const uint32_t projIdx = 3;
             projValue = m_module.opCompositeExtract(m_floatType, m_ps.in.TEXCOORD[i], 1, &projIdx);
             uint32_t insertIdx = texcoordCnt - 1;
             texcoord = m_module.opCompositeInsert(texcoord_t, projValue, texcoord, 1, &insertIdx);
           }
 
-          bool shouldProject = m_fsKey.Stages[i].Contents.Projected;
-
           if (i != 0 && (
             m_fsKey.Stages[i - 1].Contents.ColorOp == D3DTOP_BUMPENVMAP ||
             m_fsKey.Stages[i - 1].Contents.ColorOp == D3DTOP_BUMPENVMAPLUMINANCE)) {
@@ -1837,10 +1849,23 @@ namespace dxvk {
             shouldProject = false;
           }
 
-          if (shouldProject)
+          if (unlikely(stage.SampleDref)) {
+            uint32_t component = 2;
+            uint32_t reference = m_module.opCompositeExtract(m_floatType, texcoord, 1, &component);
+
+            // [D3D8] Scale Dref to [0..(2^N - 1)] for D24S8 and D16 if Dref scaling is enabled
+            if (m_options.drefScaling) {
+              uint32_t maxDref = m_module.constf32(1.0f / (float(1 << m_options.drefScaling) - 1.0f));
+              reference        = m_module.opFMul(m_floatType, reference, maxDref);
+            }
+
+            texture = m_module.opImageSampleDrefImplicitLod(m_floatType, imageVarId, texcoord, reference, imageOperands);
+            texture = ScalarReplicate(texture);
+          } else if (shouldProject) {
             texture = m_module.opImageSampleProjImplicitLod(m_vec4Type, imageVarId, texcoord, imageOperands);
-          else
+          } else {
             texture = m_module.opImageSampleImplicitLod(m_vec4Type, imageVarId, texcoord, imageOperands);
+          }
 
           if (i != 0 && m_fsKey.Stages[i - 1].Contents.ColorOp == D3DTOP_BUMPENVMAPLUMINANCE) {
             uint32_t index = m_module.constu32(D3D9SharedPSStages_Count * (i - 1) + D3D9SharedPSStages_BumpEnvLScale);
@@ -1868,11 +1893,6 @@ namespace dxvk {
         return texture;
       };
 
-      auto ScalarReplicate = [&](uint32_t reg) {
-        std::array<uint32_t, 4> replicant = { reg, reg, reg, reg };
-        return m_module.opCompositeConstruct(m_vec4Type, replicant.size(), replicant.data());
-      };
-
       auto AlphaReplicate = [&](uint32_t reg) {
         uint32_t alphaComponentId = 3;
         uint32_t alpha = m_module.opCompositeExtract(m_floatType, reg, 1, &alphaComponentId);
@@ -2269,6 +2289,11 @@ namespace dxvk {
           dimensionality = spv::Dim2D;
           sampler.texcoordCnt = 2;
           viewType       = VK_IMAGE_VIEW_TYPE_2D;
+
+          // Z coordinate for Dref sampling
+          if (m_fsKey.Stages[i].Contents.SampleDref)
+            sampler.texcoordCnt++;
+
           break;
         case D3DRTYPE_CUBETEXTURE:
           dimensionality = spv::DimCube;
diff -pruN 2.3.1-1/src/d3d9/d3d9_fixed_function.h 2.4-1/src/d3d9/d3d9_fixed_function.h
--- 2.3.1-1/src/d3d9/d3d9_fixed_function.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_fixed_function.h	2024-07-10 09:56:14.000000000 +0000
@@ -48,10 +48,15 @@ namespace dxvk {
   struct D3D9FixedFunctionOptions {
     D3D9FixedFunctionOptions(const D3D9Options* options);
 
-    bool invariantPosition;
-    bool forceSampleRateShading;
+    bool    invariantPosition;
+    bool    forceSampleRateShading;
+    int32_t drefScaling;
   };
 
+  constexpr float GetDrefScaleFactor(int32_t bitDepth) {
+    return 1.0f / (float(1 << bitDepth) - 1.0f);
+  }
+
   // Returns new oFog if VS
   // Returns new oColor if PS
   uint32_t DoFixedFunctionFog(D3D9ShaderSpecConstantManager& spec, SpirvModule& spvModule, const D3D9FogContext& fogCtx);
@@ -59,7 +64,7 @@ namespace dxvk {
   void DoFixedFunctionAlphaTest(SpirvModule& spvModule, const D3D9AlphaTestContext& ctx);
 
   // Returns a render state block
-  uint32_t SetupRenderStateBlock(SpirvModule& spvModule, uint32_t count);
+  uint32_t SetupRenderStateBlock(SpirvModule& spvModule);
 
   struct D3D9PointSizeInfoVS {
     uint32_t defaultValue;
@@ -128,9 +133,11 @@ namespace dxvk {
         uint32_t VertexBlendCount   : 3;
 
         uint32_t VertexClipping     : 1;
+
+        uint32_t Projected : 8;
       } Contents;
 
-      uint32_t Primitive[4];
+      uint32_t Primitive[5];
     };
   };
 
@@ -163,6 +170,7 @@ namespace dxvk {
         uint32_t     Projected    : 1;
 
         uint32_t     ProjectedCount : 3;
+        uint32_t     SampleDref     : 1;
 
         uint32_t     TextureBound : 1;
 
diff -pruN 2.3.1-1/src/d3d9/d3d9_format.cpp 2.4-1/src/d3d9/d3d9_format.cpp
--- 2.3.1-1/src/d3d9/d3d9_format.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_format.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -438,9 +438,15 @@ namespace dxvk {
   D3D9VkFormatTable::D3D9VkFormatTable(
     const Rc<DxvkAdapter>& adapter,
     const D3D9Options&     options) {
-    m_dfSupport = options.supportDFFormats;
+
+    const auto& props = adapter->deviceProperties();
+    uint32_t vendorId  = options.customVendorId == -1 ? props.vendorID : uint32_t(options.customVendorId);
+
+    // NVIDIA does not natively support any DF formats
+    m_dfSupport = vendorId == uint32_t(DxvkGpuVendor::Nvidia) ? false : options.supportDFFormats;
     m_x4r4g4b4Support = options.supportX4R4G4B4;
-    m_d32supportFinal = options.supportD32;
+    // Only AMD supports D16_LOCKABLE natively
+    m_d16lockableSupport = vendorId == uint32_t(DxvkGpuVendor::Amd) ? true : options.supportD16Lockable;
 
     // AMD do not support 24-bit depth buffers on Vulkan,
     // so we have to fall back to a 32-bit depth format.
@@ -473,13 +479,13 @@ namespace dxvk {
     if (Format == D3D9Format::X4R4G4B4 && !m_x4r4g4b4Support)
       return D3D9_VK_FORMAT_MAPPING();
 
-    if (Format == D3D9Format::DF16 && !m_dfSupport)
+    if (Format == D3D9Format::D16_LOCKABLE && !m_d16lockableSupport)
       return D3D9_VK_FORMAT_MAPPING();
 
-    if (Format == D3D9Format::DF24 && !m_dfSupport)
+    if (Format == D3D9Format::DF16 && !m_dfSupport)
       return D3D9_VK_FORMAT_MAPPING();
 
-    if (Format == D3D9Format::D32 && !m_d32supportFinal)
+    if (Format == D3D9Format::DF24 && !m_dfSupport)
       return D3D9_VK_FORMAT_MAPPING();
     
     if (!m_d24s8Support && mapping.FormatColor == VK_FORMAT_D24_UNORM_S8_UINT)
diff -pruN 2.3.1-1/src/d3d9/d3d9_format.h 2.4-1/src/d3d9/d3d9_format.h
--- 2.3.1-1/src/d3d9/d3d9_format.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_format.h	2024-07-10 09:56:14.000000000 +0000
@@ -217,7 +217,7 @@ namespace dxvk {
 
     bool m_dfSupport;
     bool m_x4r4g4b4Support;
-    bool m_d32supportFinal;
+    bool m_d16lockableSupport;
   };
 
   inline bool IsFourCCFormat(D3D9Format format) {
diff -pruN 2.3.1-1/src/d3d9/d3d9_format_helpers.cpp 2.4-1/src/d3d9/d3d9_format_helpers.cpp
--- 2.3.1-1/src/d3d9/d3d9_format_helpers.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_format_helpers.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -134,7 +134,7 @@ namespace dxvk {
     info.stage = VK_SHADER_STAGE_COMPUTE_BIT;
     info.bindingCount = bindings.size();
     info.bindings = bindings.data();
-    info.pushConstOffset = 0;
+    info.pushConstStages = VK_SHADER_STAGE_COMPUTE_BIT;
     info.pushConstSize = sizeof(VkExtent2D);
 
     return new DxvkShader(info, std::move(code));
diff -pruN 2.3.1-1/src/d3d9/d3d9_interface.cpp 2.4-1/src/d3d9/d3d9_interface.cpp
--- 2.3.1-1/src/d3d9/d3d9_interface.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_interface.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -354,6 +354,12 @@ namespace dxvk {
     || pPresentationParameters    == nullptr)
       return D3DERR_INVALIDCALL;
 
+    // creating a device with D3DCREATE_PUREDEVICE only works in conjunction
+    // with D3DCREATE_HARDWARE_VERTEXPROCESSING on native drivers
+    if (BehaviorFlags & D3DCREATE_PUREDEVICE &&
+    !(BehaviorFlags & D3DCREATE_HARDWARE_VERTEXPROCESSING))
+      return D3DERR_INVALIDCALL;
+
     auto* adapter = GetAdapter(Adapter);
 
     if (adapter == nullptr)
diff -pruN 2.3.1-1/src/d3d9/d3d9_options.cpp 2.4-1/src/d3d9/d3d9_options.cpp
--- 2.3.1-1/src/d3d9/d3d9_options.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_options.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -37,8 +37,8 @@ namespace dxvk {
     this->customDeviceId        = parsePciId(config.getOption<std::string>("d3d9.customDeviceId"));
     this->customDeviceDesc      =            config.getOption<std::string>("d3d9.customDeviceDesc");
 
-    const int32_t vendorId = this->customDeviceId != -1
-      ? this->customDeviceId
+    const uint32_t vendorId = this->customVendorId != -1
+      ? this->customVendorId
       : (adapter != nullptr ? adapter->deviceProperties().vendorID : 0);
 
     this->maxFrameLatency               = config.getOption<int32_t>     ("d3d9.maxFrameLatency",               0);
@@ -55,12 +55,12 @@ namespace dxvk {
     this->maxAvailableMemory            = config.getOption<int32_t>     ("d3d9.maxAvailableMemory",            4096);
     this->supportDFFormats              = config.getOption<bool>        ("d3d9.supportDFFormats",              true);
     this->supportX4R4G4B4               = config.getOption<bool>        ("d3d9.supportX4R4G4B4",               true);
-    this->supportD32                    = config.getOption<bool>        ("d3d9.supportD32",                    true);
+    this->supportD16Lockable            = config.getOption<bool>        ("d3d9.supportD16Lockable",            false);
     this->useD32forD24                  = config.getOption<bool>        ("d3d9.useD32forD24",                  false);
     this->disableA8RT                   = config.getOption<bool>        ("d3d9.disableA8RT",                   false);
     this->invariantPosition             = config.getOption<bool>        ("d3d9.invariantPosition",             true);
     this->memoryTrackTest               = config.getOption<bool>        ("d3d9.memoryTrackTest",               false);
-    this->supportVCache                 = config.getOption<bool>        ("d3d9.supportVCache",                 vendorId == 0x10de);
+    this->supportVCache                 = config.getOption<bool>        ("d3d9.supportVCache",                 vendorId == uint32_t(DxvkGpuVendor::Nvidia));
     this->enableDialogMode              = config.getOption<bool>        ("d3d9.enableDialogMode",              false);
     this->forceSamplerTypeSpecConstants = config.getOption<bool>        ("d3d9.forceSamplerTypeSpecConstants", false);
     this->forceSwapchainMSAA            = config.getOption<int32_t>     ("d3d9.forceSwapchainMSAA",            -1);
@@ -77,6 +77,10 @@ namespace dxvk {
     this->samplerLodBias                = config.getOption<float>       ("d3d9.samplerLodBias",                0.0f);
     this->clampNegativeLodBias          = config.getOption<bool>        ("d3d9.clampNegativeLodBias",          false);
     this->countLosableResources         = config.getOption<bool>        ("d3d9.countLosableResources",         true);
+    this->reproducibleCommandStream     = config.getOption<bool>        ("d3d9.reproducibleCommandStream",     false);
+
+    // D3D8 options
+    this->drefScaling                   = config.getOption<int32_t>     ("d3d8.scaleDref",                     0);
 
     // Clamp LOD bias so that people don't abuse this in unintended ways
     this->samplerLodBias = dxvk::fclamp(this->samplerLodBias, -2.0f, 1.0f);
@@ -90,8 +94,8 @@ namespace dxvk {
       d3d9FloatEmulation = D3D9FloatEmulation::Enabled;
     } else {
       bool hasMulz = adapter != nullptr
-                  && (adapter->matchesDriver(VK_DRIVER_ID_MESA_RADV, 0, 0)
-                   || adapter->matchesDriver(VK_DRIVER_ID_MESA_NVK, 0, 0));
+                  && (adapter->matchesDriver(VK_DRIVER_ID_MESA_RADV)
+                   || adapter->matchesDriver(VK_DRIVER_ID_MESA_NVK));
       d3d9FloatEmulation = hasMulz ? D3D9FloatEmulation::Strict : D3D9FloatEmulation::Enabled;
     }
 
diff -pruN 2.3.1-1/src/d3d9/d3d9_options.h 2.4-1/src/d3d9/d3d9_options.h
--- 2.3.1-1/src/d3d9/d3d9_options.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_options.h	2024-07-10 09:56:14.000000000 +0000
@@ -78,8 +78,8 @@ namespace dxvk {
     /// Support X4R4G4B4
     bool supportX4R4G4B4;
 
-    /// Support D32
-    bool supportD32;
+    /// Support D16_LOCKABLE
+    bool supportD16Lockable;
 
     /// Use D32f for D24
     bool useD32forD24;
@@ -155,6 +155,14 @@ namespace dxvk {
 
     /// Disable counting losable resources and rejecting calls to Reset() if any are still alive
     bool countLosableResources;
+
+    /// Ensure that for the same D3D commands the output VK commands
+    /// don't change between runs. Useful for comparative benchmarking,
+    /// can negatively affect performance.
+    bool reproducibleCommandStream;
+
+    /// Enable depth texcoord Z (Dref) scaling (D3D8 quirk)
+    int32_t drefScaling;
   };
 
 }
diff -pruN 2.3.1-1/src/d3d9/d3d9_query.cpp 2.4-1/src/d3d9/d3d9_query.cpp
--- 2.3.1-1/src/d3d9/d3d9_query.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_query.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -41,11 +41,6 @@ namespace dxvk {
       case D3DQUERYTYPE_TIMESTAMPFREQ:
         break;
 
-      case D3DQUERYTYPE_VERTEXSTATS:
-        m_query[0] = dxvkDevice->createGpuQuery(
-          VK_QUERY_TYPE_PIPELINE_STATISTICS, 0, 0);
-        break;
-
       default:
         throw DxvkError(str::format("D3D9Query: Unsupported query type ", m_queryType));
     }
@@ -222,12 +217,10 @@ namespace dxvk {
 
       switch (m_queryType) {
         case D3DQUERYTYPE_VCACHE:
-          // Don't know what the hell any of this means.
-          // Nor do I care. This just makes games work.
-          m_dataCache.VCache.Pattern     = MAKEFOURCC('H', 'C', 'A', 'C');
+          m_dataCache.VCache.Pattern     = MAKEFOURCC('C', 'A', 'C', 'H');
           m_dataCache.VCache.OptMethod   = 1;
-          m_dataCache.VCache.CacheSize   = 24;
-          m_dataCache.VCache.MagicNumber = 20;
+          m_dataCache.VCache.CacheSize   = 16;
+          m_dataCache.VCache.MagicNumber = 7;
           break;
 
         case D3DQUERYTYPE_OCCLUSION:
@@ -246,11 +239,6 @@ namespace dxvk {
           m_dataCache.TimestampFreq = GetTimestampQueryFrequency();
           break;
 
-        case D3DQUERYTYPE_VERTEXSTATS:
-          m_dataCache.VertexStats.NumRenderedTriangles      = queryData[0].statistic.iaPrimitives;
-          m_dataCache.VertexStats.NumExtraClippingTriangles = queryData[0].statistic.clipPrimitives;
-          break;
-
         default:
           break;
       }
@@ -276,7 +264,6 @@ namespace dxvk {
   void D3D9Query::Begin(DxvkContext* ctx) {
     switch (m_queryType) {
       case D3DQUERYTYPE_OCCLUSION:
-      case D3DQUERYTYPE_VERTEXSTATS:
         ctx->beginQuery(m_query[0]);
         break;
 
@@ -296,7 +283,6 @@ namespace dxvk {
         ctx->writeTimestamp(m_query[0]);
         break;
 
-      case D3DQUERYTYPE_VERTEXSTATS:
       case D3DQUERYTYPE_OCCLUSION:
         ctx->endQuery(m_query[0]);
         break;
@@ -314,7 +300,6 @@ namespace dxvk {
 
   bool D3D9Query::QueryBeginnable(D3DQUERYTYPE QueryType) {
     return QueryType == D3DQUERYTYPE_OCCLUSION
-        || QueryType == D3DQUERYTYPE_VERTEXSTATS
         || QueryType == D3DQUERYTYPE_TIMESTAMPDISJOINT;
   }
 
@@ -338,7 +323,6 @@ namespace dxvk {
       case D3DQUERYTYPE_TIMESTAMP:
       case D3DQUERYTYPE_TIMESTAMPDISJOINT:
       case D3DQUERYTYPE_TIMESTAMPFREQ:
-      case D3DQUERYTYPE_VERTEXSTATS:
         return D3D_OK;
 
       default:
diff -pruN 2.3.1-1/src/d3d9/d3d9_state.h 2.4-1/src/d3d9/d3d9_state.h
--- 2.3.1-1/src/d3d9/d3d9_state.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_state.h	2024-07-10 09:56:14.000000000 +0000
@@ -169,11 +169,11 @@ namespace dxvk {
 
   constexpr D3DLIGHT9 DefaultLight = {
     D3DLIGHT_DIRECTIONAL,     // Type
-    {1.0f, 1.0f, 1.0f, 1.0f}, // Diffuse
+    {1.0f, 1.0f, 1.0f, 0.0f}, // Diffuse
     {0.0f, 0.0f, 0.0f, 0.0f}, // Specular
     {0.0f, 0.0f, 0.0f, 0.0f}, // Ambient
     {0.0f, 0.0f, 0.0f},       // Position
-    {0.0f, 0.0f, 0.0f},       // Direction
+    {0.0f, 0.0f, 1.0f},       // Direction
     0.0f,                     // Range
     0.0f,                     // Falloff
     0.0f, 0.0f, 0.0f,         // Attenuations [constant, linear, quadratic]
diff -pruN 2.3.1-1/src/d3d9/d3d9_stateblock.cpp 2.4-1/src/d3d9/d3d9_stateblock.cpp
--- 2.3.1-1/src/d3d9/d3d9_stateblock.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_stateblock.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -49,7 +49,7 @@ namespace dxvk {
     if (m_captures.flags.test(D3D9CapturedStateFlag::VertexDecl))
       SetVertexDeclaration(m_deviceState->vertexDecl.ptr());
 
-    ApplyOrCapture<D3D9StateFunction::Capture>();
+    ApplyOrCapture<D3D9StateFunction::Capture, true>();
 
     return D3D_OK;
   }
@@ -61,7 +61,7 @@ namespace dxvk {
     if (m_captures.flags.test(D3D9CapturedStateFlag::VertexDecl) && m_state.vertexDecl != nullptr)
       m_parent->SetVertexDeclaration(m_state.vertexDecl.ptr());
 
-    ApplyOrCapture<D3D9StateFunction::Apply>();
+    ApplyOrCapture<D3D9StateFunction::Apply, false>();
     m_applying = false;
 
     return D3D_OK;
@@ -122,6 +122,20 @@ namespace dxvk {
   }
 
 
+  HRESULT D3D9StateBlock::SetStreamSourceWithoutOffset(
+          UINT                StreamNumber,
+          D3D9VertexBuffer*   pStreamData,
+          UINT                Stride) {
+    m_state.vertexBuffers[StreamNumber].vertexBuffer = pStreamData;
+
+    m_state.vertexBuffers[StreamNumber].stride = Stride;
+
+    m_captures.flags.set(D3D9CapturedStateFlag::VertexBuffers);
+    m_captures.vertexBuffers.set(StreamNumber, true);
+    return D3D_OK;
+  }
+
+
   HRESULT D3D9StateBlock::SetStreamSourceFreq(UINT StreamNumber, UINT Setting) {
     m_state.streamFreq[StreamNumber] = Setting;
 
@@ -572,8 +586,12 @@ namespace dxvk {
       m_captures.flags.set(D3D9CapturedStateFlag::Material);
     }
 
-    if (Type != D3D9StateBlockType::None)
-      this->Capture();
+    if (Type != D3D9StateBlockType::None) {
+      if (m_captures.flags.test(D3D9CapturedStateFlag::VertexDecl))
+        SetVertexDeclaration(m_deviceState->vertexDecl.ptr());
+
+      ApplyOrCapture<D3D9StateFunction::Capture, false>();
+    }
   }
 
 }
diff -pruN 2.3.1-1/src/d3d9/d3d9_stateblock.h 2.4-1/src/d3d9/d3d9_stateblock.h
--- 2.3.1-1/src/d3d9/d3d9_stateblock.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_stateblock.h	2024-07-10 09:56:14.000000000 +0000
@@ -115,6 +115,11 @@ namespace dxvk {
             UINT               OffsetInBytes,
             UINT               Stride);
 
+    HRESULT SetStreamSourceWithoutOffset(
+            UINT               StreamNumber,
+            D3D9VertexBuffer*  pStreamData,
+            UINT               Stride);
+
     HRESULT SetStreamSourceFreq(UINT StreamNumber, UINT Setting);
 
     HRESULT SetStateTexture(DWORD StateSampler, IDirect3DBaseTexture9* pTexture);
@@ -181,7 +186,7 @@ namespace dxvk {
       Capture
     };
 
-    template <typename Dst, typename Src>
+    template <typename Dst, typename Src, bool IgnoreStreamOffset>
     void ApplyOrCapture(Dst* dst, const Src* src) {
       if (m_captures.flags.test(D3D9CapturedStateFlag::StreamFreq)) {
         for (uint32_t idx : bit::BitMask(m_captures.streamFreq.dword(0)))
@@ -211,11 +216,19 @@ namespace dxvk {
       if (m_captures.flags.test(D3D9CapturedStateFlag::VertexBuffers)) {
         for (uint32_t idx : bit::BitMask(m_captures.vertexBuffers.dword(0))) {
           const auto& vbo = src->vertexBuffers[idx];
-          dst->SetStreamSource(
-            idx,
-            vbo.vertexBuffer.ptr(),
-            vbo.offset,
-            vbo.stride);
+          if constexpr (!IgnoreStreamOffset) {
+            dst->SetStreamSource(
+              idx,
+              vbo.vertexBuffer.ptr(),
+              vbo.offset,
+              vbo.stride);
+          } else {
+            // For whatever reason, D3D9 doesn't capture the stream offset
+            dst->SetStreamSourceWithoutOffset(
+              idx,
+              vbo.vertexBuffer.ptr(),
+              vbo.stride);
+          }
         }
       }
 
@@ -324,12 +337,12 @@ namespace dxvk {
       }
     }
 
-    template <D3D9StateFunction Func>
+    template <D3D9StateFunction Func, bool IgnoreStreamOffset>
     void ApplyOrCapture() {
       if      constexpr (Func == D3D9StateFunction::Apply)
-        ApplyOrCapture(m_parent, &m_state);
+        ApplyOrCapture<D3D9DeviceEx, D3D9CapturableState, IgnoreStreamOffset>(m_parent, &m_state);
       else if constexpr (Func == D3D9StateFunction::Capture)
-        ApplyOrCapture(this, m_deviceState);
+        ApplyOrCapture<D3D9StateBlock, D3D9DeviceState, IgnoreStreamOffset>(this, m_deviceState);
     }
 
     template <
diff -pruN 2.3.1-1/src/d3d9/d3d9_swapchain.cpp 2.4-1/src/d3d9/d3d9_swapchain.cpp
--- 2.3.1-1/src/d3d9/d3d9_swapchain.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_swapchain.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -172,6 +172,9 @@ namespace dxvk {
 
     m_lastDialog = m_dialog;
 
+    if (m_window == nullptr)
+      return D3D_OK;
+
 #ifdef _WIN32
     const bool useGDIFallback = m_partialCopy && !HasFrontBuffer();
     if (useGDIFallback)
@@ -191,6 +194,7 @@ namespace dxvk {
       if (!m_wctx->presenter->hasSwapChain())
         return D3D_OK;
 
+      UpdateTargetFrameRate(presentInterval);
       PresentImage(presentInterval);
       return D3D_OK;
     } catch (const DxvkError& e) {
@@ -927,7 +931,6 @@ namespace dxvk {
     presenterDesc.fullScreenExclusive = PickFullscreenMode();
 
     m_wctx->presenter = new Presenter(m_device, m_wctx->frameLatencySignal, presenterDesc);
-    m_wctx->presenter->setFrameRateLimit(m_parent->GetOptions()->maxFrameRate);
   }
 
 
@@ -994,6 +997,9 @@ namespace dxvk {
 
 
   void D3D9SwapChainEx::UpdateWindowCtx() {
+    if (m_window == nullptr)
+      return;
+
     if (!m_presenters.count(m_window)) {
       auto res = m_presenters.emplace(
         std::piecewise_construct,
@@ -1107,6 +1113,17 @@ namespace dxvk {
   }
 
 
+  void D3D9SwapChainEx::UpdateTargetFrameRate(uint32_t SyncInterval) {
+    double frameRateOption = double(m_parent->GetOptions()->maxFrameRate);
+    double frameRate = std::max(frameRateOption, 0.0);
+
+    if (SyncInterval && frameRateOption == 0.0)
+      frameRate = -m_displayRefreshRate / double(SyncInterval);
+
+    m_wctx->presenter->setFrameRateLimit(frameRate);
+  }
+
+
   void D3D9SwapChainEx::SyncFrameLatency() {
     // Wait for the sync event so that we respect the maximum frame latency
     m_wctx->frameLatencySignal->wait(m_wctx->frameId - GetActualFrameLatency());
@@ -1323,10 +1340,10 @@ namespace dxvk {
     || dstRect.right  - dstRect.left != LONG(width)
     || dstRect.bottom - dstRect.top  != LONG(height);
 
-    bool recreate =
-       m_wctx->presenter == nullptr
-    || m_wctx->presenter->info().imageExtent.width  != width
-    || m_wctx->presenter->info().imageExtent.height != height;
+    bool recreate = m_wctx != nullptr
+      && (m_wctx->presenter == nullptr
+      || m_wctx->presenter->info().imageExtent.width  != width
+      || m_wctx->presenter->info().imageExtent.height != height);
 
     m_swapchainExtent = { width, height };
     m_dstRect = dstRect;
diff -pruN 2.3.1-1/src/d3d9/d3d9_swapchain.h 2.4-1/src/d3d9/d3d9_swapchain.h
--- 2.3.1-1/src/d3d9/d3d9_swapchain.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_swapchain.h	2024-07-10 09:56:14.000000000 +0000
@@ -213,6 +213,8 @@ namespace dxvk {
 
     void InitRamp();
 
+    void UpdateTargetFrameRate(uint32_t SyncInterval);
+
     uint32_t GetActualFrameLatency();
 
     uint32_t PickFormats(
diff -pruN 2.3.1-1/src/d3d9/d3d9_swvp_emu.cpp 2.4-1/src/d3d9/d3d9_swvp_emu.cpp
--- 2.3.1-1/src/d3d9/d3d9_swvp_emu.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_swvp_emu.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -9,13 +9,14 @@ namespace dxvk {
 
   // Doesn't compare everything, only what we use in SWVP.
 
-  size_t D3D9VertexDeclHash::operator () (const D3D9VertexElements& key) const {
+  size_t D3D9VertexDeclHash::operator () (const D3D9CompactVertexElements& key) const {
     DxvkHashState hash;
 
     std::hash<BYTE> bytehash;
     std::hash<WORD> wordhash;
 
-    for (auto& element : key) {
+    for (uint32_t i = 0; i < key.size(); i++) {
+      const auto& element = key[i];
       hash.add(wordhash(element.Stream));
       hash.add(wordhash(element.Offset));
       hash.add(bytehash(element.Type));
@@ -27,7 +28,7 @@ namespace dxvk {
     return hash;
   }
 
-  bool D3D9VertexDeclEq::operator () (const D3D9VertexElements& a, const D3D9VertexElements& b) const {
+  bool D3D9VertexDeclEq::operator () (const D3D9CompactVertexElements& a, const D3D9CompactVertexElements& b) const {
     if (a.size() != b.size())
       return false;
 
@@ -109,7 +110,7 @@ namespace dxvk {
       m_module.opLabel(m_module.allocateId());
     }
 
-    void compile(const D3D9VertexDecl* pDecl) {
+    void compile(const D3D9CompactVertexElements& elements) {
       uint32_t uint_t     = m_module.defIntType(32, false);
       uint32_t float_t    = m_module.defFloatType(32);
       uint32_t vec4_t     = m_module.defVectorType(float_t, 4);
@@ -142,15 +143,24 @@ namespace dxvk {
       m_module.decorateBuiltIn(primitiveIdPtr, spv::BuiltInPrimitiveId);
 
       uint32_t primitiveId = m_module.opLoad(uint_t, primitiveIdPtr);
-
+      
       // The size of any given vertex
-      uint32_t vertexSize       = m_module.constu32(pDecl->GetSize(0) / sizeof(uint32_t));
+      uint32_t size = 0;
+      for (uint32_t i = 0; i < elements.size(); i++) {
+        const auto& element = elements[i];
+        if (element.Stream == 0 && element.Type != D3DDECLTYPE_UNUSED) {
+          size = std::max(size, element.Offset + GetDecltypeSize(D3DDECLTYPE(element.Type)));
+        }
+      }
+
+      uint32_t vertexSize       = m_module.constu32(size / sizeof(uint32_t));
 
       //The offset of this vertex from the beginning of the buffer
       uint32_t thisVertexOffset = m_module.opIMul(uint_t, vertexSize, primitiveId);
 
 
-      for (auto& element : pDecl->GetElements()) {
+      for (uint32_t i = 0; i < elements.size(); i++) {
+        const auto& element = elements[i];
         // Load the slot associated with this element
         DxsoSemantic semantic = { DxsoUsage(element.Usage), element.UsageIndex };
 
@@ -297,9 +307,7 @@ namespace dxvk {
 
   };
 
-  Rc<DxvkShader> D3D9SWVPEmulator::GetShaderModule(D3D9DeviceEx* pDevice, const D3D9VertexDecl* pDecl) {
-    auto& elements = pDecl->GetElements();
-
+  Rc<DxvkShader> D3D9SWVPEmulator::GetShaderModule(D3D9DeviceEx* pDevice, D3D9CompactVertexElements&& elements) {
     // Use the shader's unique key for the lookup
     { std::unique_lock<dxvk::mutex> lock(m_mutex);
       
@@ -317,7 +325,7 @@ namespace dxvk {
     // This shader has not been compiled yet, so we have to create a
     // new module. This takes a while, so we won't lock the structure.
     D3D9SWVPEmulatorGenerator generator(name);
-    generator.compile(pDecl);
+    generator.compile(elements);
     Rc<DxvkShader> shader = generator.finalize();
 
     shader->setShaderKey(key);
@@ -338,7 +346,8 @@ namespace dxvk {
     // that object instead and discard the newly created module.
     { std::unique_lock<dxvk::mutex> lock(m_mutex);
       
-      auto status = m_modules.insert({ elements, shader });
+      std::pair<D3D9CompactVertexElements, Rc<DxvkShader>> pair = { std::move(elements), shader };
+      auto status = m_modules.insert(std::move(pair));
       if (!status.second)
         return status.first->second;
     }
diff -pruN 2.3.1-1/src/d3d9/d3d9_swvp_emu.h 2.4-1/src/d3d9/d3d9_swvp_emu.h
--- 2.3.1-1/src/d3d9/d3d9_swvp_emu.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/d3d9_swvp_emu.h	2024-07-10 09:56:14.000000000 +0000
@@ -11,26 +11,41 @@ namespace dxvk {
   class D3D9VertexDecl;
   class D3D9DeviceEx;
 
+  struct D3D9CompactVertexElement {
+      uint16_t Stream : 4;
+      uint16_t Type : 5;
+      uint16_t Method : 3;
+      uint16_t Usage : 4;
+      uint16_t UsageIndex;
+      uint16_t Offset;
+
+      D3D9CompactVertexElement(const D3DVERTEXELEMENT9& element) 
+        : Stream(element.Stream), Type(element.Type), Method(element.Method),
+          Usage(element.Usage), UsageIndex(element.UsageIndex), Offset(element.Offset) {}
+  };
+
+  using D3D9CompactVertexElements = small_vector<D3D9CompactVertexElement, 4>;
+
   struct D3D9VertexDeclHash {
-    size_t operator () (const D3D9VertexElements& key) const;
+    size_t operator () (const D3D9CompactVertexElements& key) const;
   };
 
   struct D3D9VertexDeclEq {
-    bool operator () (const D3D9VertexElements& a, const D3D9VertexElements& b) const;
+    bool operator () (const D3D9CompactVertexElements& a, const D3D9CompactVertexElements& b) const;
   };
 
   class D3D9SWVPEmulator {
 
   public:
 
-    Rc<DxvkShader> GetShaderModule(D3D9DeviceEx* pDevice, const D3D9VertexDecl* pDecl);
+    Rc<DxvkShader> GetShaderModule(D3D9DeviceEx* pDevice,  D3D9CompactVertexElements&& elements);
 
   private:
 
     dxvk::mutex                               m_mutex;
 
     std::unordered_map<
-      D3D9VertexElements, Rc<DxvkShader>,
+      D3D9CompactVertexElements, Rc<DxvkShader>,
       D3D9VertexDeclHash, D3D9VertexDeclEq>   m_modules;
 
   };
diff -pruN 2.3.1-1/src/d3d9/meson.build 2.4-1/src/d3d9/meson.build
--- 2.3.1-1/src/d3d9/meson.build	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/d3d9/meson.build	2024-07-10 09:56:14.000000000 +0000
@@ -57,17 +57,24 @@ if platform != 'windows'
   d3d9_link_depends += files('d3d9.sym')
 endif
 
-d3d9_dll = shared_library('d3d9'+dll_ext, d3d9_src, glsl_generator.process(d3d9_shaders), d3d9_res,
-  name_prefix         : dxvk_name_prefix,
+d3d9_dll = shared_library(dxvk_name_prefix+'d3d9', d3d9_src, glsl_generator.process(d3d9_shaders), d3d9_res,
   dependencies        : [ dxso_dep, dxvk_dep ],
   include_directories : dxvk_include_path,
   install             : true,
   vs_module_defs      : 'd3d9'+def_spec_ext,
   link_args           : d3d9_ld_args,
   link_depends        : [ d3d9_link_depends ],
+  kwargs              : dxvk_so_version,
 )
 
 d3d9_dep = declare_dependency(
   link_with           : [ d3d9_dll ],
   include_directories : [ dxvk_include_path ],
 )
+
+if platform != 'windows'
+  pkg.generate(d3d9_dll,
+    filebase: dxvk_pkg_prefix + 'd3d9',
+    subdirs:  'dxvk',
+  )
+endif
diff -pruN 2.3.1-1/src/dxbc/dxbc_analysis.cpp 2.4-1/src/dxbc/dxbc_analysis.cpp
--- 2.3.1-1/src/dxbc/dxbc_analysis.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxbc/dxbc_analysis.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -87,9 +87,9 @@ namespace dxvk {
     }
 
     for (uint32_t i = 0; i < ins.dstCount; i++) {
-      if (ins.dst[0].type == DxbcOperandType::IndexableTemp) {
-        uint32_t index = ins.dst[0].idx[0].offset;
-        m_analysis->xRegMasks[index] |= ins.dst[0].mask;
+      if (ins.dst[i].type == DxbcOperandType::IndexableTemp) {
+        uint32_t index = ins.dst[i].idx[0].offset;
+        m_analysis->xRegMasks[index] |= ins.dst[i].mask;
       }
     }
   }
diff -pruN 2.3.1-1/src/dxbc/dxbc_chunk_isgn.cpp 2.4-1/src/dxbc/dxbc_chunk_isgn.cpp
--- 2.3.1-1/src/dxbc/dxbc_chunk_isgn.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxbc/dxbc_chunk_isgn.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -23,7 +23,11 @@ namespace dxvk {
       entry.systemValue     = static_cast<DxbcSystemValue>(reader.readu32());
       entry.componentType   = componentTypes.at(reader.readu32());
       entry.registerId      = reader.readu32();
-      entry.componentMask   = bit::extract(reader.readu32(), 0, 3);
+
+      uint32_t mask = reader.readu32();
+
+      entry.componentMask   = bit::extract(mask, 0, 3);
+      entry.componentUsed   = bit::extract(mask, 8, 11);
 
       if (hasPrecision)
         reader.readu32();
diff -pruN 2.3.1-1/src/dxbc/dxbc_chunk_isgn.h 2.4-1/src/dxbc/dxbc_chunk_isgn.h
--- 2.3.1-1/src/dxbc/dxbc_chunk_isgn.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxbc/dxbc_chunk_isgn.h	2024-07-10 09:56:14.000000000 +0000
@@ -20,6 +20,7 @@ namespace dxvk {
     uint32_t          semanticIndex;
     uint32_t          registerId;
     DxbcRegMask       componentMask;
+    DxbcRegMask       componentUsed;
     DxbcScalarType    componentType;
     DxbcSystemValue   systemValue;
     uint32_t          streamId;
diff -pruN 2.3.1-1/src/dxbc/dxbc_compiler.cpp 2.4-1/src/dxbc/dxbc_compiler.cpp
--- 2.3.1-1/src/dxbc/dxbc_compiler.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxbc/dxbc_compiler.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -257,14 +257,13 @@ namespace dxvk {
     info.outputMask = m_outputMask;
     info.uniformSize = m_immConstData.size();
     info.uniformData = m_immConstData.data();
+    info.pushConstStages = VK_SHADER_STAGE_FRAGMENT_BIT;
+    info.pushConstSize = sizeof(DxbcPushConstants);
     info.outputTopology = m_outputTopology;
 
     if (m_programInfo.type() == DxbcProgramType::HullShader)
       info.patchVertexCount = m_hs.vertexCountIn;
 
-    if (m_programInfo.type() == DxbcProgramType::PixelShader && m_ps.pushConstantId)
-      info.pushConstSize = sizeof(DxbcPushConstants);
-
     if (m_moduleInfo.xfb) {
       info.xfbRasterizedStream = m_moduleInfo.xfb->rasterizedStream;
 
@@ -1624,8 +1623,13 @@ namespace dxvk {
       
       case DxbcOpcode::Mad:
       case DxbcOpcode::DFma:
-        dst.id = m_module.opFFma(typeId,
-          src.at(0).id, src.at(1).id, src.at(2).id);
+        if (likely(!m_moduleInfo.options.longMad)) {
+          dst.id = m_module.opFFma(typeId,
+            src.at(0).id, src.at(1).id, src.at(2).id);
+        } else {
+          dst.id = m_module.opFMul(typeId, src.at(0).id, src.at(1).id);
+          dst.id = m_module.opFAdd(typeId, dst.id, src.at(2).id);
+        }
         break;
       
       case DxbcOpcode::Max:
@@ -2464,58 +2468,6 @@ namespace dxvk {
     if (m_uavs.at(registerId).ctrId == 0)
       m_uavs.at(registerId).ctrId = emitDclUavCounter(registerId);
     
-    // Only use subgroup ops on compute to avoid having to
-    // deal with helper invocations or hardware limitations
-    bool useSubgroupOps = m_moduleInfo.options.useSubgroupOpsForAtomicCounters
-      && m_programInfo.type() == DxbcProgramType::ComputeShader;
-
-    // Current block ID used in a phi later on
-    uint32_t baseBlockId = m_module.getBlockId();
-
-    // In case we have subgroup ops enabled, we need to
-    // count the number of active lanes, the lane index,
-    // and we need to perform the atomic op conditionally
-    uint32_t laneCount = 0;
-    uint32_t laneIndex = 0;
-
-    DxbcConditional elect;
-
-    if (useSubgroupOps) {
-      m_module.enableCapability(spv::CapabilityGroupNonUniform);
-      m_module.enableCapability(spv::CapabilityGroupNonUniformBallot);
-
-      uint32_t ballot = m_module.opGroupNonUniformBallot(
-        getVectorTypeId({ DxbcScalarType::Uint32, 4 }),
-        m_module.constu32(spv::ScopeSubgroup),
-        m_module.constBool(true));
-      
-      laneCount = m_module.opGroupNonUniformBallotBitCount(
-        getScalarTypeId(DxbcScalarType::Uint32),
-        m_module.constu32(spv::ScopeSubgroup),
-        spv::GroupOperationReduce, ballot);
-      
-      laneIndex = m_module.opGroupNonUniformBallotBitCount(
-        getScalarTypeId(DxbcScalarType::Uint32),
-        m_module.constu32(spv::ScopeSubgroup),
-        spv::GroupOperationExclusiveScan, ballot);
-      
-      // Elect one lane to perform the atomic op
-      uint32_t election = m_module.opGroupNonUniformElect(
-        m_module.defBoolType(),
-        m_module.constu32(spv::ScopeSubgroup));
-
-      elect.labelIf  = m_module.allocateId();
-      elect.labelEnd = m_module.allocateId();
-
-      m_module.opSelectionMerge(elect.labelEnd, spv::SelectionControlMaskNone);
-      m_module.opBranchConditional(election, elect.labelIf, elect.labelEnd);
-      
-      m_module.opLabel(elect.labelIf);
-    } else {
-      // We're going to use this for the increment
-      laneCount = m_module.constu32(1);
-    }
-
     // Get a pointer to the atomic counter in question
     DxbcRegisterInfo ptrType;
     ptrType.type.ctype   = DxbcScalarType::Uint32;
@@ -2547,13 +2499,14 @@ namespace dxvk {
     switch (ins.op) {
       case DxbcOpcode::ImmAtomicAlloc:
         value.id = m_module.opAtomicIAdd(typeId, ptrId,
-          scopeId, semanticsId, laneCount);
+          scopeId, semanticsId, m_module.constu32(1));
         break;
         
       case DxbcOpcode::ImmAtomicConsume:
         value.id = m_module.opAtomicISub(typeId, ptrId,
-          scopeId, semanticsId, laneCount);
-        value.id = m_module.opISub(typeId, value.id, laneCount);
+          scopeId, semanticsId, m_module.constu32(1));
+        value.id = m_module.opISub(typeId, value.id,
+          m_module.constu32(1));
         break;
       
       default:
@@ -2563,26 +2516,6 @@ namespace dxvk {
         return;
     }
 
-    // If we're using subgroup ops, we have to broadcast
-    // the result of the atomic op and compute the index
-    if (useSubgroupOps) {
-      m_module.opBranch(elect.labelEnd);
-      m_module.opLabel (elect.labelEnd);
-
-      uint32_t undef = m_module.constUndef(typeId);
-
-      std::array<SpirvPhiLabel, 2> phiLabels = {{
-        { value.id, elect.labelIf },
-        { undef,    baseBlockId   },
-      }};
-
-      value.id = m_module.opPhi(typeId,
-        phiLabels.size(), phiLabels.data());
-      value.id = m_module.opGroupNonUniformBroadcastFirst(typeId,
-        m_module.constu32(spv::ScopeSubgroup), value.id);
-      value.id = m_module.opIAdd(typeId, value.id, laneIndex);
-    }
-    
     // Store the result
     emitRegisterStore(ins.dst[0], value);
   }
@@ -5584,12 +5517,12 @@ namespace dxvk {
     result.type.ctype  = DxbcScalarType::Uint32;
     result.type.ccount = 1;
     
-    if (info.image.sampled == 1) {
+    if (info.image.ms == 0 && info.image.sampled == 1) {
       result.id = m_module.opImageQueryLevels(
         getVectorTypeId(result.type),
         m_module.opLoad(info.typeId, info.varId));
     } else {
-      // Report one LOD in case of UAVs
+      // Report one LOD in case of UAVs or multisampled images
       result.id = m_module.constu32(1);
     }
 
@@ -7799,6 +7732,21 @@ namespace dxvk {
     return DxbcRegMask::firstN(getTexCoordDim(imageType));
   }
   
+
+  bool DxbcCompiler::ignoreInputSystemValue(DxbcSystemValue sv) const {
+    switch (sv) {
+      case DxbcSystemValue::Position:
+      case DxbcSystemValue::IsFrontFace:
+      case DxbcSystemValue::SampleIndex:
+      case DxbcSystemValue::PrimitiveId:
+      case DxbcSystemValue::Coverage:
+        return m_programInfo.type() == DxbcProgramType::PixelShader;
+
+      default:
+        return false;
+    }
+  }
+
   
   DxbcVectorType DxbcCompiler::getInputRegType(uint32_t regIdx) const {
     switch (m_programInfo.type()) {
@@ -7829,8 +7777,25 @@ namespace dxvk {
         result.ctype  = DxbcScalarType::Float32;
         result.ccount = 4;
 
-        if (m_isgn->findByRegister(regIdx))
-          result.ccount = m_isgn->regMask(regIdx).minComponents();
+        if (m_isgn == nullptr || !m_isgn->findByRegister(regIdx))
+          return result;
+
+        DxbcRegMask mask(0u);
+        DxbcRegMask used(0u);
+
+        for (const auto& e : *m_isgn) {
+          if (e.registerId == regIdx && !ignoreInputSystemValue(e.systemValue)) {
+            mask |= e.componentMask;
+            used |= e.componentUsed;
+          }
+        }
+
+        if (m_programInfo.type() == DxbcProgramType::PixelShader) {
+          if ((used.raw() & mask.raw()) == used.raw())
+            mask = used;
+        }
+
+        result.ccount = mask.minComponents();
         return result;
       }
     }
diff -pruN 2.3.1-1/src/dxbc/dxbc_compiler.h 2.4-1/src/dxbc/dxbc_compiler.h
--- 2.3.1-1/src/dxbc/dxbc_compiler.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxbc/dxbc_compiler.h	2024-07-10 09:56:14.000000000 +0000
@@ -1221,6 +1221,9 @@ namespace dxvk {
     uint32_t getUavCoherence(
             uint32_t                registerId,
             DxbcUavFlags            flags);
+    
+    bool ignoreInputSystemValue(
+            DxbcSystemValue         sv) const;
 
     ///////////////////////////
     // Type definition methods
diff -pruN 2.3.1-1/src/dxbc/dxbc_options.cpp 2.4-1/src/dxbc/dxbc_options.cpp
--- 2.3.1-1/src/dxbc/dxbc_options.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxbc/dxbc_options.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -17,9 +17,6 @@ namespace dxvk {
 
     useDepthClipWorkaround
       = !devFeatures.extDepthClipEnable.depthClipEnable;
-    useSubgroupOpsForAtomicCounters
-      = (devInfo.vk11.subgroupSupportedStages     & VK_SHADER_STAGE_COMPUTE_BIT)
-     && (devInfo.vk11.subgroupSupportedOperations & VK_SUBGROUP_FEATURE_BALLOT_BIT);
 
     VkFormatFeatureFlags2 r32Features
       = device->getFormatFeatures(VK_FORMAT_R32_SFLOAT).optimal
@@ -41,6 +38,7 @@ namespace dxvk {
     disableMsaa              = options.disableMsaa;
     forceSampleRateShading   = options.forceSampleRateShading;
     enableSampleShadingInterlock = device->features().extFragmentShaderInterlock.fragmentShaderSampleInterlock;
+    longMad                  = options.longMad;
 
     // Figure out float control flags to match D3D11 rules
     if (options.floatControls) {
diff -pruN 2.3.1-1/src/dxbc/dxbc_options.h 2.4-1/src/dxbc/dxbc_options.h
--- 2.3.1-1/src/dxbc/dxbc_options.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxbc/dxbc_options.h	2024-07-10 09:56:14.000000000 +0000
@@ -30,10 +30,6 @@ namespace dxvk {
     /// Determines whether raw access chains are supported
     bool supportsRawAccessChains = false;
 
-    /// Use subgroup operations to reduce the number of
-    /// atomic operations for append/consume buffers.
-    bool useSubgroupOpsForAtomicCounters = false;
-
     /// Clear thread-group shared memory to zero
     bool zeroInitWorkgroupMemory = false;
 
@@ -58,6 +54,9 @@ namespace dxvk {
 
     /// Minimum storage buffer alignment
     VkDeviceSize minSsboAlignment = 0;
+
+    /// Should we make our Mads a FFma or do it the long way with an FMul and an FAdd?
+    bool longMad;
   };
   
 }
\ No newline at end of file
diff -pruN 2.3.1-1/src/dxgi/dxgi_adapter.cpp 2.4-1/src/dxgi/dxgi_adapter.cpp
--- 2.3.1-1/src/dxgi/dxgi_adapter.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxgi/dxgi_adapter.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -130,7 +130,7 @@ namespace dxvk {
     // We can't really reconstruct the version numbers
     // returned by Windows drivers from Vulkan data
     if (SUCCEEDED(hr) && pUMDVersion)
-      pUMDVersion->QuadPart = ~0ull;
+      pUMDVersion->QuadPart = INT64_MAX;
 
     if (FAILED(hr)) {
       Logger::err("DXGI: CheckInterfaceSupport: Unsupported interface");
diff -pruN 2.3.1-1/src/dxgi/dxgi_factory.cpp 2.4-1/src/dxgi/dxgi_factory.cpp
--- 2.3.1-1/src/dxgi/dxgi_factory.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxgi/dxgi_factory.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -124,13 +124,10 @@ namespace dxvk {
       }
     }
 
-
     // If any monitors are left on the list, enable the
     // fallback to always enumerate all monitors.
     if ((m_monitorFallback = !monitors.empty()))
       Logger::warn("DXGI: Found monitors not associated with any adapter, using fallback");
-    else
-      m_monitorFallback = m_options.useMonitorFallback;
   }
   
   
@@ -233,7 +230,7 @@ namespace dxvk {
     descFs.Windowed         = pDesc->Windowed;
     
     IDXGISwapChain1* swapChain = nullptr;
-    HRESULT hr = CreateSwapChainForHwnd(
+    HRESULT hr = CreateSwapChainForHwndBase(
       pDevice, pDesc->OutputWindow,
       &desc, &descFs, nullptr,
       &swapChain);
@@ -247,6 +244,19 @@ namespace dxvk {
           IUnknown*             pDevice,
           HWND                  hWnd,
     const DXGI_SWAP_CHAIN_DESC1* pDesc,
+    const DXGI_SWAP_CHAIN_FULLSCREEN_DESC* pFullscreenDesc,
+          IDXGIOutput*          pRestrictToOutput,
+          IDXGISwapChain1**     ppSwapChain) {
+    return CreateSwapChainForHwndBase(
+      pDevice, hWnd,
+      pDesc, pFullscreenDesc, pRestrictToOutput,
+      ppSwapChain);
+  }
+
+  HRESULT STDMETHODCALLTYPE DxgiFactory::CreateSwapChainForHwndBase(
+          IUnknown*             pDevice,
+          HWND                  hWnd,
+    const DXGI_SWAP_CHAIN_DESC1* pDesc,
     const DXGI_SWAP_CHAIN_FULLSCREEN_DESC* pFullscreenDesc,
           IDXGIOutput*          pRestrictToOutput,
           IDXGISwapChain1**     ppSwapChain) {
diff -pruN 2.3.1-1/src/dxgi/dxgi_factory.h 2.4-1/src/dxgi/dxgi_factory.h
--- 2.3.1-1/src/dxgi/dxgi_factory.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxgi/dxgi_factory.h	2024-07-10 09:56:14.000000000 +0000
@@ -200,6 +200,14 @@ namespace dxvk {
     UINT             m_flags;
     BOOL             m_monitorFallback;
       
+
+    HRESULT STDMETHODCALLTYPE CreateSwapChainForHwndBase(
+            IUnknown*             pDevice,
+            HWND                  hWnd,
+      const DXGI_SWAP_CHAIN_DESC1* pDesc,
+      const DXGI_SWAP_CHAIN_FULLSCREEN_DESC* pFullscreenDesc,
+            IDXGIOutput*          pRestrictToOutput,
+            IDXGISwapChain1**     ppSwapChain);
   };
   
 }
diff -pruN 2.3.1-1/src/dxgi/dxgi_interfaces.h 2.4-1/src/dxgi/dxgi_interfaces.h
--- 2.3.1-1/src/dxgi/dxgi_interfaces.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxgi/dxgi_interfaces.h	2024-07-10 09:56:14.000000000 +0000
@@ -137,6 +137,13 @@ IDXGIVkSwapChain1 : public IDXGIVkSwapCh
 };
 
 
+MIDL_INTERFACE("aed91093-e02e-458c-bdef-a97da1a7e6d2")
+IDXGIVkSwapChain2 : public IDXGIVkSwapChain1 {
+  virtual void STDMETHODCALLTYPE SetTargetFrameRate(
+          double                    FrameRate) = 0;
+};
+
+
 /**
  * \brief Private DXGI presenter factory
  */
@@ -471,5 +478,6 @@ __CRT_UUID_DECL(IDXGIVkInteropSurface,
 __CRT_UUID_DECL(IDXGIVkSurfaceFactory,     0x1e7895a1,0x1bc3,0x4f9c,0xa6,0x70,0x29,0x0a,0x4b,0xc9,0x58,0x1a);
 __CRT_UUID_DECL(IDXGIVkSwapChain,          0xe4a9059e,0xb569,0x46ab,0x8d,0xe7,0x50,0x1b,0xd2,0xbc,0x7f,0x7a);
 __CRT_UUID_DECL(IDXGIVkSwapChain1,         0x785326d4,0xb77b,0x4826,0xae,0x70,0x8d,0x08,0x30,0x8e,0xe6,0xd1);
+__CRT_UUID_DECL(IDXGIVkSwapChain2,         0xaed91093,0xe02e,0x458c,0xbd,0xef,0xa9,0x7d,0xa1,0xa7,0xe6,0xd2);
 __CRT_UUID_DECL(IDXGIVkSwapChainFactory,   0xe7d6c3ca,0x23a0,0x4e08,0x9f,0x2f,0xea,0x52,0x31,0xdf,0x66,0x33);
 #endif
diff -pruN 2.3.1-1/src/dxgi/dxgi_options.cpp 2.4-1/src/dxgi/dxgi_options.cpp
--- 2.3.1-1/src/dxgi/dxgi_options.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxgi/dxgi_options.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -26,6 +26,21 @@ namespace dxvk {
     return id;
   }
 
+  /* First generation XeSS causes crash on proton for Intel due to missing
+   * Intel interface. Avoid crash by pretending to be non-Intel if the
+   * libxess.dll module is loaded by an application.
+   */
+  static bool isXessUsed() {
+#ifdef _WIN32
+      if (GetModuleHandleA("libxess") != nullptr ||
+          GetModuleHandleA("libxess_dx11") != nullptr)
+        return true;
+      else
+        return false;
+#else
+      return false;
+#endif
+  }
 
   static bool isNvapiEnabled() {
     return env::getEnvVar("DXVK_ENABLE_NVAPI") == "1";
@@ -78,6 +93,7 @@ namespace dxvk {
     this->maxDeviceMemory = VkDeviceSize(config.getOption<int32_t>("dxgi.maxDeviceMemory", 0)) << 20;
     this->maxSharedMemory = VkDeviceSize(config.getOption<int32_t>("dxgi.maxSharedMemory", 0)) << 20;
 
+    this->maxFrameRate = config.getOption<int32_t>("dxgi.maxFrameRate", 0);
     this->syncInterval = config.getOption<int32_t>("dxgi.syncInterval", -1);
 
     // Expose Nvidia GPUs properly if NvAPI is enabled in environment
@@ -95,15 +111,17 @@ namespace dxvk {
     this->hideAmdGpu = config.getOption<Tristate>("dxgi.hideAmdGpu", Tristate::Auto) == Tristate::True;
     this->hideIntelGpu = config.getOption<Tristate>("dxgi.hideIntelGpu", Tristate::Auto) == Tristate::True;
 
+    /* Force vendor ID to non-Intel ID when XeSS is in use */
+    if (isXessUsed()) {
+      Logger::info(str::format("Detected XeSS usage, hiding Intel GPU Vendor"));
+      this->hideIntelGpu = true;
+    }
+
     this->enableHDR = config.getOption<bool>("dxgi.enableHDR", env::getEnvVar("DXVK_HDR") == "1");
     if (this->enableHDR && isHDRDisallowed()) {
       Logger::info("HDR was configured to be enabled, but has been force disabled as a UE4 DX11 game was detected.");
       this->enableHDR = false;
     }
-
-    this->useMonitorFallback = config.getOption<bool>("dxgi.useMonitorFallback", env::getEnvVar("DXVK_MONITOR_FALLBACK") == "1");
-    if (this->useMonitorFallback)
-      Logger::info("Enabled useMonitorFallback option");
   }
   
 }
diff -pruN 2.3.1-1/src/dxgi/dxgi_options.h 2.4-1/src/dxgi/dxgi_options.h
--- 2.3.1-1/src/dxgi/dxgi_options.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxgi/dxgi_options.h	2024-07-10 09:56:14.000000000 +0000
@@ -49,8 +49,8 @@ namespace dxvk {
     /// Enable HDR
     bool enableHDR;
 
-    /// Use monitor fallback to enumerating all monitors per output
-    bool useMonitorFallback;
+    /// Limit frame rate
+    int32_t maxFrameRate;
 
     /// Sync interval. Overrides the value
     /// passed to IDXGISwapChain::Present.
diff -pruN 2.3.1-1/src/dxgi/dxgi_swapchain.cpp 2.4-1/src/dxgi/dxgi_swapchain.cpp
--- 2.3.1-1/src/dxgi/dxgi_swapchain.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxgi/dxgi_swapchain.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -25,6 +25,9 @@ namespace dxvk {
     // Query updated interface versions from presenter, this
     // may fail e.g. with older vkd3d-proton builds.
     m_presenter->QueryInterface(__uuidof(IDXGIVkSwapChain1), reinterpret_cast<void**>(&m_presenter1));
+    m_presenter->QueryInterface(__uuidof(IDXGIVkSwapChain2), reinterpret_cast<void**>(&m_presenter2));
+
+    m_frameRateOption = m_factory->GetOptions()->maxFrameRate;
 
     // Query monitor info form DXVK's DXGI factory, if available
     m_factory->QueryInterface(__uuidof(IDXGIVkMonitorInfo), reinterpret_cast<void**>(&m_monitorInfo));
@@ -303,15 +306,22 @@ namespace dxvk {
   
   
   HRESULT STDMETHODCALLTYPE DxgiSwapChain::Present(UINT SyncInterval, UINT Flags) {
-    return Present1(SyncInterval, Flags, nullptr);
+    return PresentBase(SyncInterval, Flags, nullptr);
   }
-  
-  
+
   HRESULT STDMETHODCALLTYPE DxgiSwapChain::Present1(
           UINT                      SyncInterval,
           UINT                      PresentFlags,
     const DXGI_PRESENT_PARAMETERS*  pPresentParameters) {
 
+    return PresentBase(SyncInterval, PresentFlags, pPresentParameters);
+  }
+
+  HRESULT STDMETHODCALLTYPE DxgiSwapChain::PresentBase(
+          UINT                      SyncInterval,
+          UINT                      PresentFlags,
+    const DXGI_PRESENT_PARAMETERS*  pPresentParameters) {
+
     if (SyncInterval > 4)
       return DXGI_ERROR_INVALID_CALL;
 
@@ -321,6 +331,7 @@ namespace dxvk {
       SyncInterval = options->syncInterval;
 
     UpdateGlobalHDRState();
+    UpdateTargetFrameRate(SyncInterval);
 
     std::lock_guard<dxvk::recursive_mutex> lockWin(m_lockWindow);
     HRESULT hr = S_OK;
@@ -449,9 +460,7 @@ namespace dxvk {
         return E_FAIL;
       }
       
-      // If the swap chain allows it, change the display mode
-      if (m_desc.Flags & DXGI_SWAP_CHAIN_FLAG_ALLOW_MODE_SWITCH)
-        ChangeDisplayMode(output.ptr(), &newDisplayMode);
+      ChangeDisplayMode(output.ptr(), &newDisplayMode);
 
       wsi::updateFullscreenWindow(m_monitor, m_window, false);
     }
@@ -661,35 +670,33 @@ namespace dxvk {
       }
     }
 
-    const bool modeSwitch = m_desc.Flags & DXGI_SWAP_CHAIN_FLAG_ALLOW_MODE_SWITCH;
+    DXGI_MODE_DESC1 displayMode = { };
+    displayMode.Width            = m_desc.Width;
+    displayMode.Height           = m_desc.Height;
+    displayMode.RefreshRate      = m_descFs.RefreshRate;
+    displayMode.Format           = m_desc.Format;
+    // Ignore these two, games usually use them wrong and we don't
+    // support any scaling modes except UNSPECIFIED anyway.
+    displayMode.ScanlineOrdering = DXGI_MODE_SCANLINE_ORDER_UNSPECIFIED;
+    displayMode.Scaling          = DXGI_MODE_SCALING_UNSPECIFIED;
     
-    if (modeSwitch) {
-      DXGI_MODE_DESC1 displayMode = { };
-      displayMode.Width            = m_desc.Width;
-      displayMode.Height           = m_desc.Height;
-      displayMode.RefreshRate      = m_descFs.RefreshRate;
-      displayMode.Format           = m_desc.Format;
-      // Ignore these two, games usually use them wrong and we don't
-      // support any scaling modes except UNSPECIFIED anyway.
-      displayMode.ScanlineOrdering = DXGI_MODE_SCANLINE_ORDER_UNSPECIFIED;
-      displayMode.Scaling          = DXGI_MODE_SCALING_UNSPECIFIED;
-      
-      if (FAILED(ChangeDisplayMode(output.ptr(), &displayMode))) {
-        Logger::err("DXGI: EnterFullscreenMode: Failed to change display mode");
-        return DXGI_ERROR_NOT_CURRENTLY_AVAILABLE;
-      }
+    if (FAILED(ChangeDisplayMode(output.ptr(), &displayMode))) {
+      Logger::err("DXGI: EnterFullscreenMode: Failed to change display mode");
+      return DXGI_ERROR_NOT_CURRENTLY_AVAILABLE;
     }
     
     // Update swap chain description
     m_descFs.Windowed = FALSE;
     
     // Move the window so that it covers the entire output
+    bool modeSwitch = (m_desc.Flags & DXGI_SWAP_CHAIN_FLAG_ALLOW_MODE_SWITCH) != 0u;
+
     DXGI_OUTPUT_DESC desc;
     output->GetDesc(&desc);
 
     if (!wsi::enterFullscreenMode(desc.Monitor, m_window, &m_windowState, modeSwitch)) {
-        Logger::err("DXGI: EnterFullscreenMode: Failed to enter fullscreen mode");
-        return DXGI_ERROR_NOT_CURRENTLY_AVAILABLE;
+      Logger::err("DXGI: EnterFullscreenMode: Failed to enter fullscreen mode");
+      return DXGI_ERROR_NOT_CURRENTLY_AVAILABLE;
     }
     
     m_monitor = desc.Monitor;
@@ -753,14 +760,19 @@ namespace dxvk {
     pOutput->GetDesc(&outputDesc);
     
     DXGI_MODE_DESC1 preferredMode = *pDisplayMode;
-    DXGI_MODE_DESC1 selectedMode;
+    DXGI_MODE_DESC1 selectedMode = { };
+
+    if (!(m_desc.Flags & DXGI_SWAP_CHAIN_FLAG_ALLOW_MODE_SWITCH)) {
+      preferredMode.Width = 0;
+      preferredMode.Height = 0;
+    }
 
     if (preferredMode.Format == DXGI_FORMAT_UNKNOWN)
       preferredMode.Format = m_desc.Format;
     
     HRESULT hr = pOutput->FindClosestMatchingMode1(
       &preferredMode, &selectedMode, nullptr);
-    
+
     if (FAILED(hr)) {
       Logger::err(str::format(
         "DXGI: Failed to query closest mode:",
@@ -770,6 +782,9 @@ namespace dxvk {
       return hr;
     }
 
+    if (!selectedMode.RefreshRate.Denominator)
+      selectedMode.RefreshRate.Denominator = 1;
+
     if (!wsi::setWindowMode(outputDesc.Monitor, m_window, ConvertDisplayMode(selectedMode)))
       return DXGI_ERROR_NOT_CURRENTLY_AVAILABLE;
 
@@ -791,6 +806,8 @@ namespace dxvk {
       ReleaseMonitorData();
     }
 
+    m_frameRateRefresh = double(selectedMode.RefreshRate.Numerator)
+                       / double(selectedMode.RefreshRate.Denominator);
     return S_OK;
   }
   
@@ -802,6 +819,7 @@ namespace dxvk {
     if (!wsi::restoreDisplayMode())
       return DXGI_ERROR_NOT_CURRENTLY_AVAILABLE;
 
+    m_frameRateRefresh = 0.0;
     return S_OK;
   }
   
@@ -945,4 +963,23 @@ namespace dxvk {
     return hr;
   }
 
+
+  void DxgiSwapChain::UpdateTargetFrameRate(
+          UINT                    SyncInterval) {
+    if (m_presenter2 == nullptr)
+      return;
+
+    // Use a negative number to indicate that the limiter should only
+    // be engaged if the target frame rate is actually exceeded
+    double frameRate = std::max(m_frameRateOption, 0.0);
+
+    if (SyncInterval && m_frameRateOption == 0.0)
+      frameRate = -m_frameRateRefresh / double(SyncInterval);
+
+    if (m_frameRateLimit != frameRate) {
+      m_frameRateLimit = frameRate;
+      m_presenter2->SetTargetFrameRate(frameRate);
+    }
+  }
+
 }
diff -pruN 2.3.1-1/src/dxgi/dxgi_swapchain.h 2.4-1/src/dxgi/dxgi_swapchain.h
--- 2.3.1-1/src/dxgi/dxgi_swapchain.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxgi/dxgi_swapchain.h	2024-07-10 09:56:14.000000000 +0000
@@ -187,12 +187,17 @@ namespace dxvk {
 
     Com<IDXGIVkSwapChain>           m_presenter;
     Com<IDXGIVkSwapChain1>          m_presenter1;
+    Com<IDXGIVkSwapChain2>          m_presenter2;
     
     HMONITOR                        m_monitor;
     bool                            m_monitorHasOutput = true;
     bool                            m_frameStatisticsDisjoint = true;
     wsi::DxvkWindowState            m_windowState;
 
+    double                          m_frameRateOption = 0.0;
+    double                          m_frameRateRefresh = 0.0;
+    double                          m_frameRateLimit = 0.0;
+
     DXGI_COLOR_SPACE_TYPE           m_colorSpace = DXGI_COLOR_SPACE_RGB_FULL_G22_NONE_P709;
 
     uint32_t                        m_globalHDRStateSerial = 0;
@@ -233,6 +238,13 @@ namespace dxvk {
             DXGI_FORMAT             Format,
             DXGI_COLOR_SPACE_TYPE   ColorSpace);
 
+    void UpdateTargetFrameRate(
+            UINT                    SyncInterval);
+
+    HRESULT STDMETHODCALLTYPE PresentBase(
+            UINT                      SyncInterval,
+            UINT                      PresentFlags,
+      const DXGI_PRESENT_PARAMETERS*  pPresentParameters);
   };
   
 }
diff -pruN 2.3.1-1/src/dxgi/meson.build 2.4-1/src/dxgi/meson.build
--- 2.3.1-1/src/dxgi/meson.build	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxgi/meson.build	2024-07-10 09:56:14.000000000 +0000
@@ -21,17 +21,24 @@ if platform != 'windows'
   dxgi_link_depends += files('dxgi.sym')
 endif
 
-dxgi_dll = shared_library('dxgi'+dll_ext, dxgi_src, dxgi_res,
-  name_prefix         : dxvk_name_prefix,
+dxgi_dll = shared_library(dxvk_name_prefix+'dxgi', dxgi_src, dxgi_res,
   dependencies        : [ dxvk_dep ],
   include_directories : dxvk_include_path,
   install             : true,
   vs_module_defs      : 'dxgi'+def_spec_ext,
   link_args           : dxgi_ld_args,
   link_depends        : [ dxgi_link_depends ],
+  kwargs              : dxvk_so_version,
 )
 
 dxgi_dep = declare_dependency(
   link_with           : [ dxgi_dll ],
   include_directories : [ dxvk_include_path ],
 )
+
+if platform != 'windows'
+  pkg.generate(dxgi_dll,
+    filebase: dxvk_pkg_prefix + 'dxgi',
+    subdirs:  'dxvk',
+  )
+endif
diff -pruN 2.3.1-1/src/dxso/dxso_compiler.cpp 2.4-1/src/dxso/dxso_compiler.cpp
--- 2.3.1-1/src/dxso/dxso_compiler.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxso/dxso_compiler.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -228,8 +228,8 @@ namespace dxvk {
     info.bindings = m_bindings.data();
     info.inputMask = m_inputMask;
     info.outputMask = m_outputMask;
-    info.pushConstOffset = m_pushConstOffset;
-    info.pushConstSize = m_pushConstSize;
+    info.pushConstStages = VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT;
+    info.pushConstSize = sizeof(D3D9RenderStateInfo);
 
     if (m_programInfo.type() == DxsoProgramTypes::PixelShader)
       info.flatShadingInputs = m_ps.flatShadingMask;
@@ -2909,6 +2909,14 @@ void DxsoCompiler::emitControlFlowGeneri
         uint32_t component = sampler.dimensions;
         reference = m_module.opCompositeExtract(
           fType, texcoordVar.id, 1, &component);
+
+        // [D3D8] Scale Dref from [0..(2^N - 1)] for D24S8 and D16 if Dref scaling is enabled
+        if (m_moduleInfo.options.drefScaling) {
+          uint32_t drefScale       = m_module.constf32(GetDrefScaleFactor(m_moduleInfo.options.drefScaling));
+          reference                = m_module.opFMul(fType, reference, drefScale);
+        }
+
+        // Clamp Dref to [0..1] for D32F emulating UNORM textures 
         uint32_t clampDref = m_spec.get(m_module, m_specUbo, SpecDrefClamp, samplerIdx, 1);
         clampDref = m_module.opINotEqual(bool_t, clampDref, m_module.constu32(0));
         uint32_t clampedDref = m_module.opFClamp(fType, reference, m_module.constf32(0.0f), m_module.constf32(1.0f));
@@ -3561,30 +3569,7 @@ void DxsoCompiler::emitControlFlowGeneri
 
 
   void DxsoCompiler::setupRenderStateInfo() {
-    uint32_t count;
-
-    // Only need alpha ref for PS 3.
-    // No FF fog component.
-    if (m_programInfo.type() == DxsoProgramType::PixelShader) {
-      if (m_programInfo.majorVersion() == 3) {
-        m_pushConstOffset = offsetof(D3D9RenderStateInfo, alphaRef);
-        m_pushConstSize   = sizeof(float);
-      }
-      else {
-        m_pushConstOffset = 0;
-        m_pushConstSize   = offsetof(D3D9RenderStateInfo, pointSize);
-      }
-
-      count = 5;
-    }
-    else {
-      m_pushConstOffset = offsetof(D3D9RenderStateInfo, pointSize);
-      // Point scale never triggers on programmable
-      m_pushConstSize   = sizeof(float) * 3;
-      count = 8;
-    }
-
-    m_rsBlock = SetupRenderStateBlock(m_module, count);
+    m_rsBlock = SetupRenderStateBlock(m_module);
   }
 
 
diff -pruN 2.3.1-1/src/dxso/dxso_compiler.h 2.4-1/src/dxso/dxso_compiler.h
--- 2.3.1-1/src/dxso/dxso_compiler.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxso/dxso_compiler.h	2024-07-10 09:56:14.000000000 +0000
@@ -344,8 +344,6 @@ namespace dxvk {
     // covers vertex input and fragment output.
     uint32_t m_inputMask = 0u;
     uint32_t m_outputMask = 0u;
-    uint32_t m_pushConstOffset = 0u;
-    uint32_t m_pushConstSize = 0u;
 
     ///////////////////////////////////
     // Shader-specific data structures
diff -pruN 2.3.1-1/src/dxso/dxso_options.cpp 2.4-1/src/dxso/dxso_options.cpp
--- 2.3.1-1/src/dxso/dxso_options.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxso/dxso_options.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -31,6 +31,8 @@ namespace dxvk {
 
     longMad = options.longMad;
     robustness2Supported = devFeatures.extRobustness2.robustBufferAccess2;
+
+    drefScaling         = options.drefScaling;
   }
 
 }
\ No newline at end of file
diff -pruN 2.3.1-1/src/dxso/dxso_options.h 2.4-1/src/dxso/dxso_options.h
--- 2.3.1-1/src/dxso/dxso_options.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxso/dxso_options.h	2024-07-10 09:56:14.000000000 +0000
@@ -49,6 +49,12 @@ namespace dxvk {
 
     /// Whether or not we can rely on robustness2 to handle oob constant access
     bool robustness2Supported;
+
+    /// Whether runtime to apply Dref scaling for depth textures of specified bit depth
+    /// (24: D24S8, 16: D16, 0: Disabled). This allows compatability with games
+    /// that expect a different depth test range, which was typically a D3D8 quirk on
+    /// early NVIDIA hardware.
+    int32_t drefScaling = 0;
   };
 
 }
\ No newline at end of file
diff -pruN 2.3.1-1/src/dxvk/dxvk_adapter.cpp 2.4-1/src/dxvk/dxvk_adapter.cpp
--- 2.3.1-1/src/dxvk/dxvk_adapter.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_adapter.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -346,6 +346,11 @@ namespace dxvk {
     enabledFeatures.vk13.synchronization2 = VK_TRUE;
     enabledFeatures.vk13.dynamicRendering = VK_TRUE;
 
+    // Maintenance4 may cause performance problems on amdvlk in some cases
+    if (m_deviceInfo.vk12.driverID != VK_DRIVER_ID_AMD_OPEN_SOURCE
+     && m_deviceInfo.vk12.driverID != VK_DRIVER_ID_AMD_PROPRIETARY)
+      enabledFeatures.vk13.maintenance4 = VK_TRUE;
+
     // We expose depth clip rather than depth clamp to client APIs
     enabledFeatures.extDepthClipEnable.depthClipEnable =
       m_deviceFeatures.extDepthClipEnable.depthClipEnable;
@@ -410,11 +415,15 @@ namespace dxvk {
       m_deviceFeatures.khrPresentWait.presentWait;
 
     // Unless we're on an Nvidia driver where these extensions are known to be broken
-    if (matchesDriver(VK_DRIVER_ID_NVIDIA_PROPRIETARY, 0, VK_MAKE_VERSION(535, 0, 0))) {
+    if (matchesDriver(VK_DRIVER_ID_NVIDIA_PROPRIETARY, Version(), Version(535, 0, 0))) {
       enabledFeatures.khrPresentId.presentId = VK_FALSE;
       enabledFeatures.khrPresentWait.presentWait = VK_FALSE;
     }
 
+    // Enable descriptor pool overallocation if supported
+    enabledFeatures.nvDescriptorPoolOverallocation.descriptorPoolOverallocation =
+      m_deviceFeatures.nvDescriptorPoolOverallocation.descriptorPoolOverallocation;
+
     // Enable raw access chains for shader backends
     enabledFeatures.nvRawAccessChains.shaderRawAccessChains =
       m_deviceFeatures.nvRawAccessChains.shaderRawAccessChains;
@@ -425,10 +434,7 @@ namespace dxvk {
     // Log feature support info an extension list
     Logger::info(str::format("Device properties:"
       "\n  Device : ", m_deviceInfo.core.properties.deviceName,
-      "\n  Driver : ", m_deviceInfo.vk12.driverName, " ",
-      VK_VERSION_MAJOR(m_deviceInfo.core.properties.driverVersion), ".",
-      VK_VERSION_MINOR(m_deviceInfo.core.properties.driverVersion), ".",
-      VK_VERSION_PATCH(m_deviceInfo.core.properties.driverVersion)));
+      "\n  Driver : ", m_deviceInfo.vk12.driverName, " ", m_deviceInfo.driverVersion.toString()));
 
     Logger::info("Enabled device extensions:");
     this->logNameList(extensionNameList);
@@ -610,6 +616,10 @@ namespace dxvk {
           enabledFeatures.khrPresentWait = *reinterpret_cast<const VkPhysicalDevicePresentWaitFeaturesKHR*>(f);
           break;
 
+        case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_POOL_OVERALLOCATION_FEATURES_NV:
+          enabledFeatures.nvDescriptorPoolOverallocation = *reinterpret_cast<const VkPhysicalDeviceDescriptorPoolOverallocationFeaturesNV*>(f);
+          break;
+
         case VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAW_ACCESS_CHAINS_FEATURES_NV:
           enabledFeatures.nvRawAccessChains = *reinterpret_cast<const VkPhysicalDeviceRawAccessChainsFeaturesNV*>(f);
           break;
@@ -626,9 +636,7 @@ namespace dxvk {
     Logger::info(str::format("Device properties:"
       "\n  Device name: ", m_deviceInfo.core.properties.deviceName,
       "\n  Driver:      ", m_deviceInfo.vk12.driverName, " ",
-      VK_VERSION_MAJOR(m_deviceInfo.core.properties.driverVersion), ".",
-      VK_VERSION_MINOR(m_deviceInfo.core.properties.driverVersion), ".",
-      VK_VERSION_PATCH(m_deviceInfo.core.properties.driverVersion)));
+      m_deviceInfo.driverVersion.toString()));
 
     Logger::info("Enabled device extensions:");
     this->logNameList(extensionNameList);
@@ -664,26 +672,29 @@ namespace dxvk {
 
   bool DxvkAdapter::matchesDriver(
           VkDriverIdKHR       driver,
-          uint32_t            minVer,
-          uint32_t            maxVer) const {
+          Version             minVer,
+          Version             maxVer) const {
     bool driverMatches = driver == m_deviceInfo.vk12.driverID;
 
-    if (minVer) driverMatches &= m_deviceInfo.core.properties.driverVersion >= minVer;
-    if (maxVer) driverMatches &= m_deviceInfo.core.properties.driverVersion <  maxVer;
+    if (minVer) driverMatches &= m_deviceInfo.driverVersion >= minVer;
+    if (maxVer) driverMatches &= m_deviceInfo.driverVersion <  maxVer;
 
     return driverMatches;
   }
   
   
+  bool DxvkAdapter::matchesDriver(
+          VkDriverIdKHR       driver) const {
+    return driver == m_deviceInfo.vk12.driverID;
+  }
+  
+  
   void DxvkAdapter::logAdapterInfo() const {
     const auto deviceInfo = this->devicePropertiesExt();
     const auto memoryInfo = this->memoryProperties();
     
     Logger::info(str::format(deviceInfo.core.properties.deviceName, ":",
-      "\n  Driver : ", deviceInfo.vk12.driverName, " ",
-      VK_VERSION_MAJOR(deviceInfo.core.properties.driverVersion), ".",
-      VK_VERSION_MINOR(deviceInfo.core.properties.driverVersion), ".",
-      VK_VERSION_PATCH(deviceInfo.core.properties.driverVersion)));
+      "\n  Driver : ", deviceInfo.vk12.driverName, " ", deviceInfo.driverVersion.toString()));
 
     for (uint32_t i = 0; i < memoryInfo.memoryHeapCount; i++) {
       constexpr VkDeviceSize mib = 1024 * 1024;
@@ -785,22 +796,8 @@ namespace dxvk {
     m_vki->vkGetPhysicalDeviceProperties2(m_handle, &m_deviceInfo.core);
     
     // Some drivers reports the driver version in a slightly different format
-    switch (m_deviceInfo.vk12.driverID) {
-      case VK_DRIVER_ID_NVIDIA_PROPRIETARY:
-        m_deviceInfo.core.properties.driverVersion = VK_MAKE_VERSION(
-          (m_deviceInfo.core.properties.driverVersion >> 22) & 0x3ff,
-          (m_deviceInfo.core.properties.driverVersion >> 14) & 0x0ff,
-          (m_deviceInfo.core.properties.driverVersion >>  6) & 0x0ff);
-        break;
-
-      case VK_DRIVER_ID_INTEL_PROPRIETARY_WINDOWS:
-        m_deviceInfo.core.properties.driverVersion = VK_MAKE_VERSION(
-          m_deviceInfo.core.properties.driverVersion >> 14,
-          m_deviceInfo.core.properties.driverVersion & 0x3fff, 0);
-        break;
-
-      default:;
-    }
+    m_deviceInfo.driverVersion = decodeDriverVersion(
+      m_deviceInfo.vk12.driverID, m_deviceInfo.core.properties.driverVersion);
   }
 
 
@@ -935,6 +932,11 @@ namespace dxvk {
       m_deviceFeatures.khrPresentWait.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.khrPresentWait);
     }
 
+    if (m_deviceExtensions.supports(VK_NV_DESCRIPTOR_POOL_OVERALLOCATION_EXTENSION_NAME)) {
+      m_deviceFeatures.nvDescriptorPoolOverallocation.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_POOL_OVERALLOCATION_FEATURES_NV;
+      m_deviceFeatures.nvDescriptorPoolOverallocation.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.nvDescriptorPoolOverallocation);
+    }
+
     if (m_deviceExtensions.supports(VK_NV_RAW_ACCESS_CHAINS_EXTENSION_NAME)) {
       m_deviceFeatures.nvRawAccessChains.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAW_ACCESS_CHAINS_FEATURES_NV;
       m_deviceFeatures.nvRawAccessChains.pNext = std::exchange(m_deviceFeatures.core.pNext, &m_deviceFeatures.nvRawAccessChains);
@@ -1007,6 +1009,7 @@ namespace dxvk {
       &devExtensions.khrPresentWait,
       &devExtensions.khrSwapchain,
       &devExtensions.khrWin32KeyedMutex,
+      &devExtensions.nvDescriptorPoolOverallocation,
       &devExtensions.nvRawAccessChains,
       &devExtensions.nvxBinaryImport,
       &devExtensions.nvxImageViewHandle,
@@ -1147,6 +1150,11 @@ namespace dxvk {
       enabledFeatures.khrPresentWait.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.khrPresentWait);
     }
 
+    if (devExtensions.nvDescriptorPoolOverallocation) {
+      enabledFeatures.nvDescriptorPoolOverallocation.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_DESCRIPTOR_POOL_OVERALLOCATION_FEATURES_NV;
+      enabledFeatures.nvDescriptorPoolOverallocation.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.nvDescriptorPoolOverallocation);
+    }
+
     if (devExtensions.nvRawAccessChains) {
       enabledFeatures.nvRawAccessChains.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_RAW_ACCESS_CHAINS_FEATURES_NV;
       enabledFeatures.nvRawAccessChains.pNext = std::exchange(enabledFeatures.core.pNext, &enabledFeatures.nvRawAccessChains);
@@ -1298,6 +1306,8 @@ namespace dxvk {
       "\n  presentId                              : ", features.khrPresentId.presentId ? "1" : "0",
       "\n", VK_KHR_PRESENT_WAIT_EXTENSION_NAME,
       "\n  presentWait                            : ", features.khrPresentWait.presentWait ? "1" : "0",
+      "\n", VK_NV_DESCRIPTOR_POOL_OVERALLOCATION_EXTENSION_NAME,
+      "\n  descriptorPoolOverallocation           : ", features.nvDescriptorPoolOverallocation.descriptorPoolOverallocation ? "1" : "0",
       "\n", VK_NV_RAW_ACCESS_CHAINS_EXTENSION_NAME,
       "\n  shaderRawAccessChains                  : ", features.nvRawAccessChains.shaderRawAccessChains ? "1" : "0",
       "\n", VK_NVX_BINARY_IMPORT_EXTENSION_NAME,
@@ -1316,4 +1326,25 @@ namespace dxvk {
       "\n  Sparse   : ", queues.sparse != VK_QUEUE_FAMILY_IGNORED ? str::format(queues.sparse) : "n/a"));
   }
   
+
+  Version DxvkAdapter::decodeDriverVersion(VkDriverId driverId, uint32_t version) {
+    switch (driverId) {
+      case VK_DRIVER_ID_NVIDIA_PROPRIETARY:
+        return Version(
+          (version >> 22) & 0x3ff,
+          (version >> 14) & 0x0ff,
+          (version >>  6) & 0x0ff);
+        break;
+
+      case VK_DRIVER_ID_INTEL_PROPRIETARY_WINDOWS:
+        return Version(version >> 14, version & 0x3fff, 0);
+
+      default:
+        return Version(
+          VK_API_VERSION_MAJOR(version),
+          VK_API_VERSION_MINOR(version),
+          VK_API_VERSION_PATCH(version));
+    }
+  }
+
 }
diff -pruN 2.3.1-1/src/dxvk/dxvk_adapter.h 2.4-1/src/dxvk/dxvk_adapter.h
--- 2.3.1-1/src/dxvk/dxvk_adapter.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_adapter.h	2024-07-10 09:56:14.000000000 +0000
@@ -255,8 +255,17 @@ namespace dxvk {
      */
     bool matchesDriver(
             VkDriverIdKHR       driver,
-            uint32_t            minVer,
-            uint32_t            maxVer) const;
+            Version             minVer,
+            Version             maxVer) const;
+
+    /**
+     * \brief Tests if the driver matches certain criteria
+     *
+     * \param [in] driver Driver ID
+     * \returns \c True if the driver matches these criteria
+     */
+    bool matchesDriver(
+            VkDriverIdKHR       driver) const;
     
     /**
      * \brief Logs DXVK adapter info
@@ -343,6 +352,8 @@ namespace dxvk {
     static void logFeatures(const DxvkDeviceFeatures& features);
     static void logQueueFamilies(const DxvkAdapterQueueIndices& queues);
     
+    static Version decodeDriverVersion(VkDriverId driverId, uint32_t version);
+
   };
   
 }
diff -pruN 2.3.1-1/src/dxvk/dxvk_compute.cpp 2.4-1/src/dxvk/dxvk_compute.cpp
--- 2.3.1-1/src/dxvk/dxvk_compute.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_compute.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -48,9 +48,7 @@ namespace dxvk {
       // Retrieve actual pipeline handle on first use. This
       // may wait for an ongoing compile job to finish, or
       // compile the pipeline immediately on the calling thread.
-      m_libraryHandle = m_library->acquirePipelineHandle(
-        DxvkShaderPipelineLibraryCompileArgs());
-
+      m_libraryHandle = m_library->acquirePipelineHandle().handle;
       return m_libraryHandle;
     } else {
       // Slow path for compute shaders that do use spec constants
diff -pruN 2.3.1-1/src/dxvk/dxvk_context.cpp 2.4-1/src/dxvk/dxvk_context.cpp
--- 2.3.1-1/src/dxvk/dxvk_context.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_context.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -1997,7 +1997,7 @@ namespace dxvk {
     depthOp.loadOpS       = VK_ATTACHMENT_LOAD_OP_LOAD;
     depthOp.loadLayout    = imageView->imageInfo().layout;
     depthOp.storeLayout   = imageView->imageInfo().layout;
-    
+
     if (clearAspects & VK_IMAGE_ASPECT_COLOR_BIT)
       colorOp.loadOp = VK_ATTACHMENT_LOAD_OP_CLEAR;
     else if (discardAspects & VK_IMAGE_ASPECT_COLOR_BIT)
@@ -2010,7 +2010,7 @@ namespace dxvk {
     
     if (clearAspects & VK_IMAGE_ASPECT_STENCIL_BIT)
       depthOp.loadOpS = VK_ATTACHMENT_LOAD_OP_CLEAR;
-    else if (discardAspects & VK_IMAGE_ASPECT_DEPTH_BIT)
+    else if (discardAspects & VK_IMAGE_ASPECT_STENCIL_BIT)
       depthOp.loadOpS = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
 
     if (attachmentIndex >= 0 && !m_state.om.framebufferInfo.isWritable(attachmentIndex, clearAspects | discardAspects)) {
@@ -2041,6 +2041,8 @@ namespace dxvk {
     }
     
     if (attachmentIndex < 0) {
+      bool hasViewFormatMismatch = imageView->info().format != imageView->imageInfo().format;
+
       if (m_execBarriers.isImageDirty(imageView->image(), imageView->imageSubresources(), DxvkAccess::Write))
         m_execBarriers.recordCommands(m_cmd);
 
@@ -2075,6 +2077,11 @@ namespace dxvk {
 
         attachmentInfo.loadOp = colorOp.loadOp;
 
+        // We can't use LOAD_OP_CLEAR if the view format does not match the
+        // underlying image format, so just discard here and use clear later.
+        if (hasViewFormatMismatch && attachmentInfo.loadOp == VK_ATTACHMENT_LOAD_OP_CLEAR)
+          attachmentInfo.loadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
+
         renderingInfo.colorAttachmentCount = 1;
         renderingInfo.pColorAttachments = &attachmentInfo;
 
@@ -2110,6 +2117,20 @@ namespace dxvk {
       }
 
       m_cmd->cmdBeginRendering(&renderingInfo);
+
+      if (hasViewFormatMismatch) {
+        VkClearAttachment clearInfo = { };
+        clearInfo.aspectMask = imageView->info().aspect;
+        clearInfo.clearValue = clearValue;
+
+        VkClearRect clearRect = { };
+        clearRect.rect.extent.width = extent.width;
+        clearRect.rect.extent.height = extent.height;
+        clearRect.layerCount = imageView->info().numLayers;
+
+        m_cmd->cmdClearAttachments(1, &clearInfo, 1, &clearRect);
+      }
+
       m_cmd->cmdEndRendering();
 
       m_execBarriers.accessImage(
@@ -4741,8 +4762,10 @@ namespace dxvk {
     this->renderPassEmitPostBarriers(framebufferInfo, ops);
 
     uint32_t colorInfoCount = 0;
+    uint32_t lateClearCount = 0;
 
     std::array<VkRenderingAttachmentInfo, MaxNumRenderTargets> colorInfos;
+    std::array<VkClearAttachment, MaxNumRenderTargets> lateClears;
 
     for (uint32_t i = 0; i < MaxNumRenderTargets; i++) {
       const auto& colorTarget = framebufferInfo.getColorTarget(i);
@@ -4754,9 +4777,21 @@ namespace dxvk {
         colorInfos[i].loadOp = ops.colorOps[i].loadOp;
         colorInfos[i].storeOp = VK_ATTACHMENT_STORE_OP_STORE;
 
-        if (ops.colorOps[i].loadOp == VK_ATTACHMENT_LOAD_OP_CLEAR)
+        if (ops.colorOps[i].loadOp == VK_ATTACHMENT_LOAD_OP_CLEAR) {
           colorInfos[i].clearValue.color = ops.colorOps[i].clearValue;
 
+          // We can't use LOAD_OP_CLEAR if the view format does not match the
+          // underlying image format, so just discard here and use clear later.
+          if (colorTarget.view->info().format != colorTarget.view->imageInfo().format) {
+            colorInfos[i].loadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
+
+            auto& clear = lateClears[lateClearCount++];
+            clear.colorAttachment = i;
+            clear.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
+            clear.clearValue.color = ops.colorOps[i].clearValue;
+          }
+        }
+
         colorInfoCount = i + 1;
       }
     }
@@ -4804,6 +4839,15 @@ namespace dxvk {
 
     m_cmd->cmdBeginRendering(&renderingInfo);
     
+    if (lateClearCount) {
+      VkClearRect clearRect = { };
+      clearRect.rect.extent.width   = fbSize.width;
+      clearRect.rect.extent.height  = fbSize.height;
+      clearRect.layerCount          = fbSize.layers;
+
+      m_cmd->cmdClearAttachments(lateClearCount, lateClears.data(), 1, &clearRect);
+    }
+
     for (uint32_t i = 0; i < framebufferInfo.numAttachments(); i++) {
       m_cmd->trackResource<DxvkAccess::None> (framebufferInfo.getAttachment(i).view);
       m_cmd->trackResource<DxvkAccess::Write>(framebufferInfo.getAttachment(i).view->image());
@@ -4940,7 +4984,7 @@ namespace dxvk {
     // Mark compute resources and push constants as dirty
     m_descriptorState.dirtyStages(VK_SHADER_STAGE_COMPUTE_BIT);
 
-    if (newPipeline->getBindings()->layout().getPushConstantRange().size)
+    if (newPipeline->getBindings()->layout().getPushConstantRange(true).size)
       m_flags.set(DxvkContextFlag::DirtyPushConstants);
 
     m_flags.clr(DxvkContextFlag::CpDirtyPipelineState);
@@ -5014,7 +5058,7 @@ namespace dxvk {
 
     m_descriptorState.dirtyStages(VK_SHADER_STAGE_ALL_GRAPHICS);
 
-    if (newPipeline->getBindings()->layout().getPushConstantRange().size)
+    if (newPipeline->getBindings()->layout().getPushConstantRange(true).size)
       m_flags.set(DxvkContextFlag::DirtyPushConstants);
 
     m_flags.clr(DxvkContextFlag::GpDirtyPipeline);
@@ -5881,16 +5925,19 @@ namespace dxvk {
     auto bindings = BindPoint == VK_PIPELINE_BIND_POINT_GRAPHICS
       ? m_state.gp.pipeline->getBindings()
       : m_state.cp.pipeline->getBindings();
-    
-    VkPushConstantRange pushConstRange = bindings->layout().getPushConstantRange();
+
+    // Optimized pipelines may have push constants trimmed, so look up
+    // the exact layout used for the currently bound pipeline.
+    bool independentSets = BindPoint == VK_PIPELINE_BIND_POINT_GRAPHICS
+      && m_flags.test(DxvkContextFlag::GpIndependentSets);
+
+    VkPushConstantRange pushConstRange = bindings->layout().getPushConstantRange(independentSets);
 
     if (!pushConstRange.size)
       return;
 
-    // Push constants should be compatible between complete and
-    // independent layouts, so always ask for the complete one
     m_cmd->cmdPushConstants(
-      bindings->getPipelineLayout(false),
+      bindings->getPipelineLayout(independentSets),
       pushConstRange.stageFlags,
       pushConstRange.offset,
       pushConstRange.size,
diff -pruN 2.3.1-1/src/dxvk/dxvk_descriptor.cpp 2.4-1/src/dxvk/dxvk_descriptor.cpp
--- 2.3.1-1/src/dxvk/dxvk_descriptor.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_descriptor.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -72,7 +72,7 @@ namespace dxvk {
     // memory bloat. This may be necessary for off-screen
     // rendering applications, or in situations where games
     // pre-render a lot of images without presenting in between.
-    return m_descriptorPools.size() >= 8;
+    return m_descriptorPools.size() > MaxDesiredPoolCount;
   }
 
 
@@ -100,29 +100,25 @@ namespace dxvk {
 
 
   void DxvkDescriptorPool::reset() {
-    // As a heuristic to save memory, check how many descriptors
-    // have actively been used in the past couple of submissions.
-    bool isLowUsageFrame = false;
-
+    // As a heuristic to save memory, check how many descriptor
+    // sets were actually being used in past submissions.
     size_t poolCount = m_descriptorPools.size();
+    bool needsReset = poolCount > MaxDesiredPoolCount;
 
     if (poolCount > 1 || m_setsAllocated > m_manager->getMaxSetCount() / 2) {
       double factor = std::max(11.0 / 3.0 - double(poolCount) / 3.0, 1.0);
-      isLowUsageFrame = double(m_setsUsed) * factor < double(m_setsAllocated);
+      needsReset = double(m_setsUsed) * factor < double(m_setsAllocated);
     }
 
-    m_lowUsageFrames = isLowUsageFrame
-      ? m_lowUsageFrames + 1
-      : 0;
     m_setsUsed = 0;
 
-    if (m_lowUsageFrames < 16) {
+    if (!needsReset) {
       for (auto& entry : m_setLists)
         entry.second.reset();
     } else {
-      // If most sets are no longer being used, reset and destroy
-      // descriptor pools and reset all lookup tables in order to
-      // accomodate more descriptors of different layouts.
+      // If most sets are no longer needed, reset and destroy
+      // descriptor pools and reset all lookup tables in order
+      // to accomodate more descriptors of different layouts.
       for (auto pool : m_descriptorPools)
         m_manager->recycleVulkanDescriptorPool(pool);
 
@@ -131,7 +127,6 @@ namespace dxvk {
       m_setMaps.clear();
 
       m_setsAllocated = 0;
-      m_lowUsageFrames = 0;
     }
 
     m_cachedEntry = { nullptr, nullptr };
@@ -311,7 +306,12 @@ namespace dxvk {
     info.maxSets       = m_maxSets;
     info.poolSizeCount = pools.size();
     info.pPoolSizes    = pools.data();
-    
+
+    if (m_device->features().nvDescriptorPoolOverallocation.descriptorPoolOverallocation) {
+      info.flags |= VK_DESCRIPTOR_POOL_CREATE_ALLOW_OVERALLOCATION_POOLS_BIT_NV
+                 |  VK_DESCRIPTOR_POOL_CREATE_ALLOW_OVERALLOCATION_SETS_BIT_NV;
+    }
+
     VkDescriptorPool pool = VK_NULL_HANDLE;
 
     if (vk->vkCreateDescriptorPool(vk->device(), &info, nullptr, &pool) != VK_SUCCESS)
diff -pruN 2.3.1-1/src/dxvk/dxvk_descriptor.h 2.4-1/src/dxvk/dxvk_descriptor.h
--- 2.3.1-1/src/dxvk/dxvk_descriptor.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_descriptor.h	2024-07-10 09:56:14.000000000 +0000
@@ -80,7 +80,7 @@ namespace dxvk {
    * to be updated.
    */
   class DxvkDescriptorPool : public RcObject {
-
+    constexpr static uint32_t MaxDesiredPoolCount = 2;
   public:
 
     DxvkDescriptorPool(
@@ -155,8 +155,6 @@ namespace dxvk {
 
     uint32_t m_prevSetsAllocated = 0;
 
-    uint32_t m_lowUsageFrames = 0;
-
     DxvkDescriptorSetMap* getSetMapCached(
       const DxvkBindingLayoutObjects*           layout);
 
diff -pruN 2.3.1-1/src/dxvk/dxvk_device.cpp 2.4-1/src/dxvk/dxvk_device.cpp
--- 2.3.1-1/src/dxvk/dxvk_device.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_device.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -82,7 +82,7 @@ namespace dxvk {
 
         // Disable lifetime tracking for drivers that do not have any
         // significant issues with 32-bit address space to begin with
-        if (m_adapter->matchesDriver(VK_DRIVER_ID_MESA_RADV_KHR, 0, 0))
+        if (m_adapter->matchesDriver(VK_DRIVER_ID_MESA_RADV_KHR))
           return false;
 
         return true;
@@ -325,12 +325,12 @@ namespace dxvk {
   DxvkDevicePerfHints DxvkDevice::getPerfHints() {
     DxvkDevicePerfHints hints;
     hints.preferFbDepthStencilCopy = m_features.extShaderStencilExport
-      && (m_adapter->matchesDriver(VK_DRIVER_ID_MESA_RADV_KHR, 0, 0)
-       || m_adapter->matchesDriver(VK_DRIVER_ID_AMD_OPEN_SOURCE_KHR, 0, 0)
-       || m_adapter->matchesDriver(VK_DRIVER_ID_AMD_PROPRIETARY_KHR, 0, 0));
+      && (m_adapter->matchesDriver(VK_DRIVER_ID_MESA_RADV_KHR)
+       || m_adapter->matchesDriver(VK_DRIVER_ID_AMD_OPEN_SOURCE_KHR)
+       || m_adapter->matchesDriver(VK_DRIVER_ID_AMD_PROPRIETARY_KHR));
     hints.preferFbResolve = m_features.amdShaderFragmentMask
-      && (m_adapter->matchesDriver(VK_DRIVER_ID_AMD_OPEN_SOURCE_KHR, 0, 0)
-       || m_adapter->matchesDriver(VK_DRIVER_ID_AMD_PROPRIETARY_KHR, 0, 0));
+      && (m_adapter->matchesDriver(VK_DRIVER_ID_AMD_OPEN_SOURCE_KHR)
+       || m_adapter->matchesDriver(VK_DRIVER_ID_AMD_PROPRIETARY_KHR));
     return hints;
   }
 
diff -pruN 2.3.1-1/src/dxvk/dxvk_device_filter.cpp 2.4-1/src/dxvk/dxvk_device_filter.cpp
--- 2.3.1-1/src/dxvk/dxvk_device_filter.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_device_filter.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -2,11 +2,16 @@
 
 namespace dxvk {
   
-  DxvkDeviceFilter::DxvkDeviceFilter(DxvkDeviceFilterFlags flags)
+  DxvkDeviceFilter::DxvkDeviceFilter(
+          DxvkDeviceFilterFlags flags,
+    const DxvkOptions&          options)
   : m_flags(flags) {
     m_matchDeviceName = env::getEnvVar("DXVK_FILTER_DEVICE_NAME");
-    
-    if (m_matchDeviceName.size() != 0)
+
+    if (m_matchDeviceName.empty())
+      m_matchDeviceName = options.deviceFilter;
+
+    if (!m_matchDeviceName.empty())
       m_flags.set(DxvkDeviceFilterFlag::MatchDeviceName);
   }
   
@@ -17,10 +22,10 @@ namespace dxvk {
   
   
   bool DxvkDeviceFilter::testAdapter(const VkPhysicalDeviceProperties& properties) const {
-    if (properties.apiVersion < VK_MAKE_VERSION(1, 3, 0)) {
+    if (properties.apiVersion < VK_MAKE_API_VERSION(0, 1, 3, 0)) {
       Logger::warn(str::format("Skipping Vulkan ",
-        VK_VERSION_MAJOR(properties.apiVersion), ".",
-        VK_VERSION_MINOR(properties.apiVersion), " adapter: ",
+        VK_API_VERSION_MAJOR(properties.apiVersion), ".",
+        VK_API_VERSION_MINOR(properties.apiVersion), " adapter: ",
         properties.deviceName));
       return false;
     }
diff -pruN 2.3.1-1/src/dxvk/dxvk_device_filter.h 2.4-1/src/dxvk/dxvk_device_filter.h
--- 2.3.1-1/src/dxvk/dxvk_device_filter.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_device_filter.h	2024-07-10 09:56:14.000000000 +0000
@@ -1,6 +1,7 @@
 #pragma once
 
 #include "dxvk_adapter.h"
+#include "dxvk_options.h"
 
 namespace dxvk {
   
@@ -31,7 +32,10 @@ namespace dxvk {
     
   public:
     
-    DxvkDeviceFilter(DxvkDeviceFilterFlags flags);
+    DxvkDeviceFilter(
+            DxvkDeviceFilterFlags flags,
+      const DxvkOptions&          options);
+
     ~DxvkDeviceFilter();
     
     /**
diff -pruN 2.3.1-1/src/dxvk/dxvk_device_info.h 2.4-1/src/dxvk/dxvk_device_info.h
--- 2.3.1-1/src/dxvk/dxvk_device_info.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_device_info.h	2024-07-10 09:56:14.000000000 +0000
@@ -2,6 +2,8 @@
 
 #include "dxvk_include.h"
 
+#include "../util/util_version.h"
+
 namespace dxvk {
 
   /**
@@ -13,6 +15,7 @@ namespace dxvk {
    * so before using them, check whether they are supported.
    */
   struct DxvkDeviceInfo {
+    Version                                                   driverVersion;
     VkPhysicalDeviceProperties2                               core;
     VkPhysicalDeviceVulkan11Properties                        vk11;
     VkPhysicalDeviceVulkan12Properties                        vk12;
@@ -68,6 +71,7 @@ namespace dxvk {
     VkPhysicalDeviceMaintenance5FeaturesKHR                   khrMaintenance5;
     VkPhysicalDevicePresentIdFeaturesKHR                      khrPresentId;
     VkPhysicalDevicePresentWaitFeaturesKHR                    khrPresentWait;
+    VkPhysicalDeviceDescriptorPoolOverallocationFeaturesNV    nvDescriptorPoolOverallocation;
     VkPhysicalDeviceRawAccessChainsFeaturesNV                 nvRawAccessChains;
     VkBool32                                                  nvxBinaryImport;
     VkBool32                                                  nvxImageViewHandle;
diff -pruN 2.3.1-1/src/dxvk/dxvk_extensions.h 2.4-1/src/dxvk/dxvk_extensions.h
--- 2.3.1-1/src/dxvk/dxvk_extensions.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_extensions.h	2024-07-10 09:56:14.000000000 +0000
@@ -325,6 +325,7 @@ namespace dxvk {
     DxvkExt khrPresentWait                    = { VK_KHR_PRESENT_WAIT_EXTENSION_NAME,                       DxvkExtMode::Optional };
     DxvkExt khrSwapchain                      = { VK_KHR_SWAPCHAIN_EXTENSION_NAME,                          DxvkExtMode::Required };
     DxvkExt khrWin32KeyedMutex                = { VK_KHR_WIN32_KEYED_MUTEX_EXTENSION_NAME,                  DxvkExtMode::Optional };
+    DxvkExt nvDescriptorPoolOverallocation    = { VK_NV_DESCRIPTOR_POOL_OVERALLOCATION_EXTENSION_NAME,      DxvkExtMode::Optional };
     DxvkExt nvRawAccessChains                 = { VK_NV_RAW_ACCESS_CHAINS_EXTENSION_NAME,                   DxvkExtMode::Optional };
     DxvkExt nvxBinaryImport                   = { VK_NVX_BINARY_IMPORT_EXTENSION_NAME,                      DxvkExtMode::Disabled };
     DxvkExt nvxImageViewHandle                = { VK_NVX_IMAGE_VIEW_HANDLE_EXTENSION_NAME,                  DxvkExtMode::Disabled };
diff -pruN 2.3.1-1/src/dxvk/dxvk_graphics.cpp 2.4-1/src/dxvk/dxvk_graphics.cpp
--- 2.3.1-1/src/dxvk/dxvk_graphics.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_graphics.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -1111,7 +1111,9 @@ namespace dxvk {
 
     if (doCreateBasePipeline)
       baseHandle = this->getBasePipeline(state);
-    else
+
+    // Fast-linking may fail in some situations
+    if (!baseHandle)
       fastHandle = this->getOptimizedPipeline(state);
 
     // Log pipeline state if requested, or on failure
@@ -1148,6 +1150,13 @@ namespace dxvk {
      || (state.rs.lineMode() != VK_LINE_RASTERIZATION_MODE_DEFAULT_EXT && isLineRendering))
       return false;
 
+    // Depth clip is assumed to be enabled. If the driver does not
+    // support dynamic depth clip, we'd have to late-compile anyway
+    // unless the pipeline is used multiple times.
+    if (!m_device->features().extExtendedDynamicState3.extendedDynamicState3DepthClipEnable
+     && !state.rs.depthClipEnable())
+      return false;
+
     if (m_shaders.tcs != nullptr) {
       // If tessellation shaders are present, the input patch
       // vertex count must match the shader's definition.
@@ -1219,9 +1228,6 @@ namespace dxvk {
     key.viLibrary = m_manager->createVertexInputLibrary(viState);
     key.foLibrary = m_manager->createFragmentOutputLibrary(foState);
 
-    if (!m_device->features().extExtendedDynamicState3.extendedDynamicState3DepthClipEnable)
-      key.args.depthClipEnable = state.rs.depthClipEnable();
-
     auto entry = m_basePipelines.find(key);
     if (entry != m_basePipelines.end())
       return entry->second;
@@ -1236,10 +1242,11 @@ namespace dxvk {
     const DxvkGraphicsPipelineBaseInstanceKey& key) const {
     auto vk = m_device->vkd();
 
+    DxvkShaderPipelineLibraryHandle vs = m_vsLibrary->acquirePipelineHandle();
+    DxvkShaderPipelineLibraryHandle fs = m_fsLibrary->acquirePipelineHandle();
+
     std::array<VkPipeline, 4> libraries = {{
-      key.viLibrary->getHandle(),
-      m_vsLibrary->acquirePipelineHandle(key.args),
-      m_fsLibrary->acquirePipelineHandle(key.args),
+      key.viLibrary->getHandle(), vs.handle, fs.handle,
       key.foLibrary->getHandle(),
     }};
 
@@ -1248,13 +1255,14 @@ namespace dxvk {
     libInfo.pLibraries      = libraries.data();
 
     VkGraphicsPipelineCreateInfo info = { VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO, &libInfo };
+    info.flags              = vs.linkFlags | fs.linkFlags;
     info.layout             = m_bindings->getPipelineLayout(true);
     info.basePipelineIndex  = -1;
 
     VkPipeline pipeline = VK_NULL_HANDLE;
     VkResult vr = vk->vkCreateGraphicsPipelines(vk->device(), VK_NULL_HANDLE, 1, &info, nullptr, &pipeline);
 
-    if (vr != VK_SUCCESS)
+    if (vr && vr != VK_PIPELINE_COMPILE_REQUIRED_EXT)
       Logger::err(str::format("DxvkGraphicsPipeline: Failed to create base pipeline: ", vr));
 
     return pipeline;
diff -pruN 2.3.1-1/src/dxvk/dxvk_graphics.h 2.4-1/src/dxvk/dxvk_graphics.h
--- 2.3.1-1/src/dxvk/dxvk_graphics.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_graphics.h	2024-07-10 09:56:14.000000000 +0000
@@ -381,19 +381,16 @@ namespace dxvk {
   struct DxvkGraphicsPipelineBaseInstanceKey {
     const DxvkGraphicsPipelineVertexInputLibrary*     viLibrary = nullptr;
     const DxvkGraphicsPipelineFragmentOutputLibrary*  foLibrary = nullptr;
-    DxvkShaderPipelineLibraryCompileArgs              args;
 
     bool eq(const DxvkGraphicsPipelineBaseInstanceKey& other) const {
       return viLibrary == other.viLibrary
-          && foLibrary == other.foLibrary
-          && args      == other.args;
+          && foLibrary == other.foLibrary;
     }
 
     size_t hash() const {
       DxvkHashState hash;
       hash.add(size_t(viLibrary));
       hash.add(size_t(foLibrary));
-      hash.add(args.hash());
       return hash;
     }
   };
diff -pruN 2.3.1-1/src/dxvk/dxvk_instance.cpp 2.4-1/src/dxvk/dxvk_instance.cpp
--- 2.3.1-1/src/dxvk/dxvk_instance.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_instance.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -4,6 +4,7 @@
 #include "dxvk_openvr.h"
 #include "dxvk_openxr.h"
 #include "dxvk_platform_exts.h"
+#include "../wsi/wsi_platform.h"
 
 #include <algorithm>
 #include <sstream>
@@ -20,6 +21,8 @@ namespace dxvk {
     Logger::info(str::format("Game: ", env::getExeName()));
     Logger::info(str::format("DXVK: ", DXVK_VERSION));
 
+    wsi::init();
+
     m_config = Config::getUserConfig();
     m_config.merge(Config::getAppConfig(env::getExePath()));
     m_config.logOptions();
@@ -64,6 +67,8 @@ namespace dxvk {
   DxvkInstance::~DxvkInstance() {
     if (m_messenger)
       m_vki->vkDestroyDebugUtilsMessengerEXT(m_vki->instance(), m_messenger, nullptr);
+
+    wsi::quit();
   }
   
   
@@ -177,8 +182,8 @@ namespace dxvk {
       appInfo.pApplicationName      = appName.c_str();
       appInfo.applicationVersion    = flags.raw();
       appInfo.pEngineName           = "DXVK";
-      appInfo.engineVersion         = VK_MAKE_VERSION(2, 3, 1);
-      appInfo.apiVersion            = VK_MAKE_VERSION(1, 3, 0);
+      appInfo.engineVersion         = VK_MAKE_API_VERSION(0, 2, 4, 0);
+      appInfo.apiVersion            = VK_MAKE_API_VERSION(0, 1, 3, 0);
 
       VkInstanceCreateInfo info = { VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO };
       info.pApplicationInfo         = &appInfo;
@@ -251,7 +256,7 @@ namespace dxvk {
         filterFlags.set(DxvkDeviceFilterFlag::SkipCpuDevices);
     }
 
-    DxvkDeviceFilter filter(filterFlags);
+    DxvkDeviceFilter filter(filterFlags, m_options);
     std::vector<Rc<DxvkAdapter>> result;
 
     uint32_t numDGPU = 0;
diff -pruN 2.3.1-1/src/dxvk/dxvk_options.cpp 2.4-1/src/dxvk/dxvk_options.cpp
--- 2.3.1-1/src/dxvk/dxvk_options.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_options.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -13,6 +13,7 @@ namespace dxvk {
     hud                   = config.getOption<std::string>("dxvk.hud", "");
     tearFree              = config.getOption<Tristate>("dxvk.tearFree",               Tristate::Auto);
     hideIntegratedGraphics = config.getOption<bool>   ("dxvk.hideIntegratedGraphics", false);
+    deviceFilter          = config.getOption<std::string>("dxvk.deviceFilter",        "");
   }
 
 }
diff -pruN 2.3.1-1/src/dxvk/dxvk_options.h 2.4-1/src/dxvk/dxvk_options.h
--- 2.3.1-1/src/dxvk/dxvk_options.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_options.h	2024-07-10 09:56:14.000000000 +0000
@@ -41,6 +41,9 @@ namespace dxvk {
     // present. May be necessary for some games that
     // incorrectly assume monitor layouts.
     bool hideIntegratedGraphics;
+
+    // Device name
+    std::string deviceFilter;
   };
 
 }
diff -pruN 2.3.1-1/src/dxvk/dxvk_pipelayout.cpp 2.4-1/src/dxvk/dxvk_pipelayout.cpp
--- 2.3.1-1/src/dxvk/dxvk_pipelayout.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_pipelayout.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -205,7 +205,7 @@ namespace dxvk {
 
 
   DxvkBindingLayout::DxvkBindingLayout(VkShaderStageFlags stages)
-  : m_pushConst { 0, 0, 0 }, m_stages(stages) {
+  : m_pushConst { 0, 0, 0 }, m_pushConstStages(0), m_stages(stages) {
 
   }
 
@@ -249,11 +249,17 @@ namespace dxvk {
   }
 
 
+  void DxvkBindingLayout::addPushConstantStage(VkShaderStageFlagBits stage) {
+    m_pushConstStages |= stage;
+  }
+
+
   void DxvkBindingLayout::merge(const DxvkBindingLayout& layout) {
     for (uint32_t i = 0; i < layout.m_bindings.size(); i++)
       m_bindings[i].merge(layout.m_bindings[i]);
 
     addPushConstantRange(layout.m_pushConst);
+    m_pushConstStages |= layout.m_pushConstStages;
   }
 
 
@@ -266,6 +272,9 @@ namespace dxvk {
         return false;
     }
 
+    if (m_pushConstStages != other.m_pushConstStages)
+      return false;
+
     if (m_pushConst.stageFlags != other.m_pushConst.stageFlags
      || m_pushConst.offset     != other.m_pushConst.offset
      || m_pushConst.size       != other.m_pushConst.size)
@@ -282,6 +291,7 @@ namespace dxvk {
     for (uint32_t i = 0; i < m_bindings.size(); i++)
       hash.add(m_bindings[i].hash());
 
+    hash.add(m_pushConstStages);
     hash.add(m_pushConst.stageFlags);
     hash.add(m_pushConst.offset);
     hash.add(m_pushConst.size);
@@ -334,15 +344,16 @@ namespace dxvk {
     }
 
     // Create pipeline layout objects
-    VkPushConstantRange pushConst = m_layout.getPushConstantRange();
+    VkPushConstantRange pushConstComplete = m_layout.getPushConstantRange(false);
+    VkPushConstantRange pushConstIndependent = m_layout.getPushConstantRange(true);
 
     VkPipelineLayoutCreateInfo pipelineLayoutInfo = { VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO };
     pipelineLayoutInfo.setLayoutCount = setCount;
     pipelineLayoutInfo.pSetLayouts = setLayouts.data();
 
-    if (pushConst.stageFlags && pushConst.size) {
+    if (pushConstComplete.stageFlags && pushConstComplete.size) {
       pipelineLayoutInfo.pushConstantRangeCount = 1;
-      pipelineLayoutInfo.pPushConstantRanges = &pushConst;
+      pipelineLayoutInfo.pPushConstantRanges = &pushConstComplete;
     }
 
     // If the full set is defined, create a layout without INDEPENDENT_SET_BITS
@@ -356,6 +367,11 @@ namespace dxvk {
     if (m_device->canUseGraphicsPipelineLibrary() && (m_layout.getStages() & VK_SHADER_STAGE_ALL_GRAPHICS)) {
       pipelineLayoutInfo.flags = VK_PIPELINE_LAYOUT_CREATE_INDEPENDENT_SETS_BIT_EXT;
 
+      if (pushConstIndependent.stageFlags && pushConstIndependent.size) {
+        pipelineLayoutInfo.pushConstantRangeCount = 1;
+        pipelineLayoutInfo.pPushConstantRanges = &pushConstIndependent;
+      }
+
       if (vk->vkCreatePipelineLayout(vk->device(), &pipelineLayoutInfo, nullptr, &m_independentLayout))
         throw DxvkError("DxvkBindingLayoutObjects: Failed to create pipeline layout");
     }
diff -pruN 2.3.1-1/src/dxvk/dxvk_pipelayout.h 2.4-1/src/dxvk/dxvk_pipelayout.h
--- 2.3.1-1/src/dxvk/dxvk_pipelayout.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_pipelayout.h	2024-07-10 09:56:14.000000000 +0000
@@ -292,8 +292,19 @@ namespace dxvk {
      * \brief Retrieves push constant range
      * \returns Push constant range
      */
-    VkPushConstantRange getPushConstantRange() const {
-      return m_pushConst;
+    VkPushConstantRange getPushConstantRange(bool independent) const {
+      VkPushConstantRange result = m_pushConst;
+
+      if (!independent) {
+        result.stageFlags &= m_pushConstStages;
+
+        if (!result.stageFlags) {
+          result.offset = 0;
+          result.size = 0;
+        }
+      }
+
+      return result;
     }
 
     /**
@@ -325,6 +336,12 @@ namespace dxvk {
     void addPushConstantRange(VkPushConstantRange range);
 
     /**
+     * \brief Adds a stage that actively uses push constants
+     * \param [in] stage Shader stage
+     */
+    void addPushConstantStage(VkShaderStageFlagBits stage);
+
+    /**
      * \brief Merges binding layouts
      *
      * Adds bindings and push constant range from another layout to
@@ -353,6 +370,7 @@ namespace dxvk {
 
     std::array<DxvkBindingList, DxvkDescriptorSets::SetCount> m_bindings;
     VkPushConstantRange                                       m_pushConst;
+    VkShaderStageFlags                                        m_pushConstStages;
     VkShaderStageFlags                                        m_stages;
 
   };
diff -pruN 2.3.1-1/src/dxvk/dxvk_platform_exts.cpp 2.4-1/src/dxvk/dxvk_platform_exts.cpp
--- 2.3.1-1/src/dxvk/dxvk_platform_exts.cpp	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_platform_exts.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,40 @@
+#include "dxvk_platform_exts.h"
+#include "../wsi/wsi_platform.h"
+
+namespace dxvk {
+
+  DxvkPlatformExts DxvkPlatformExts::s_instance;
+
+  std::string_view DxvkPlatformExts::getName() {
+    return "Platform WSI";
+  }
+
+
+  DxvkNameSet DxvkPlatformExts::getInstanceExtensions() {
+    std::vector<const char *> extensionNames = wsi::getInstanceExtensions();
+
+    DxvkNameSet names;
+    for (const char* name : extensionNames)
+      names.add(name);
+
+    return names;
+  }
+
+
+  DxvkNameSet DxvkPlatformExts::getDeviceExtensions(
+          uint32_t      adapterId) {
+    return DxvkNameSet();
+  }
+
+
+  void DxvkPlatformExts::initInstanceExtensions() {
+
+  }
+
+
+  void DxvkPlatformExts::initDeviceExtensions(
+    const DxvkInstance* instance) {
+
+  }
+
+}
diff -pruN 2.3.1-1/src/dxvk/dxvk_presenter.cpp 2.4-1/src/dxvk/dxvk_presenter.cpp
--- 2.3.1-1/src/dxvk/dxvk_presenter.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_presenter.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -130,7 +130,7 @@ namespace dxvk {
       m_frameQueue.push(frame);
       m_frameCond.notify_one();
     } else {
-      applyFrameRateLimit(mode);
+      m_fpsLimiter.delay();
       m_signal->signal(frameId);
     }
 
@@ -382,6 +382,9 @@ namespace dxvk {
 
 
   bool Presenter::supportsColorSpace(VkColorSpaceKHR colorspace) {
+    if (!m_surface)
+      return false;
+
     std::vector<VkSurfaceFormatKHR> surfaceFormats;
     getSupportedFormats(surfaceFormats, VK_FULL_SCREEN_EXCLUSIVE_DEFAULT_EXT);
 
@@ -645,14 +648,6 @@ namespace dxvk {
   }
 
 
-  void Presenter::applyFrameRateLimit(VkPresentModeKHR mode) {
-    bool vsync = mode == VK_PRESENT_MODE_FIFO_KHR
-              || mode == VK_PRESENT_MODE_FIFO_RELAXED_KHR;
-
-    m_fpsLimiter.delay(vsync);
-  }
-
-
   void Presenter::runFrameThread() {
     env::setThreadName("dxvk-frame");
 
@@ -672,10 +667,6 @@ namespace dxvk {
       if (!frame.frameId)
         return;
 
-      // Apply the FPS limiter before signaling the frame event in
-      // order to reduce latency if the app uses it for frame pacing.
-      applyFrameRateLimit(frame.mode);
-
       // If the present operation has succeeded, actually wait for it to complete.
       // Don't bother with it on MAILBOX / IMMEDIATE modes since doing so would
       // restrict us to the display refresh rate on some platforms (XWayland).
@@ -687,6 +678,11 @@ namespace dxvk {
           Logger::err(str::format("Presenter: vkWaitForPresentKHR failed: ", vr));
       }
 
+      // Apply FPS limtier here to align it as closely with scanout as we can,
+      // and delay signaling the frame latency event to emulate behaviour of a
+      // low refresh rate display as closely as we can.
+      m_fpsLimiter.delay();
+
       // Always signal even on error, since failures here
       // are transparent to the front-end.
       m_signal->signal(frame.frameId);
diff -pruN 2.3.1-1/src/dxvk/dxvk_presenter.h 2.4-1/src/dxvk/dxvk_presenter.h
--- 2.3.1-1/src/dxvk/dxvk_presenter.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_presenter.h	2024-07-10 09:56:14.000000000 +0000
@@ -294,9 +294,6 @@ namespace dxvk {
 
     void destroySurface();
 
-    void applyFrameRateLimit(
-            VkPresentModeKHR          mode);
-
     void runFrameThread();
 
   };
diff -pruN 2.3.1-1/src/dxvk/dxvk_shader.cpp 2.4-1/src/dxvk/dxvk_shader.cpp
--- 2.3.1-1/src/dxvk/dxvk_shader.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_shader.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -59,8 +59,8 @@ namespace dxvk {
 
     if (info.pushConstSize) {
       VkPushConstantRange pushConst;
-      pushConst.stageFlags = info.stage;
-      pushConst.offset = info.pushConstOffset;
+      pushConst.stageFlags = info.pushConstStages;
+      pushConst.offset = 0;
       pushConst.size = info.pushConstSize;
 
       m_bindings.addPushConstantRange(pushConst);
@@ -75,6 +75,8 @@ namespace dxvk {
 
     // Run an analysis pass over the SPIR-V code to gather some
     // info that we may need during pipeline compilation.
+    bool usesPushConstants = false;
+
     std::vector<BindingOffsets> bindingOffsets;
     std::vector<uint32_t> varIds;
     std::vector<uint32_t> sampleMaskIds;
@@ -154,6 +156,9 @@ namespace dxvk {
           if (std::find(sampleMaskIds.begin(), sampleMaskIds.end(), ins.arg(2)) != sampleMaskIds.end())
             m_flags.set(DxvkShaderFlag::ExportsSampleMask);
         }
+
+        if (ins.arg(3) == spv::StorageClassPushConstant)
+          usesPushConstants = true;
       }
 
       // Ignore the actual shader code, there's nothing interesting for us in there.
@@ -169,6 +174,11 @@ namespace dxvk {
         m_bindingOffsets.push_back(info);
     }
 
+    // Set flag for stages that actually use push constants
+    // so that they can be trimmed for optimized pipelines.
+    if (usesPushConstants)
+      m_bindings.addPushConstantStage(info.stage);
+
     // Don't set pipeline library flag if the shader
     // doesn't actually support pipeline libraries
     m_needsLibraryCompile = canUsePipelineLibrary(true);
@@ -1012,7 +1022,7 @@ namespace dxvk {
 
 
   DxvkShaderPipelineLibrary::~DxvkShaderPipelineLibrary() {
-    this->destroyShaderPipelinesLocked();
+    this->destroyShaderPipelineLocked();
   }
 
 
@@ -1032,22 +1042,17 @@ namespace dxvk {
   }
 
 
-  VkPipeline DxvkShaderPipelineLibrary::acquirePipelineHandle(
-    const DxvkShaderPipelineLibraryCompileArgs& args) {
+  DxvkShaderPipelineLibraryHandle DxvkShaderPipelineLibrary::acquirePipelineHandle() {
     std::lock_guard lock(m_mutex);
 
     if (m_device->mustTrackPipelineLifetime())
       m_useCount += 1;
 
-    VkPipeline& pipeline = (m_shaders.vs && !args.depthClipEnable)
-      ? m_pipelineNoDepthClip
-      : m_pipeline;
-
-    if (pipeline)
-      return pipeline;
+    if (m_pipeline.handle)
+      return m_pipeline;
 
-    pipeline = compileShaderPipelineLocked(args);
-    return pipeline;
+    m_pipeline = compileShaderPipelineLocked();
+    return m_pipeline;
   }
 
 
@@ -1056,7 +1061,7 @@ namespace dxvk {
       std::lock_guard lock(m_mutex);
 
       if (!(--m_useCount))
-        this->destroyShaderPipelinesLocked();
+        this->destroyShaderPipelineLocked();
     }
   }
 
@@ -1069,17 +1074,16 @@ namespace dxvk {
       return;
 
     // Compile the pipeline with default args
-    VkPipeline pipeline = compileShaderPipelineLocked(
-      DxvkShaderPipelineLibraryCompileArgs());
+    DxvkShaderPipelineLibraryHandle pipeline = compileShaderPipelineLocked();
 
     // On 32-bit, destroy the pipeline immediately in order to
     // save memory. We should hit the driver's disk cache once
     // we need to recreate the pipeline.
     if (m_device->mustTrackPipelineLifetime()) {
       auto vk = m_device->vkd();
-      vk->vkDestroyPipeline(vk->device(), pipeline, nullptr);
+      vk->vkDestroyPipeline(vk->device(), pipeline.handle, nullptr);
 
-      pipeline = VK_NULL_HANDLE;
+      pipeline.handle = VK_NULL_HANDLE;
     }
 
     // Write back pipeline handle for future use
@@ -1087,37 +1091,32 @@ namespace dxvk {
   }
 
 
-  void DxvkShaderPipelineLibrary::destroyShaderPipelinesLocked() {
+  void DxvkShaderPipelineLibrary::destroyShaderPipelineLocked() {
     auto vk = m_device->vkd();
 
-    vk->vkDestroyPipeline(vk->device(), m_pipeline, nullptr);
-    vk->vkDestroyPipeline(vk->device(), m_pipelineNoDepthClip, nullptr);
+    vk->vkDestroyPipeline(vk->device(), m_pipeline.handle, nullptr);
 
-    m_pipeline = VK_NULL_HANDLE;
-    m_pipelineNoDepthClip = VK_NULL_HANDLE;
+    m_pipeline.handle = VK_NULL_HANDLE;
   }
 
 
-  VkPipeline DxvkShaderPipelineLibrary::compileShaderPipelineLocked(
-    const DxvkShaderPipelineLibraryCompileArgs& args) {
+  DxvkShaderPipelineLibraryHandle DxvkShaderPipelineLibrary::compileShaderPipelineLocked() {
     this->notifyLibraryCompile();
 
     // If this is not the first time we're compiling the pipeline,
     // try to get a cache hit using the shader module identifier
     // so that we don't have to decompress our SPIR-V shader again.
-    VkPipeline pipeline = VK_NULL_HANDLE;
+    DxvkShaderPipelineLibraryHandle pipeline = { VK_NULL_HANDLE, 0 };
 
-    if (m_compiledOnce && canUsePipelineCacheControl()) {
-      pipeline = this->compileShaderPipeline(args,
-        VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT);
-    }
+    if (m_compiledOnce && canUsePipelineCacheControl())
+      pipeline = this->compileShaderPipeline(VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT);
 
-    if (!pipeline)
-      pipeline = this->compileShaderPipeline(args, 0);
+    if (!pipeline.handle)
+      pipeline = this->compileShaderPipeline(0);
 
     // Well that didn't work
-    if (!pipeline)
-      return VK_NULL_HANDLE;
+    if (!pipeline.handle)
+      return { VK_NULL_HANDLE, 0 };
 
     // Increment stat counter the first time this
     // shader pipeline gets compiled successfully
@@ -1134,8 +1133,7 @@ namespace dxvk {
   }
 
 
-  VkPipeline DxvkShaderPipelineLibrary::compileShaderPipeline(
-    const DxvkShaderPipelineLibraryCompileArgs& args,
+  DxvkShaderPipelineLibraryHandle DxvkShaderPipelineLibrary::compileShaderPipeline(
           VkPipelineCreateFlags                 flags) {
     DxvkShaderStageInfo stageInfo(m_device);
     VkShaderStageFlags stageMask = getShaderStages();
@@ -1151,7 +1149,7 @@ namespace dxvk {
           // Fail if we have no idenfitier for whatever reason, caller
           // should fall back to the slow path if this happens
           if (!identifier->identifierSize)
-            return VK_NULL_HANDLE;
+            return { VK_NULL_HANDLE, 0 };
 
           stageInfo.addStage(stage, *identifier, nullptr);
         } else {
@@ -1168,22 +1166,21 @@ namespace dxvk {
       }
     }
 
-    if (stageMask & VK_SHADER_STAGE_VERTEX_BIT)
-      return compileVertexShaderPipeline(args, stageInfo, flags);
-
-    if (stageMask & VK_SHADER_STAGE_FRAGMENT_BIT)
-      return compileFragmentShaderPipeline(stageInfo, flags);
+    VkPipeline pipeline = VK_NULL_HANDLE;
 
-    if (stageMask & VK_SHADER_STAGE_COMPUTE_BIT)
-      return compileComputeShaderPipeline(stageInfo, flags);
+    if (stageMask & VK_SHADER_STAGE_VERTEX_BIT)
+      pipeline = compileVertexShaderPipeline(stageInfo, flags);
+    else if (stageMask & VK_SHADER_STAGE_FRAGMENT_BIT)
+      pipeline = compileFragmentShaderPipeline(stageInfo, flags);
+    else if (stageMask & VK_SHADER_STAGE_COMPUTE_BIT)
+      pipeline = compileComputeShaderPipeline(stageInfo, flags);
 
     // Should be unreachable
-    return VK_NULL_HANDLE;
+    return { pipeline, flags };
   }
 
 
   VkPipeline DxvkShaderPipelineLibrary::compileVertexShaderPipeline(
-    const DxvkShaderPipelineLibraryCompileArgs& args,
     const DxvkShaderStageInfo&          stageInfo,
           VkPipelineCreateFlags         flags) {
     auto vk = m_device->vkd();
@@ -1230,10 +1227,10 @@ namespace dxvk {
       // Only use the fixed depth clip state if we can't make it dynamic
       if (!m_device->features().extExtendedDynamicState3.extendedDynamicState3DepthClipEnable) {
         rsDepthClipInfo.pNext = std::exchange(rsInfo.pNext, &rsDepthClipInfo);
-        rsDepthClipInfo.depthClipEnable = args.depthClipEnable;
+        rsDepthClipInfo.depthClipEnable = VK_TRUE;
       }
     } else {
-      rsInfo.depthClampEnable = !args.depthClipEnable;
+      rsInfo.depthClampEnable = VK_FALSE;
     }
 
     // Only the view mask is used as input, and since we do not use MultiView, it is always 0
@@ -1256,7 +1253,7 @@ namespace dxvk {
     VkPipeline pipeline = VK_NULL_HANDLE;
     VkResult vr = vk->vkCreateGraphicsPipelines(vk->device(), VK_NULL_HANDLE, 1, &info, nullptr, &pipeline);
 
-    if (vr && !(flags & VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT))
+    if (vr && vr != VK_PIPELINE_COMPILE_REQUIRED_EXT)
       Logger::err(str::format("DxvkShaderPipelineLibrary: Failed to create vertex shader pipeline: ", vr));
 
     return vr ? VK_NULL_HANDLE : pipeline;
@@ -1367,7 +1364,7 @@ namespace dxvk {
     VkPipeline pipeline = VK_NULL_HANDLE;
     VkResult vr = vk->vkCreateComputePipelines(vk->device(), VK_NULL_HANDLE, 1, &info, nullptr, &pipeline);
 
-    if (vr && !(flags & VK_PIPELINE_CREATE_FAIL_ON_PIPELINE_COMPILE_REQUIRED_BIT))
+    if (vr && vr != VK_PIPELINE_COMPILE_REQUIRED_EXT)
       Logger::err(str::format("DxvkShaderPipelineLibrary: Failed to create compute shader pipeline: ", vr));
 
     return vr ? VK_NULL_HANDLE : pipeline;
diff -pruN 2.3.1-1/src/dxvk/dxvk_shader.h 2.4-1/src/dxvk/dxvk_shader.h
--- 2.3.1-1/src/dxvk/dxvk_shader.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_shader.h	2024-07-10 09:56:14.000000000 +0000
@@ -52,7 +52,7 @@ namespace dxvk {
     /// Flat shading input mask
     uint32_t flatShadingInputs = 0;
     /// Push constant range
-    uint32_t pushConstOffset = 0;
+    VkShaderStageFlags pushConstStages = 0;
     uint32_t pushConstSize = 0;
     /// Uniform buffer data
     uint32_t uniformSize = 0;
@@ -369,26 +369,6 @@ namespace dxvk {
 
 
   /**
-   * \brief Shader pipeline library compile args
-   */
-  struct DxvkShaderPipelineLibraryCompileArgs {
-    VkBool32 depthClipEnable = VK_TRUE;
-
-    bool operator == (const DxvkShaderPipelineLibraryCompileArgs& other) const {
-      return depthClipEnable == other.depthClipEnable;
-    }
-
-    bool operator != (const DxvkShaderPipelineLibraryCompileArgs& other) const {
-      return !this->operator == (other);
-    }
-
-    size_t hash() const {
-      return size_t(depthClipEnable);
-    }
-  };
-
-
-  /**
    * \brief Shader set
    *
    * Stores a set of shader pointers
@@ -483,6 +463,17 @@ namespace dxvk {
 
 
   /**
+   * \brief Pipeline library handle
+   *
+   * Stores a pipeline library handle and the necessary link flags.
+   */
+  struct DxvkShaderPipelineLibraryHandle {
+    VkPipeline            handle;
+    VkPipelineCreateFlags linkFlags;
+  };
+
+
+  /**
    * \brief Shader pipeline library
    *
    * Stores a pipeline object for either a complete compute
@@ -521,11 +512,9 @@ namespace dxvk {
      * Either returns an already compiled pipeline library object, or
      * performs the compilation step if that has not happened yet.
      * Increments the use count by one.
-     * \param [in] args Compile arguments
      * \returns Vulkan pipeline handle
      */
-    VkPipeline acquirePipelineHandle(
-      const DxvkShaderPipelineLibraryCompileArgs& args);
+    DxvkShaderPipelineLibraryHandle acquirePipelineHandle();
 
     /**
      * \brief Releases pipeline
@@ -552,26 +541,22 @@ namespace dxvk {
           DxvkShaderSet             m_shaders;
     const DxvkBindingLayoutObjects* m_layout;
 
-    dxvk::mutex     m_mutex;
-    VkPipeline      m_pipeline             = VK_NULL_HANDLE;
-    VkPipeline      m_pipelineNoDepthClip  = VK_NULL_HANDLE;
-    uint32_t        m_useCount             = 0u;
-    bool            m_compiledOnce         = false;
-
-    dxvk::mutex                 m_identifierMutex;
-    DxvkShaderIdentifierSet     m_identifiers;
+    dxvk::mutex                     m_mutex;
+    DxvkShaderPipelineLibraryHandle m_pipeline      = { VK_NULL_HANDLE, 0 };
+    uint32_t                        m_useCount      = 0u;
+    bool                            m_compiledOnce  = false;
+
+    dxvk::mutex                     m_identifierMutex;
+    DxvkShaderIdentifierSet         m_identifiers;
 
-    void destroyShaderPipelinesLocked();
+    void destroyShaderPipelineLocked();
 
-    VkPipeline compileShaderPipelineLocked(
-      const DxvkShaderPipelineLibraryCompileArgs& args);
+    DxvkShaderPipelineLibraryHandle compileShaderPipelineLocked();
 
-    VkPipeline compileShaderPipeline(
-      const DxvkShaderPipelineLibraryCompileArgs& args,
+    DxvkShaderPipelineLibraryHandle compileShaderPipeline(
             VkPipelineCreateFlags                 flags);
 
     VkPipeline compileVertexShaderPipeline(
-      const DxvkShaderPipelineLibraryCompileArgs& args,
       const DxvkShaderStageInfo&          stageInfo,
             VkPipelineCreateFlags         flags);
 
diff -pruN 2.3.1-1/src/dxvk/dxvk_swapchain_blitter.cpp 2.4-1/src/dxvk/dxvk_swapchain_blitter.cpp
--- 2.3.1-1/src/dxvk/dxvk_swapchain_blitter.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/dxvk_swapchain_blitter.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -329,6 +329,8 @@ namespace dxvk {
 
     DxvkShaderCreateInfo vsInfo;
     vsInfo.stage = VK_SHADER_STAGE_VERTEX_BIT;
+    vsInfo.pushConstStages = VK_SHADER_STAGE_FRAGMENT_BIT;
+    vsInfo.pushConstSize = sizeof(PresenterArgs);
     vsInfo.outputMask = 0x1;
     m_vs = new DxvkShader(vsInfo, std::move(vsCode));
     
@@ -336,6 +338,7 @@ namespace dxvk {
     fsInfo.stage = VK_SHADER_STAGE_FRAGMENT_BIT;
     fsInfo.bindingCount = fsBindings.size();
     fsInfo.bindings = fsBindings.data();
+    fsInfo.pushConstStages = VK_SHADER_STAGE_FRAGMENT_BIT;
     fsInfo.pushConstSize = sizeof(PresenterArgs);
     fsInfo.inputMask = 0x1;
     fsInfo.outputMask = 0x1;
diff -pruN 2.3.1-1/src/dxvk/hud/dxvk_hud_item.cpp 2.4-1/src/dxvk/hud/dxvk_hud_item.cpp
--- 2.3.1-1/src/dxvk/hud/dxvk_hud_item.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/hud/dxvk_hud_item.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -129,12 +129,8 @@ namespace dxvk::hud {
 
     std::string driverInfo = props.vk12.driverInfo;
 
-    if (driverInfo.empty()) {
-      driverInfo = str::format(
-        VK_VERSION_MAJOR(props.core.properties.driverVersion), ".",
-        VK_VERSION_MINOR(props.core.properties.driverVersion), ".",
-        VK_VERSION_PATCH(props.core.properties.driverVersion));
-    }
+    if (driverInfo.empty())
+      driverInfo = props.driverVersion.toString();
 
     m_deviceName = props.core.properties.deviceName;
     m_driverName = str::format("Driver:  ", props.vk12.driverName);
diff -pruN 2.3.1-1/src/dxvk/hud/dxvk_hud_renderer.cpp 2.4-1/src/dxvk/hud/dxvk_hud_renderer.cpp
--- 2.3.1-1/src/dxvk/hud/dxvk_hud_renderer.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/hud/dxvk_hud_renderer.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -183,14 +183,17 @@ namespace dxvk::hud {
     vsInfo.stage = VK_SHADER_STAGE_VERTEX_BIT;
     vsInfo.bindingCount = vsBindings.size();
     vsInfo.bindings = vsBindings.data();
-    vsInfo.outputMask = 0x3;
+    vsInfo.pushConstStages = VK_SHADER_STAGE_VERTEX_BIT;
     vsInfo.pushConstSize = sizeof(HudTextPushConstants);
+    vsInfo.outputMask = 0x3;
     result.vert = new DxvkShader(vsInfo, std::move(vsCode));
 
     DxvkShaderCreateInfo fsInfo;
     fsInfo.stage = VK_SHADER_STAGE_FRAGMENT_BIT;
     fsInfo.bindingCount = fsBindings.size();
     fsInfo.bindings = fsBindings.data();
+    fsInfo.pushConstStages = VK_SHADER_STAGE_VERTEX_BIT;
+    fsInfo.pushConstSize = sizeof(HudTextPushConstants);
     fsInfo.inputMask = 0x3;
     fsInfo.outputMask = 0x1;
     result.frag = new DxvkShader(fsInfo, std::move(fsCode));
@@ -212,6 +215,7 @@ namespace dxvk::hud {
     DxvkShaderCreateInfo vsInfo;
     vsInfo.stage = VK_SHADER_STAGE_VERTEX_BIT;
     vsInfo.outputMask = 0x1;
+    vsInfo.pushConstStages = VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT;
     vsInfo.pushConstSize = sizeof(HudGraphPushConstants);
     result.vert = new DxvkShader(vsInfo, std::move(vsCode));
     
@@ -221,6 +225,7 @@ namespace dxvk::hud {
     fsInfo.bindings = fsBindings.data();
     fsInfo.inputMask = 0x1;
     fsInfo.outputMask = 0x1;
+    fsInfo.pushConstStages = VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT;
     fsInfo.pushConstSize = sizeof(HudGraphPushConstants);
     result.frag = new DxvkShader(fsInfo, std::move(fsCode));
     
diff -pruN 2.3.1-1/src/dxvk/meson.build 2.4-1/src/dxvk/meson.build
--- 2.3.1-1/src/dxvk/meson.build	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/meson.build	2024-07-10 09:56:14.000000000 +0000
@@ -89,6 +89,7 @@ dxvk_src = [
   'dxvk_options.cpp',
   'dxvk_pipelayout.cpp',
   'dxvk_pipemanager.cpp',
+  'dxvk_platform_exts.cpp',
   'dxvk_presenter.cpp',
   'dxvk_queue.cpp',
   'dxvk_resource.cpp',
@@ -117,20 +118,6 @@ if platform == 'windows'
   ]
 endif
 
-if dxvk_wsi == 'win32'
-  dxvk_src += [
-    'platform/dxvk_win32_exts.cpp'
-  ]
-elif dxvk_wsi == 'sdl2'
-  dxvk_src += [
-    'platform/dxvk_sdl2_exts.cpp'
-  ]
-elif dxvk_wsi == 'glfw'
-  dxvk_src += [
-    'platform/dxvk_glfw_exts.cpp'
-  ]
-endif
-
 dxvk_extra_deps = [ dependency('threads') ]
 if platform == 'linux'
   dxvk_extra_deps += [ cpp.find_library('dl', required: false) ]
diff -pruN 2.3.1-1/src/dxvk/platform/dxvk_glfw_exts.cpp 2.4-1/src/dxvk/platform/dxvk_glfw_exts.cpp
--- 2.3.1-1/src/dxvk/platform/dxvk_glfw_exts.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/platform/dxvk_glfw_exts.cpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,49 +0,0 @@
-#include "../dxvk_platform_exts.h"
-
-#include "../../vulkan/vulkan_loader.h"
-#include <GLFW/glfw3.h>
-
-namespace dxvk {
-
-  DxvkPlatformExts DxvkPlatformExts::s_instance;
-
-  std::string_view DxvkPlatformExts::getName() {
-    return "GLFW WSI";
-  }
-
-  DxvkNameSet DxvkPlatformExts::getInstanceExtensions() {
-    if (!glfwVulkanSupported()) 
-      throw DxvkError(str::format("GLFW WSI: Vulkan is not supported in any capacity!"));
-
-    uint32_t extensionCount = 0;
-    const char** extensionArray = glfwGetRequiredInstanceExtensions(&extensionCount);
-
-    if (extensionCount == 0)
-      throw DxvkError(str::format("GLFW WSI: Failed to get required instance extensions"));
-        
-    DxvkNameSet names;
-    for (uint32_t i = 0; i < extensionCount; ++i) { 
-      names.add(extensionArray[i]);
-    }
-
-    return names;
-  }
-
-
-  DxvkNameSet DxvkPlatformExts::getDeviceExtensions(
-          uint32_t      adapterId) {
-    return DxvkNameSet();
-  }
-
-
-  void DxvkPlatformExts::initInstanceExtensions() {
-    //Nothing needs to be done here on GLFW
-  }
-
-
-  void DxvkPlatformExts::initDeviceExtensions(
-    const DxvkInstance* instance) {
-    //Nothing needs to be done here on GLFW
-  }
-
-}
\ No newline at end of file
diff -pruN 2.3.1-1/src/dxvk/platform/dxvk_sdl2_exts.cpp 2.4-1/src/dxvk/platform/dxvk_sdl2_exts.cpp
--- 2.3.1-1/src/dxvk/platform/dxvk_sdl2_exts.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/platform/dxvk_sdl2_exts.cpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,50 +0,0 @@
-#include "../dxvk_platform_exts.h"
-
-#include <SDL2/SDL.h>
-#include <SDL2/SDL_vulkan.h>
-
-namespace dxvk {
-
-  DxvkPlatformExts DxvkPlatformExts::s_instance;
-
-  std::string_view DxvkPlatformExts::getName() {
-    return "SDL2 WSI";
-  }
-
-
-  DxvkNameSet DxvkPlatformExts::getInstanceExtensions() {
-    SDL_Vulkan_LoadLibrary(nullptr);
-
-    uint32_t extensionCount = 0;
-    if (!SDL_Vulkan_GetInstanceExtensions(nullptr, &extensionCount, nullptr))
-      throw DxvkError(str::format("SDL2 WSI: Failed to get instance extension count. ", SDL_GetError()));
-
-    auto extensionNames = std::vector<const char *>(extensionCount);
-    if (!SDL_Vulkan_GetInstanceExtensions(nullptr, &extensionCount, extensionNames.data()))
-      throw DxvkError(str::format("SDL2 WSI: Failed to get instance extensions. ", SDL_GetError()));
-
-    DxvkNameSet names;
-    for (const char* name : extensionNames)
-      names.add(name);
-
-    return names;
-  }
-
-
-  DxvkNameSet DxvkPlatformExts::getDeviceExtensions(
-          uint32_t      adapterId) {
-    return DxvkNameSet();
-  }
-  
-
-  void DxvkPlatformExts::initInstanceExtensions() {
-
-  }
-
-
-  void DxvkPlatformExts::initDeviceExtensions(
-    const DxvkInstance* instance) {
-
-  }
-
-}
diff -pruN 2.3.1-1/src/dxvk/platform/dxvk_win32_exts.cpp 2.4-1/src/dxvk/platform/dxvk_win32_exts.cpp
--- 2.3.1-1/src/dxvk/platform/dxvk_win32_exts.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/dxvk/platform/dxvk_win32_exts.cpp	1970-01-01 00:00:00.000000000 +0000
@@ -1,36 +0,0 @@
-#include "../dxvk_platform_exts.h"
-
-namespace dxvk {
-
-  DxvkPlatformExts DxvkPlatformExts::s_instance;
-
-  std::string_view DxvkPlatformExts::getName() {
-    return "Win32 WSI";
-  }
-
-
-  DxvkNameSet DxvkPlatformExts::getInstanceExtensions() {
-    DxvkNameSet names;
-    names.add(VK_KHR_WIN32_SURFACE_EXTENSION_NAME);
-
-    return names;
-  }
-
-
-  DxvkNameSet DxvkPlatformExts::getDeviceExtensions(
-          uint32_t      adapterId) {
-    return DxvkNameSet();
-  }
-
-
-  void DxvkPlatformExts::initInstanceExtensions() {
-
-  }
-
-
-  void DxvkPlatformExts::initDeviceExtensions(
-    const DxvkInstance* instance) {
-
-  }
-
-}
\ No newline at end of file
diff -pruN 2.3.1-1/src/meson.build 2.4-1/src/meson.build
--- 2.3.1-1/src/meson.build	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/meson.build	2024-07-10 09:56:14.000000000 +0000
@@ -31,7 +31,14 @@ if get_option('enable_d3d9')
   subdir('d3d9')
 endif
 
+if get_option('enable_d3d8')
+  if not get_option('enable_d3d9')
+    error('D3D9 is required for D3D8.')
+  endif
+  subdir('d3d8')
+endif
+
 # Nothing selected
-if not get_option('enable_d3d9') and not get_option('enable_dxgi')
+if not get_option('enable_d3d8') and not get_option('enable_d3d9') and not get_option('enable_dxgi')
   warning('Nothing selected to be built.?')
 endif
diff -pruN 2.3.1-1/src/util/config/config.cpp 2.4-1/src/util/config/config.cpp
--- 2.3.1-1/src/util/config/config.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/util/config/config.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -13,7 +13,10 @@
 
 namespace dxvk {
 
-  const static std::vector<std::pair<const char*, Config>> g_appDefaults = {{
+  using ProfileList = std::vector<std::pair<const char*, Config>>;
+
+
+  const static ProfileList g_profiles = {{
     /* Assassin's Creed Syndicate: amdags issues  */
     { R"(\\ACS\.exe$)", {{
       { "dxgi.customVendorId",              "10de" },
@@ -122,8 +125,6 @@ namespace dxvk {
     }} },
     /* NieR Replicant                             */
     { R"(\\NieR Replicant ver\.1\.22474487139\.exe)", {{
-      { "dxgi.syncInterval",                "1"    },
-      { "dxgi.maxFrameRate",                "60"   },
       { "d3d11.cachedDynamicResources",     "vi"   },
     }} },
     /* SteamVR performance test                   */
@@ -138,6 +139,10 @@ namespace dxvk {
     { R"(\\h1(_[ms]p64_ship|-mod)\.exe$)", {{
       { "dxgi.customVendorId",              "10de" },
     }} },
+    /* H2M-Mod - Modern Warfare Remastered        */
+    { R"(\\h2m-mod\.exe$)", {{
+      { "dxgi.customVendorId",              "10de" },
+    }} },
     /* Modern Warfare 2 Campaign Remastered       *
      * AMD AGS crash same as above                */
     { R"(\\MW2CR\.exe$)", {{
@@ -331,11 +336,6 @@ namespace dxvk {
     { R"(\\SonicFrontiers\.exe$)", {{
       { "dxgi.maxFrameLatency",             "1" },
     }} },
-    /* TRAHA Global                               *
-     * Shadow issues when it sees AMD/Nvidia      */
-    { R"(\\RapaNui-Win64-Shipping\.exe$)", {{
-      { "dxgi.customVendorId",              "8086" },
-    }} },
     /* SpellForce 3 Reforced & expansions         *
      * Greatly improves CPU bound performance     */
     { R"(\\SF3ClientFinal\.exe$)", {{
@@ -370,6 +370,14 @@ namespace dxvk {
     { R"(\\EADesktop\.exe$)", {{
       { "dxvk.maxChunkSize",                "1"   },
     }} },
+    /* Origin app (legacy EA Desktop)             */
+    { R"(\\Origin\\(Origin|OriginWebHelperService)\.exe$)", {{
+      { "dxvk.maxChunkSize",                "1"   },
+    }} },
+    /* Ubisoft Connect (UPlay)                    */
+    { R"(\\Ubisoft\\Ubisoft Game Launcher\\(UbisoftConnect|upc)\.exe$)", {{
+      { "dxvk.maxChunkSize",                "1"   },
+    }} },
     /* GOG Galaxy                                 */
     { R"(\\GalaxyClient\.exe$)", {{
       { "dxvk.maxChunkSize",                "1"   },
@@ -411,11 +419,6 @@ namespace dxvk {
     { R"(\\RidersRepublic(_BE)?\.exe$)", {{
       { "dxgi.hideAmdGpu",                "True"   },
     }} },
-    /* HoloCure - Save the Fans!
-       Same as Cyberpunk 2077                     */
-    { R"(\\HoloCure\.exe$)", {{
-      { "dxgi.useMonitorFallback",          "True" },
-    }} },
     /* Kenshi                                     *
      * Helps CPU bound performance                */
     { R"(\\kenshi_x64\.exe$)", {{
@@ -429,6 +432,31 @@ namespace dxvk {
       { "d3d11.exposeDriverCommandLists",   "False" },
       { "dxgi.hideNvidiaGpu",               "False" },
     }} },
+    /* Red Faction Guerrilla Re-Mars-tered        *
+     * Broken skybox                              */
+    { R"(\\rfg\.exe$)", {{
+      { "d3d11.longMad",                  "True"    },
+    }} },
+    /* Guild Wars 2 - Fixes invisibility effect   *
+     * flicker when invariantPosition is enabled  */
+    { R"(\\Gw2-64\.exe$)", {{
+      { "d3d11.longMad",                  "True"    },
+    }} },
+    /* Ghostbusters: The Video Game Remastered    *
+     * Flickering on character faces              */
+    { R"(\\ghost\.exe$)", {{
+      { "d3d11.longMad",                  "True"    },
+    }} },
+    /* Watch_Dogs series - Some objects flicker  */
+    { R"(\\watch(_)?dogs(2|Legion)?\.exe$)", {{
+      { "d3d11.longMad",                  "True"    },
+    }} },
+    /* Crysis 1/Warhead - Game bug in d3d10 makes *
+     * it select lowest supported refresh rate    */
+    { R"(\\Crysis(64)?\.exe$)", {{
+      { "d3d9.maxFrameRate",              "-1"      },
+      { "dxgi.maxFrameRate",              "-1"      },
+    }} },
 
     /**********************************************/
     /* D3D9 GAMES                                 */
@@ -474,6 +502,7 @@ namespace dxvk {
       { "d3d9.supportX4R4G4B4",             "False" },
       { "d3d9.maxAvailableMemory",          "2048" },
       { "d3d9.memoryTrackTest",             "True" },
+      { "d3d9.cachedDynamicBuffers",        "True" },
     }} },
     /* Dead Space uses the a NULL render target instead
        of a 1x1 one if DF24 is NOT supported
@@ -484,6 +513,14 @@ namespace dxvk {
       { "d3d9.maxFrameRate",                     "60" },
       { "d3d9.presentInterval",                  "1" },
     }} },
+    /* Dead Space 2
+       Physics issues above 60 FPS
+       Built-in Vsync Locks the game to 30 FPS
+    */
+    { R"(\\deadspace2\.exe$)", {{
+      { "d3d9.maxFrameRate",                     "60" },
+      { "d3d9.presentInterval",                  "1" },
+    }} },
     /* Halo CE/HaloPC                             */
     { R"(\\halo(ce)?\.exe$)", {{
       // Game enables minor decal layering fixes
@@ -519,10 +556,6 @@ namespace dxvk {
     { R"(\\SKShinoviVersus\.exe$)", {{
       { "d3d9.forceAspectRatio",            "16:9" },
     }} },
-    /* Metal Slug X                               */
-    { R"(\\mslugx\.exe$)", {{
-      { "d3d9.supportD32",                  "False" },
-    }} },
     /* Skyrim (NVAPI)                             */
     { R"(\\TESV\.exe$)", {{
       { "d3d9.customVendorId",              "1002" },
@@ -544,15 +577,15 @@ namespace dxvk {
       { "d3d9.supportDFFormats",            "False" },
       { "d3d9.deviceLossOnFocusLoss",       "True" },
     }} },
-    /* Battlefield 2                              *
-     * Bad z-pass and ingame GUI loss on alt tab  */
-    { R"(\\BF2\.exe$)", {{
-      { "d3d9.longMad",                     "True" },  
-      { "d3d9.deviceLossOnFocusLoss",       "True" },
-    }} },
-    /* Battlefield 2142 - Same GUI issue as BF2   */
-    { R"(\\BF2142\.exe$)", {{ 
+    /* Battlefield 2 & Battlefield 2142           *
+     * Bad z-pass and ingame GUI loss on alt tab  *
+     * Also hang when alt tabbing which seems     *
+     * like a game bug that d3d9 drivers work     *
+     * around.                                    */
+    { R"(\\(BF2|BF2142)\.exe$)", {{
+      { "d3d9.longMad",                     "True" },
       { "d3d9.deviceLossOnFocusLoss",       "True" },
+      { "d3d9.countLosableResources",       "False"},
     }} },
     /* SpellForce 2 Series                        */
     { R"(\\SpellForce2.*\.exe$)", {{
@@ -561,10 +594,12 @@ namespace dxvk {
     /* Tomb Raider: Legend, Anniversary, Underworld  *
      * Read from a buffer created with:              *
      * D3DPOOL_DEFAULT,                              *
-     * D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY buffer  */
+     * D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY buffer. *
+     * Legend flickers with next gen content option. */
     { R"(\\(trl|tra|tru)\.exe$)", {{
       { "d3d9.cachedDynamicBuffers",        "True" },
       { "d3d9.maxFrameRate",                "60" },
+      { "d3d9.longMad",                     "True" },
     }} },
     /* Everquest                                 */
     { R"(\\eqgame\.exe$)", {{
@@ -666,10 +701,6 @@ namespace dxvk {
     { R"(\\SWTFU2\.exe$)", {{
       { "d3d9.forceSamplerTypeSpecConstants",  "True" },
     }} },
-    /* Scrapland (Remastered)                   */
-    { R"(\\Scrap\.exe$)", {{
-      { "d3d9.deferSurfaceCreation",        "True" },
-    }} },
     /* Majesty 2 (Collection)                   *
      * Crashes on UMA without a memory limit,   *
      * since the game(s) will allocate all      *
@@ -678,12 +709,14 @@ namespace dxvk {
       { "d3d9.memoryTrackTest",             "True" },
       { "d3d9.maxAvailableMemory",          "2048" },
     }} },
-    /* Myst V End of Ages
-       Game has white textures on amd radv.
-       Expects Nvidia, Intel or ATI VendorId.
-       "Radeon" in gpu description also works   */
+    /* Myst V End of Ages                       *
+     * White textures unless it sees Nvidia,    *
+     * Intel or ATI VendorId.                   *
+     * "Radeon" in gpu description also works.  *
+     * countLosable for resolution change crash.*/
     { R"(\\eoa\.exe$)", {{
       { "d3d9.customVendorId",              "10de" },
+      { "d3d9.countLosableResources",       "False" },
     }} },
     /* Supreme Commander & Forged Alliance Forever */
     { R"(\\(SupremeCommander|ForgedAlliance)\.exe$)", {{
@@ -698,10 +731,6 @@ namespace dxvk {
     { R"(\\bionic_commando\.exe$)", {{
       { "d3d9.maxFrameRate",                "60" },
     }} },
-    /* Need For Speed 3 modern patch            */
-    { R"(\\nfs3\.exe$)", {{
-      { "d3d9.enableDialogMode",            "True" },
-    }} },
     /* Beyond Good And Evil                     *
      * UI breaks at high fps                     */
     { R"(\\BGE\.exe$)", {{
@@ -877,6 +906,56 @@ namespace dxvk {
     { R"(\\ShippingPC-SkyGame\.exe$)", {{
       { "d3d9.maxFrameRate",                "60" },
     }} },
+    /* 9th Dawn II                               *
+     * OpenGL game that also spins up d3d9       *
+     * Black screens without config              */
+    { R"(\\ninthdawnii\.exe$)", {{
+      { "d3d9.deferSurfaceCreation",        "True" },
+    }} },
+    /* Delta Force: Xtreme 1 & 2                 *
+     * Black screen on Alt-Tab and performance   */
+    { R"(\\(DFX|dfx2)\.exe$)", {{
+      { "d3d9.deviceLossOnFocusLoss",       "True" },
+      { "d3d9.cachedDynamicBuffers",        "True" },
+    }} },
+    /* The Sims 3 - Black screen on alt-tab      */
+    { R"(\\TS3(W)?\.exe$)", {{ 
+      { "d3d9.deviceLossOnFocusLoss",       "True" },
+    }} },
+    /* Prototype                                 *
+     * Incorrect shadows on AMD & Intel          */
+    { R"(\\prototypef\.exe$)", {{ 
+      { "d3d9.supportDFFormats",            "False" },
+    }} },
+    /* STAR WARS: The Force Unleashed            *
+     * Prevents black screen on each alt-tab     */
+    { R"(\\SWTFU\.exe$)", {{ 
+      { "d3d9.deviceLossOnFocusLoss",       "True" },
+    }} },
+    /* Fallout New Vegas - Various visual issues *
+     * with mods such as New Vegas Reloaded      */
+    { R"(\\FalloutNV\.exe$)", {{ 
+      { "d3d9.floatEmulation",              "Strict" },
+    }} },
+    /* Dungeons and Dragons: Dragonshard         *
+     * Massive FPS decreases in some scenes      */
+    { R"(\\Dragonshard\.exe$)", {{ 
+      { "d3d9.cachedDynamicBuffers",        "True" },
+    }} },
+    /* Guild Wars 1 - Alt-tab black screen when  *
+     * fullscreen with non native resolution     */
+    { R"(\\Gw\.exe$)", {{ 
+      { "d3d9.deviceLossOnFocusLoss",       "True" },
+    }} },
+    /* Battle for Middle-earth 2 and expansion   *
+     * Slowdowns in certain scenarios            */
+    { R"(\\(The Battle for Middle-earth (\(tm\))? II( Demo)?|The Lord of the Rings, The Rise of the Witch-king)\\game\.dat$)", {{
+      { "d3d9.cachedDynamicBuffers",        "True" },
+    }} },
+    /* WRC4 - Audio breaks above 60fps */
+    { R"(\\WRC4\.exe$)", {{
+      { "d3d9.maxFrameRate",                "60" },
+    }} },
 
 
     /**********************************************/
@@ -901,13 +980,6 @@ namespace dxvk {
     { R"(\\RiftApart\.exe$)", {{
       { "dxgi.hideNvidiaGpu",               "False" },
     }} },
-    /* CP2077 enumerates display outputs each frame.
-     * Avoid using QueryDisplayConfig to avoid
-     * performance degradation until the
-     * optimization of that function is in Proton. */
-    { R"(\\Cyberpunk2077\.exe$)", {{
-      { "dxgi.useMonitorFallback",          "True" },
-    }} },
     /* Metro Exodus Enhanced Edition picks GPU adapters
      * by available VRAM, which causes issues on some
      * systems with integrated graphics. */
@@ -919,9 +991,162 @@ namespace dxvk {
     { R"(\\P3R\.exe$)", {{
       { "dxgi.syncInterval",                "1" },
     }} },
+
+    /**********************************************/
+    /* D3D8 GAMES                                 */
+    /**********************************************/
+
+    /* Duke Nukem Forever (2001)                  */
+    { R"(\\DukeForever\.exe$)", {{
+      { "d3d9.maxFrameRate",                "60"   },
+    }} },
+    /* Indiana Jones and the Emperor's Tomb      *
+     * Fixes intro window being stuck on screen  */
+    { R"(\\indy\.exe$)", {{
+      { "d3d9.enableDialogMode",            "True" },
+    }} },
+    /* Tom Clancy's Splinter Cell                *
+     * Supports shadow buffers                   */
+    { R"(\\splintercell\.exe$)", {{
+      { "d3d8.useShadowBuffers",            "True" },
+    }} },
+    /* Anito: Defend a Land Enraged              */
+    { R"(\\Anito\.exe$)", {{
+      { "d3d9.memoryTrackTest",             "True" },
+      { "d3d9.maxAvailableMemory",          "1024" },
+    }} },
+    /* Red Faction                               *
+     * Fixes crashing when starting a new game   */
+    { R"(\\RF\.exe$)", {{
+      { "d3d9.allowDirectBufferMapping",   "False" },
+    }} },
+    /* Commandos 3                               *
+     * The game doesn't use NOOVERWRITE properly *
+     * and reads from actively modified buffers, *
+     * which causes graphical glitches at times  */
+    { R"(\\Commandos3\.exe$)", {{
+      { "d3d9.allowDirectBufferMapping",   "False" },
+    }} },
+    /* Motor City Online                         */
+    { R"(\\MCity_d\.exe$)", {{
+      { "d3d9.cachedDynamicBuffers",        "True" },
+      { "d3d8.batching",                    "True" },
+    }} },
+    /* Railroad Tycoon 3                         */
+    { R"(\\RT3\.exe$)", {{
+      { "d3d9.maxFrameRate",                  "60" },
+    }} },
+    /* Pure Pinball 2.0 REDUX                    *
+     * This game reads from undeclared vs inputs *
+     * but somehow works on native. Let's just   *
+     * change its declaration to make them work. */
+    { R"(\\Pure Pinball 2\.0 REDUX\.exe$)", {{
+      { "d3d8.forceVsDecl",  "0:2,4:2,7:4,9:1,8:1" },
+    }} },
+    /* Need for Speed III: Hot Pursuit           *
+       (with the "Modern Patch")                 */
+    { R"(\\nfs3\.exe$)", {{
+      { "d3d9.enableDialogMode",            "True" },
+      { "d3d9.cachedDynamicBuffers",        "True" },
+      { "d3d8.batching",                    "True" },
+    }} },
+    /* Need for Speed: High Stakes / Road         *
+       Challenge (with the "Modern Patch") -      *
+       Won't actually render anything in game     *
+       without a memory limit in place            */
+    { R"(\\nfs4\.exe$)", {{
+      { "d3d9.enableDialogMode",            "True" },
+      { "d3d9.cachedDynamicBuffers",        "True" },
+      { "d3d9.memoryTrackTest",             "True" },
+      { "d3d9.maxAvailableMemory",           "256" },
+      { "d3d8.batching",                    "True" },
+    }} },
+    /* Need for Speed: Hot Pursuit 2              */
+    { R"(\\NFSHP2\.exe$)", {{
+      { "d3d9.cachedDynamicBuffers",        "True" },
+    }} },
+    /* Project I.G.I. 2: Covert Strike            */
+    { R"(\\igi2\.exe$)", {{
+      { "d3d9.cachedDynamicBuffers",        "True" },
+    }} },
+    /* Treasure Planet: Battle at Procyon        *
+     * Declares v5 as color but shader uses v6   */
+    { R"(\\TP_Win32\.exe$)", {{
+      { "d3d8.forceVsDecl",      "0:2,3:2,6:4,7:1" },
+    }} },
+    /* Scrapland (Remastered)                   */
+    { R"(\\Scrap\.exe$)", {{
+      { "d3d9.deferSurfaceCreation",        "True" },
+    }} },
+    /* V-Rally 3                                  */
+    { R"(\\VRally3(Demo)?\.exe$)", {{
+      { "d3d9.maxFrameRate",                  "60" },
+    }} },
+    /* Soldiers: Heroes Of World War II           *
+     * Fills up all available memory and hangs    *
+     * while loading the main menu otherwise      */
+    { R"(\\Soldiers\.exe$)", {{
+      { "d3d9.memoryTrackTest",             "True" },
+      { "d3d9.maxAvailableMemory",          "512"  },
+    }} },
+    /* Cossacks II: Napoleonic Wars &             *
+     * Battle for Europe                          */
+    { R"(\\Cossacks II.*\\engine\.exe$)", {{
+      { "d3d9.maxFrameRate",                  "60" },
+    }} },
+    /* Alexander                                  */
+    { R"(\\Alexander\\Data\\engine\.exe$)", {{
+      { "d3d9.maxFrameRate",                  "60" },
+    }} },
+    /* 3DMark2001 (SE)                            *
+     * Fixes a drastic performance drop in the    *
+     * "Car Chase - High Detail" benchmark        */
+    { R"(\\3DMark2001(SE)?\.exe$)", {{
+      { "d3d9.allowDirectBufferMapping",   "False" },
+    }} },
+    /* Delta Force: Black Hawk Down               */
+    { R"(\\dfbhd\.exe$)", {{
+      { "d3d9.cachedDynamicBuffers",        "True" },
+    }} },
+    /* X2: The Threat                             */
+    { R"(\\X2\.exe$)", {{
+      { "d3d9.cachedDynamicBuffers",        "True" },
+    }} },
+    /* The Lord of the Rings:                     *
+     * The Fellowship of the Ring                 */
+    { R"(\\Fellowship\.exe$)", {{
+      { "d3d9.maxFrameRate",                  "60" },
+      { "d3d8.placeP8InScratch",            "True" },
+    }} },
+    /* Rise of Nations + Expansion - alt-tab crash*/
+    { R"(\\(nations|patriots)\.exe$)", {{
+      { "d3d9.deviceLossOnFocusLoss",       "True" },
+    }} },
   }};
 
 
+  const static ProfileList g_deckProfiles = {{
+    /* Fallout 4: Defaults to 45 FPS on OLED, but also breaks above 60 FPS */
+    { R"(\\Fallout4\.exe$)", {{
+      { "dxgi.syncInterval",                "1" },
+      { "dxgi.maxFrameRate",                "60" },
+    }} },
+  }};
+
+
+  const Config* findProfile(const ProfileList& profiles, const std::string& appName) {
+    auto appConfig = std::find_if(profiles.begin(), profiles.end(),
+      [&appName] (const std::pair<const char*, Config>& pair) {
+        std::regex expr(pair.first, std::regex::extended | std::regex::icase);
+        return std::regex_search(appName, expr);
+      });
+
+    return appConfig != profiles.end()
+      ? &appConfig->second
+      : nullptr;
+  }
+
+
   static bool isWhitespace(char ch) {
     return ch == ' ' || ch == '\x9' || ch == '\r';
   }
@@ -1171,20 +1396,22 @@ namespace dxvk {
 
 
   Config Config::getAppConfig(const std::string& appName) {
-    auto appConfig = std::find_if(g_appDefaults.begin(), g_appDefaults.end(),
-      [&appName] (const std::pair<const char*, Config>& pair) {
-        std::regex expr(pair.first, std::regex::extended | std::regex::icase);
-        return std::regex_search(appName, expr);
-      });
+    const Config* config = nullptr;
+
+    if (env::getEnvVar("SteamDeck") == "1")
+      config = findProfile(g_deckProfiles, appName);
+
+    if (!config)
+      config = findProfile(g_profiles, appName);
 
-    if (appConfig != g_appDefaults.end()) {
+    if (config) {
       // Inform the user that we loaded a default config
       Logger::info(str::format("Found built-in config:"));
 
-      for (auto& pair : appConfig->second.m_options)
+      for (auto& pair : config->m_options)
         Logger::info(str::format("  ", pair.first, " = ", pair.second));
 
-      return appConfig->second;
+      return *config;
     }
 
     return Config();
diff -pruN 2.3.1-1/src/util/thread.h 2.4-1/src/util/thread.h
--- 2.3.1-1/src/util/thread.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/util/thread.h	2024-07-10 09:56:14.000000000 +0000
@@ -316,7 +316,11 @@ namespace dxvk {
       switch (priority) {
         default:
         case ThreadPriority::Normal: policy = SCHED_OTHER; break;
+#ifndef __linux__
+        case ThreadPriority::Lowest: policy = SCHED_OTHER; break;
+#else
         case ThreadPriority::Lowest: policy = SCHED_IDLE;  break;
+#endif
       }
       ::pthread_setschedparam(this->native_handle(), policy, &param);
     }
diff -pruN 2.3.1-1/src/util/util_env.cpp 2.4-1/src/util/util_env.cpp
--- 2.3.1-1/src/util/util_env.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/util/util_env.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -6,6 +6,10 @@
 #ifdef __linux__
 #include <unistd.h>
 #include <limits.h>
+#elif defined(__FreeBSD__)
+#include <sys/sysctl.h>
+#include <unistd.h>
+#include <limits.h>
 #endif
 
 #include "util_env.h"
@@ -85,6 +89,17 @@ namespace dxvk::env {
     size_t count = readlink("/proc/self/exe", exePath.data(), exePath.size());
 
     return std::string(exePath.begin(), exePath.begin() + count);
+#elif defined(__FreeBSD__)
+    int mib[4] = {CTL_KERN, KERN_PROC, KERN_PROC_PATHNAME, getpid()};
+    char exePath[PATH_MAX] = {};
+    size_t size = PATH_MAX;
+
+    if (sysctl(mib, 4, exePath, &size, NULL, 0) != 0) {
+        // throw error here?
+        return "";
+    }
+
+    return std::string(exePath);
 #endif
   }
   
diff -pruN 2.3.1-1/src/util/util_flush.cpp 2.4-1/src/util/util_flush.cpp
--- 2.3.1-1/src/util/util_flush.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/util/util_flush.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -2,6 +2,12 @@
 
 namespace dxvk {
 
+  GpuFlushTracker::GpuFlushTracker(
+          bool ensureReproducibleHeuristic)
+  : m_ensureReproducibleHeuristic(ensureReproducibleHeuristic) {
+
+  }
+
   bool GpuFlushTracker::considerFlush(
           GpuFlushType          flushType,
           uint64_t              chunkId,
@@ -17,6 +23,9 @@ namespace dxvk {
     if (!chunkCount)
       return false;
 
+    if (m_ensureReproducibleHeuristic && flushType != GpuFlushType::ExplicitFlush)
+      return false;
+
     // Take any earlier missed flush with a stronger hint into account, so
     // that we still flush those as soon as possible. Ignore synchronization
     // commands since they will either perform a flush or not need it at all.
diff -pruN 2.3.1-1/src/util/util_flush.h 2.4-1/src/util/util_flush.h
--- 2.3.1-1/src/util/util_flush.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/util/util_flush.h	2024-07-10 09:56:14.000000000 +0000
@@ -34,6 +34,8 @@ namespace dxvk {
 
   public:
 
+    GpuFlushTracker(bool ensureReproducibleHeuristic);
+
     /**
      * \brief Checks whether a context flush should be performed
      *
@@ -61,6 +63,8 @@ namespace dxvk {
 
   private:
 
+    bool          m_ensureReproducibleHeuristic;
+
     GpuFlushType  m_lastMissedType        = GpuFlushType::ImplicitWeakHint;
 
     uint64_t      m_lastFlushChunkId      = 0ull;
diff -pruN 2.3.1-1/src/util/util_fps_limiter.cpp 2.4-1/src/util/util_fps_limiter.cpp
--- 2.3.1-1/src/util/util_fps_limiter.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/util/util_fps_limiter.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -35,52 +35,83 @@ namespace dxvk {
     std::lock_guard<dxvk::mutex> lock(m_mutex);
 
     if (!m_envOverride) {
-      m_targetInterval = frameRate > 0.0
+      TimerDuration interval = frameRate != 0.0
         ? TimerDuration(int64_t(double(TimerDuration::period::den) / frameRate))
         : TimerDuration::zero();
 
-      if (isEnabled() && !m_initialized)
-        initialize();
+      if (m_targetInterval != interval) {
+        m_targetInterval = interval;
+
+        m_heuristicFrameCount = 0;
+        m_heuristicEnable = false;
+      }
     }
   }
 
 
-  void FpsLimiter::delay(bool vsyncEnabled) {
-    std::lock_guard<dxvk::mutex> lock(m_mutex);
+  void FpsLimiter::delay() {
+    std::unique_lock<dxvk::mutex> lock(m_mutex);
+    auto interval = m_targetInterval;
 
-    if (!isEnabled())
+    if (interval == TimerDuration::zero()) {
+      m_nextFrame = TimePoint();
       return;
+    }
 
-    auto t0 = m_lastFrame;
     auto t1 = dxvk::high_resolution_clock::now();
 
-    auto frameTime = std::chrono::duration_cast<TimerDuration>(t1 - t0);
+    if (interval < TimerDuration::zero()) {
+      interval = -interval;
 
-    if (frameTime * 100 > m_targetInterval * 103 - m_deviation * 100) {
-      // If we have a slow frame, reset the deviation since we
-      // do not want to compensate for low performance later on
-      m_deviation = TimerDuration::zero();
-    } else {
-      // Don't call sleep if the amount of time to sleep is shorter
-      // than the time the function calls are likely going to take
-      TimerDuration sleepDuration = m_targetInterval - m_deviation - frameTime;
-      t1 = Sleep::sleepFor(t1, sleepDuration);
-
-      // Compensate for any sleep inaccuracies in the next frame, and
-      // limit cumulative deviation in order to avoid stutter in case we
-      // have a number of slow frames immediately followed by a fast one.
-      frameTime = std::chrono::duration_cast<TimerDuration>(t1 - t0);
-      m_deviation += frameTime - m_targetInterval;
-      m_deviation = std::min(m_deviation, m_targetInterval / 16);
+      if (!testRefreshHeuristic(interval, t1))
+        return;
     }
 
-    m_lastFrame = t1;
+    // Subsequent code must not access any class members
+    // that can be written by setTargetFrameRate
+    lock.unlock();
+
+    if (t1 < m_nextFrame)
+      Sleep::sleepUntil(t1, m_nextFrame);
+
+    m_nextFrame = (t1 < m_nextFrame + interval)
+      ? m_nextFrame + interval
+      : t1 + interval;
   }
 
 
-  void FpsLimiter::initialize() {
-    m_lastFrame = dxvk::high_resolution_clock::now();
-    m_initialized = true;
+  bool FpsLimiter::testRefreshHeuristic(TimerDuration interval, TimePoint now) {
+    if (m_heuristicEnable)
+      return true;
+
+    // Use a sliding window to determine whether the current
+    // frame rate is higher than the targeted refresh rate
+    uint32_t heuristicWindow = m_heuristicFrameTimes.size();
+    auto windowStart = m_heuristicFrameTimes[m_heuristicFrameCount % heuristicWindow];
+    auto windowDuration = std::chrono::duration_cast<TimerDuration>(now - windowStart);
+
+    m_heuristicFrameTimes[m_heuristicFrameCount % heuristicWindow] = now;
+    m_heuristicFrameCount += 1;
+
+    // The first window of frames may contain faster frames as the
+    // internal swap chain queue fills up, so we should ignore it.
+    if (m_heuristicFrameCount < 2 * heuristicWindow)
+      return false;
+
+    // Test whether we should engage the frame rate limiter. It will
+    // stay enabled until the refresh rate or vsync enablement change.
+    m_heuristicEnable = (103 * windowDuration) < (100 * heuristicWindow) * interval;
+
+    if (m_heuristicEnable) {
+      double got = (double(heuristicWindow) * double(TimerDuration::period::den))
+                 / (double(windowDuration.count()) * double(TimerDuration::period::num));
+      double refresh = double(TimerDuration::period::den) / (double(TimerDuration::period::num) * double(interval.count()));
+
+      Logger::info(str::format("Detected frame rate (~", uint32_t(got), ") higher than selected refresh rate of ~",
+        uint32_t(refresh), " Hz.\n", "Engaging frame rate limiter."));
+    }
+
+    return m_heuristicEnable;
   }
 
 }
diff -pruN 2.3.1-1/src/util/util_fps_limiter.h 2.4-1/src/util/util_fps_limiter.h
--- 2.3.1-1/src/util/util_fps_limiter.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/util/util_fps_limiter.h	2024-07-10 09:56:14.000000000 +0000
@@ -1,5 +1,7 @@
 #pragma once
 
+#include <array>
+
 #include "thread.h"
 #include "util_time.h"
 
@@ -34,17 +36,8 @@ namespace dxvk {
      * Blocks the calling thread if the limiter is enabled
      * and the time since the last call to \ref delay is
      * shorter than the target interval.
-     * \param [in] vsyncEnabled \c true if vsync is enabled
      */
-    void delay(bool vsyncEnabled);
-
-    /**
-     * \brief Checks whether the frame rate limiter is enabled
-     * \returns \c true if the target frame rate is non-zero.
-     */
-    bool isEnabled() const {
-      return m_targetInterval != TimerDuration::zero();
-    }
+    void delay();
 
   private:
 
@@ -54,13 +47,15 @@ namespace dxvk {
     dxvk::mutex     m_mutex;
 
     TimerDuration   m_targetInterval  = TimerDuration::zero();
-    TimerDuration   m_deviation       = TimerDuration::zero();
-    TimePoint       m_lastFrame;
+    TimePoint       m_nextFrame       = TimePoint();
 
-    bool            m_initialized     = false;
     bool            m_envOverride     = false;
 
-    void initialize();
+    uint32_t                  m_heuristicFrameCount = 0;
+    std::array<TimePoint, 16> m_heuristicFrameTimes = { };
+    bool                      m_heuristicEnable = false;
+
+    bool testRefreshHeuristic(TimerDuration interval, TimePoint now);
 
   };
 
diff -pruN 2.3.1-1/src/util/util_matrix.cpp 2.4-1/src/util/util_matrix.cpp
--- 2.3.1-1/src/util/util_matrix.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/util/util_matrix.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -205,6 +205,10 @@ namespace dxvk {
     Vector4 dot0    = { m[0] * row0 };
     float dot1      = (dot0.x + dot0.y) + (dot0.z + dot0.w);
 
+    if (unlikely(std::abs(dot1) <= 0.000001f)) {
+      return m;
+    }
+
     return inverse * (1.0f / dot1);
   }
 
diff -pruN 2.3.1-1/src/util/util_small_vector.h 2.4-1/src/util/util_small_vector.h
--- 2.3.1-1/src/util/util_small_vector.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/util/util_small_vector.h	2024-07-10 09:56:14.000000000 +0000
@@ -13,8 +13,37 @@ namespace dxvk {
 
     small_vector() { }
 
-    small_vector             (const small_vector&) = delete;
-    small_vector& operator = (const small_vector&) = delete;
+    small_vector(const small_vector& other) {
+      reserve(other.m_size);
+      for (size_t i = 0; i < other.m_size; i++) {
+        *ptr(i) = *other.ptr(i);
+      }
+      m_size = other.m_size;
+    };
+
+    small_vector& operator = (const small_vector& other) {
+      for (size_t i = 0; i < m_size; i++)
+        ptr(i)->~T();
+
+      reserve(other.m_size);
+      for (size_t i = 0; i < other.m_size; i++) {
+        *ptr(i) = *other.ptr(i);
+      }
+      m_size = other.m_size;
+    };
+
+    small_vector(small_vector&& other) {
+      if (other.m_capacity == N) {
+        for (size_t i = 0; i < other.m_size; i++) {
+          u.m_data[i] = std::move(other.u.m_data[i]);
+        }
+      } else {
+        u.m_ptr = other.u.m_ptr;
+        other.m_capacity = N;
+      }
+      m_size = other.m_size;
+      other.m_size = 0;
+    }
 
     ~small_vector() {
       for (size_t i = 0; i < m_size; i++)
@@ -23,7 +52,7 @@ namespace dxvk {
       if (m_capacity > N)
         delete[] u.m_ptr;
     }
-    
+
     size_t size() const {
       return m_size;
     }
@@ -43,7 +72,7 @@ namespace dxvk {
 
       if (m_capacity > N)
         delete[] u.m_ptr;
-      
+
       m_capacity = n;
       u.m_ptr = data;
     }
@@ -56,7 +85,7 @@ namespace dxvk {
 
       for (size_t i = n; i < m_size; i++)
         ptr(i)->~T();
-      
+
       for (size_t i = m_size; i < n; i++)
         new (ptr(i)) T();
 
diff -pruN 2.3.1-1/src/util/util_version.h 2.4-1/src/util/util_version.h
--- 2.3.1-1/src/util/util_version.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/util/util_version.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,48 @@
+#pragma once
+
+#include <cstdint>
+
+#include "../vulkan/vulkan_loader.h"
+
+#include "util_string.h"
+
+namespace dxvk {
+
+  /**
+   * \brief Decoded driver version
+   */
+  class Version {
+
+  public:
+
+    Version() = default;
+
+    Version(uint32_t major, uint32_t minor, uint32_t patch)
+    : m_raw((uint64_t(major) << 48) | (uint64_t(minor) << 24) | uint64_t(patch)) { }
+
+    uint32_t major() const { return uint32_t(m_raw >> 48); }
+    uint32_t minor() const { return uint32_t(m_raw >> 24) & 0xffffffu; }
+    uint32_t patch() const { return uint32_t(m_raw) & 0xffffffu; }
+
+    bool operator == (const Version& other) const { return m_raw == other.m_raw; }
+    bool operator != (const Version& other) const { return m_raw != other.m_raw; }
+    bool operator >= (const Version& other) const { return m_raw >= other.m_raw; }
+    bool operator <= (const Version& other) const { return m_raw <= other.m_raw; }
+    bool operator >  (const Version& other) const { return m_raw >  other.m_raw; }
+    bool operator <  (const Version& other) const { return m_raw <  other.m_raw; }
+
+    std::string toString() const {
+      return str::format(major(), ".", minor(), ".", patch());
+    }
+
+    explicit operator bool () const {
+      return m_raw != 0;
+    }
+
+  private:
+
+    uint64_t m_raw = 0;
+
+  };
+
+}
diff -pruN 2.3.1-1/src/util/util_win32_compat.h 2.4-1/src/util/util_win32_compat.h
--- 2.3.1-1/src/util/util_win32_compat.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/util/util_win32_compat.h	2024-07-10 09:56:14.000000000 +0000
@@ -1,6 +1,6 @@
 #pragma once
 
-#if defined(__linux__)
+#if defined(__unix__)
 
 #include <windows.h>
 #include <dlfcn.h>
diff -pruN 2.3.1-1/src/vulkan/vulkan_util.h 2.4-1/src/vulkan/vulkan_util.h
--- 2.3.1-1/src/vulkan/vulkan_util.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/vulkan/vulkan_util.h	2024-07-10 09:56:14.000000000 +0000
@@ -4,6 +4,11 @@
 
 #include "vulkan_loader.h"
 
+#if defined(_MSC_VER)
+// Unary minus on unsigned type
+#pragma warning( disable : 4146 )
+#endif
+
 namespace dxvk::vk {
 
   inline VkImageSubresourceRange makeSubresourceRange(
diff -pruN 2.3.1-1/src/wsi/glfw/wsi_monitor_glfw.cpp 2.4-1/src/wsi/glfw/wsi_monitor_glfw.cpp
--- 2.3.1-1/src/wsi/glfw/wsi_monitor_glfw.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/wsi/glfw/wsi_monitor_glfw.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -1,3 +1,5 @@
+#if defined(DXVK_WSI_GLFW)
+
 #include "../wsi_monitor.h"
 
 #include "wsi/native_wsi.h"
@@ -11,22 +13,22 @@
 
 namespace dxvk::wsi {
 
-  HMONITOR getDefaultMonitor() {
+  HMONITOR GlfwWsiDriver::getDefaultMonitor() {
     return enumMonitors(0);
   }
 
 
-  HMONITOR enumMonitors(uint32_t index) {
+  HMONITOR GlfwWsiDriver::enumMonitors(uint32_t index) {
     return isDisplayValid(int32_t(index))
          ? toHmonitor(index)
          : nullptr;
   }
 
-  HMONITOR enumMonitors(const LUID *adapterLUID[], uint32_t numLUIDs, uint32_t index) {
+  HMONITOR GlfwWsiDriver::enumMonitors(const LUID *adapterLUID[], uint32_t numLUIDs, uint32_t index) {
     return enumMonitors(index);
   }
 
-  bool getDisplayName(
+  bool GlfwWsiDriver::getDisplayName(
       HMONITOR hMonitor,
       WCHAR            (&Name)[32]) {
     const int32_t displayId = fromHmonitor(hMonitor);
@@ -46,7 +48,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool getDesktopCoordinates(
+  bool GlfwWsiDriver::getDesktopCoordinates(
       HMONITOR hMonitor,
       RECT* pRect) {
     const int32_t displayId = fromHmonitor(hMonitor);
@@ -97,7 +99,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool getDisplayMode(
+  bool GlfwWsiDriver::getDisplayMode(
       HMONITOR hMonitor,
       uint32_t ModeNumber,
       WsiMode* pMode) {
@@ -121,7 +123,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool getCurrentDisplayMode(
+  bool GlfwWsiDriver::getCurrentDisplayMode(
       HMONITOR hMonitor,
       WsiMode* pMode) {
     const int32_t displayId = fromHmonitor(hMonitor);
@@ -141,7 +143,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool getDesktopDisplayMode(
+  bool GlfwWsiDriver::getDesktopDisplayMode(
       HMONITOR hMonitor,
       WsiMode* pMode) {
     const int32_t displayId = fromHmonitor(hMonitor);
@@ -159,9 +161,11 @@ namespace dxvk::wsi {
     return true;
   }
 
-  std::vector<uint8_t> getMonitorEdid(HMONITOR hMonitor) {
+  std::vector<uint8_t> GlfwWsiDriver::getMonitorEdid(HMONITOR hMonitor) {
     Logger::err("getMonitorEdid not implemented on this platform.");
     return {};
   }
 
 }
+
+#endif
diff -pruN 2.3.1-1/src/wsi/glfw/wsi_platform_glfw.cpp 2.4-1/src/wsi/glfw/wsi_platform_glfw.cpp
--- 2.3.1-1/src/wsi/glfw/wsi_platform_glfw.cpp	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/wsi/glfw/wsi_platform_glfw.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,71 @@
+#if defined(DXVK_WSI_GLFW)
+
+#include "wsi_platform_glfw.h"
+#include "../../util/util_error.h"
+#include "../../util/util_string.h"
+#include "../../util/util_win32_compat.h"
+
+namespace dxvk::wsi {
+
+  GlfwWsiDriver::GlfwWsiDriver() {
+    libglfw = LoadLibraryA( // FIXME: Get soname as string from meson
+#if defined(_WIN32)
+        "glfw.dll"
+#elif defined(__APPLE__)
+        "libglfw.3.dylib"
+#else
+        "libglfw.so.3"
+#endif
+      );
+    if (libglfw == nullptr)
+      throw DxvkError("GLFW WSI: Failed to load GLFW DLL.");
+
+    #define GLFW_PROC(ret, name, params) \
+      name = reinterpret_cast<pfn_##name>(GetProcAddress(libglfw, #name)); \
+      if (name == nullptr) { \
+        FreeLibrary(libglfw); \
+        libglfw = nullptr; \
+        throw DxvkError("GLFW WSI: Failed to load " #name "."); \
+      }
+    #include "wsi_platform_glfw_funcs.h"
+  }
+
+  GlfwWsiDriver::~GlfwWsiDriver() {
+    FreeLibrary(libglfw);
+  }
+
+  std::vector<const char *> GlfwWsiDriver::getInstanceExtensions() {
+    if (!glfwVulkanSupported())
+      throw DxvkError(str::format("GLFW WSI: Vulkan is not supported in any capacity!"));
+
+    uint32_t extensionCount = 0;
+    const char** extensionArray = glfwGetRequiredInstanceExtensions(&extensionCount);
+
+    if (extensionCount == 0)
+      throw DxvkError(str::format("GLFW WSI: Failed to get required instance extensions"));
+
+    std::vector<const char *> names;
+    for (uint32_t i = 0; i < extensionCount; ++i) {
+      names.push_back(extensionArray[i]);
+    }
+
+    return names;
+  }
+
+  static bool createGlfwWsiDriver(WsiDriver **driver) {
+    try {
+      *driver = new GlfwWsiDriver();
+    } catch (const DxvkError& e) {
+      return false;
+    }
+    return true;
+  }
+
+  WsiBootstrap GlfwWSI = {
+    "GLFW",
+    createGlfwWsiDriver
+  };
+
+}
+
+#endif
diff -pruN 2.3.1-1/src/wsi/glfw/wsi_platform_glfw.h 2.4-1/src/wsi/glfw/wsi_platform_glfw.h
--- 2.3.1-1/src/wsi/glfw/wsi_platform_glfw.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/wsi/glfw/wsi_platform_glfw.h	2024-07-10 09:56:14.000000000 +0000
@@ -3,21 +3,108 @@
 #include "../../vulkan/vulkan_loader.h"
 #include <GLFW/glfw3.h>
 
-#include "../wsi_monitor.h"
+#include "../wsi_platform.h"
 
 namespace dxvk::wsi {
 
-  /**
-    * \brief Impl-dependent state
-    */
-  struct DxvkWindowState {
-  };
-  
-  inline bool isDisplayValid(int32_t displayId) {
-    int32_t displayCount = 0;
-    glfwGetMonitors(&displayCount);
+  class GlfwWsiDriver : public WsiDriver {
+  private:
+    HMODULE libglfw;
+    #define GLFW_PROC(ret, name, params) \
+      typedef ret (*pfn_##name) params; \
+      pfn_##name name;
+    #include "wsi_platform_glfw_funcs.h"
+
+    inline bool isDisplayValid(int32_t displayId) {
+      int32_t displayCount = 0;
+      glfwGetMonitors(&displayCount);
+
+      return displayId < displayCount && displayId >= 0;
+    }
+
+  public:
+    GlfwWsiDriver();
+    ~GlfwWsiDriver();
+
+    // Platform
+    virtual std::vector<const char *> getInstanceExtensions();
+
+    // Monitor
+    virtual HMONITOR getDefaultMonitor();
+
+    virtual HMONITOR enumMonitors(uint32_t index);
+
+    virtual HMONITOR enumMonitors(const LUID *adapterLUID[], uint32_t numLUIDs, uint32_t index);
+
+    virtual bool getDisplayName(
+            HMONITOR         hMonitor,
+            WCHAR            (&Name)[32]);
+
+    virtual bool getDesktopCoordinates(
+            HMONITOR         hMonitor,
+            RECT*            pRect);
+
+    virtual bool getDisplayMode(
+            HMONITOR         hMonitor,
+            uint32_t         modeNumber,
+            WsiMode*         pMode);
+
+    virtual bool getCurrentDisplayMode(
+            HMONITOR         hMonitor,
+            WsiMode*         pMode);
+
+    virtual bool getDesktopDisplayMode(
+            HMONITOR         hMonitor,
+            WsiMode*         pMode);
 
-    return displayId < displayCount && displayId >= 0;
-  }
+    virtual WsiEdidData getMonitorEdid(HMONITOR hMonitor);
+
+    // Window
+
+    virtual void getWindowSize(
+            HWND      hWindow,
+            uint32_t* pWidth,
+            uint32_t* pWeight);
+
+    virtual void resizeWindow(
+            HWND             hWindow,
+            DxvkWindowState* pState,
+            uint32_t         width,
+            uint32_t         weight);
+
+    virtual bool setWindowMode(
+            HMONITOR         hMonitor,
+            HWND             hWindow,
+      const WsiMode&         mode);
+
+    virtual bool enterFullscreenMode(
+            HMONITOR         hMonitor,
+            HWND             hWindow,
+            DxvkWindowState* pState,
+            [[maybe_unused]]
+            bool             modeSwitch);
+
+    virtual bool leaveFullscreenMode(
+            HWND             hWindow,
+            DxvkWindowState* pState,
+            bool             restoreCoordinates);
+
+    virtual bool restoreDisplayMode();
+
+    virtual HMONITOR getWindowMonitor(HWND hWindow);
+
+    virtual bool isWindow(HWND hWindow);
+
+    virtual void updateFullscreenWindow(
+            HMONITOR hMonitor,
+            HWND     hWindow,
+            bool     forceTopmost);
+
+    virtual VkResult createSurface(
+            HWND                hWindow,
+            PFN_vkGetInstanceProcAddr pfnVkGetInstanceProcAddr,
+            VkInstance          instance,
+            VkSurfaceKHR*       pSurface);
+  };
 
-}
\ No newline at end of file
+}
diff -pruN 2.3.1-1/src/wsi/glfw/wsi_platform_glfw_funcs.h 2.4-1/src/wsi/glfw/wsi_platform_glfw_funcs.h
--- 2.3.1-1/src/wsi/glfw/wsi_platform_glfw_funcs.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/wsi/glfw/wsi_platform_glfw_funcs.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,12 @@
+GLFW_PROC(VkResult, glfwCreateWindowSurface, (VkInstance, GLFWwindow*, const VkAllocationCallbacks*, VkSurfaceKHR*))
+GLFW_PROC(GLFWmonitor**, glfwGetMonitors, (int*))
+GLFW_PROC(void, glfwGetMonitorWorkarea, (GLFWmonitor*, int*, int*, int*, int*))
+GLFW_PROC(GLFWmonitor*, glfwGetPrimaryMonitor, (void))
+GLFW_PROC(const char**, glfwGetRequiredInstanceExtensions, (uint32_t*))
+GLFW_PROC(const GLFWvidmode*, glfwGetVideoMode, (GLFWmonitor*))
+GLFW_PROC(const GLFWvidmode*, glfwGetVideoModes, (GLFWmonitor*, int*))
+GLFW_PROC(void, glfwGetWindowSize, (GLFWwindow*, int*, int*))
+GLFW_PROC(void, glfwSetWindowMonitor, (GLFWwindow*, GLFWmonitor*, int, int, int, int, int))
+GLFW_PROC(void, glfwSetWindowSize, (GLFWwindow*, int, int))
+GLFW_PROC(int, glfwVulkanSupported, (void))
+#undef GLFW_PROC
diff -pruN 2.3.1-1/src/wsi/glfw/wsi_window_glfw.cpp 2.4-1/src/wsi/glfw/wsi_window_glfw.cpp
--- 2.3.1-1/src/wsi/glfw/wsi_window_glfw.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/wsi/glfw/wsi_window_glfw.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -1,6 +1,8 @@
+#if defined(DXVK_WSI_GLFW)
+
 #include "../wsi_window.h"
 
-#include "native/wsi/native_wsi.h"
+#include "native/wsi/native_glfw.h"
 #include "wsi_platform_glfw.h"
 
 #include "../../util/util_string.h"
@@ -12,7 +14,7 @@
 
 namespace dxvk::wsi {
 
-  void getWindowSize(
+  void GlfwWsiDriver::getWindowSize(
       HWND hWindow,
       uint32_t* pWidth,
       uint32_t* pHeight) {
@@ -29,7 +31,7 @@ namespace dxvk::wsi {
   }
 
 
-  void resizeWindow(
+  void GlfwWsiDriver::resizeWindow(
       HWND hWindow,
       DxvkWindowState* pState,
       uint32_t Width,
@@ -40,7 +42,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool setWindowMode(
+  bool GlfwWsiDriver::setWindowMode(
       HMONITOR hMonitor,
       HWND hWindow,
       const WsiMode& pMode) {
@@ -67,7 +69,7 @@ namespace dxvk::wsi {
     return true;
   }
 
-  bool enterFullscreenMode(
+  bool GlfwWsiDriver::enterFullscreenMode(
       HMONITOR hMonitor,
       HWND hWindow,
       DxvkWindowState* pState,
@@ -89,7 +91,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool leaveFullscreenMode(
+  bool GlfwWsiDriver::leaveFullscreenMode(
       HWND hWindow,
       DxvkWindowState* pState,
       bool             restoreCoordinates) {
@@ -103,13 +105,13 @@ namespace dxvk::wsi {
   }
 
 
-  bool restoreDisplayMode() {
+  bool GlfwWsiDriver::restoreDisplayMode() {
     // Don't need to do anything with GLFW here.
     return true;
   }
 
 
-  HMONITOR getWindowMonitor(HWND hWindow) {
+  HMONITOR GlfwWsiDriver::getWindowMonitor(HWND hWindow) {
     // TODO: implement this with glfwGetWindowMonitor 
     //  (or maybe not? glfwGetWindowMonitor only seems to reference *fullscreen* windows)
     // GLFWwindow* window = fromHwnd(hWindow);
@@ -119,19 +121,19 @@ namespace dxvk::wsi {
   }
 
 
-  bool isWindow(HWND hWindow) {
+  bool GlfwWsiDriver::isWindow(HWND hWindow) {
     GLFWwindow* window = fromHwnd(hWindow);
     return window != nullptr;
   }
 
-  void updateFullscreenWindow(
+  void GlfwWsiDriver::updateFullscreenWindow(
       HMONITOR hMonitor,
       HWND     hWindow,
       bool     forceTopmost) {
     // Don't need to do anything with GLFW here.
   }
 
-  VkResult createSurface(
+  VkResult GlfwWsiDriver::createSurface(
       HWND hWindow,
       PFN_vkGetInstanceProcAddr pfnVkGetInstanceProcAddr,
       VkInstance                instance,
@@ -141,4 +143,6 @@ namespace dxvk::wsi {
     return glfwCreateWindowSurface(instance, window, nullptr, pSurface);
   }
 
-}
\ No newline at end of file
+}
+
+#endif
diff -pruN 2.3.1-1/src/wsi/meson.build 2.4-1/src/wsi/meson.build
--- 2.3.1-1/src/wsi/meson.build	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/wsi/meson.build	2024-07-10 09:56:14.000000000 +0000
@@ -1,33 +1,24 @@
-wsi_common_src = [
+wsi_src = [
   'wsi_edid.cpp',
-]
-
-wsi_win32_src = [
+  'wsi_platform.cpp',
   'win32/wsi_monitor_win32.cpp',
+  'win32/wsi_platform_win32.cpp',
   'win32/wsi_window_win32.cpp',
-]
-
-wsi_sdl2_src = [
   'sdl2/wsi_monitor_sdl2.cpp',
+  'sdl2/wsi_platform_sdl2.cpp',
   'sdl2/wsi_window_sdl2.cpp',
-]
-
-wsi_glfw_src = [
   'glfw/wsi_monitor_glfw.cpp',
+  'glfw/wsi_platform_glfw.cpp',
   'glfw/wsi_window_glfw.cpp',
 ]
 
-if dxvk_wsi == 'win32'
-  wsi_src  = wsi_common_src + wsi_win32_src
-  wsi_deps = [ dep_displayinfo ]
-elif dxvk_wsi == 'sdl2'
-  wsi_src  = wsi_common_src + wsi_sdl2_src
-  wsi_deps = [ dep_displayinfo, lib_sdl2 ]
-elif dxvk_wsi == 'glfw'
-  wsi_src  = wsi_common_src + wsi_glfw_src
-  wsi_deps = [ dep_displayinfo, lib_glfw ]
-else
-  error('Unknown wsi')
+wsi_deps = [ dep_displayinfo ]
+
+if platform != 'windows'
+  wsi_deps += [
+    lib_sdl2.partial_dependency(compile_args: true, includes: true),
+    lib_glfw.partial_dependency(compile_args: true, includes: true),
+  ]
 endif
 
 wsi_lib = static_library('wsi', wsi_src,
diff -pruN 2.3.1-1/src/wsi/sdl2/wsi_monitor_sdl2.cpp 2.4-1/src/wsi/sdl2/wsi_monitor_sdl2.cpp
--- 2.3.1-1/src/wsi/sdl2/wsi_monitor_sdl2.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/wsi/sdl2/wsi_monitor_sdl2.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -1,3 +1,5 @@
+#if defined(DXVK_WSI_SDL2)
+
 #include "../wsi_monitor.h"
 
 #include "wsi/native_wsi.h"
@@ -12,22 +14,22 @@
 
 namespace dxvk::wsi {
 
-  HMONITOR getDefaultMonitor() {
+  HMONITOR Sdl2WsiDriver::getDefaultMonitor() {
     return enumMonitors(0);
   }
 
 
-  HMONITOR enumMonitors(uint32_t index) {
+  HMONITOR Sdl2WsiDriver::enumMonitors(uint32_t index) {
     return isDisplayValid(int32_t(index))
       ? toHmonitor(index)
       : nullptr;
   }
 
-  HMONITOR enumMonitors(const LUID *adapterLUID[], uint32_t numLUIDs, uint32_t index) {
+  HMONITOR Sdl2WsiDriver::enumMonitors(const LUID *adapterLUID[], uint32_t numLUIDs, uint32_t index) {
     return enumMonitors(index);
   }
 
-  bool getDisplayName(
+  bool Sdl2WsiDriver::getDisplayName(
           HMONITOR         hMonitor,
           WCHAR            (&Name)[32]) {
     const int32_t displayId    = fromHmonitor(hMonitor);
@@ -47,7 +49,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool getDesktopCoordinates(
+  bool Sdl2WsiDriver::getDesktopCoordinates(
           HMONITOR         hMonitor,
           RECT*            pRect) {
     const int32_t displayId    = fromHmonitor(hMonitor);
@@ -90,7 +92,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool getDisplayMode(
+  bool Sdl2WsiDriver::getDisplayMode(
           HMONITOR         hMonitor,
           uint32_t         ModeNumber,
           WsiMode*         pMode) {
@@ -109,7 +111,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool getCurrentDisplayMode(
+  bool Sdl2WsiDriver::getCurrentDisplayMode(
           HMONITOR         hMonitor,
           WsiMode*         pMode) {
     const int32_t displayId    = fromHmonitor(hMonitor);
@@ -129,7 +131,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool getDesktopDisplayMode(
+  bool Sdl2WsiDriver::getDesktopDisplayMode(
           HMONITOR         hMonitor,
           WsiMode*         pMode) {
     const int32_t displayId    = fromHmonitor(hMonitor);
@@ -148,9 +150,11 @@ namespace dxvk::wsi {
     return true;
   }
 
-  std::vector<uint8_t> getMonitorEdid(HMONITOR hMonitor) {
+  std::vector<uint8_t> Sdl2WsiDriver::getMonitorEdid(HMONITOR hMonitor) {
     Logger::err("getMonitorEdid not implemented on this platform.");
     return {};
   }
 
 }
+
+#endif
diff -pruN 2.3.1-1/src/wsi/sdl2/wsi_platform_sdl2.cpp 2.4-1/src/wsi/sdl2/wsi_platform_sdl2.cpp
--- 2.3.1-1/src/wsi/sdl2/wsi_platform_sdl2.cpp	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/wsi/sdl2/wsi_platform_sdl2.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,69 @@
+#if defined(DXVK_WSI_SDL2)
+
+#include "wsi_platform_sdl2.h"
+#include "../../util/util_error.h"
+#include "../../util/util_string.h"
+#include "../../util/util_win32_compat.h"
+
+#include <SDL2/SDL_vulkan.h>
+
+namespace dxvk::wsi {
+
+  Sdl2WsiDriver::Sdl2WsiDriver() {
+    libsdl = LoadLibraryA( // FIXME: Get soname as string from meson
+#if defined(_WIN32)
+        "SDL2.dll"
+#elif defined(__APPLE__)
+        "libSDL2-2.0.0.dylib"
+#else
+        "libSDL2-2.0.so.0"
+#endif
+      );
+    if (libsdl == nullptr)
+      throw DxvkError("SDL2 WSI: Failed to load SDL2 DLL.");
+
+    #define SDL_PROC(ret, name, params) \
+      name = reinterpret_cast<pfn_##name>(GetProcAddress(libsdl, #name)); \
+      if (name == nullptr) { \
+        FreeLibrary(libsdl); \
+        libsdl = nullptr; \
+        throw DxvkError("SDL2 WSI: Failed to load " #name "."); \
+      }
+    #include "wsi_platform_sdl2_funcs.h"
+  }
+
+  Sdl2WsiDriver::~Sdl2WsiDriver() {
+    FreeLibrary(libsdl);
+  }
+
+  std::vector<const char *> Sdl2WsiDriver::getInstanceExtensions() {
+    SDL_Vulkan_LoadLibrary(nullptr);
+
+    uint32_t extensionCount = 0;
+    if (!SDL_Vulkan_GetInstanceExtensions(nullptr, &extensionCount, nullptr))
+      throw DxvkError(str::format("SDL2 WSI: Failed to get instance extension count. ", SDL_GetError()));
+
+    auto extensionNames = std::vector<const char *>(extensionCount);
+    if (!SDL_Vulkan_GetInstanceExtensions(nullptr, &extensionCount, extensionNames.data()))
+      throw DxvkError(str::format("SDL2 WSI: Failed to get instance extensions. ", SDL_GetError()));
+
+    return extensionNames;
+  }
+
+  static bool createSdl2WsiDriver(WsiDriver **driver) {
+    try {
+      *driver = new Sdl2WsiDriver();
+    } catch (const DxvkError& e) {
+      return false;
+    }
+    return true;
+  }
+
+  WsiBootstrap Sdl2WSI = {
+    "SDL2",
+    createSdl2WsiDriver
+  };
+
+}
+
+#endif
diff -pruN 2.3.1-1/src/wsi/sdl2/wsi_platform_sdl2.h 2.4-1/src/wsi/sdl2/wsi_platform_sdl2.h
--- 2.3.1-1/src/wsi/sdl2/wsi_platform_sdl2.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/wsi/sdl2/wsi_platform_sdl2.h	2024-07-10 09:56:14.000000000 +0000
@@ -1,21 +1,108 @@
 #pragma once
 
-#include <SDL2/SDL.h>
+#include <SDL.h>
 
-#include "../wsi_monitor.h"
+#include "../wsi_platform.h"
 
 namespace dxvk::wsi {
 
-  /**
-    * \brief Impl-dependent state
-    */
-  struct DxvkWindowState {
+  class Sdl2WsiDriver : public WsiDriver {
+  private:
+    HMODULE libsdl;
+    #define SDL_PROC(ret, name, params) \
+      typedef ret (SDLCALL *pfn_##name) params; \
+      pfn_##name name;
+    #include "wsi_platform_sdl2_funcs.h"
+
+    inline bool isDisplayValid(int32_t displayId) {
+      const int32_t displayCount = SDL_GetNumVideoDisplays();
+
+      return displayId < displayCount && displayId >= 0;
+    }
+
+  public:
+    Sdl2WsiDriver();
+    ~Sdl2WsiDriver();
+
+    // Platform
+    virtual std::vector<const char *> getInstanceExtensions();
+
+    // Monitor
+    virtual HMONITOR getDefaultMonitor();
+
+    virtual HMONITOR enumMonitors(uint32_t index);
+
+    virtual HMONITOR enumMonitors(const LUID *adapterLUID[], uint32_t numLUIDs, uint32_t index);
+
+    virtual bool getDisplayName(
+            HMONITOR         hMonitor,
+            WCHAR            (&Name)[32]);
+
+    virtual bool getDesktopCoordinates(
+            HMONITOR         hMonitor,
+            RECT*            pRect);
+
+    virtual bool getDisplayMode(
+            HMONITOR         hMonitor,
+            uint32_t         modeNumber,
+            WsiMode*         pMode);
+
+    virtual bool getCurrentDisplayMode(
+            HMONITOR         hMonitor,
+            WsiMode*         pMode);
+
+    virtual bool getDesktopDisplayMode(
+            HMONITOR         hMonitor,
+            WsiMode*         pMode);
+
+    virtual WsiEdidData getMonitorEdid(HMONITOR hMonitor);
+
+    // Window
+
+    virtual void getWindowSize(
+            HWND      hWindow,
+            uint32_t* pWidth,
+            uint32_t* pWeight);
+
+    virtual void resizeWindow(
+            HWND             hWindow,
+            DxvkWindowState* pState,
+            uint32_t         width,
+            uint32_t         weight);
+
+    virtual bool setWindowMode(
+            HMONITOR         hMonitor,
+            HWND             hWindow,
+      const WsiMode&         mode);
+
+    virtual bool enterFullscreenMode(
+            HMONITOR         hMonitor,
+            HWND             hWindow,
+            DxvkWindowState* pState,
+            [[maybe_unused]]
+            bool             modeSwitch);
+
+    virtual bool leaveFullscreenMode(
+            HWND             hWindow,
+            DxvkWindowState* pState,
+            bool             restoreCoordinates);
+
+    virtual bool restoreDisplayMode();
+
+    virtual HMONITOR getWindowMonitor(HWND hWindow);
+
+    virtual bool isWindow(HWND hWindow);
+
+    virtual void updateFullscreenWindow(
+            HMONITOR hMonitor,
+            HWND     hWindow,
+            bool     forceTopmost);
+
+    virtual VkResult createSurface(
+            HWND                hWindow,
+            PFN_vkGetInstanceProcAddr pfnVkGetInstanceProcAddr,
+            VkInstance          instance,
+            VkSurfaceKHR*       pSurface);
   };
-
-  inline bool isDisplayValid(int32_t displayId) {
-    const int32_t displayCount = SDL_GetNumVideoDisplays();
-
-    return displayId < displayCount && displayId >= 0;
-  }
   
-}
\ No newline at end of file
+}
diff -pruN 2.3.1-1/src/wsi/sdl2/wsi_platform_sdl2_funcs.h 2.4-1/src/wsi/sdl2/wsi_platform_sdl2_funcs.h
--- 2.3.1-1/src/wsi/sdl2/wsi_platform_sdl2_funcs.h	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/wsi/sdl2/wsi_platform_sdl2_funcs.h	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,16 @@
+SDL_PROC(SDL_DisplayMode*, SDL_GetClosestDisplayMode, (int, const SDL_DisplayMode*, SDL_DisplayMode*))
+SDL_PROC(int, SDL_GetCurrentDisplayMode, (int, SDL_DisplayMode*))
+SDL_PROC(int, SDL_GetDesktopDisplayMode, (int, SDL_DisplayMode*))
+SDL_PROC(int, SDL_GetDisplayBounds, (int, SDL_Rect*))
+SDL_PROC(int, SDL_GetDisplayMode, (int, int, SDL_DisplayMode*))
+SDL_PROC(const char*, SDL_GetError, (void))
+SDL_PROC(int, SDL_GetNumVideoDisplays, (void))
+SDL_PROC(int, SDL_GetWindowDisplayIndex, (SDL_Window*))
+SDL_PROC(int, SDL_SetWindowDisplayMode, (SDL_Window*, const SDL_DisplayMode*))
+SDL_PROC(int, SDL_SetWindowFullscreen, (SDL_Window*, Uint32))
+SDL_PROC(void, SDL_GetWindowSize, (SDL_Window*, int*, int*))
+SDL_PROC(void, SDL_SetWindowSize, (SDL_Window*, int, int))
+SDL_PROC(SDL_bool, SDL_Vulkan_CreateSurface, (SDL_Window*, VkInstance, VkSurfaceKHR*))
+SDL_PROC(SDL_bool, SDL_Vulkan_GetInstanceExtensions, (SDL_Window*, unsigned int*, const char**))
+SDL_PROC(int, SDL_Vulkan_LoadLibrary, (const char*))
+#undef SDL_PROC
diff -pruN 2.3.1-1/src/wsi/sdl2/wsi_window_sdl2.cpp 2.4-1/src/wsi/sdl2/wsi_window_sdl2.cpp
--- 2.3.1-1/src/wsi/sdl2/wsi_window_sdl2.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/wsi/sdl2/wsi_window_sdl2.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -1,17 +1,19 @@
+#if defined(DXVK_WSI_SDL2)
+
 #include "../wsi_window.h"
 
-#include "native/wsi/native_wsi.h"
+#include "native/wsi/native_sdl2.h"
 #include "wsi_platform_sdl2.h"
 
 #include "../../util/util_string.h"
 #include "../../util/log/log.h"
 
 #include <windows.h>
-#include <SDL2/SDL_vulkan.h>
+#include <SDL_vulkan.h>
 
 namespace dxvk::wsi {
 
-  void getWindowSize(
+  void Sdl2WsiDriver::getWindowSize(
         HWND      hWindow,
         uint32_t* pWidth,
         uint32_t* pHeight) {
@@ -28,7 +30,7 @@ namespace dxvk::wsi {
   }
 
 
-  void resizeWindow(
+  void Sdl2WsiDriver::resizeWindow(
           HWND             hWindow,
           DxvkWindowState* pState,
           uint32_t         Width,
@@ -39,7 +41,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool setWindowMode(
+  bool Sdl2WsiDriver::setWindowMode(
           HMONITOR         hMonitor,
           HWND             hWindow,
     const WsiMode&         pMode) {
@@ -73,7 +75,7 @@ namespace dxvk::wsi {
 
 
 
-  bool enterFullscreenMode(
+  bool Sdl2WsiDriver::enterFullscreenMode(
           HMONITOR         hMonitor,
           HWND             hWindow,
           DxvkWindowState* pState,
@@ -99,7 +101,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool leaveFullscreenMode(
+  bool Sdl2WsiDriver::leaveFullscreenMode(
           HWND             hWindow,
           DxvkWindowState* pState,
           bool             restoreCoordinates) {
@@ -114,13 +116,13 @@ namespace dxvk::wsi {
   }
 
 
-  bool restoreDisplayMode() {
+  bool Sdl2WsiDriver::restoreDisplayMode() {
     // Don't need to do anything with SDL2 here.
     return true;
   }
 
 
-  HMONITOR getWindowMonitor(HWND hWindow) {
+  HMONITOR Sdl2WsiDriver::getWindowMonitor(HWND hWindow) {
     SDL_Window* window      = fromHwnd(hWindow);
     const int32_t displayId = SDL_GetWindowDisplayIndex(window);
 
@@ -128,13 +130,13 @@ namespace dxvk::wsi {
   }
 
 
-  bool isWindow(HWND hWindow) {
+  bool Sdl2WsiDriver::isWindow(HWND hWindow) {
     SDL_Window* window = fromHwnd(hWindow);
     return window != nullptr;
   }
 
 
-  void updateFullscreenWindow(
+  void Sdl2WsiDriver::updateFullscreenWindow(
           HMONITOR hMonitor,
           HWND     hWindow,
           bool     forceTopmost) {
@@ -142,7 +144,7 @@ namespace dxvk::wsi {
   }
 
 
-  VkResult createSurface(
+  VkResult Sdl2WsiDriver::createSurface(
           HWND                      hWindow,
           PFN_vkGetInstanceProcAddr pfnVkGetInstanceProcAddr,
           VkInstance                instance,
@@ -155,3 +157,5 @@ namespace dxvk::wsi {
   }
 
 }
+
+#endif
diff -pruN 2.3.1-1/src/wsi/win32/wsi_monitor_win32.cpp 2.4-1/src/wsi/win32/wsi_monitor_win32.cpp
--- 2.3.1-1/src/wsi/win32/wsi_monitor_win32.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/wsi/win32/wsi_monitor_win32.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -1,4 +1,6 @@
-#include "../wsi_monitor.h"
+#if defined(DXVK_WSI_WIN32)
+
+#include "wsi_platform_win32.h"
 
 #include "../../util/util_string.h"
 #include "../../util/log/log.h"
@@ -13,7 +15,7 @@
 
 namespace dxvk::wsi {
 
-  HMONITOR getDefaultMonitor() {
+  HMONITOR Win32WsiDriver::getDefaultMonitor() {
     return ::MonitorFromPoint({ 0, 0 }, MONITOR_DEFAULTTOPRIMARY);
   }
 
@@ -45,7 +47,7 @@ namespace dxvk::wsi {
     return FALSE;
   }
 
-  HMONITOR enumMonitors(uint32_t index) {
+  HMONITOR Win32WsiDriver::enumMonitors(uint32_t index) {
     MonitorEnumInfo info;
     info.iMonitorId = index;
     info.oMonitor   = nullptr;
@@ -58,7 +60,7 @@ namespace dxvk::wsi {
     return info.oMonitor;
   }
 
-  HMONITOR enumMonitors(const LUID *adapterLUID[], uint32_t numLUIDs, uint32_t index) {
+  HMONITOR Win32WsiDriver::enumMonitors(const LUID *adapterLUID[], uint32_t numLUIDs, uint32_t index) {
     if (!numLUIDs)
       return enumMonitors(index);
 
@@ -132,7 +134,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool getDisplayName(
+  bool Win32WsiDriver::getDisplayName(
           HMONITOR         hMonitor,
           WCHAR            (&Name)[32]) {
     // Query monitor info to get the device name
@@ -150,7 +152,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool getDesktopCoordinates(
+  bool Win32WsiDriver::getDesktopCoordinates(
           HMONITOR         hMonitor,
           RECT*            pRect) {
     ::MONITORINFOEXW monInfo;
@@ -200,7 +202,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool getDisplayMode(
+  bool Win32WsiDriver::getDisplayMode(
           HMONITOR         hMonitor,
           uint32_t         modeNumber,
           WsiMode*         pMode) {
@@ -208,14 +210,14 @@ namespace dxvk::wsi {
   }
 
 
-  bool getCurrentDisplayMode(
+  bool Win32WsiDriver::getCurrentDisplayMode(
           HMONITOR         hMonitor,
           WsiMode*         pMode) {
     return retrieveDisplayMode(hMonitor, ENUM_CURRENT_SETTINGS, pMode);
   }
 
 
-  bool getDesktopDisplayMode(
+  bool Win32WsiDriver::getDesktopDisplayMode(
           HMONITOR         hMonitor,
           WsiMode*         pMode) {
     return retrieveDisplayMode(hMonitor, ENUM_REGISTRY_SETTINGS, pMode);
@@ -308,7 +310,7 @@ namespace dxvk::wsi {
     wchar_t extraChars[MAX_DEVICE_ID_LEN];
   };
 
-  WsiEdidData getMonitorEdid(HMONITOR hMonitor) {
+  WsiEdidData Win32WsiDriver::getMonitorEdid(HMONITOR hMonitor) {
     static constexpr GUID GUID_DEVINTERFACE_MONITOR = { 0xe6f07b5f, 0xee97, 0x4a90, 0xb0, 0x76, 0x33, 0xf5, 0x7b, 0xf4, 0xea, 0xa7 };
     static auto pfnSetupDiGetClassDevsW             = reinterpret_cast<decltype(SetupDiGetClassDevsW)*>            (::GetProcAddress(::GetModuleHandleW(L"setupapi.dll"), "SetupDiGetClassDevsW"));
     static auto pfnSetupDiEnumDeviceInterfaces      = reinterpret_cast<decltype(SetupDiEnumDeviceInterfaces)*>     (::GetProcAddress(::GetModuleHandleW(L"setupapi.dll"), "SetupDiEnumDeviceInterfaces"));
@@ -370,3 +372,5 @@ namespace dxvk::wsi {
   }
 
 }
+
+#endif
diff -pruN 2.3.1-1/src/wsi/win32/wsi_platform_win32.cpp 2.4-1/src/wsi/win32/wsi_platform_win32.cpp
--- 2.3.1-1/src/wsi/win32/wsi_platform_win32.cpp	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/wsi/win32/wsi_platform_win32.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,23 @@
+#if defined(DXVK_WSI_WIN32)
+
+#include "wsi_platform_win32.h"
+
+namespace dxvk::wsi {
+
+  std::vector<const char *> Win32WsiDriver::getInstanceExtensions() {
+    return { VK_KHR_WIN32_SURFACE_EXTENSION_NAME };
+  }
+
+  static bool createWin32WsiDriver(WsiDriver **driver) {
+    *driver = new Win32WsiDriver();
+    return true;
+  }
+
+  WsiBootstrap Win32WSI = {
+    "Win32",
+    createWin32WsiDriver
+  };
+
+}
+
+#endif
diff -pruN 2.3.1-1/src/wsi/win32/wsi_platform_win32.h 2.4-1/src/wsi/win32/wsi_platform_win32.h
--- 2.3.1-1/src/wsi/win32/wsi_platform_win32.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/wsi/win32/wsi_platform_win32.h	2024-07-10 09:56:14.000000000 +0000
@@ -2,15 +2,91 @@
 
 #include <windows.h>
 
+#include "../wsi_platform.h"
+
 namespace dxvk::wsi {
 
-  /**
-    * \brief Impl-dependent state
-    */
-  struct DxvkWindowState {
-    LONG style   = 0;
-    LONG exstyle = 0;
-    RECT rect    = { 0, 0, 0, 0 };
+  class Win32WsiDriver : public WsiDriver {
+  public:
+    // Platform
+    virtual std::vector<const char *> getInstanceExtensions();
+
+    // Monitor
+    virtual HMONITOR getDefaultMonitor();
+
+    virtual HMONITOR enumMonitors(uint32_t index);
+
+    virtual HMONITOR enumMonitors(const LUID *adapterLUID[], uint32_t numLUIDs, uint32_t index);
+
+    virtual bool getDisplayName(
+            HMONITOR         hMonitor,
+            WCHAR            (&Name)[32]);
+
+    virtual bool getDesktopCoordinates(
+            HMONITOR         hMonitor,
+            RECT*            pRect);
+
+    virtual bool getDisplayMode(
+            HMONITOR         hMonitor,
+            uint32_t         modeNumber,
+            WsiMode*         pMode);
+
+    virtual bool getCurrentDisplayMode(
+            HMONITOR         hMonitor,
+            WsiMode*         pMode);
+
+    virtual bool getDesktopDisplayMode(
+            HMONITOR         hMonitor,
+            WsiMode*         pMode);
+
+    virtual WsiEdidData getMonitorEdid(HMONITOR hMonitor);
+
+    // Window
+
+    virtual void getWindowSize(
+            HWND      hWindow,
+            uint32_t* pWidth,
+            uint32_t* pWeight);
+
+    virtual void resizeWindow(
+            HWND             hWindow,
+            DxvkWindowState* pState,
+            uint32_t         width,
+            uint32_t         weight);
+
+    virtual bool setWindowMode(
+            HMONITOR         hMonitor,
+            HWND             hWindow,
+      const WsiMode&         mode);
+
+    virtual bool enterFullscreenMode(
+            HMONITOR         hMonitor,
+            HWND             hWindow,
+            DxvkWindowState* pState,
+            [[maybe_unused]]
+            bool             modeSwitch);
+
+    virtual bool leaveFullscreenMode(
+            HWND             hWindow,
+            DxvkWindowState* pState,
+            bool             restoreCoordinates);
+
+    virtual bool restoreDisplayMode();
+
+    virtual HMONITOR getWindowMonitor(HWND hWindow);
+
+    virtual bool isWindow(HWND hWindow);
+
+    virtual void updateFullscreenWindow(
+            HMONITOR hMonitor,
+            HWND     hWindow,
+            bool     forceTopmost);
+
+    virtual VkResult createSurface(
+            HWND                hWindow,
+            PFN_vkGetInstanceProcAddr pfnVkGetInstanceProcAddr,
+            VkInstance          instance,
+            VkSurfaceKHR*       pSurface);
   };
 
-}
\ No newline at end of file
+}
diff -pruN 2.3.1-1/src/wsi/win32/wsi_window_win32.cpp 2.4-1/src/wsi/win32/wsi_window_win32.cpp
--- 2.3.1-1/src/wsi/win32/wsi_window_win32.cpp	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/wsi/win32/wsi_window_win32.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -1,5 +1,6 @@
-#include "../wsi_window.h"
-#include "../wsi_monitor.h"
+#if defined(DXVK_WSI_WIN32)
+
+#include "wsi_platform_win32.h"
 
 #include "../../util/util_string.h"
 #include "../../util/log/log.h"
@@ -94,7 +95,7 @@ namespace dxvk::wsi {
   }
 
 
-  void getWindowSize(
+  void Win32WsiDriver::getWindowSize(
         HWND      hWindow,
         uint32_t* pWidth,
         uint32_t* pHeight) {
@@ -109,7 +110,7 @@ namespace dxvk::wsi {
   }
 
 
-  void resizeWindow(
+  void Win32WsiDriver::resizeWindow(
           HWND             hWindow,
           DxvkWindowState* pState,
           uint32_t         width,
@@ -130,7 +131,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool setWindowMode(
+  bool Win32WsiDriver::setWindowMode(
           HMONITOR                hMonitor,
           HWND                    hWindow,
     const WsiMode&                mode) {
@@ -163,21 +164,21 @@ namespace dxvk::wsi {
   }
 
 
-  bool enterFullscreenMode(
+  bool Win32WsiDriver::enterFullscreenMode(
           HMONITOR         hMonitor,
           HWND             hWindow,
           DxvkWindowState* pState,
           [[maybe_unused]]
           bool             modeSwitch) {
     // Find a display mode that matches what we need
-    ::GetWindowRect(hWindow, &pState->rect);
+    ::GetWindowRect(hWindow, &pState->win.rect);
 
     // Change the window flags to remove the decoration etc.
     LONG style   = ::GetWindowLongW(hWindow, GWL_STYLE);
     LONG exstyle = ::GetWindowLongW(hWindow, GWL_EXSTYLE);
     
-    pState->style = style;
-    pState->exstyle = exstyle;
+    pState->win.style = style;
+    pState->win.exstyle = exstyle;
     
     style   &= ~WS_OVERLAPPEDWINDOW;
     exstyle &= ~WS_EX_OVERLAPPEDWINDOW;
@@ -196,7 +197,7 @@ namespace dxvk::wsi {
   }
 
 
-  bool leaveFullscreenMode(
+  bool Win32WsiDriver::leaveFullscreenMode(
           HWND             hWindow,
           DxvkWindowState* pState,
           bool             restoreCoordinates) {
@@ -205,27 +206,27 @@ namespace dxvk::wsi {
     LONG curStyle   = ::GetWindowLongW(hWindow, GWL_STYLE)   & ~WS_VISIBLE;
     LONG curExstyle = ::GetWindowLongW(hWindow, GWL_EXSTYLE) & ~WS_EX_TOPMOST;
 
-    if (curStyle   == (pState->style   & ~(WS_VISIBLE    | WS_OVERLAPPEDWINDOW))
-     && curExstyle == (pState->exstyle & ~(WS_EX_TOPMOST | WS_EX_OVERLAPPEDWINDOW))) {
-      ::SetWindowLongW(hWindow, GWL_STYLE,   pState->style);
-      ::SetWindowLongW(hWindow, GWL_EXSTYLE, pState->exstyle);
+    if (curStyle   == (pState->win.style   & ~(WS_VISIBLE    | WS_OVERLAPPEDWINDOW))
+     && curExstyle == (pState->win.exstyle & ~(WS_EX_TOPMOST | WS_EX_OVERLAPPEDWINDOW))) {
+      ::SetWindowLongW(hWindow, GWL_STYLE,   pState->win.style);
+      ::SetWindowLongW(hWindow, GWL_EXSTYLE, pState->win.exstyle);
     }
 
     // Restore window position and apply the style
     UINT flags = SWP_FRAMECHANGED | SWP_NOACTIVATE;
-    const RECT rect = pState->rect;
+    const RECT rect = pState->win.rect;
 
     if (!restoreCoordinates)
       flags |= SWP_NOSIZE | SWP_NOMOVE;
     
-    ::SetWindowPos(hWindow, (pState->exstyle & WS_EX_TOPMOST) ? HWND_TOPMOST : HWND_NOTOPMOST,
+    ::SetWindowPos(hWindow, (pState->win.exstyle & WS_EX_TOPMOST) ? HWND_TOPMOST : HWND_NOTOPMOST,
       rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, flags);
 
     return true;
   }
 
 
-  bool restoreDisplayMode() {
+  bool Win32WsiDriver::restoreDisplayMode() {
     bool success = true;
     bool result = ::EnumDisplayMonitors(nullptr, nullptr,
       &restoreDisplayModeCallback,
@@ -235,7 +236,7 @@ namespace dxvk::wsi {
   }
 
 
-  HMONITOR getWindowMonitor(HWND hWindow) {
+  HMONITOR Win32WsiDriver::getWindowMonitor(HWND hWindow) {
     RECT windowRect = { 0, 0, 0, 0 };
     ::GetWindowRect(hWindow, &windowRect);
     
@@ -248,12 +249,12 @@ namespace dxvk::wsi {
   }
 
 
-  bool isWindow(HWND hWindow) {
+  bool Win32WsiDriver::isWindow(HWND hWindow) {
     return ::IsWindow(hWindow);
   }
 
 
-  void updateFullscreenWindow(
+  void Win32WsiDriver::updateFullscreenWindow(
           HMONITOR hMonitor,
           HWND     hWindow,
           bool     forceTopmost) {
@@ -274,7 +275,7 @@ namespace dxvk::wsi {
   }
 
 
-  VkResult createSurface(
+  VkResult Win32WsiDriver::createSurface(
           HWND                hWindow,
           PFN_vkGetInstanceProcAddr pfnVkGetInstanceProcAddr,
           VkInstance          instance,
@@ -296,3 +297,5 @@ namespace dxvk::wsi {
   }
 
 }
+
+#endif
diff -pruN 2.3.1-1/src/wsi/wsi_platform.cpp 2.4-1/src/wsi/wsi_platform.cpp
--- 2.3.1-1/src/wsi/wsi_platform.cpp	1970-01-01 00:00:00.000000000 +0000
+++ 2.4-1/src/wsi/wsi_platform.cpp	2024-07-10 09:56:14.000000000 +0000
@@ -0,0 +1,177 @@
+#include "wsi_platform.h"
+#include "wsi_monitor.h"
+#include "wsi_window.h"
+#include "../util/util_env.h"
+#include "../util/util_error.h"
+
+namespace dxvk::wsi {
+  static WsiDriver* s_driver = nullptr;
+  static int s_refcount = 0;
+
+  static const WsiBootstrap *wsiBootstrap[] = {
+#if defined(DXVK_WSI_WIN32)
+    &Win32WSI,
+#endif
+#if defined(DXVK_WSI_SDL2)
+    &Sdl2WSI,
+#endif
+#if defined(DXVK_WSI_GLFW)
+    &GlfwWSI,
+#endif
+  };
+
+  void init() {
+    if (s_refcount++ > 0)
+      return;
+
+    std::string hint = dxvk::env::getEnvVar("DXVK_WSI_DRIVER");
+    if (hint == "") {
+        // At least for Windows, it is reasonable to fall back to a default;
+        // for other platforms however we _need_ to know which WSI to use!
+#if defined(DXVK_WSI_WIN32)
+        hint = "Win32";
+#else
+        throw DxvkError("DXVK_WSI_DRIVER environment variable unset");
+#endif
+    }
+
+    bool success = false;
+    for (const WsiBootstrap *b : wsiBootstrap) {
+      if (hint == b->name && b->createDriver(&s_driver)) {
+        success = true;
+        break;
+      }
+    }
+
+    if (!success)
+      throw DxvkError("Failed to initialize WSI.");
+  }
+
+  void quit() {
+    if (s_refcount == 0)
+      return;
+
+    s_refcount--;
+    if (s_refcount == 0) {
+      delete s_driver;
+      s_driver = nullptr;
+    }
+  }
+
+  std::vector<const char *> getInstanceExtensions() {
+    return s_driver->getInstanceExtensions();
+  }
+
+  void getWindowSize(
+          HWND      hWindow,
+          uint32_t* pWidth,
+          uint32_t* pHeight) {
+    s_driver->getWindowSize(hWindow, pWidth, pHeight);
+  }
+
+  void resizeWindow(
+          HWND             hWindow,
+          DxvkWindowState* pState,
+          uint32_t         width,
+          uint32_t         height) {
+    s_driver->resizeWindow(hWindow, pState, width, height);
+  }
+
+  bool setWindowMode(
+          HMONITOR         hMonitor,
+          HWND             hWindow,
+    const WsiMode&         mode) {
+    return s_driver->setWindowMode(hMonitor, hWindow, mode);
+  }
+
+  bool enterFullscreenMode(
+          HMONITOR         hMonitor,
+          HWND             hWindow,
+          DxvkWindowState* pState,
+          [[maybe_unused]]
+          bool             modeSwitch) {
+    return s_driver->enterFullscreenMode(hMonitor, hWindow, pState, modeSwitch);
+  }
+
+  bool leaveFullscreenMode(
+          HWND             hWindow,
+          DxvkWindowState* pState,
+          bool             restoreCoordinates) {
+    return s_driver->leaveFullscreenMode(hWindow, pState, restoreCoordinates);
+  }
+
+  bool restoreDisplayMode() {
+    return s_driver->restoreDisplayMode();
+  }
+
+  HMONITOR getWindowMonitor(HWND hWindow) {
+    return s_driver->getWindowMonitor(hWindow);
+  }
+
+  bool isWindow(HWND hWindow) {
+    return s_driver->isWindow(hWindow);
+  }
+
+  void updateFullscreenWindow(
+          HMONITOR hMonitor,
+          HWND     hWindow,
+          bool     forceTopmost) {
+    s_driver->updateFullscreenWindow(hMonitor, hWindow, forceTopmost);
+  }
+
+  VkResult createSurface(
+          HWND                hWindow,
+          PFN_vkGetInstanceProcAddr pfnVkGetInstanceProcAddr,
+          VkInstance          instance,
+          VkSurfaceKHR*       pSurface) {
+    return s_driver->createSurface(hWindow, pfnVkGetInstanceProcAddr, instance, pSurface);
+  }
+
+  HMONITOR getDefaultMonitor() {
+    return s_driver->getDefaultMonitor();
+  }
+
+  HMONITOR enumMonitors(uint32_t index) {
+    return s_driver->enumMonitors(index);
+  }
+
+  HMONITOR enumMonitors(const LUID *adapterLUID[], uint32_t numLUIDs, uint32_t index) {
+    return s_driver->enumMonitors(adapterLUID, numLUIDs, index);
+  }
+
+  bool getDisplayName(
+          HMONITOR         hMonitor,
+          WCHAR            (&Name)[32]) {
+    return s_driver->getDisplayName(hMonitor, Name);
+  }
+
+  bool getDesktopCoordinates(
+          HMONITOR         hMonitor,
+          RECT*            pRect) {
+    return s_driver->getDesktopCoordinates(hMonitor, pRect);
+  }
+
+  bool getDisplayMode(
+          HMONITOR         hMonitor,
+          uint32_t         modeNumber,
+          WsiMode*         pMode) {
+    return s_driver->getDisplayMode(hMonitor, modeNumber, pMode);
+  }
+
+  bool getCurrentDisplayMode(
+          HMONITOR         hMonitor,
+          WsiMode*         pMode) {
+    return s_driver->getCurrentDisplayMode(hMonitor, pMode);
+  }
+
+  bool getDesktopDisplayMode(
+          HMONITOR         hMonitor,
+          WsiMode*         pMode) {
+    return s_driver->getDesktopDisplayMode(hMonitor, pMode);
+  }
+
+  WsiEdidData getMonitorEdid(HMONITOR hMonitor) {
+    return s_driver->getMonitorEdid(hMonitor);
+  }
+
+}
diff -pruN 2.3.1-1/src/wsi/wsi_platform.h 2.4-1/src/wsi/wsi_platform.h
--- 2.3.1-1/src/wsi/wsi_platform.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/wsi/wsi_platform.h	2024-07-10 09:56:14.000000000 +0000
@@ -1,9 +1,114 @@
 #pragma once
 
+#include "wsi_window.h"
+
+#include <vector>
+
+namespace dxvk::wsi {
+
+  class WsiDriver {
+  public:
+    virtual ~WsiDriver() {
+    }
+
+    // Platform
+    virtual std::vector<const char *> getInstanceExtensions() = 0;
+
+    // Monitor
+    virtual HMONITOR getDefaultMonitor() = 0;
+
+    virtual HMONITOR enumMonitors(uint32_t index) = 0;
+
+    virtual HMONITOR enumMonitors(const LUID *adapterLUID[], uint32_t numLUIDs, uint32_t index) = 0;
+
+    virtual bool getDisplayName(
+            HMONITOR         hMonitor,
+            WCHAR            (&Name)[32]) = 0;
+
+    virtual bool getDesktopCoordinates(
+            HMONITOR         hMonitor,
+            RECT*            pRect) = 0;
+
+    virtual bool getDisplayMode(
+            HMONITOR         hMonitor,
+            uint32_t         modeNumber,
+            WsiMode*         pMode) = 0;
+
+    virtual bool getCurrentDisplayMode(
+            HMONITOR         hMonitor,
+            WsiMode*         pMode) = 0;
+
+    virtual bool getDesktopDisplayMode(
+            HMONITOR         hMonitor,
+            WsiMode*         pMode) = 0;
+
+    virtual WsiEdidData getMonitorEdid(HMONITOR hMonitor) = 0;
+
+    // Window
+
+    virtual void getWindowSize(
+            HWND      hWindow,
+            uint32_t* pWidth,
+            uint32_t* pWeight) = 0;
+
+    virtual void resizeWindow(
+            HWND             hWindow,
+            DxvkWindowState* pState,
+            uint32_t         width,
+            uint32_t         weight) = 0;
+
+    virtual bool setWindowMode(
+            HMONITOR         hMonitor,
+            HWND             hWindow,
+      const WsiMode&         mode) = 0;
+
+    virtual bool enterFullscreenMode(
+            HMONITOR         hMonitor,
+            HWND             hWindow,
+            DxvkWindowState* pState,
+            [[maybe_unused]]
+            bool             modeSwitch) = 0;
+
+    virtual bool leaveFullscreenMode(
+            HWND             hWindow,
+            DxvkWindowState* pState,
+            bool             restoreCoordinates) = 0;
+
+    virtual bool restoreDisplayMode() = 0;
+
+    virtual HMONITOR getWindowMonitor(HWND hWindow) = 0;
+
+    virtual bool isWindow(HWND hWindow) = 0;
+
+    virtual void updateFullscreenWindow(
+            HMONITOR hMonitor,
+            HWND     hWindow,
+            bool     forceTopmost) = 0;
+
+    virtual VkResult createSurface(
+            HWND                hWindow,
+            PFN_vkGetInstanceProcAddr pfnVkGetInstanceProcAddr,
+            VkInstance          instance,
+            VkSurfaceKHR*       pSurface) = 0;
+  };
+
+  struct WsiBootstrap {
+    const std::string name;
+    bool (*createDriver)(WsiDriver **driver);
+  };
+
 #if defined(DXVK_WSI_WIN32)
-#include "win32/wsi_platform_win32.h"
-#elif defined(DXVK_WSI_SDL2)
-#include "sdl2/wsi_platform_sdl2.h"
-#elif defined(DXVK_WSI_GLFW)
-#include "glfw/wsi_platform_glfw.h"
+  extern WsiBootstrap Win32WSI;
+#endif
+#if defined(DXVK_WSI_SDL2)
+  extern WsiBootstrap Sdl2WSI;
 #endif
+#if defined(DXVK_WSI_GLFW)
+  extern WsiBootstrap GlfwWSI;
+#endif
+
+  void init();
+  void quit();
+  std::vector<const char *> getInstanceExtensions();
+
+}
diff -pruN 2.3.1-1/src/wsi/wsi_window.h 2.4-1/src/wsi/wsi_window.h
--- 2.3.1-1/src/wsi/wsi_window.h	2024-03-20 12:48:30.000000000 +0000
+++ 2.4-1/src/wsi/wsi_window.h	2024-07-10 09:56:14.000000000 +0000
@@ -3,13 +3,31 @@
 #include <windows.h>
 
 #include "wsi_monitor.h"
-#include "wsi_platform.h"
 
 #include "../vulkan/vulkan_loader.h"
 
 namespace dxvk::wsi {
 
   /**
+    * \brief Impl-dependent state
+    */
+  struct DxvkWindowState {
+#if defined(DXVK_WSI_WIN32)
+    struct {
+      LONG style   = 0;
+      LONG exstyle = 0;
+      RECT rect    = { 0, 0, 0, 0 };
+    } win;
+#endif
+#if defined(DXVK_WSI_SDL2)
+    // Nothing to store
+#endif
+#if defined(DXVK_WSI_GLFW)
+    // Nothing to store
+#endif
+  };
+
+  /**
     * \brief The size of the window
     * 
     * \param [in] hWindow The window
